#!/bin/sh
#
# Script to automate rebuilding all python-dependent packages.
# Anything with Python modules needs to be rebuilt when the x.y
# version changes.  This script iterates over all packages in
# build-dependent order and when it finds a python package it does the
# following:
#
# 1) Bumps the REL value.
# 2) Rebuilds the package.
# 3) If #2 is successful, update the plist.
# 4) Rebuild again and commit changes.
#
# Any failing package stops the process.  Since this may be common,
# you are allowed to specify a package to start *after* on the command
# line.  For example, if you run this script for hours and it rebuilds
# about 20 python packages and fails on base/lib/pystuff then you can
# manually fix pystuff and come back and run this script with this
# syntax:
#
#     ./rebuild-dependent-packages.sh --start-after=base/lib/pystuff
#
# The script will pick up with the next python package *after*
# base/lib/pystuff.

PATH=/usr/bin
CWD="$(pwd)"
PROG="$(basename "$0")"
TOPSRC="$(make show var=TOPSRC)"
TMPDIR="$(mktemp -d)"
START_AFTER=
SHOW_ONLY=

trap 'rm -rf "${TMPDIR}"' EXIT

usage() {
    echo "Rebuild Python-dependent packages"
    echo "Usage: ${PROG} [options]"
    echo "Options:"
    echo "    -a PKGPATH, --start-after=PKGPATH  Relative package path to start build after"
    echo "    -s, --show-only                    Just display Python-dependent packages"
    echo "    -h, --help                         Display usage information"
    echo
    echo "Example:"
    echo "    ${PROG}"
    echo "    ${PROG} --start-after=base/lib/libxml"
}

OPTS=$(getopt -o 'a:sh' --long 'start-after:,show-only,help' -n "${PROG}" -- "$@")
RET=$?

if [ "${RET}" -ne 0 ]; then
    echo "Terminating..." >&2
    exit 1
fi

eval set -- "${OPTS}"
unset OPTS

while true ; do
    case "$1" in
        -a|--start-after)
            START_AFTER="$2"
            shift 2
            continue
            ;;
        -s|--show-only)
            SHOW_ONLY=y
            shift
            continue
            ;;
        -h|--help)
            usage
            exit 0
            ;;
        --)
            shift
            break
            ;;
        *)
            echo "Internal error!" >&2
            exit 1
            ;;
    esac
done

make -C "${TOPSRC}" --no-print-directory order | grep -v "base/devel/python" > "${TMPDIR}"/order.txt
total_lines=$(wc -l "${TMPDIR}"/order.txt | cut -d ' ' -f 1)

if [ -n "${START_AFTER}" ]; then
    mark=$(grep -n "${START_AFTER}" "${TMPDIR}"/order.txt | cut -d ':' -f 1)
    tail -n $((total_lines - mark)) "${TMPDIR}"/order.txt > "${TMPDIR}"/order2.txt
    mv "${TMPDIR}"/order2.txt "${TMPDIR}"/order.txt
fi

while read -r pkgpath ; do
    if ! grep -q ^usr/lib/python "${TOPSRC}"/"${pkgpath}"/plist* >/dev/null 2>&1 ; then
        continue
    fi

    if [ "${SHOW_ONLY}" = "y" ]; then
        echo "${pkgpath}"
    else
        cd "${TOPSRC}"/"${pkgpath}" || exit 1
        ( make clean bumprel && make world && make update-plist && make world ) || exit 1
        git add .
        git commit -m "[$(echo "${pkgpath}" | rev | cut -d '/' -f 1 | rev)] Rebuilt for new Python"
    fi
done < "${TMPDIR}"/order.txt

cd "${CWD}" || exit 1
