{-# OPTIONS_GHC -Wall #-}
{-# LANGUAGE MultiWayIf #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RecordWildCards #-}

-- A program to generate reports that help maintain the Haskell module packages in Peachtree.
--
-- This program only uses Haskell modules that we have packaged, so you can just use 'runhaskell'
-- on it.  There's no modules to install or building to do.
--
-- You can run this program in two ways:
--
-- (1) To generate a list of all the modules and what they are used by.  Unused modules will be
--     listed in a special section at the top.  This is the default mode of operation.
--
--     To generate this report, you give it a list of .conf files in a package.conf.d directory.
--     This could be the installed directory (/usr/lib/ghc/lib/package.conf.d/*conf) on the
--     running system.  It could also be an alternate package root that you set up while doing
--     development (/tmp/somewhere/usr/lib/ghc/lib/package.conf.d/*conf).
--
--     Run like so:
--
--     $ runhaskell module-report.hs /usr/lib/ghc/lib/package.conf.d/*conf | less
--
--     Anything in the UNUSED section can probably be removed, but double check first.  Modules
--     that only provide a binary, and no library, will not have a .conf file installed so we
--     can't check what they use.  It's best to use ldd for these - see the comment in 'isUsed'.
--
-- (2) To generate the order that modules should be listed in the 'upstream' file of one of the
--     big rollup packages.
--
--     To generate this report, you cd into a package directory and 'make extract'.  You then
--     give it all the .cabal files.  The report lists modules ONLY IN THIS PACKAGE in the order
--     they need to be built, earliest things (those the most depend upon) first.  We don't have
--     enough information to generate the entire file dynamically, unfortunately.
--
--     Run like so:
--
--     $ runhaskell module-report.hs --upstream-report $(find /usr/obj/haskell-lua-9.2.5-20.x86_64 -name '*.cabal')

import           Control.Monad(forM_)
import           Data.Bifunctor(second)
import qualified Data.ByteString as BS
import           Data.Graph(Graph, Vertex, graphFromEdges, topSort)
import           Data.List((\\), isInfixOf, sort)
import qualified Data.Map.Strict as Map
import qualified Data.Text as T
import qualified Data.Text.IO as TIO
import           Distribution.InstalledPackageInfo(InstalledPackageInfo(..), parseInstalledPackageInfo)
import           Distribution.PackageDescription.Parsec(parseGenericPackageDescriptionMaybe)
import           Distribution.Types.CondTree(CondTree(..))
import           Distribution.Types.Dependency(Dependency, depPkgName)
import           Distribution.Types.GenericPackageDescription(GenericPackageDescription(..))
import           Distribution.Types.PackageDescription(PackageDescription(..))
import           Distribution.Types.PackageId(PackageIdentifier(..))
import           Distribution.Text(display)
import           Options.Applicative
import           Text.Regex.TDFA((=~))

data Options = Options { optFiles :: [String],
                         optUpstreamReport :: Bool
                       }

type DepMap = Map.Map T.Text [T.Text]

parseOpts :: Parser Options
parseOpts = Options <$> some (argument str (metavar "FILES..."))
                    <*> switch (long "upstream-report" <> help "Generate a tsorted list of only the given modules")

-- Given a module config file, return a tuple of its name and a list of all the modules
-- it depends upon.
getDependencies :: FilePath -> IO (T.Text, [T.Text])
getDependencies confFile = do
    s <- BS.readFile confFile

    case parseInstalledPackageInfo s of
        Left err        -> do print err
                              return ("ERROR", [])
        Right (_, ipi)  -> do let packageName = display $ installedUnitId ipi
                              return (T.pack packageName, map (T.pack . display) $ depends ipi)

addOneInvertedDependency :: T.Text -> [T.Text] -> DepMap -> DepMap
addOneInvertedDependency _    []         depmap = depmap
addOneInvertedDependency name (dep:deps) depmap =
    case Map.lookup dep depmap of
        Nothing -> addOneInvertedDependency name deps (Map.insert dep [name] depmap)
        Just d  -> addOneInvertedDependency name deps (Map.insert dep (name:d) depmap)

-- Convert a list of modules and other modules it depends upon into a list of module names
-- and other modules that depend upon it.
invertDependencies :: [(T.Text, [T.Text])] -> DepMap -> DepMap
invertDependencies []                 depmap = depmap
invertDependencies ((name, deps):lst) depmap = invertDependencies lst (addOneInvertedDependency name deps depmap)

-- Is this some top-level module that isn't used by anything that we want to keep?  Typically,
-- this is something like a module that provides a binary.  Strangely, we also need to put
-- ghc in this list.
isUsed :: T.Text -> Bool
isUsed p = let
    keepers = [ "hscolour"          -- Program for syntax highlighting documentation
              , "hspec-discover"    -- Program for finding unit tests automatically
              , "pandoc"            -- Program for converting document formats
              , "shake"             -- Used in building ghc
              , "shellcheck"        -- Program for linting shell scripts
              , "xmobar"            -- Task manager thing used in xmonad
              , "xmonad-contrib"    -- Potentially used by your local xmonad.hs

              -- These modules ship with ghc, so we're not going to get rid of them
              -- even though nothing uses them
              , "ghc"               -- The compiler, of course
              , "ghc-compact"       -- Core library
              , "haskeline"         -- readline for haskell
              , "libiserv"          -- Internal compiler library
              , "xhtml"             -- Used by haddock

              -- These modules are used by packages that provide only a binary, thus there
              -- is no package.conf.d file to inspect.  I don't know of any way to figure
              -- this out automatically.  So, I've made this list by:
              --
              -- (1) Installing all haskell packages to a temp directory, say /tmp/blah.
              -- (2) ldd /tmp/blah/usr/X11/bin/* /tmp/blah/usr/bin/* 2>/dev/null | /bin/grep libHS | cut -d' ' -f1 | sort | uniq | less
              -- (3) Check the output of this program against that output.  Anything in the
              --     UNUSED section that's not in the ldd output can be removed.

              , "HTTP"              -- used by cabal-install
              , "echo"              -- used by cabal-install
              , "edit-distance"     -- used by cabal-install
              , "gi-gtk"            -- used by quitxmonad (and all my radio stuff)
              , "hackage-security"  -- used by cabal-install
              , "language-c"        -- used by c2hs
              , "resolv"            -- used by cabal-install
              ]
 in
    any (\r -> p =~ (r `T.append` "-[0-9]"))
        keepers

reportUnused :: [T.Text] -> IO ()
reportUnused []     = return ()
reportUnused unused = do
    TIO.putStrLn "UNUSED:"
    forM_ unused $ \u -> TIO.putStrLn $ "\t" `T.append` u

reportUsers :: [(T.Text, [T.Text])] -> IO ()
reportUsers lst = forM_ lst $ \(name, users) -> do
    TIO.putStrLn $ T.concat ["### ", name, " is used by:"]
    forM_ (sort users) $ \u -> TIO.putStrLn ("\t" `T.append` u)

-- Print a report on all modules and what uses them.
moduleReport :: [String] -> IO ()
moduleReport files = do
    -- deps is a list of tuples where each tuple is:  the package name, and a list of
    -- all the other packages it depends on.  If a package has no dependencies, it still
    -- gets a tuple but with an empty list.
    --
    -- We need to keep this list around so we can grab all the package names we
    -- examined.  This helps us know what's unused.
    deps <- mapM getDependencies files

    -- Invert that list - now each tuple is: the package name, a list of all the other
    -- packages that depend on it.  In this list, if a package is not depended on by
    -- anything, it will not be included.
    let invertedList = Map.toList $ invertDependencies deps Map.empty

    -- Unused packages are therefore anything in the first list that's not in the second.
    -- Well, basically - we need to apply some additional checks here.
        unused = filter (not . isUsed)
                        (map fst deps \\ map fst invertedList)

    -- And now do the reports
    reportUnused unused
    reportUsers invertedList

-- Extract the dependency names from a CondTree
depFromCondTree :: CondTree v [Dependency] a -> [T.Text]
depFromCondTree (CondNode _ deps _) = map (T.pack . display . depPkgName) deps

-- This isn't great - it will include dependencies we may or may not need, because
-- a GenericPackageDescription hasn't evaluated any conditionals yet.  So we basically
-- get everything here.  I can't find any way to turn a GPD into a PackageDescription
-- that would have conditionals processed.  On the other hand, it doesn't matter.  The
-- extra dependencies we get will be filtered out because they aren't part of the
-- set we're scanning.
genericDependencies :: GenericPackageDescription -> [T.Text]
genericDependencies GenericPackageDescription{..} = let
    libDeps    = maybe [] depFromCondTree condLibrary
    subLibDeps = concatMap (depFromCondTree . snd) condSubLibraries
    execDeps   = concatMap (depFromCondTree . snd) condExecutables
 in
    libDeps ++ subLibDeps ++ execDeps

getBuildDependencies :: FilePath -> IO (T.Text, [T.Text])
getBuildDependencies cabalFile = do
    s <- BS.readFile cabalFile

    case parseGenericPackageDescriptionMaybe s of
        Nothing  -> error $ "Could not read cabal file " ++ cabalFile
        Just gpd -> do let name = T.pack $ display $ pkgName $ package $ packageDescription gpd
                       return (name, genericDependencies gpd)

depToEdge :: (T.Text, [T.Text]) -> (T.Text, T.Text, [T.Text])
depToEdge (node, deps) = (node, node, deps)

toGraph :: [(T.Text, [T.Text])] -> (Graph, Vertex -> (T.Text, T.Text, [T.Text]))
toGraph deps = let
    (graph, nodeFromVertex, _) = graphFromEdges (map depToEdge deps)
 in
    (graph, nodeFromVertex)

-- Print a tsorted list of modules suitable for including in a package's "upstream" file.
-- This list will only include those modules that are given on the command line - none of
-- their other dependencies will be printed out.  Unused modules will still be included
-- on the assumption that something somewhere else uses them.
upstreamReport :: [String] -> IO ()
upstreamReport files = do
    -- deps is a list of tuples where each tuple is: the package name, and a list of
    -- all the other packages it depends on.  This includes packages outside the set we
    -- wanted to scan (like, those provided by other peachtree packages).
    deps <- mapM getBuildDependencies files

    -- Filter out all those dependencies that aren't in the set we're scanning.
    let pkgNames = map fst deps
        deps'    = map (second (filter (`elem` pkgNames))) deps

    let (graph, nodeFromVertex) = toGraph deps'
        graph'                  = reverse $ topSort graph

    forM_ graph' $ \v -> do
        let (node, _, _) = nodeFromVertex v
        TIO.putStrLn node

{- HLINT ignore "Redundant multi-way if" -}
reportMode :: Options -> IO ()
reportMode Options{..} = do
    -- Filter out the base module - everything uses it.
    let files = filter (\fp -> not $ "/base-" `isInfixOf` fp) optFiles

    if | optUpstreamReport -> upstreamReport files
       | otherwise         -> moduleReport files

main :: IO ()
main = execParser opts >>= reportMode
 where
    opts = info (parseOpts <**> helper)
                (fullDesc <> progDesc "Examine Haskell module dependencies")
