--
-- The Peachtree Linux default xmonad configuration
-- (based on David Cantrell's <dcantrell@burdell.org> xmonad.hs,
-- which is based on xmonad.hs from Chris Lumens <clumens@bangmoney.org>)
--

-- The UrgencyHook functions handle the URGENT window manager hints, which
-- allow for notifications on the status bar to work.  Applications are
-- configured individually to control their usage of the URGENT hint.
--
--
-- * urxvt will set the URGENT hint when a bell character appears:
--       URxvt.urgentOnBell: true
-- * Or for xterm:
--       XTerm.vt100.bellIsUrgent: true
--
-- * irssi sets the bell in its terminal on various highlights (e.g., a
--   privmsg or someone typing your nick in a channel):
--       /set bell_beeps ON
--       /set beep_msg_level MSGS NOTICES DCC DCCMSGS HILIGHT

-- Symbolic color names come from /usr/share/X11/rgb.txt
-- The chart shows decimal values for RGB and the symbolic name for that
-- color.
-- NOTE:  Using the RGB values requires conversion to hexadecimal.

import qualified Data.Map as M
import qualified XMonad.StackSet as W
import qualified XMonad as XM

import XMonad.Actions.Search(SearchEngine, google, images, maps, promptSearch, searchEngine, selectSearch)
import XMonad.Actions.TopicSpace(TopicConfig(..), checkTopicConfig, currentTopicDir, switchTopic)
import XMonad.Config.Desktop(desktopConfig)
import XMonad.Core(XConfig(..), spawn)
import XMonad.Hooks.DynamicLog(PP(..), statusBar, xmobarColor, xmobarPP)
import XMonad.Hooks.ManageDocks(avoidStruts, manageDocks)
import XMonad.Hooks.ManageHelpers(doCenterFloat, isDialog)
import XMonad.Hooks.UrgencyHook(NoUrgencyHook(..), withUrgencyHook)
import XMonad.Layout((|||), Full(..), Mirror(..), Tall(..))
import XMonad.Layout.Grid(Grid(..))
import XMonad.Layout.IM(Property(..), withIM)
import XMonad.Layout.IndependentScreens(countScreens)
import XMonad.Layout.NoBorders(smartBorders)
import XMonad.Layout.PerWorkspace(onWorkspace)
import XMonad.Layout.Reflect(reflectHoriz)
import XMonad.Layout.Tabbed(simpleTabbed)
import XMonad.Main(xmonad)
import XMonad.ManageHook((<+>), (-->), (=?), className, composeAll, doF, doFloat, doShift, stringProperty, title)
import XMonad.Prompt(XPConfig(..), def)
import XMonad.Prompt.Shell(shellPrompt)
import XMonad.Prompt.Ssh(sshPrompt)
import XMonad.Util.EZConfig(additionalKeysP, mkNamedKeymap)
import XMonad.Util.NamedActions((^++^), addDescrKeys, addName, noName, showKm, subtitle, xMessage)

-- Additional search engines
rhbz = searchEngine "rhbz" "https://bugzilla.redhat.com/buglist.cgi?quicksearch="

-- Preferred terminal program
myTerminal = "urxvt"

-- Topic (workspace) count and names.
myTopics = ["1", "2", "3", "4", "5", "6", "7", "8", "9"]

-- Automatically start programs when we start xmonad and change workspaces
myTopicConfig = TopicConfig {
    defaultTopicAction  = const spawnShell,
    defaultTopic        = "1",
    maxTopicHistory     = 10,
    -- Don't do anything automatically on workspace 3.  That's for radio
    -- programs that have special needs.  We'll spawn things there as
    -- needed.
    topicActions        = M.fromList [("3", return ()),
                                      ("4", spawn "firefox")],
    topicDirs           = M.fromList $ zip myTopics (repeat "${HOME}")
 }
 where spawnShell = currentTopicDir myTopicConfig >>= spawnShellIn
       spawnShellIn dir = spawn (myTerminal ++ " -cd " ++ dir)

-- Workspace layouts
-- NOTES:
--     avoidStruts keeps status bars always visible (xmobar, ...)
--     smartBorders disables window borders when they don't make sense
--     onWorkspace controls the layouts for that specific workspace
--     standardLayouts are applied to all other workspaces and can also
--        be cycled through on any workspace
--
myLayoutHook =
    avoidStruts $
    smartBorders $
    onWorkspace "3" wsjtx $
    onWorkspace "4" Full $
    onWorkspace "8" (gimp ||| standardLayouts)
    standardLayouts
    where standardLayouts = tiled ||| Mirror tiled ||| Full ||| simpleTabbed
          tiled           = Tall 1 (3/100) (1/2)
          gimp            = withIM 0.11 (Role "gimp-toolbox") $
                            reflectHoriz $
                            withIM 0.15 (Role "gimp-dock") Full
          wsjtx           = Mirror $
                            withIM 0.4 (Title "WSJT-X - Wide Graph") $
                            Mirror $
                            withIM 0.4 (ClassName "jt-slog") $
                            Grid

-- Window hooks.  These are called when windows are opened that match the
-- condition (e.g., title or className).  These hooks make sure programs
-- start up as expected.
--
-- Some programs belong on certain workspaces, so move them if necessary.
-- See myWorkspaces above for workspace descriptions.
myManageHook = composeAll
    [ isDialog                             --> doCenterFloat,
      title     =? "Firefox Preferences"   --> doCenterFloat,
      title     =? "About Mozilla Firefox" --> doCenterFloat,
      className =? "Vncviewer"             --> doCenterFloat,
      className =? "Firefox"               --> doShift "4",
      title     =? "alsamixer"             --> doCenterFloat,
      title     =? "YAD"                   --> doCenterFloat,

      -- move Gimp to workspace 8 & do not float main Gimp windows
      className =? "Gimp"                  --> doShift "8",
      role      =? "gimp-startup"          --> doCenterFloat,
      role      =? "gimp-toolbox"          --> unfloat,
      role      =? "gimp-image-window"     --> unfloat,
      role      =? "gimp-dock"             --> unfloat
    ] <+> manageDocks <+> manageHook desktopConfig
    where unfloat = XM.ask >>= doF . W.sink
          role    = stringProperty "WM_WINDOW_ROLE"

-- Prompt configuration
myXPConfig :: XPConfig
myXPConfig = def {
    font   = "xft:Liberation-Sans-12",
    height = 30
 }

-- Customize keyboard commands
-- NOTE:  This map customization uses XMonad.Utils.EZConfig.  See the
--        mkKeymap section in the XMonad.Utils.EZConfig docs for more
--        info.  Keys are mostly specified in Emacs style.
myKeys conf = let
    subKeys s ks = subtitle s : mkNamedKeymap conf ks
 in
    subKeys "custom keys" [
        ("M-S-<Print>", addName "Take a screenshot" $   spawn "gm import -window root ${HOME}/screenshot-$$.jpg"),

        ("M-m",     addName "Audio mixer"       $ spawn "urxvt -title alsamixer -e alsamixer"),
        ("M-q",     addName "Restart xmonad"      restartXmonad),

        -- Present a logout/shutdown dialog
        ("M-S-q",   addName "Quit xmonad"       $ spawn "quitxmonad"),
        ("M-S-x",   addName "Lock the screen"   $ spawn "xsecurelock")]

    ^++^

    subKeys "prompts" [
        ("M-p",     addName "Run a command"           $ shellPrompt myXPConfig),
        ("M-S-s",   addName "ssh to a remote host"    $ sshPrompt myXPConfig)]

    ^++^

    subKeys "M-/ : searching" [
        ("M-/ b",   addName "RH bugzilla"   $ promptSearch myXPConfig rhbz),
        ("M-/ g",   addName "Google"        $ promptSearch myXPConfig google),
        ("M-/ i",   addName "Google Images" $ promptSearch myXPConfig images),
        ("M-/ m",   addName "Google Maps"   $ promptSearch myXPConfig maps)]

    ^++^

    -- Start programs when switching to a topic.
    subKeys "switching workspaces" (
        zipWith (\k i -> ("M-" ++ k, noName $ switchTopic myTopicConfig (myTopics !! (i-1))))
                myTopics [1..])

-- Status bar
myBar = "xmobar"

-- Status bar output
-- NOTES:
--     * ppCurrent is the current workspace name
--     * ppHidden are other workspaces _with_ windows
--     * ppHiddenNoWindows are other workspaces without windows
--     * ppLayout is the current layout name
--     * ppUrgent are hidden workspaces with windows that have the URGENT hint
--     * ppTitle is the current window title
--     * ppWsSep is the separator character(s) between workspace names
--     * ppSep is the separator character(s) between objects
myPP = xmobarPP {
    ppHiddenNoWindows = xmobarColor "#444444" ""
 }

-- Startup applications
-- These programs are run each time XMonad is started or restarted.
myStartupHook = do
    spawn "xset -b"
    spawn "xsetroot -cursor_name left_ptr"

-- Main xmonad configuration
myConfig = addDescrKeys ((XM.mod4Mask, XM.xK_F1), xMessage) myKeys $ XM.def {
    layoutHook         = myLayoutHook,
    manageHook         = myManageHook,

    modMask            = XM.mod4Mask,
    borderWidth        = 1,
    terminal           = myTerminal,
    workspaces         = myTopics,
    startupHook        = myStartupHook
 }

-- Avoids turning xmobar in to a zombie process on recompile/restart.
restartXmonad = do
    nScreens <- countScreens
    -- Can't use "XM.recompile True" here because we need to set LD_LIBRARY_PATH.
    -- Otherwise xmonad won't build.  We should figure out why that is.
    spawn "killall -9 xmobar && env LD_LIBRARY_PATH=/usr/lib xmonad --recompile"
    XM.restart "xmonad" True

-- Build the command line for the primary xmobar process.  On a single monitor
-- system (a laptop, or a desktop with just one display), the config file to
-- use is xmobarrc-single and contains pretty much everything.  On a multi
-- monitor system, it's xmobarrc-right and contains just the stuff that we
-- need xmonad to pipe into it.
primaryXMobar screens | screens < 2 = myBar ++ " ~/.xmobar/xmobarrc-single"
                      | otherwise   = myBar ++ " ~/.xmobar/xmobarrc-right"

-- Run the secondary xmobar process.  On a single monitor system, there is
-- no secondary process.  On a multi monitor system, use xmobarrc-left as
-- the config file and start the process up.  xmonad does not communicate
-- with this process, so it just runs its own plugins.  We use this process
-- mainly for system monitoring.
spawnSecondaryXMobar screens | screens < 2 = return ()
                             | otherwise   = XM.io $ spawn $ myBar ++ " -x 1 ~/.xmobar/xmobarrc-left"

-- Run xmonad with xmobar status bar, support for custom urgency hook,
-- and the above configuration.
main = do
    nScreens <- countScreens
    spawnSecondaryXMobar nScreens

    checkTopicConfig myTopics myTopicConfig
    xmonad =<< statusBar (primaryXMobar nScreens) myPP toggleStrutsKey (withUrgencyHook NoUrgencyHook myConfig)
 where
    toggleStrutsKey XConfig {XM.modMask = modMask} = (modMask, XM.xK_b)
