#!/bin/sh
#
# Usage: ./find-badopts.sh [config file]
#
# Find and report any enabled kernel configuration options that are
# marked as EXPERIMENTAL, NEW, or DEPRECATED in the kernel source
# tree.  This script does not actually make the config file changes,
# it just reports what it found in the config file you provide it.
#
# This script requires an available $WRKSRC tree.
#

PATH=/usr/bin
WRKSRC="$(make show var=WRKSRC)"
CFGFILE="$(realpath "${1}" 2>/dev/null)"

if [ -z "${CFGFILE}" ]; then
    echo "Usage: $(basename "${0}") [kernel config file]"
    exit 1
fi

if [ ! -d "${WRKSRC}" ]; then
    echo "*** Missing ${WRKSRC}" >&2
    echo "*** Please run at least 'make patch' before using this script." >&2
    exit 1
fi

if [ ! -f "${CFGFILE}" ]; then
    echo "*** ${1} not found" >&2
    exit 1
fi

if grep "Kernel Configuration" x86_64.config 2>/dev/null | grep -q "^# Linux" >/dev/null 2>&1 ; then
    echo "*** ${1} is not a kernel configuration file" >&2
    exit 1
fi

echo "Looking for EXPERIMENTAL/NEW/DEPRECATED kernel config options..."

# Find all of the experimental kernel configuration options
find "${WRKSRC}" -type f -name "Kconfig*" -print0 | xargs grep -n -E "\((EXPERIMENTAL|NEW|DEPRECATED)\)" | while read -r line ; do
    filename="$(echo "${line}" | cut -d ':' -f 1)"
    linenum="$(echo "${line}" | cut -d ':' -f 2)"
    desc="$(echo "${line}" | cut -d ':' -f 3)"

    # filter out arch-specifics
    if echo "${line}" | grep -q "/arch/" >/dev/null 2>&1 ; then
        mach="$(uname -m)"

        if [ "${mach}" = "x86_64" ]; then
            # x86_64 platforms use /arch/x86 and /arch/x86_64
            archname="x86"
        elif [ "${mach}" = "ppc64" ]; then
            # ppc64 uses /arch/powerpc
            archname="powerpc"
        elif [ "${mach}" = "aarch64" ]; then
            # aarch64 uses /arch/arm and /arch/arm64
            archname="arm"
        elif [ "${mach}" = "s390x" ]; then
            # s390x uses /arch/s390
            archname="s390"
        elif [ "${mach}" = "sparc64" ]; then
            # sparc64 uses /arch/sparc
            archname="sparc"
        elif [ "${mach}" = "ia64" ] || [ "${mach}" = "alpha" ] || [ "${mach}" = "m68k" ] || [ "${mach}" = "parisc" ]; then
            # these uname machine names match the kernel arch subdir
            archname="${mach}"
        else
            # unhandled architecture
            archname=
        fi

        if echo "${line}" | grep -q "/arch/${archname}" >/dev/null 2>&1 ; then
            continue
        fi
    fi

    # get the option name which appears (or should) above the description line
    opt="$(head -n "${linenum}" "${filename}" | tail -n 2 | head -n 1 | cut -d ' ' -f 2)"

    if [ -z "${opt}" ]; then
        echo "*** Unable to find the option for ${desc}"
        echo "*** It should be in ${filename}"
    else
        grep -E "^${opt}" "${CFGFILE}"
    fi
done

exit 0

