/*
 * Simple program to extract Microsoft CAB content.
 * (Most of this code comes from Peachtree Linux dist-tools.)
 *
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <errno.h>
#include <err.h>
#include <archive.h>
#include <archive_entry.h>

/* Local prototypes */
static int copy_data(struct archive *, struct archive *);
static int extract_entry(struct archive *, struct archive *, struct archive_entry *);
static int unpack_archive(const char *, const char *, const int);

/*
 * From:
 * https://github.com/libarchive/libarchive/wiki/Examples
 * Reads data from the input archive stream and writes it to the output
 * archive stream.
 */
static int copy_data(struct archive *ar, struct archive *aw)
{
    int r;
    const void *buf = NULL;
    size_t s;
    off_t o;

    while ((r = archive_read_data_block(ar, &buf, &s, &o)) != ARCHIVE_EOF) {
        if (r != ARCHIVE_OK) {
            return r;
        }

        r = archive_write_data_block(aw, buf, s, o);

        if (r != ARCHIVE_OK) {
            warnx("%s", archive_error_string(aw));
            return r;
        }
    }

    return ARCHIVE_OK;
}

/* Extract a single entry from an archive and write it to disk. */
static int extract_entry(struct archive *input, struct archive *output, struct archive_entry *entry)
{
    int r = 0;
    int ret = 0;

    printf("%s\n", archive_entry_pathname(entry));

    r = archive_write_header(output, entry);

    if (r != ARCHIVE_OK) {
        warnx("%s", archive_error_string(output));
        ret = -1;
    } else if (archive_entry_size(entry) > 0) {
        if (copy_data(input, output) != ARCHIVE_OK) {
            ret = -1;
        }

        if (r != ARCHIVE_OK) {
            warnx("%s", archive_error_string(output));
        } else if (r < ARCHIVE_WARN) {
            ret = -1;
        }
    }

    r = archive_write_finish_entry(output);

    if (r != ARCHIVE_OK) {
        warnx("%s", archive_error_string(output));
    } else if (r < ARCHIVE_WARN) {
        ret = -1;
    }

    return ret;
}

/*
 * Unpack an archive file to a destination directory.  The directory must
 * exist before calling this function.
 */
static int unpack_archive(const char *archive, const char *dest, const int force)
{
    int flags, r, ret = 0;
    char *rfilename = NULL;
    char *cwd = NULL;
    struct archive *input = NULL;
    struct archive *output = NULL;
    struct archive_entry *entry = NULL;

    assert(archive != NULL);
    assert(dest != NULL);

    /* attributes to restore */
    flags = ARCHIVE_EXTRACT_TIME;
    flags |= ARCHIVE_EXTRACT_PERM;
    flags |= ARCHIVE_EXTRACT_ACL;
    flags |= ARCHIVE_EXTRACT_FFLAGS;

    if (force) {
        /* user passed a -f flag, so try to force past errors unpacking */
        flags |= ARCHIVE_EXTRACT_UNLINK;
    }

    /* full location to the archive file */
    if ((rfilename = realpath(archive, rfilename)) == NULL) {
        warnx("*** unable to find real path to %s", archive);
        return -1;
    }

    /* archive file reader */
    input = archive_read_new();
#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_read_support_compression_all(input);
#else
    archive_read_support_filter_all(input);
#endif
    archive_read_support_format_all(input);
    r = archive_read_open_filename(input, archive, 10240);

    if (r != ARCHIVE_OK) {
        warnx("*** error opening %s", archive);
        return -1;
    }

    /* change to dest */
    cwd = getcwd(cwd, 0);

    if (chdir(dest) != 0) {
        free(cwd);
        warn("chdir");
        return -1;
    }

    /* handler to write archive file members to disk */
    output = archive_write_disk_new();
    archive_write_disk_set_options(output, flags);
    archive_write_disk_set_standard_lookup(output);

    /* extract each archive file member */
    while ((r = archive_read_next_header(input, &entry)) != ARCHIVE_EOF) {
        if (r != ARCHIVE_OK) {
            warnx("%s", archive_error_string(input));
        } else if (r < ARCHIVE_WARN) {
            ret = -1;
        }

        if (extract_entry(input, output, entry)) {
            ret = -1;
        }
    }

    archive_read_free(input);
#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_write_finish(output);
#else
    archive_write_free(output);
#endif

    /* change back to original directory */
    if (chdir(cwd) != 0) {
        free(cwd);
        warn("chdir");
        return -1;
    }

    free(cwd);
    free(rfilename);
    return ret;
}

static void usage(const char *progname)
{
    printf("Usage: %s [options] archive\n", progname);
    printf("Options:\n");
    printf("    -f            Force unpack/overwrite existing files.\n");
    printf("    -d PATH       Destination path to extract to.\n");
    printf("    -?            This screen.\n");

    return;
}

int main(int argc, char **argv)
{
    int i = 0;
    int force = 0;
    char c;
    char *walker = NULL;
    char *dest = NULL;
    char *resolved_dest = NULL;
    char *prog = NULL;
    char *progname = NULL;
    struct stat sb;

    prog = strdup(argv[0]);
    progname = basename(prog);

    /* we need options */
    if (argc == 1) {
        usage(progname);
        return EXIT_FAILURE;
    }

    /* parse those options */
    while ((i = getopt(argc, argv, "fd:")) != -1) {
        switch (i) {
            case 'f':
                force = -1;
                break;
            case 'd':
                dest = strdup(optarg);
                break;
            default:
                usage(progname);
                return EXIT_FAILURE;
        }
    }

    /* verify the directory argument is usable */
    if (dest == NULL) {
        errx(EXIT_FAILURE, "ERROR: Missing -d argument.");
    }

    resolved_dest = realpath(dest, NULL);
    memset(&sb, 0, sizeof(sb));

    if (stat(resolved_dest, &sb) == -1) {
        if (errno == ENOENT) {
            /* try to create the path in 'mkdir -p' style */
            walker = resolved_dest;

            while (*walker != '\0') {
                walker++;

                if (*walker == '/') {
                    c = *walker;
                    *walker = '\0';

                    if (mkdir(resolved_dest, S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH) == -1) {
                        err(EXIT_FAILURE, "mkdir");
                    }

                    printf("mkdir |%s|\n", resolved_dest);
                    *walker = c;
                }
            }
        } else {
            err(EXIT_FAILURE, "stat");
        }
    } else if (!S_ISDIR(sb.st_mode)) {
        errx(EXIT_FAILURE, "ERROR: %s exists and is not a directory.", dest);
    }

    /* and extract all the archives to the destination */
    i = optind;

    while (i < argc) {
        if (unpack_archive(argv[i], resolved_dest, force)) {
            warnx("ERROR: unable to unpack %s", argv[i]);
        }

        i++;
    }

    free(prog);
    free(dest);
    free(resolved_dest);

    return EXIT_SUCCESS;
}
