#!/bin/sh
#
# General purpose ACPI event handler for Peachtree Linux.  This script
# tries to handle as many of the events as possible on the system.
#

# Current things handled by this script:
# 1) If you plug in the AC power, the CPU frequency scaling governor is
#    changed to the 'performance' one.  This only works on Intel CPUs.
# 2) If you unplug the AC power and go to battery, the CPU frequency
#    scaling governor is changed to the 'powersave' one.  This only works
#    on Intel CPUs.
# 3) If you press the power button, system shutdown is invoked to change
#    to runlevel 0 and power off the system if possible.
#
# Supported systems:
# Panasonic ToughBook CF-19
# Asus Z10PA dual Intel Xeon
# Lenovo ThinkPad X260

SYSCPU=/sys/devices/system/cpu
SYSRFKILL=/sys/class/rfkill

if [ -r /sys/class/backlight/intel_backlight ]; then
    SYSBACKLIGHT="/sys/class/backlight/intel_backlight"
elif [ -r /sys/class/backlight/acpi_video0 ]; then
    SYSBACKLIGHT="/sys/class/backlight/acpi_video0"
else
    SYSBACKLIGHT=
fi
if [ -f $SYSBACKLIGHT/max_brightness ]; then
    BACKLIGHTSTEP=$(($(cat $SYSBACKLIGHT/max_brightness) / 20))
else
    BACKLIGHTSTEP=1
fi

if [ $# != 1 ]; then
    exit 1
fi
set $*

ACPI_GROUP="$1"
ACPI_ACTION="$2"

case "$ACPI_GROUP" in
    processor)
        case "$ACPI_ACTION" in
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    battery)
        case "$ACPI_ACTION" in
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    ac_adapter)
        case "$ACPI_ACTION" in
            AC*)
                case "$4" in
                    00000000)
                        # AC power is disconnected, system continues on battery.
                        # Put CPUs in power saving mode.
                        if [ -d $SYSCPU ]; then
                            for cpudir in $SYSCPU/cpu* ; do
                                [ -d $cpudir ] || continue
                                [ -f $cpudir/cpufreq/scaling_governor ] || continue
                                echo "powersave" > $cpudir/cpufreq/scaling_governor
                            done
                        fi
                        ;;
                    00000001)
                        # AC power is connected, battery goes to charging mode.
                        # Put CPUs in performance mode since power is infinite.
                        if [ -d $SYSCPU ]; then
                            for cpudir in $SYSCPU/cpu* ; do
                                [ -d $cpudir ] || continue
                                [ -f $cpudir/cpufreq/scaling_governor ] || continue
                                echo "performance" > $cpudir/cpufreq/scaling_governor
                            done
                        fi
                        ;;
                esac
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    button/lid)
        case "$ACPI_ACTION" in
            LID)
                case "$3" in
                    close)
                        # Suspend system when the lid is closed.
                        /usr/sbin/pm-suspend &
                        ;;
                    open)
                        # Run /etc/acpi/resume.sh if it exists
                        [ -x /etc/acpi/resume.sh ] && /etc/acpi/resume.sh
                        ;;
                esac
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    button/mute)
        case "$ACPI_ACTION" in
            MUTE)
                # Toggle mute on and off.
                # NOTE: amixer has a mode called 'toggle', but it behaves
                # inconsistently across devices.  The following block is
                # more consistent across systems.
                /usr/bin/amixer -c 0 get Master | grep -q '[off]' >/dev/null 2>&1
                if [ $? -eq 0 ]; then
                    /usr/bin/amixer set Master unmute
                else
                    /usr/bin/amixer set Master mute
                fi
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    button/volumedown)
        case "$ACPI_ACTION" in
            VOLDN)
                # Decrease audio volume.  When the volume hits 0%, subsequent attempts
                # to decrease the volume are a no-op.
                /usr/bin/amixer -q sset Master 3%-
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    button/volumeup)
        case "$ACPI_ACTION" in
            VOLUP)
                # Increase audio volume by 3%.  When the volume hits 100%, subsequent
                # attempts to increase the volume are a no-op.
                /usr/bin/amixer -q sset Master 3%+
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    button/f20)
        case "$ACPI_ACTION" in
            F20)
                # Toggle mic input on and off (on ThinkPad X260)
                status="$(amixer sget Capture | grep "\[on\]")"
                ret=$?
                if [ $ret -eq 0 -a ! -z "$status" ]; then
                    amixer -q sset Cature nocap
                else
                    amixer -q sset Capture cap
                fi
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    button/power)
        case "$ACPI_ACTION" in
            LNXPWRBN:*)
                # User pressed the power button on the system, go to
                # runlevel 0 and attempt to power off the system.
                if [ -r /var/run/slim.auth ]; then
                    # there's an X session running, ask that user
                    /usr/bin/env XAUTHORITY=/var/run/slim.auth DISPLAY=:0.0 /usr/bin/quitxmonad
                else
                    # no detected X session, so schedule a shutdown in 5 mins
                    /usr/sbin/shutdown -P -h +5 "Power button pressed, shutting down."
                fi
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    button/sleep)
        case "$ACPI_ACTION" in
            SBTN)
                # Sleep is suspend to RAM.
                /usr/sbin/pm-suspend &
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    button/suspend)
        case "$ACPI_ACTION" in
            SUSP)
                # Hibernate is suspend to disk (which may or may not work).
                /usr/sbin/pm-hibernate &
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    video/brightnessdown)
        case "$ACPI_ACTION" in
            BRTDN)
                # decrease LCD brightness
                new=$(($(cat $SYSBACKLIGHT/actual_brightness) - $BACKLIGHTSTEP))
                [ $new -lt 0 ] && new=0
                echo $new > $SYSBACKLIGHT/brightness
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    video/brightnessup)
        case "$ACPI_ACTION" in
            BRTUP)
                # increase LCD brightness
                new=$(($(cat $SYSBACKLIGHT/actual_brightness) + $BACKLIGHTSTEP))
                max=$(cat $SYSBACKLIGHT/max_brightness)
                [ $new -gt $max ] && new=$max
                echo $new > $SYSBACKLIGHT/brightness
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    video/switchmode)
        case "$ACPI_ACTION" in
            VMOD)
                # Toggle external displays on and off as a mirror of the main
                mirror-all-displays
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    button/wlan)
        case "$ACPI_ACTION" in
            WLAN)
                # Toggle wireless network on and off (requires kernel rfkill)
                if [ -d $SYSRFKILL ]; then
                    for rfkill in $SYSRFKILL/* ; do
                        current=$(cat $rfkill/state)
                        [ $current -eq 1 ] && new=0 || new=1
                        echo $new > $rfkill/soft
                    done
                fi
                ;;
            *)
                logger "$ACPI_GROUP action $ACPI_ACTION is not defined"
                ;;
        esac
        ;;
    jack/lineout)
        break
        ;;
    jack/videoout)
        break
        ;;
    *)
        logger "ACPI group $ACPI_GROUP is not defined"
        ;;
esac
