/* dist-tools database definitions header file */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright David Shea <david@reallylongword.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#ifndef _DIST_TOOLS_DB_H
#define _DIST_TOOLS_DB_H

#include <stdio.h>
#include <sqlite3.h>

/* SQL schemas */
#define BOM_INFO_SCHEMA                                          \
    "CREATE TABLE IF NOT EXISTS info                             \
        (name TEXT NOT NULL ON CONFLICT FAIL,                    \
         ver TEXT NOT NULL ON CONFLICT FAIL,                     \
         rel INTEGER NOT NULL ON CONFLICT FAIL CHECK (rel >= 0), \
         arch TEXT NOT NULL ON CONFLICT FAIL,                    \
         repo TEXT,                                              \
         size BIG INTEGER NOT NULL ON CONFLICT FAIL,             \
         debug BOOLEAN,                                          \
         desc TEXT,                                              \
         maintainer TEXT,                                        \
         license TEXT,                                           \
         depspath TEXT)"

#define BOM_DEPS_SCHEMA \
    "CREATE TABLE IF NOT EXISTS deps (name TEXT NOT NULL ON CONFLICT FAIL)"

#define BOM_PROVS_SCHEMA \
    "CREATE TABLE IF NOT EXISTS provs (name TEXT NOT NULL ON CONFLICT FAIL)"

#define BOM_CONTENTS_SCHEMA \
    "CREATE TABLE IF NOT EXISTS contents                           \
        (path TEXT NOT NULL ON CONFLICT FAIL,                      \
         ownername TEXT NOT NULL ON CONFLICT FAIL,                 \
         uid INTEGER NOT NULL ON CONFLICT FAIL CHECK (uid >= 0),   \
         groupname TEXT NOT NULL ON CONFLICT FAIL,                 \
         gid INTEGER NOT NULL ON CONFLICT FAIL CHECK (gid >= 0),   \
         mode BLOB NOT NULL ON CONFLICT FAIL,                      \
         size BIG INTEGER,                                         \
         mtime BIG INTEGER,                                        \
         filetype INTEGER CHECK (filetype >= 0 AND filetype <= 9), \
         linkdest TEXT,                                            \
         digest TEXT)"

#define DB_PACKAGES_SCHEMA                                       \
    "CREATE TABLE IF NOT EXISTS packages                         \
        (id INTEGER PRIMARY KEY,                                 \
         name TEXT NOT NULL ON CONFLICT FAIL,                    \
         ver TEXT NOT NULL ON CONFLICT FAIL,                     \
         rel INTEGER NOT NULL ON CONFLICT FAIL CHECK (rel >= 0), \
         arch TEXT NOT NULL ON CONFLICT FAIL,                    \
         repo TEXT,                                              \
         size BIG INTEGER NOT NULL ON CONFLICT FAIL,             \
         debug BOOLEAN,                                          \
         desc TEXT,                                              \
         maintainer TEXT,                                        \
         license TEXT,                                           \
         depspath TEXT,                                          \
         digest TEXT NOT NULL)"

#define DB_CONTENTS_SCHEMA                                         \
    "CREATE TABLE IF NOT EXISTS contents                           \
        (packageid INTEGER NOT NULL,                               \
         path TEXT NOT NULL ON CONFLICT FAIL,                      \
         ownername TEXT NOT NULL ON CONFLICT FAIL,                 \
         uid INTEGER NOT NULL ON CONFLICT FAIL CHECK (uid >= 0),   \
         groupname TEXT NOT NULL ON CONFLICT FAIL,                 \
         gid INTEGER NOT NULL ON CONFLICT FAIL CHECK (gid >= 0),   \
         mode BLOB NOT NULL ON CONFLICT FAIL,                      \
         size BIG INTEGER,                                         \
         mtime BIG INTEGER,                                        \
         filetype INTEGER CHECK (filetype >= 0 AND filetype <= 9), \
         linkdest TEXT,                                            \
         digest TEXT,                                              \
         FOREIGN KEY(packageid) REFERENCES packages(id))"

/* SQL statements */
#define BOM_INFO_INSERT     "INSERT INTO info VALUES (?,?,?,?,?,?,?,?,?,?,?)"
#define BOM_DEPS_INSERT     "INSERT INTO deps VALUES (?)"
#define BOM_PROVS_INSERT    "INSERT INTO provs VALUES (?)"
#define BOM_CONTENTS_INSERT "INSERT INTO contents VALUES (?,?,?,?,?,?,?,?,?,?,?)"
#define BOM_CONTENTS_ATTACH "ATTACH DATABASE '%s' AS bom"
#define BOM_CONTENTS_DETACH "DETACH DATABASE bom"
#define BOM_CONTENTS_SELECT "SELECT * FROM bom.contents"

#define DB_CONTENTS_INSERT  "INSERT INTO contents SELECT '%d',* FROM bom.contents"
#define DB_CONTENTS_GENDIST "INSERT INTO contents VALUES (?,?,?,?,?,?,?,?,?,?,?,?)"
#define DB_PACKAGES_INSERT  "INSERT INTO packages VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?)"
#define DB_PACKAGES_UPDATE  "UPDATE packages SET name=?, ver=?, rel=?, arch=?, repo=?, size=?, debug=?, desc=?, maintainer=?, license=?, depspath=?, digest=? WHERE id='%d'"
#define DB_CONTENTS_DELETE  "DELETE FROM contents WHERE packageid='%d'"

/* used during 'distadd -u' operations */
/* selects entries that have been removed in the newer version of a package */
#define DB_CONTENTS_GONE    "SELECT c.packageid,c.filetype,c.path,c.digest,c.linkdest FROM contents AS c WHERE c.packageid=? AND (c.path NOT IN (SELECT path FROM bom.contents))"

#define INFO_SELECT         "SELECT * FROM info"
#define CONTENTS_SELECT     "SELECT * FROM contents"
#define DEPS_SELECT         "SELECT * FROM deps"
#define PROVS_SELECT        "SELECT * FROM provs"

/* Used by 'distinfo -q' operations */
#define DB_INFO_ETC_QUERY   "SELECT name,ver,rel,arch,digest FROM packages WHERE debug='0'"
#define DB_INFO_QUERY       "SELECT packages.name,packages.ver,packages.rel,packages.arch,packages.debug FROM packages,contents WHERE (packages.id=contents.packageid AND contents.path=?) ORDER BY packages.name"

/* For 'distinfo -f' on installed packages */
#define DB_INFO_FILELIST    "SELECT path FROM contents WHERE packageid=?"

/* For 'distinfo -f' on package files */
#define BOM_INFO_FILELIST   "SELECT path FROM contents"

/* For 'distinfo --provs' on package files */
#define BOM_INFO_PROVS      "SELECT * FROM provs"

/* For 'distinfo --deps' on package files */
#define BOM_INFO_DEPS       "SELECT * FROM deps"

/* For 'etcmerge' getting a list of all packages */
#define ETCMERGE_ALLPKGS    "SELECT name,ver,rel,arch,debug,digest FROM packages ORDER BY name"

/* For 'distchk' and 'distrm' getting package contents */
#define PACKAGE_CONTENTS    "SELECT * FROM contents WHERE packageid=%d"

/* For scanpkg */
#define SCANPKG_SELECT      "SELECT * FROM packages WHERE id IN ("

/* For depsorting */
#define SYS_DEP_CHECK       "SELECT packageid,path FROM contents WHERE path LIKE '%%%s%%'"

#endif /* _DIST_TOOLS_DB_H */
