/* Main dist-tools header file. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright David Shea <david@reallylongword.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#ifndef _DIST_TOOLS_H
#define _DIST_TOOLS_H

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <sys/stat.h>
#include <errno.h>
#include <sqlite3.h>
#include "dist.h"
#include "plist.h"
#include "strlist.h"
#include "pkglist.h"
#include "deplist.h"
#include "finfolist.h"
#include "signature.h"

/* Version information is set in src/Makefile */

/*
 * DB_PATH is where installed package information is recorded.
 * This includes the DB_FILE and subdirectories containing package
 * scripts, configuration files, and other metadata.
 */
#define DB_PATH "/var/db/dist"

/*
 * DB_FILE is the name of the main installed package database.
 */
#define DB_FILE DB_PATH"/installed.db"

/*
 * Path to the etcmerge(8) always-install list.
 */
#define ETC_INSTALL "/etc/dist/install.list"

/*
 * Path to the etcmerge(8) always-skip list.
 */
#define ETC_SKIP "/etc/dist/skip.list"

/*
 * Programs that dist-tools relies on.
 */
#define LDCONFIG "/usr/sbin/ldconfig"
#define SDIFF    "/usr/bin/sdiff"
#define EDITOR   "/usr/bin/vim"
#define ETCMERGE "/usr/sbin/etcmerge"

/*
 * .dist archive and/or database members
 */
#define DIST_SIGNATURE   "signature"
#define DIST_BOM         "bom.db"
#define DIST_LEGAL       "legal"
#define DIST_RELOC       "reloc.payload"
#define DIST_ETC         "etc.payload"
#define DIST_PREINSTALL  "preinstall"
#define DIST_POSTINSTALL "postinstall"
#define DIST_PREREMOVE   "preremove"
#define DIST_POSTREMOVE  "postremove"

/* Subdirectory where payloads are temporarily constructed or extracted */
#define PAYLOAD_SUBDIR "reloc"
#define STAGING_SUBDIR "staging"

/* Prefix identifying configuration files and directories */
#define CONFDIR_PREFIX "etc"

/* Where shared libraries are installed */
#define LIBDIR "/usr/lib:/usr/X11/lib:/usr/local/lib"

/* Where system configuration files live */
#define ETCDIR "/etc"

/* X11 prefix */
#define X11_PREFIX "/usr/X11"

/* Wayland prefix */
#define WAYLAND_PREFIX "/usr/wayland"

/* Documentation directory */
#define DOCDIR "/usr/share/doc"

/* Licenses directory */
#define LICENSESDIR "/usr/share/licenses"

/* Interesting stat(2) mode bits. */
#define S_INTER (S_ISUID | S_ISGID | S_ISVTX | S_IRWXU | S_IRWXG | S_IRWXO)

/* Common directory creation mode used by dist-tools. */
#define _DIR_MODE S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH

/* Package types */
typedef enum _pkg_type_t {
    PKG_NULL = 0,
    PKG_DISTFILE = 1,
    PKG_DBENTRY = 2
} pkg_type_t;

/* Payload types */
enum { PAYLOAD_NULL, PAYLOAD_DIST, PAYLOAD_RELOC, PAYLOAD_ETC };

/* Package scripts */
enum { SCRIPT_NULL,
       SCRIPT_PREINSTALL,
       SCRIPT_POSTINSTALL,
       SCRIPT_PREREMOVE,
       SCRIPT_POSTREMOVE };

/* Use the SHA-256 digest algorithm */
#define DIGEST_ALGORITHM SHA512SUM

/* Prototypes */

/* checkopts.c */
int checkneg(int, char *);
int checknull(char *, char *);

/* checksums.c */
char *checksum(const char *, enum checksum);

/* copyfile.c */
int copyfile(const char *, const char *, bool, bool);

/* is_elf.c */
bool is_elf(const char *);

/* checksig.c */
bool checksig(char *);

/* targetarch.c */
int check_targetarch(char *);
char *getdefaultarch(void);

/* gendist.c */
int gen_contents(struct finfo *, enum checksum, const char *);
int write_one_entry(const int, struct finfo *, sqlite3_stmt *, sigver_t);
int write_contents(int, sigver_t, sqlite3 *, finfolist_t *);
int get_file_type(struct finfo *, int *);

/* genattrs.c */
int writedeps(struct dist *, sqlite3 *);
int writeprovs(struct dist *, sqlite3 *);
int makelists(const char *, strlist_t **, strlist_t **);
int genattrs(char *, sqlite3 *, plist_t *, sigver_t, enum checksum, bool, const char *);

/* getpkginfo.c */
int getpkginfo(struct dist *, const int, const char *, const char *, sqlite3_stmt *);

/* checkfile.c */
int checkfile(const struct finfo *, const char *, const bool);
int read_contents(sqlite3 *, sigver_t, finfolist_t **, finfolist_t **, finfolist_t **);

/* scanpkg.c */
pkglist_t *scanpkg(const char *, const char *);

/* rmtree.c */
int rmtree(const char *, const bool, const bool);

/* freepkg.c */
void freepkg(struct dist *);

/* initpkg.c */
struct dist *initpkg(void);

/* remove.c */
int checkunlink(struct finfo *, const char *, const uint64_t);

/* strmode.c */
void strmode(mode_t, char *);

/* ver.c */
int ver(char *);

/* plist.c */
plist_t *read_plist(char *, char *);
int plist_length(plist_t *);
plist_t *init_plist(void);
void free_plist(plist_t *);

/* inodes.c */
char *lookup_inode(ino_t);
int add_inodes(char *);
void free_inodes(void);

/* unpack.c */
int unpack_metadata2str(const char *, const char *, char *, bool);
int unpack_metadata2file(const char *, const char *, const char *, bool);
int unpack_archive(const char *, const char *, const bool);

/* tmpdir.c */
int set_tmpdir(char *);

/* runscript.c */
int runscript(const struct dist *, const struct dist *, const unsigned int, const bool, char *, char *);

/* parsing.c */
int split_name_id(const char *, char **, uint64_t *);

/* getpath.c */
char *get_path(char *, const char *);

/* sizelen.c */
unsigned int size_len(ssize_t);

/* query.c */
int insert_pkg_info(const struct dist *, sqlite3 *, const char *, const int, const bool);
int insert_contents(int, const char *, struct sqlite3 *);
int result_count(sqlite3 *, const char *);

/* rmentry.c */
int rmentry(const struct dist *, const char *);

/* initdb.c */
int initdb(char *, const char *, const char *);

/* finfo.c */
int read_finfo(int, sigver_t, sqlite3_stmt *, struct finfo *, char *);
void free_finfo(struct finfo *);

/* finfolist.c */
finfolist_t *init_finfolist(void);
void free_finfolist(finfolist_t *);
int length_finfolist(finfolist_t *);
finfolist_entry_t *find_finfolist_filename(finfolist_t *, const char *);
finfolist_t *remove_finfolist_filename(finfolist_t *, const char *);
finfolist_t *sorted_finfolist_insert(finfolist_t *, struct finfo *);

/* checkpkg.c */
int checkpkg(char *, struct dist *, char *);

/* strsplit.c */
char **strsplit(char *, char *);
int strvlen(char **);
void strvfree(char **);

/* strtrim.c */
char *strtrim(const char *);

/* joinpath.c */
char *joinpath(const char *, ...);

/* strappend.c */
char *strappend(char *, char *);
char *strappend_int(char *, int);

/* strlist.c */
strlist_t *init_strlist(void);
void free_strlist(strlist_t *);
bool is_str_in_list(strlist_t *, const char *);
int length_strlist(strlist_t *);
strlist_t *split_strlist(char *, char *);

/* mkdirp.c */
int mkdirp(char *, mode_t);

/* pkglist.c */
pkglist_t *init_pkglist(void);
void free_pkglist(pkglist_t *);
int length_pkglist(pkglist_t *);
pkglist_t *sort_pkglist(pkglist_t *);
struct dist *find_pkg_by_hash(pkglist_t *, const char *);
bool have_pkg(const pkglist_t *list, const struct dist *pkg);

/* unquote.c */
char *unquote(const char *);

/* buflen.c */
int buflen(void);

/* db.c */
sqlite3 *open_db(const char *);

/* Macros */
#define xasprintf(dest, ...)                                 \
    *(dest) = NULL;                                          \
    if (asprintf((dest), __VA_ARGS__) == -1) {               \
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, \
                strerror(errno));                            \
        abort();                                             \
    }

#define xvasprintf(dest, ...)                                \
    *(dest) = NULL;                                          \
    if (vasprintf((dest), __VA_ARGS__) == -1) {              \
        fprintf(stderr, "%s (%d): %s\n", __func__, __LINE__, \
                strerror(errno));                            \
        abort();                                             \
    }

#endif /* _DIST_TOOLS_H */
