/* Package types and definitions. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright David Shea <david@reallylongword.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#ifndef _DIST_H
#define _DIST_H

#include <stdbool.h>
#include "signature.h"
#include "strlist.h"

/*
 * Supported checksum types.
 */
enum checksum { NULLSUM, MD5SUM, SHA1SUM, SHA256SUM, SHA384SUM, SHA512SUM };

/*
 * Package structure.  A dist stores information that will be written to
 * a package or extracted from it, as well as paths to the different
 * elements that make up a dist.
 */
struct dist {
    sigver_t sigver;              /* package signature version */
    int id;                    /* package ID (unique to sqlite) */

    char *tree;                /* the root tree (for making a package) */
    char *meta;                /* the path for the package meta files */
    char *name;                /* the name of the program in the package */
    char *ver;                 /* the version of the program in the
                                * package */
    int rel;                   /* the release or build of the package */
    char *plist;               /* the path to the packing list file */
    char *arch;                /* arch for which program in package was
                                * built */
    char *maintainer;          /* the name of the package maintainer */
    char *desc;                /* a one line description of the package */
    char *repo;                /* package repository (e.g. base) */
    char *license;             /* common short form of the license (e.g.,
                                * GPL) */
    bool debug;                /* is this a debug package? */

    int64_t pkgsize;           /* the size of the package file */
    char *pkgdigest;           /* the digest of the actual package itself */

    int64_t size;              /* the installed size of the package */

    int compression;           /* the compression type we are using */
    enum checksum sumtype;     /* the type of checksums we are using */
    bool do_cleanup;           /* clean up temp files after operation? */

    char *filename;            /* the package filename string */

    strlist_t *deps;           /* ELF shared library dependencies */
    strlist_t *provs;          /* ELF shared libraries provided by this package */

    char *bom;                 /* the path to the bill of materials db */
    char *reloc;               /* the path to the reloc payload file */
    char *etc;                 /* the path to the etc payload file */
    char *distname;            /* the path to the final .dist file */
    char *depspath;            /* optional - colon delimited list of dirs
                                * to search for library dependencies */
    char *legal;               /* the path to the legal file */
};

/*
 * finfo structure for metadata
 */
enum filetype {
    TYPE_NONE,                  /* 0 */
    TYPE_REG,                   /* 1 */
    TYPE_DIR,                   /* 2 */
    TYPE_HARD,                  /* 3 */
    TYPE_SOFT,                  /* 4 */
    TYPE_CHAR,                  /* 5 */
    TYPE_BLOCK,                 /* 6 */
    TYPE_FIFO,                  /* 7 */
    TYPE_SOCK,                  /* 8 */
    TYPE_OTHER                  /* 9 */
};

/*
 * finfo contains properties for each file in a package payload, the
 * checksum as well as other attributes are here.
 */
struct finfo {
    int packageid;             /* the package ID owning this file */
                               /* this is -1 for non-installed packages */
    sigver_t sigver;              /* package signature version */
    enum filetype type;        /* type of file */
    char *filename;            /* name of the file minus the staging root */
    char *fullpath;            /* full path to file in the staging root */
    char *checksum;            /* e.g., a SHA-1 checksum */
    char *owner;               /* the owner name */
    char *group;               /* the group name */
    struct stat *attr;         /* stat data for attrs */
    char *linkname;            /* if not NULL, the linked file for
                                 * TYPE_{HARD,SOFT} */
    bool has_checksum;         /* whether or not checksum is used */
    int refcnt;                /* reference count for finfolist_t tracking */
};

#endif /* _DIST_H */
