#
# Package Makefile defintions and targets
# (include this file at the end of package Makefiles)
#
# Copyright David Cantrell <dcantrell@burdell.org>
# Copyright Elliott Potter <epotter.e.potter@nasa.gov>
# Copyright David Shea <david@reallylongword.org>
# Copyright Chris Lumens <chris@bangmoney.org>
# Copyright Michael Imamura <zoogie@lugatgt.org>
# SPDX-License-Identifier: Apache-2.0
#

# The top level source Makefile should set this to 'y' before including pkg.mk
TOP_LEVEL ?= n

# Find the top level directory.  Project directory trees using this build
# system should have a top level '.peachtree' file.  This file does not
# need to contain anything, it just needs to exist.
TOPSRCERR = "*** Unable to find top level directory."

ifeq ($(TOP_LEVEL),y)
TOPSRC := $(shell pwd)
else
TOPSRC := $(shell while [ ! -f .peachtree ]; do if [ "$$(pwd)" = "/" ]; then /bin/echo $(TOPSRCERR) >&2; exit 1; fi; cd ..; done; /bin/echo "$$(pwd)")
endif

# By default the build system will try to resolve buildreqs in the
# project directory.  You can add other project directories to the
# list by appending paths to the BUILDREQS_PATHS list.  It's a space
# delimited list of directory paths.
BUILDREQS_PATHS = $(TOPSRC)

# Current working directory
CWD := $(shell pwd)

# Site-specific overrides to build system settings.
-include /etc/default/make.conf
-include /etc/make.conf
-include $(TOPSRC)/mk/make.conf
-include $(HOME)/.peachtree/make.conf
-include $(CWD)/make.conf

# Commands
# We only need to list commands here if we want to explicitly call a
# particular one.  Keep this list small.
ECHO ?= /bin/echo
SUDO ?= sudo -S

# Distribution name
DISTRIBUTION_NAME ?= Peachtree Linux

# Distribution version (only for non-kernel builds)
DISTRIBUTION_VERSION ?= 3.7.0

# Target architecture (can be set in a package Makefile to 'noarch')
ifndef TARGET_MACH
TARGET_MACH := $(shell uname -m | sed -e "s/i[[:digit:]]86/i386/g")
endif

# For noarch packages, this is "noarch-peachtree-linux", but we don't care.
TARGET_NAME = $(TARGET_MACH)-peachtree-linux

# Kernel architecture name (used in the kernel source directory)
ifeq ($(TARGET_MACH),x86_64)
KERN_ARCH = x86
endif

ifeq ($(TARGET_MACH),ppc64)
KERN_ARCH = powerpc
endif

# Number of parallel make processes to run (equals # of CPUs)
# If the nproc command is not found, set a default value.
ifeq ($(PARALLEL_JOBS),)
PARALLEL_JOBS := $(shell nproc 2>/dev/null)
endif

ifeq ($(PARALLEL_JOBS),)
PARALLEL_JOBS := 3
endif

# Common make options
MAKEFLAGS += --no-print-directory

# Default installation paths
PREFIX             ?= /usr
BINDIR             ?= $(PREFIX)/bin
SBINDIR            ?= $(PREFIX)/sbin
LIBEXECDIR         ?= $(PREFIX)/libexec
LIBDIR             ?= $(PREFIX)/lib
INCLUDEDIR         ?= $(PREFIX)/include
DATADIR            ?= $(PREFIX)/share
MANDIR             ?= $(DATADIR)/man
INFODIR            ?= $(DATADIR)/info

X11_PREFIX         ?= /usr/X11
X11_BINDIR         ?= $(X11_PREFIX)/bin
X11_SBINDIR        ?= $(X11_PREFIX)/sbin
X11_LIBEXECDIR     ?= $(X11_PREFIX)/libexec
X11_LIBDIR         ?= $(X11_PREFIX)/lib
X11_INCLUDEDIR     ?= $(X11_PREFIX)/include
X11_DATADIR        ?= $(X11_PREFIX)/share
X11_MANDIR         ?= $(X11_PREFIX)/man
X11_INFODIR        ?= $(X11_PREFIX)/info

WAYLAND_PREFIX     ?= /usr/wayland
WAYLAND_BINDIR     ?= $(WAYLAND_PREFIX)/bin
WAYLAND_SBINDIR    ?= $(WAYLAND_PREFIX)/sbin
WAYLAND_LIBEXECDIR ?= $(WAYLAND_PREFIX)/libexec
WAYLAND_LIBDIR     ?= $(WAYLAND_PREFIX)/lib
WAYLAND_INCLUDEDIR ?= $(WAYLAND_PREFIX)/include
WAYLAND_DATADIR    ?= $(WAYLAND_PREFIX)/share
WAYLAND_MANDIR     ?= $(WAYLAND_PREFIX)/man
WAYLAND_INFODIR    ?= $(WAYLAND_PREFIX)/info

DOCDIR             ?= /usr/share/doc/$(PKG)
LICENSEDIR         ?= /usr/share/licenses/$(PKG)
SYSCONFDIR         ?= /etc
INITDIR            ?= $(SYSCONFDIR)/init.d
LOCALSTATEDIR      ?= /var

# Where to find helper scripts for the build system.
# Some of these are stored in the source tree while some are
# part of dist-tools because they are more generally useful
# for packaging.
LOCAL_PKGBINDIR   ?= $(TOPSRC)/mk/bin

# Favor the in-tree copies of the packaging tools and fall back on the
# system installed ones.
SYSTEM_PKGBINDIR  := $(shell if [ -d "$(TOPSRC)/base/system/dist-tools/libexec" ]; then /bin/echo "$(TOPSRC)/base/system/dist-tools/libexec"; else /bin/echo "/usr/libexec/dist-tools"; fi)
SYSTEM_PKGDATADIR := $(shell if [ -d "$(TOPSRC)/base/system/dist-tools/share" ]; then /bin/echo "$(TOPSRC)/base/system/dist-tools/share"; else /bin/echo "/usr/share/dist-tools"; fi)

# Distribution-wide helper scripts
# These are used for building releases and distribution-wide things.
worldbuilder      ?= $(LOCAL_PKGBINDIR)/world-builder.sh
cleanchroot       ?= $(LOCAL_PKGBINDIR)/clean-chroot.sh
installchroot     ?= $(LOCAL_PKGBINDIR)/install-chroot.sh
trimchroot        ?= $(LOCAL_PKGBINDIR)/trim-chroot.sh
makerelease       ?= $(LOCAL_PKGBINDIR)/make-release.sh
showorphans       ?= $(LOCAL_PKGBINDIR)/show-orphans.sh
mergereleases     ?= $(LOCAL_PKGBINDIR)/merge-releases.sh
incver            ?= $(LOCAL_PKGBINDIR)/incver.sh
order             ?= $(LOCAL_PKGBINDIR)/order.py
lastcommit        ?= $(LOCAL_PKGBINDIR)/last-commit.sh
mergebundles      ?= $(LOCAL_PKGBINDIR)/merge-bundles.sh

# Helpers used in pkg.mk
parsebool         ?= $(SYSTEM_PKGBINDIR)/parsebool.sh
ishttpurl         ?= $(SYSTEM_PKGBINDIR)/is-http-url.sh
issourceforgeurl  ?= $(SYSTEM_PKGBINDIR)/is-sourceforge-url.sh
getpkgcollection  ?= $(SYSTEM_PKGBINDIR)/get-pkg-collection.sh

# Helper commands for each target
prunedistfiles    ?= $(SYSTEM_PKGBINDIR)/prune-distfiles.sh
checkupstream     ?= $(SYSTEM_PKGBINDIR)/check-upstream.sh
fetch             ?= $(SYSTEM_PKGBINDIR)/fetch.sh
checksum          ?= $(SYSTEM_PKGBINDIR)/checksum.sh
extract           ?= $(SYSTEM_PKGBINDIR)/extract.sh
patch             ?= $(SYSTEM_PKGBINDIR)/patch.sh
build             ?= $(SYSTEM_PKGBINDIR)/build.sh
test              ?= $(SYSTEM_PKGBINDIR)/test.sh
fakeinst          ?= $(SYSTEM_PKGBINDIR)/fakeinst.sh
updatepatches     ?= $(SYSTEM_PKGBINDIR)/update-patches.sh
updateplist       ?= $(SYSTEM_PKGBINDIR)/update-plist.sh
updatekeyring     ?= $(SYSTEM_PKGBINDIR)/update-keyring.sh
makesum           ?= $(SYSTEM_PKGBINDIR)/makesum.sh
cleansum          ?= $(SYSTEM_PKGBINDIR)/cleansum.sh
buildreqs         ?= $(SYSTEM_PKGBINDIR)/buildreqs.py
usereqs           ?= $(SYSTEM_PKGBINDIR)/usereqs.sh
package           ?= $(SYSTEM_PKGBINDIR)/package.sh
missingreqs       ?= $(SYSTEM_PKGBINDIR)/missingreqs.sh
bundle            ?= $(SYSTEM_PKGBINDIR)/bundle.sh
showprevpkg       ?= $(SYSTEM_PKGBINDIR)/showprevpkg.sh

# Build paths
RELEASE_BASE   ?= release
DISTFILES_BASE ?= distfiles
OBJ_BASE       ?= obj

ifndef RELEASE
RELEASE := $(shell readlink -f $(TOPSRC)/$(RELEASE_BASE))
endif

ifndef DISTFILES
DISTFILES := $(shell readlink -f $(TOPSRC)/$(DISTFILES_BASE))
endif

ifndef OBJ
OBJ := $(shell readlink -f $(TOPSRC)/$(OBJ_BASE))
endif

ifndef MINIROOT
MINIROOT := $(OBJ)/miniroot
endif

# Package Makefiles can set this to 'y' for inclusion in the install
# media miniroot.  By default this is 'n'.  Only packages in the base
# collection can be in the miniroot.
MINIROOT_PKG ?= n

# Package subdirectories are versioned and split by architecture.
# This may seem pointless, but it allows for a local $(RELEASE)
# tree to contain multiple releases and it allows noarch packages
# to be shared among all supported architectures.  The architecture
# split is handled in the helper scripts and inst, primarily
# because the collection name is added before the architecture
# split.
RELEASE_DIST = $(RELEASE)/dist/$(DISTRIBUTION_VERSION)

# Digest algorithms for archive files
# This can be md5, sha1, sha256, sha384, or sha512
DIGEST_ALGORITHM ?= sha512

# Global CFLAGS and make customizations
ifeq ($(findstring -Wall,$(CFLAGS)),)
CFLAGS += -Wall
endif

ifeq ($(findstring -g,$(CFLAGS)),)
CFLAGS += -g
endif

ifeq ($(findstring -O2,$(CFLAGS)),)
CFLAGS += -O2
endif

ifeq ($(findstring -pipe,$(CFLAGS)),)
CFLAGS += -pipe
endif

ifeq ($(findstring -mtune=,$(CFLAGS)),)
CFLAGS += -mtune=generic
endif

# Build with -fPIC by default.  Disable with POSITION_INDEPENDENT_CODE=n in the
# package Makefile.
ifdef POSITION_INDEPENDENT_CODE
POSITION_INDEPENDENT_CODE := $(shell $(parsebool) $(POSITION_INDEPENDENT_CODE))
else
POSITION_INDEPENDENT_CODE ?= y
endif

ifeq ($(POSITION_INDEPENDENT_CODE),y)
CFLAGS += -fPIC
endif

# Build with -fstack-protector-strong by default.  Disable with
# STACK_PROTECTOR_STRONG=n in the package Makefile.
ifdef STACK_PROTECTOR_STRONG
STACK_PROTECTOR_STRONG := $(shell $(parsebool) $(STACK_PROTECTOR_STRONG))
else
STACK_PROTECTOR_STRONG ?= y
endif

ifeq ($(STACK_PROTECTOR_STRONG),y)
CFLAGS += -fstack-protector-strong
endif

# Build with _FORTIFY_SOURCE=2 by default.  Disable with FORTIFY_SOURCE=n in
# the package Makefile.
ifdef FORTIFY_SOURCE
FORTIFY_SOURCE := $(shell $(parsebool) $(FORTIFY_SOURCE))
else
FORTIFY_SOURCE ?= y
endif

ifeq ($(FORTIFY_SOURCE),y)
CFLAGS += -D_FORTIFY_SOURCE=2
endif

# Build without BIND_NOW in the linker.  Enable BIND_NOW by setting
# LAZY_EVALUATIONS=n in the package Makefile.
ifdef LAZY_EVALUATIONS
LAZY_EVALUATIONS := $(shell $(parsebool) $(LAZY_EVALUATIONS))
else
LAZY_EVALUATIONS ?= n
endif

ifeq ($(LAZY_EVALUATIONS),n)
LDFLAGS += -z now
CFLAGS += -Wl,-z,now
endif

# The x86_64 systems we have span two generations of Intel instruction set
# architecture.  Intel uses a name for each ISA, so it's hard to tell what came
# first and so on.  Here is a rough list of recent ISAs from Intel:
#
#     Core                    core2
#     Nehalem                 nehalem
#     Sandy Bridge            sandybridge
#     Ivy Bridge              ivybridge
#     Haswell                 haswell
#     Broadwell               broadwell
#     Skylake                 skylake
#     Bonnell                 bonnell
#     Silvermont              silvermont
#     Knight's Landing        knl
#     Skylake Server          skylake-avx512
#
# With 'Core' being first and the bottom of the list being most recent.  Note
# that gcc has dedicated definitions for AMD processors and other manufacturers
# processors, but they are not listed here as we are 100% Intel for our x86_64
# systems.  All of our current systems consist of the following processors:
#
#     Intel Xeon E5-2630      Haswell
#     Intel Core i5-3320M     Ivy Bridge
#     Intel Core i7-4600U     Haswell
#     Intel Core i5-4570      Haswell
#     Intel Core i7-6600U     Skylake
#
# When setting default compiler flags, we want to restrict the ISA to the
# lowest common denominator so we don't get builds that will throw illegal
# instruction signals on the older systems.  This is most easily seen in
# programs that make heavy use of floating point operations.  The gmp package
# is a common culprit.
#
# WHERE DID THESE VALUES COME FROM?
# Reading /proc/cpuinfo and using Wikipedia to figure out the ISA.
#
# WON'T BUILDING FOR AN OLDER ISA RESULT IN SLOWER CODE ON A NEWER ISA?
# No.
#
# BUT WHAT ABOUT ALL THOSE FEATURES THAT WE WON'T BE USING ON THE NEW ISA?
# It doesn't matter.
#
# We set the -march flag to the ISA and then always set -mtune=generic.  It is
# important to note that gcc does not have an -march=generic option.  These
# values can be found here:
#
#  http://gcc.gnu.org/onlinedocs/gcc/x86-Options.html
#
ifeq ($(TARGET_MACH),x86_64)
CFLAGS += -mtune=generic
endif

# The only ppc64 systems we have are Apple Power Mac G5 towers
ifeq ($(TARGET_MACH),ppc64)
CFLAGS += -mtune=G5
endif

CXXFLAGS ?= $(CFLAGS)

# Default to Python 3.x
PYTHON ?= $(WRKSYSROOT)$(BINDIR)/python3

# Some environment variables that we may have set for our own personal use
# may cause problems with the build system.  Unset them here.
GREP_OPTIONS =

# Our mirrors, which we should try before the main sites.
_SUBDIR := $(shell $(ECHO) "$(PKG)" | cut -c1 | tr [A-Z] [a-z])
OUR_MIRRORS += https://www.burdell.org/peachtree/distfiles/$(_SUBDIR)/$(PKG)/

# Base URL for downloading packages.  Under this directory should be a
# subdirectory named 'dist'.
GIT_BRANCH := $(shell git rev-parse --abbrev-ref HEAD 2>/dev/null)
ifeq ($(GIT_BRANCH),master)
DIST_URL ?= https://www.burdell.org/peachtree/releases/devel/
else
DIST_URL = https://www.burdell.org/peachtree/releases/$(DISTRIBUTION_VERSION)/
endif

# Should world builds attempt to download a built version first?
DIST_DOWNLOAD ?= y

# World build defaults

# If set, the world build will be wrapped in "nice -n $(WORLD_PRIORITY)".
# See the man page for nice(1) for more information.
WORLD_PRIORITY ?= 19

# Command used to time the build (comment out to disable).
WORLD_TIME_CMD ?= /usr/bin/time

# Temporary directory for distadd(8) to use during builds.
DISTADD_TMPDIR ?= /var/tmp

# Keep each package $(WRKDIR) around after building?  (y/n)
# If you have limited space for $(OBJ), you should set this to 'n'.  The
# only things kept after a world build for each package are the package
# and the debug package (if applicable).
KEEP_WRKDIR ?= y

# Common download sites
FTPGNU      = https://ftp.gnu.org/gnu
GNOME       = https://download.gnome.org/sources
HACKAGE     = http://hackage.haskell.org/package
SOURCEFORGE = http://downloads.sourceforge.net/project
XORG        = https://www.x.org/releases/individual

# Each STAGE can have a pre or post leaf stage provided by the packager.
# If they exist, we execute them in pre-STAGE, STAGE, post-STAGE order.
LEAF_STAGES = pre post
STAGES      = update-patches update-plist fetch checksum extract \
              patch build test fakeinst package clean

# Create a list of all targets and declare as .PHONY so the make does
# not error if the packager has not provided them.
ALL_TARGETS  = $(foreach s, $(LEAF_STAGES), $(addprefix $(s)-, $(STAGES)))
ALL_TARGETS += $(STAGES)
.PHONY: $(ALL_TARGETS)

# Where to find these helper scripts
HELPERS          ?= $(SYSTEM_PKGBINDIR)/helpers.sh
HELPERSDB        ?= $(SYSTEM_PKGBINDIR)/helpers-db.sh

# Templates
MERGE_SH_IN      ?= $(SYSTEM_PKGDATADIR)/merge.sh.in

# Optional package metadata
# Pick up these values from a pkginfo file if it exists and if the
# package Makefile has not already set them.
ifeq ($(DESCRIPTION),)
DESCRIPTION := $(shell grep ^desc= pkginfo 2>/dev/null | head -n 1 | cut -c6- | sed -e 's|"||g')
endif

ifeq ($(MAINTAINER),)
MAINTAINER  := $(shell grep ^maintainer= pkginfo 2>/dev/null | head -n 1 | cut -c12- | sed -e 's|"||g')
endif

ifeq ($(LICENSE),)
LICENSE     := $(shell grep ^license= pkginfo 2>/dev/null | head -n 1 | cut -c9- | sed -e 's|"||g')
endif

ifeq ($(DEPSPATH),)
DEPSPATH    := $(shell grep ^depspath= pkginfo 2>/dev/null | head -n 1 | cut -c10- | sed -e 's|"||g')
endif

# plist file names
PLIST     = $(CWD)/plist
PLISTMACH = $(CWD)/plist.$(TARGET_MACH)

# patches subdirectory
PATCHES   = $(CWD)/patches

# Full package name
PACKAGE       = $(PKG)-$(VER)-$(REL).$(TARGET_MACH).dist
DEBUGPACKAGE  = $(PKG)-$(VER)-$(REL).debug.$(TARGET_MACH).dist

ifneq ($(SUBPACKAGES),)
PACKAGE      += $(foreach subpkg,$(SUBPACKAGES),$(PKG)-$(subpkg)-$(VER)-$(REL).$(TARGET_MACH).dist)
DEBUGPACKAGE += $(foreach subpkg,$(SUBPACKAGES),$(PKG)-$(subpkg)-$(VER)-$(REL).debug.$(TARGET_MACH).dist)
endif

# Debug path for gdb
DEBUGDIR = $(PREFIX)/debug

# The package may override this for complicated source trees.
ifeq ($(WRKDIST),)
WRKDIST     ?= $(PKG)-$(VER)
else
# Trim all trailing slashes from the path
_WRKDIST    := $(shell echo $(WRKDIST) | sed -e 's/\/*$$//')
WRKDIST     := $(_WRKDIST)
endif

# Package-specific build directories
WRKDIR       = $(OBJ)/$(PKG)-$(VER)-$(REL).$(TARGET_MACH)
WRKSRC      ?= $(WRKDIR)/$(WRKDIST)
WRKBLD       = $(WRKDIR)/build-$(TARGET_MACH)
WRKINST      = $(WRKDIR)/pkgroot-$(TARGET_MACH)
WRKDEBUG     = $(WRKDIR)/debug-$(TARGET_MACH)
WRKSYSROOT   = $(OBJ)/sysroot

# Package location specifics
PKGDISTFILES     := $(DISTFILES)/$(shell $(ECHO) $(PKG) | cut -c1 | tr [A-Z] [a-z])/$(PKG)
PKGCOLLECTION    := $(shell $(getpkgcollection))

RELPKGPATH        = $(PKGCOLLECTION)/$(TARGET_MACH)

FULLPKGPATH       = $(RELEASE_DIST)/$(RELPKGPATH)/$(PKG)-$(VER)-$(REL).$(TARGET_MACH).dist
FULLDEBUGPKGPATH  = $(RELEASE_DIST)/debug/$(RELPKGPATH)/$(PKG)-$(VER)-$(REL).debug.$(TARGET_MACH).dist

ifneq ($(SUBPACKAGES),)
FULLPKGPATH      += $(foreach subpkg,$(SUBPACKAGES),$(RELEASE_DIST)/$(RELPKGPATH)/$(PKG)-$(subpkg)-$(VER)-$(REL).$(TARGET_MACH).dist)
FULLDEBUGPKGPATH += $(foreach subpkg,$(SUBPACKAGES),$(RELEASE_DIST)/debug/$(RELPKGPATH)/$(PKG)-$(subpkg)-$(VER)-$(REL).debug.$(TARGET_MACH).dist)
endif

# Are we skipping this package for now?
SKIP_REASON ?=
ifdef SKIP
SKIP := $(shell $(parsebool) $(SKIP))
else
ifneq ($(TARGET_MACH),noarch)
ifneq ($(TARGET_MACH),$(shell uname -m))
SKIP ?= y
SKIP_REASON = ($(TARGET_MACH) only)
else
SKIP ?= n
endif
else
SKIP ?= n
endif
endif

# Default GITPKG and GITVER to match the package name and version.
GITPKG ?= $(PKG)
GITVER ?= $(VER)

########################################################################
# WORLD                                                                #
# Packages are part of 'world' which is the operating system and the   #
# idea that we build a package with 'make world' is to build it        #
# against the rest of the operating system regardless of what the host #
# has.  It's nomenclature from the BSD systems that is common.         #
########################################################################

# Add --sysroot= to CFLAGS and CXXFLAGS?
ifdef WORLD_CFLAGS_SYSROOT
WORLD_CFLAGS_SYSROOT := $(shell $(parsebool) $(WORLD_CFLAGS_SYSROOT))
else
WORLD_CFLAGS_SYSROOT ?= y
endif

ifdef WORLD_CXXFLAGS_SYSROOT
WORLD_CXXFLAGS_SYSROOT := $(shell $(parsebool) $(WORLD_CXXFLAGS_SYSROOT))
else
WORLD_CXXFLAGS_SYSROOT ?= y
endif

# Add --with-sysroot= to CONFIGURE_OPTS?
ifdef WORLD_WITH_SYSROOT
WORLD_WITH_SYSROOT := $(shell $(parsebool) $(WORLD_WITH_SYSROOT))
else
WORLD_WITH_SYSROOT ?= y
endif

# Add -I to CFLAGS?
ifdef WORLD_CFLAGS_INCLUDEDIR
WORLD_CFLAGS_INCLUDEDIR := $(shell $(parsebool) $(WORLD_CFLAGS_INCLUDEDIR))
else
WORLD_CFLAGS_INCLUDEDIR ?= y
endif

# Add -L to LDFLAGS?
ifdef WORLD_LDFLAGS_LIBDIR
WORLD_LDFLAGS_LIBDIR := $(shell $(parsebool) $(WORLD_LDFLAGS_LIBDIR))
else
WORLD_LDFLAGS_LIBDIR ?= y
endif

# Add -Wl,--sysroot= to LDFLAGS?
ifdef WORLD_LDFLAGS_WL_SYSROOT
WORLD_LDFLAGS_WL_SYSROOT := $(shell $(parsebool) $(WORLD_LDFLAGS_WL_SYSROOT))
else
WORLD_LDFLAGS_WL_SYSROOT ?= y
endif

# Make sure world builds know about sysroot bin and lib directories.
ifeq ($(CONFIGURE_STYLE),gnu)
WORLD_CONFIGURE_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),haskell)
WORLD_CONFIGURE_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),qmake)
WORLD_CONFIGURE_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),cmake)
WORLD_CONFIGURE_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),meson)
WORLD_CONFIGURE_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),perl)
WORLD_CONFIGURE_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),cargo)
WORLD_CONFIGURE_ENV_PATH ?= y
endif

ifdef WORLD_CONFIGURE_ENV_PATH
WORLD_CONFIGURE_ENV_PATH := $(shell $(parsebool) $(WORLD_CONFIGURE_ENV_PATH))
endif

ifeq ($(CONFIGURE_STYLE),gnu)
WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),haskell)
WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),qmake)
WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),meson)
WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),perl)
WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),cargo)
WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH ?= y
endif

ifdef WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH
WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH := $(shell $(parsebool) $(WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH))
endif

# Add path overrides to build, test, and fakeinst environments?
ifdef WORLD_BUILD_ENV_PATH
WORLD_BUILD_ENV_PATH := $(shell $(parsebool) $(WORLD_BUILD_ENV_PATH))
else
WORLD_BUILD_ENV_PATH ?= y
endif

ifdef WORLD_BUILD_ENV_LD_LIBRARY_PATH
WORLD_BUILD_ENV_LD_LIBRARY_PATH := $(shell $(parsebool) $(WORLD_BUILD_ENV_LD_LIBRARY_PATH))
else
WORLD_BUILD_ENV_LD_LIBRARY_PATH ?= y
endif

ifdef WORLD_BUILD_ENV_PKG_CONFIG
WORLD_BUILD_ENV_PKG_CONFIG := $(shell $(parsebool) $(WORLD_BUILD_ENV_PKG_CONFIG))
else
WORLD_BUILD_ENV_PKG_CONFIG ?= y
endif

ifdef WORLD_TEST_ENV_LD_LIBRARY_PATH
WORLD_TEST_ENV_LD_LIBRARY_PATH := $(shell $(parsebool) $(WORLD_TEST_ENV_LD_LIBRARY_PATH))
else
WORLD_TEST_ENV_LD_LIBRARY_PATH ?= y
endif

ifdef WORLD_TEST_ENV_PATH
WORLD_TEST_ENV_PATH := $(shell $(parsebool) $(WORLD_TEST_ENV_PATH))
else
WORLD_TEST_ENV_PATH ?= y
endif

ifdef WORLD_FAKEINST_ENV_PKG_CONFIG
WORLD_FAKEINST_ENV_PKG_CONFIG := $(shell $(parsebool) $(WORLD_FAKEINST_ENV_PKG_CONFIG))
else
WORLD_FAKEINST_ENV_PKG_CONFIG ?= y
endif

ifdef WORLD_FAKEINST_ENV_PATH
WORLD_FAKEINST_ENV_PATH := $(shell $(parsebool) $(WORLD_FAKEINST_ENV_PATH))
else
ifeq ($(CONFIGURE_STYLE),python)
WORLD_FAKEINST_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),pyproject)
WORLD_FAKEINST_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),haskell)
WORLD_FAKEINST_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),meson)
WORLD_FAKEINST_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),perl)
WORLD_FAKEINST_ENV_PATH ?= y
endif
ifeq ($(CONFIGURE_STYLE),cargo)
WORLD_FAKEINST_ENV_PATH ?= y
endif

WORLD_FAKEINST_ENV_PATH ?= n
endif

ifdef WORLD_FAKEINST_ENV_LD_LIBRARY_PATH
WORLD_FAKEINST_ENV_LD_LIBRARY_PATH := $(shell $(parsebool) $(WORLD_FAKEINST_ENV_LD_LIBRARY_PATH))
else
WORLD_FAKEINST_ENV_LD_LIBRARY_PATH ?= y
endif

# Always add the pkg-config override variables to CONFIGURE_ENV
ifdef WORLD_CONFIGURE_ENV_PKG_CONFIG
WORLD_CONFIGURE_ENV_PKG_CONFIG := $(shell $(parsebool) $(WORLD_CONFIGURE_ENV_PKG_CONFIG))
else
WORLD_CONFIGURE_ENV_PKG_CONFIG ?= y
endif

# Pass SYSROOT= to the environments of certain stages, but allow the
# package to disable it.
#
# gobject-introspection honors the $SYSROOT environment variable
ifdef WORLD_CONFIGURE_ENV_SYSROOT
WORLD_CONFIGURE_ENV_SYSROOT := $(shell $(parsebool) $(WORLD_CONFIGURE_ENV_SYSROOT))
else
WORLD_CONFIGURE_ENV_SYSROOT ?= y
endif

ifdef WORLD_BUILD_ENV_SYSROOT
WORLD_BUILD_ENV_SYSROOT := $(shell $(parsebool) $(WORLD_BUILD_ENV_SYSROOT))
else
WORLD_BUILD_ENV_SYSROOT ?= y
endif

ifdef WORLD_TEST_ENV_SYSROOT
WORLD_TEST_ENV_SYSROOT := $(shell $(parsebool) $(WORLD_TEST_ENV_SYSROOT))
else
WORLD_TEST_ENV_SYSROOT ?= y
endif

ifdef WORLD_FAKEINST_ENV_SYSROOT
WORLD_FAKEINST_ENV_SYSROOT := $(shell $(parsebool) $(WORLD_FAKEINST_ENV_SYSROOT))
else
WORLD_FAKEINST_ENV_SYSROOT ?= y
endif

# If configure style is gnu, run autoreconf on the source?
# (If 'y', we need to make sure the build uses sysroot Perl libs.)
ifeq ($(CONFIGURE_STYLE),gnu)
ifdef AUTORECONF
AUTORECONF := $(shell $(parsebool) $(AUTORECONF))
else
AUTORECONF ?= n
endif

ifeq ($(AUTORECONF),y)
WORLD_USE_PERL5LIB ?= y
CONFIGURE_ENV += AUTOM4TE=$(WRKSYSROOT)$(BINDIR)/autom4te \
                 AUTOCONF=$(WRKSYSROOT)$(BINDIR)/autoconf \
                 AUTOHEADER=$(WRKSYSROOT)$(BINDIR)/autoheader \
                 AUTOMAKE=$(WRKSYSROOT)$(BINDIR)/automake \
                 ACLOCAL=$(WRKSYSROOT)$(BINDIR)/aclocal \
                 ACLOCAL_PATH=$(WRKSYSROOT)$(DATADIR)/aclocal:$(WRKSYSROOT)$(X11_DATADIR)/aclocal:$(WRKSYSROOT)$(WAYLAND_DATADIR)/aclocal \
                 AUTOPOINT=$(WRKSYSROOT)$(BINDIR)/autopoint \
                 LIBTOOLIZE=$(WRKSYSROOT)$(BINDIR)/libtoolize \
                 M4=$(WRKSYSROOT)$(BINDIR)/m4
BUILD_ENV += AUTOM4TE=$(WRKSYSROOT)$(BINDIR)/autom4te \
             AUTOCONF=$(WRKSYSROOT)$(BINDIR)/autoconf \
             AUTOHEADER=$(WRKSYSROOT)$(BINDIR)/autoheader \
             AUTOMAKE=$(WRKSYSROOT)$(BINDIR)/automake \
             ACLOCAL=$(WRKSYSROOT)$(BINDIR)/aclocal \
             ACLOCAL_PATH=$(WRKSYSROOT)$(DATADIR)/aclocal:$(WRKSYSROOT)$(X11_DATADIR)/aclocal:$(WRKSYSROOT)$(WAYLAND_DATADIR)/aclocal \
             AUTOPOINT=$(WRKSYSROOT)$(BINDIR)/autopoint \
             LIBTOOLIZE=$(WRKSYSROOT)$(BINDIR)/libtoolize \
             M4=$(WRKSYSROOT)$(BINDIR)/m4

ifdef GETTEXTIZE
GETTEXTIZE := $(shell $(parsebool) $(GETTEXTIZE))
else
GETTEXTIZE ?= y
endif
endif
endif

# Python needs to know where to find its modules during world builds.  The
# easy way to do that is tell Python to replace its ${exec_prefix} and
# ${prefix} variables throughout its configuration with a temporary path.
# That path can be specified in the PYTHONHOME environment variable making
# this process way way easy.
ifeq ($(CONFIGURE_STYLE),python)
WORLD_USE_PYTHONHOME ?= y
else
WORLD_USE_PYTHONHOME ?= n
endif

ifeq ($(CONFIGURE_STYLE),pyproject)
WORLD_USE_PYTHONHOME ?= y
else
WORLD_USE_PYTHONHOME ?= n
endif

ifdef WORLD_USE_PYTHONHOME
WORLD_USE_PYTHONHOME := $(shell $(parsebool) $(WORLD_USE_PYTHONHOME))
endif

# If meson is used, it needs to find the sysroot Python
ifeq ($(CONFIGURE_STYLE),meson)
WORLD_USE_PYTHONHOME := y
endif

# Perl needs to know where its library files are during world builds.
# By default this is disabled.  If you set it to 'y', the PERL5LIB
# variable is available to you to append or override.  This variable
# is passed in the env variables for the configure, build, test, and
# fakeinst stages.  See the perlrun(1) man page for information on the
# PERL5LIB variable.
ifeq ($(CONFIGURE_STYLE),perl)
WORLD_USE_PERL5LIB ?= y
endif

ifdef WORLD_USE_PERL5LIB
WORLD_USE_PERL5LIB := $(shell $(parsebool) $(WORLD_USE_PERL5LIB))
else
WORLD_USE_PERL5LIB ?= n
endif

# Add --sysroot= to CFLAGS and CXXFLAGS?
ifeq ($(WORLD_CFLAGS_SYSROOT),y)
CFLAGS += --sysroot=$(WRKSYSROOT)
endif

ifeq ($(WORLD_CXXFLAGS_SYSROOT),y)
CXXFLAGS += --sysroot=$(WRKSYSROOT)
endif

# Add --with-sysroot= to CONFIGURE_OPTS?
ifeq ($(CONFIGURE_STYLE),gnu)
ifeq ($(WORLD_WITH_SYSROOT),y)
CONFIGURE_OPTS += --with-sysroot=$(WRKSYSROOT)
endif
endif

# Add -I to CFLAGS?
ifeq ($(WORLD_CFLAGS_INCLUDEDIR),y)
CFLAGS += -I$(WRKSYSROOT)$(INCLUDEDIR)
CXXFLAGS += -I$(WRKSYSROOT)$(INCLUDEDIR)

ifeq ($(X11),y)
CFLAGS += -I$(WRKSYSROOT)$(X11_INCLUDEDIR)
CXXFLAGS += -I$(WRKSYSROOT)$(X11_INCLUDEDIR)
endif

ifeq ($(WAYLAND),y)
CFLAGS += -I$(WRKSYSROOT)$(WAYLAND_INCLUDEDIR)
CXXFLAGS += -I$(WRKSYSROOT)$(WAYLAND_INCLUDEDIR)
endif
endif

# Add -L to LDFLAGS?
ifeq ($(WORLD_LDFLAGS_LIBDIR),y)
LDFLAGS += -L$(WRKSYSROOT)$(LIBDIR)

ifeq ($(X11),y)
LDFLAGS += -L$(WRKSYSROOT)$(X11_LIBDIR)
endif

ifeq ($(WAYLAND),y)
LDFLAGS += -L$(WRKSYSROOT)$(WAYLAND_LIBDIR)
endif
endif

# Add -Wl,--sysroot= to LDFLAGS?
ifeq ($(WORLD_LDFLAGS_WL_SYSROOT),y)
LDFLAGS += -Wl,--sysroot=$(WRKSYSROOT)
endif

# Make sure world builds know about sysroot bin and lib directories.
ifeq ($(WORLD_CONFIGURE_ENV_PATH),y)
CONFIGURE_ENV += PATH=$(WRKSYSROOT)$(BINDIR):$(WRKSYSROOT)$(X11_BINDIR):$(WRKSYSROOT)$(WAYLAND_BINDIR):$(PATH)
endif

ifeq ($(WORLD_CONFIGURE_ENV_LD_LIBRARY_PATH),y)
CONFIGURE_ENV += LD_LIBRARY_PATH=$(WRKSYSROOT)$(LIBDIR):$(WRKSYSROOT)$(X11_LIBDIR):$(WRKSYSROOT)$(WAYLAND_LIBDIR)
endif

# Add path overrides to build, test, and fakeinst environments?
ifeq ($(WORLD_BUILD_ENV_PATH),y)
BUILD_ENV += PATH=$(WRKSYSROOT)$(BINDIR):$(WRKSYSROOT)$(X11_BINDIR):$(WRKSYSROOT)$(WAYLAND_BINDIR):$(PATH)
endif

ifeq ($(WORLD_BUILD_ENV_LD_LIBRARY_PATH),y)
BUILD_ENV += LD_LIBRARY_PATH=$(WRKSYSROOT)$(LIBDIR):$(WRKSYSROOT)$(X11_LIBDIR):$(WRKSYSROOT)$(WAYLAND_LIBDIR)
endif

ifeq ($(WORLD_TEST_ENV_PATH),y)
TEST_ENV += PATH=$(WRKSYSROOT)$(BINDIR):$(WRKSYSROOT)$(X11_BINDIR):$(WRKSYSROOT)$(WAYLAND_BINDIR):$(PATH)
endif

ifeq ($(WORLD_TEST_ENV_LD_LIBRARY_PATH),y)
TEST_ENV += LD_LIBRARY_PATH=$(WRKSYSROOT)$(LIBDIR):$(WRKSYSROOT)$(X11_LIBDIR):$(WRKSYSROOT)$(WAYLAND_LIBDIR)
endif

ifeq ($(WORLD_FAKEINST_ENV_PATH),y)
FAKEINST_ENV += PATH=$(WRKSYSROOT)$(BINDIR):$(WRKSYSROOT)$(X11_BINDIR):$(WRKSYSROOT)$(WAYLAND_BINDIR):$(PATH)
endif

ifeq ($(WORLD_FAKEINST_ENV_LD_LIBRARY_PATH),y)
FAKEINST_ENV += LD_LIBRARY_PATH=$(WRKSYSROOT)$(LIBDIR):$(WRKSYSROOT)$(X11_LIBDIR):$(WRKSYSROOT)$(WAYLAND_LIBDIR)
endif

# Add the pkg-config override variables to CONFIGURE_ENV
ifeq ($(WORLD_CONFIGURE_ENV_PKG_CONFIG),y)
CONFIGURE_ENV += PKG_CONFIG_SYSROOT_DIR=$(WRKSYSROOT) \
                 PKG_CONFIG_PATH=$(WRKSYSROOT)$(LIBDIR)/pkgconfig:$(WRKSYSROOT)$(X11_LIBDIR)/pkgconfig:$(WRKSYSROOT)$(WAYLAND_LIBDIR)/pkgconfig:$(LIBDIR)/pkgconfig:$(DATADIR)/pkgconfig$(EXTRA_PKG_CONFIG_PATH)
endif

# Add the pkg-config override variables to BUILD_ENV
ifeq ($(WORLD_BUILD_ENV_PKG_CONFIG),y)
BUILD_ENV += PKG_CONFIG_SYSROOT_DIR=$(WRKSYSROOT) \
             PKG_CONFIG_PATH=$(WRKSYSROOT)$(LIBDIR)/pkgconfig:$(WRKSYSROOT)$(X11_LIBDIR)/pkgconfig:$(WRKSYSROOT)$(WAYLAND_LIBDIR)/pkgconfig:$(LIBDIR)/pkgconfig:$(DATADIR)/pkgconfig$(EXTRA_PKG_CONFIG_PATH)
endif

ifeq ($(WORLD_FAKEINST_ENV_PKG_CONFIG),y)
FAKEINST_ENV += PKG_CONFIG_SYSROOT_DIR=$(WRKSYSROOT) \
                PKG_CONFIG_PATH=$(WRKSYSROOT)$(LIBDIR)/pkgconfig:$(WRKSYSROOT)$(X11_LIBDIR)/pkgconfig:$(WRKSYSROOT)$(WAYLAND_LIBDIR)/pkgconfig:$(LIBDIR)/pkgconfig:$(DATADIR)/pkgconfig$(EXTRA_PKG_CONFIG_PATH)
endif

# Pass SYSROOT= to the environments of certain stages, but allow the
# package to disable it.
#
# gobject-introspection honors the $SYSROOT environment variable
ifeq ($(WORLD_CONFIGURE_ENV_SYSROOT),y)
CONFIGURE_ENV += SYSROOT=$(WRKSYSROOT)
export SYSROOT = $(WRKSYSROOT)
endif

ifeq ($(WORLD_BUILD_ENV_SYSROOT),y)
BUILD_ENV += SYSROOT=$(WRKSYSROOT)
export SYSROOT = $(WRKSYSROOT)
endif

ifeq ($(WORLD_TEST_ENV_SYSROOT),y)
TEST_ENV += SYSROOT=$(WRKSYSROOT)
export SYSROOT = $(WRKSYSROOT)
endif

ifeq ($(WORLD_FAKEINST_ENV_SYSROOT),y)
FAKEINST_ENV += SYSROOT=$(WRKSYSROOT)
export SYSROOT = $(WRKSYSROOT)
endif

# Python needs to know where to find its modules during world builds.  The
# easy way to do that is tell Python to replace its ${exec_prefix} and
# ${prefix} variables throughout its configuration with a temporary path.
# That path can be specified in the PYTHONHOME environment variable making
# this process way way easy.
ifeq ($(WORLD_USE_PYTHONHOME),y)
CONFIGURE_ENV += PYTHONHOME=$(WRKSYSROOT)$(PREFIX)
BUILD_ENV += PYTHONHOME=$(WRKSYSROOT)$(PREFIX)
TEST_ENV += PYTHONHOME=$(WRKSYSROOT)$(PREFIX)
FAKEINST_ENV += PYTHONHOME=$(WRKSYSROOT)$(PREFIX)
endif

# Perl needs to know where its library files are during world builds.
# By default this is disabled.  If you set it to 'y', the PERL5LIB
# variable is available to you to append or override.  This variable
# is passed in the env variables for the configure, build, test, and
# fakeinst stages.  See the perlrun(1) man page for information on the
# PERL5LIB variable.
ifeq ($(WORLD_USE_PERL5LIB),y)
PERL5LIB ?= $(WRKSYSROOT)$(LIBDIR)/perl:$(WRKSYSROOT)$(DATADIR)/perl:$(WRKSYSROOT)$(LIBDIR)/perl/vendor_perl:$(WRKSYSROOT)$(DATADIR)/perl/vendor_perl
CONFIGURE_ENV += PERL5LIB=$(PERL5LIB) \
                 PERL_LIB=$(WRKSYSROOT)$(DATADIR)/perl \
                 PERL_ARCHLIB=$(WRKSYSROOT)$(LIBDIR)/perl \
                 PERL_INC=$(WRKSYSROOT)$(LIBDIR)/perl/CORE
BUILD_ENV += PERL5LIB=$(PERL5LIB) \
             PERL_LIB=$(WRKSYSROOT)$(DATADIR)/perl \
             PERL_ARCHLIB=$(WRKSYSROOT)$(LIBDIR)/perl \
             PERL_INC=$(WRKSYSROOT)$(LIBDIR)/perl/CORE
TEST_ENV += PERL5LIB=$(PERL5LIB) \
            PERL_LIB=$(WRKSYSROOT)$(DATADIR)/perl \
            PERL_ARCHLIB=$(WRKSYSROOT)$(LIBDIR)/perl \
            PERL_INC=$(WRKSYSROOT)$(LIBDIR)/perl/CORE
FAKEINST_ENV += PERL5LIB=$(PERL5LIB) \
                PERL_LIB=$(WRKSYSROOT)$(DATADIR)/perl \
                PERL_ARCHLIB=$(WRKSYSROOT)$(LIBDIR)/perl \
                PERL_INC=$(WRKSYSROOT)$(LIBDIR)/perl/CORE
endif

# Unset LD_RUN_PATH
BUILD_ENV += --unset=LD_RUN_PATH
TEST_ENV += --unset=LD_RUN_PATH
FAKEINST_ENV += --unset=LD_RUN_PATH

# Make sure we use the sysroot gconv files
BUILD_ENV += GCONV_PATH=$(WRKSYSROOT)$(LIBDIR)/gconv
TEST_ENV += GCONV_PATH=$(WRKSYSROOT)$(LIBDIR)/gconv
FAKEINST_ENV += GCONV_PATH=$(WRKSYSROOT)$(LIBDIR)/gconv

# Set QT_SYSROOT for qmake builds
ifeq ($(CONFIGURE_STYLE),qmake)
BUILD_ENV += QT_SYSROOT=$(WRKSYSROOT)
endif

# Set CMAKE_PREFIX_PATH
ifeq ($(CONFIGURE_STYLE),cmake)
CMAKE_PREFIX_PATH ?= $(WRKSYSROOT)$(PREFIX):$(WRKSYSROOT)$(X11_PREFIX):$(WRKSYSROOT)$(WAYLAND_PREFIX)
BUILD_ENV += CMAKE_PREFIX_PATH=$(CMAKE_PREFIX_PATH)
endif

# Compression algorithm used by distmk for the distribution
COMPRESSION ?= zstd

# Main driver for packages, we want a 'make' with no arguments to just
# run through build and stop.  Users have to explicitly tell it to fakeinst,
# package, or clean.
all: do-post-build

# Utility targets (helpful to other targets or to packagers)
do-mkdir-wrkdir:
	@if [ ! "$(WRKDIR)" = "--.$(TARGET_MACH)" ]; then \
		mkdir -p "$(WRKDIR)" ; \
	fi

do-mkdir-wrkinst:
	@mkdir -p $(WRKINST)

do-mkdir-distfiles:
	@mkdir -p $(DISTFILES)

do-mkdir-release:
	@mkdir -p $(RELEASE)

do-mkdir-patches:
	@mkdir -p $(PATCHES)

do-mkdir-wrksysroot:
	@mkdir -p $(WRKSYSROOT)

do-pp-archives: do-mkdir-wrkdir
	@if [ -f "$(CWD)/archives" ]; then \
		cpp -E -dU -pipe -fdirectives-only -nostdinc -I$(TOPSRC) \
			$(CWD)/archives -o $(WRKDIR)/archives ; \
	fi

# This target is called during 'make world' runs.  It installs the packages
# listed in the buildreqs file to a sysroot tree.  The 'make world' stage will
# pass --sysroot to gcc so that the package build uses headers and libraries
# from that tree rather than the active system.  This is similar to doing a
# chroot build except that the runtime commands are called from the running
# system, it's only headers and libraries that are read from the sysroot tree.
usereqs: do-mkdir-wrksysroot do-mkdir-wrkdir
	@if [ "$(CACHE_SYSROOT)" = "n" ]; then \
		env TOPSRC='$(TOPSRC)' \
		    BUILDREQS_PATHS='$(BUILDREQS_PATHS)' \
		    WRKDIR='$(WRKDIR)' \
		    WRKSYSROOT='$(WRKSYSROOT)' \
		    PKGCOLLECTION='$(PKGCOLLECTION)' \
		    MAKE='$(MAKE)' \
		    DIGEST_ALGORITHM='$(DIGEST_ALGORITHM)' \
		    RELEASE='$(RELEASE)' \
		    DIST_URL='$(DIST_URL)' \
		    DIST_DOWNLOAD='$(DIST_DOWNLOAD)' \
		    DISTADD_TMPDIR='$(DISTADD_TMPDIR)' \
		    HELPERS='$(HELPERS)' \
		$(usereqs) || exit $$? ; \
	fi

# Recursively run 'make world' for all of the package's missing buildreqs
missingreqs: do-mkdir-wrksysroot do-mkdir-wrkdir
	@if [ "$(CACHE_SYSROOT)" = "n" ]; then \
		env TOPSRC='$(TOPSRC)' \
		    BUILDREQS_PATHS='$(BUILDREQS_PATHS)' \
		    WRKDIR='$(WRKDIR)' \
		    WRKSYSROOT='$(WRKSYSROOT)' \
		    PKGCOLLECTION='$(PKGCOLLECTION)' \
		    MAKE='$(MAKE)' \
		    DIST_URL='$(DIST_URL)' \
		    DIST_DOWNLOAD='$(DIST_DOWNLOAD)' \
		    DIGEST_ALGORITHM='$(DIGEST_ALGORITHM)' \
		    RELEASE='$(RELEASE)' \
		    DISTADD_TMPDIR='$(DISTADD_TMPDIR)' \
		    HELPERS='$(HELPERS)' \
		$(missingreqs) || exit $$? ; \
	fi

# Helper target for world-builder to show the contents of buildreqs.
show-buildreqs: do-mkdir-wrkdir
	@env TOPSRC='$(TOPSRC)' \
	     BUILDREQS_PATHS='$(BUILDREQS_PATHS)' \
	     WRKDIR='$(WRKDIR)' \
	     REDUCE='n' \
	     IGNORE_REQS='$(IGNORE_REQS)' \
	     TARGET_MACH='$(TARGET_MACH)' \
	$(buildreqs) || exit $$?

# Utility function to reduce an existing buildreqs file.
reduce-buildreqs: do-mkdir-wrkdir
	@env TOPSRC='$(TOPSRC)' \
	     BUILDREQS_PATHS='$(BUILDREQS_PATHS)' \
	     WRKDIR='$(WRKDIR)' \
	     REDUCE='y' \
	     IGNORE_REQS='$(IGNORE_REQS)' \
	     TARGET_MACH='$(TARGET_MACH)' \
	$(buildreqs) || exit $$?

# Perform a world-style build of the package.  This is intended for use
# when a world build fails and you want to work on the immediate package
# that failed.  Do not perform a top level 'make realclean' because those
# built packages will be needed by this test build.

ifeq ($(CACHE_SYSROOT),)
CACHE_SYSROOT ?= n
endif

# Gather number of file descriptors available since some package
# builds are gigantic (nodejs).
FILE_DESCRIPTORS_LIMIT := $(shell ulimit -a | grep "nofiles" | awk '{ print $$2; }')

world:
ifeq ($(TOP_LEVEL),y)
	@if [ $(FILE_DESCRIPTORS_LIMIT) -lt 2048 ]; then \
		echo >&2 ; \
		echo "*** nodejs is huge, you must run:" >&2 ; \
		echo "***     ulimit -n 2048" >&2 ; \
		echo "*** before building" >&2 ; \
		echo >&2 ; \
		exit 1 ; \
	fi ; \
	$(MAKE) _world_builder_helper
else
	@if [ "$(SKIP)" = "y" ]; then \
		echo "Skipping $(PACKAGE) $(SKIP_REASON), set SKIP=n to override." ; \
	else \
		env CACHE_SYSROOT='$(CACHE_SYSROOT)' $(MAKE) clean || exit $$? ; \
		env CACHE_SYSROOT='$(CACHE_SYSROOT)' $(MAKE) usereqs || exit $$? ; \
		env CACHE_SYSROOT='$(CACHE_SYSROOT)' $(MAKE) package || exit $$? ; \
		if [ "$(KEEP_WRKDIR)" = "n" ]; then \
			env CACHE_SYSROOT='$(CACHE_SYSROOT)' $(MAKE) clean || exit $$? ; \
		fi ; \
	fi
endif

quickworld:
	@env CACHE_SYSROOT='y' $(MAKE) world || exit $$?

bundle:
	@env TOPSRC='$(TOPSRC)' \
	     WRKDIR='$(WRKDIR)' \
	     MAKE='$(MAKE)' \
	     PKG='$(PKG)' \
	     FULLPKGPATH='$(FULLPKGPATH)' \
	     FULLDEBUGPKGPATH='$(FULLDEBUGPKGPATH)' \
	     PACKAGE='$(PACKAGE)' \
	     DEBUGPACKAGE='$(DEBUGPACKAGE)' \
	     DISTRIBUTION_VERSION='$(DISTRIBUTION_VERSION)' \
	     PKGCOLLECTION='$(PKGCOLLECTION)' \
	     TARGET_MACH='$(TARGET_MACH)' \
	     RELEASE='$(RELEASE)' \
	     DIGEST_ALGORITHM='$(DIGEST_ALGORITHM)' \
	     HELPERSDB='$(HELPERSDB)' \
	     MERGE_SH_IN='$(MERGE_SH_IN)' \
	$(bundle) || exit $$?

showprevpkg:
	@env TOPSRC='$(TOPSRC)' $(showprevpkg) || exit $$?

comma := ,

show:
	@if [ -z "$(var)" ]; then \
		$(ECHO) "Usage: make show var=VARIABLE[,VARIABLE2,...]" ; \
	else \
		$(ECHO) $(foreach v,$(subst $(comma), ,$(var)),"$($(v))") ; \
	fi

# Increment the release number and apply the change to the Makefile
bumprel:
	@sed -i -e "s/^REL\ *\=\ *$(REL)/REL\ =\ $$(($(REL) + 1))/g" Makefile
	@echo "$(PKG)-$(VER)-$(REL) => $(PKG)-$(VER)-$$(($(REL) + 1))"

# Show the PKG-VER-REL of a package
verrel:
	@echo "$(PKG)-$(VER)-$(REL)"

# Show the license information for the package
license:
	@if [ -f $(CWD)/pkginfo ]; then \
		. $(CWD)/pkginfo ; \
		[ -z "$${license}" ] && license="(missing license)" ; \
		printf "%-35s %-35s\n" $(PKG)-$(VER)-$(REL) "$${license}" ; \
	else \
		[ -z "$(LICENSE)" ] && license="(missing license)"
		printf "%-35s %-35s\n" $(PKG)-$(VER)-$(REL) "$${license}" ; \
	fi

# Remove old distfiles, keep current ones
prune-distfiles: do-pp-archives do-mkdir-distfiles
	@if [ "$(TOP_LEVEL)" = "y" ]; then \
		$(MAKE) _prune-all-distfiles ; \
	else \
		env WRKDIR='$(WRKDIR)' \
		    DISTFILES='$(DISTFILES)' \
		    PKGDISTFILES='$(PKGDISTFILES)' \
		    TOPSRC='$(TOPSRC)' \
		    GITPKG='$(GITPKG)' \
		    GITVER='$(GITVER)' \
		    GITCOMMIT='$(GITCOMMIT)' \
		$(prunedistfiles) ; \
	fi

# Variables specific to the upstream version checking target.
# Packages can override these settings in 'upstream' file.

# Default URL for Haskell packages is different than other packages.
ifeq ($(CONFIGURE_STYLE),haskell)
UPSTREAM_URLS ?= $(HACKAGE_UPSTREAM)
else
UPSTREAM_URLS ?= $(URLS)
endif

# hackage.haskell.org no longer offers directory browsing, so we're going to
# scrape the HTML output.  It's ugly, but that's really the easiest option
# here.  This will likely change again, so we need to keep an eye on it.  I
# noticed it this time around because 'make check-upstream' in every haskell
# package failed.
ifeq ($(UPSTREAM_URLS),$(HACKAGE_UPSTREAM))
UPSTREAM_REGEXP ?= ^[0-9\.]+$
endif

ifeq ($(shell $(ishttpurl) $(UPSTREAM_URLS)),y)
UPSTREAM_REGEXP ?= $(PKG)[-_\.]+[0-9\.]+.+\.(tar|t[bglx]z|tbz2|zip)+(\.[gx]z|\.bz2)*
endif

# Default upstream regexps
UPSTREAM_REGEXP ?= ^$(PKG)[-_\.]+[0-9\.]+.+\.(tar|t[bglx]z|tbz2|zip)+(\.[gx]z|\.bz2)*$$
UPSTREAM_PREFIX ?= $(PKG)[-_\.]+
UPSTREAM_SUFFIX ?= \.(tar|t[bglx]z|tbz2|zip)+(\.[gx]z|\.bz2)*.*

# Check for new upstream version of the package's source.
# See upstream.template for more information.
check-upstream:
	@env PKG='$(PKG)' \
	     VER='$(VER)' \
	     UPSTREAM_URLS='$(UPSTREAM_URLS)' \
	     UPSTREAM_REGEXP='$(UPSTREAM_REGEXP)' \
	     UPSTREAM_PREFIX='$(UPSTREAM_PREFIX)' \
	     UPSTREAM_SUFFIX='$(UPSTREAM_SUFFIX)' \
	     TOPSRC='$(TOPSRC)' \
	$(checkupstream)

# Apply the new upstream version information to the package source
# files.  Generally this should be used after a maintainer runs "make
# check-upstream".  If things look good, then run "make take-upstream"
# to apply those changes to the source files.
take-upstream:
	@env PKG='$(PKG)' \
	     VER='$(VER)' \
	     UPSTREAM_URLS='$(UPSTREAM_URLS)' \
	     UPSTREAM_REGEXP='$(UPSTREAM_REGEXP)' \
	     UPSTREAM_PREFIX='$(UPSTREAM_PREFIX)' \
	     UPSTREAM_SUFFIX='$(UPSTREAM_SUFFIX)' \
	     TOPSRC='$(TOPSRC)' \
	     UPDATE='y' \
	$(checkupstream)

# Given an archives file with new entries of just filenames, expand
# to lines for each supported hash type.  e.g., if we have:
#
#     bash-4.2.tar.gz
#
# In the archives file, replace it with one of:
#
#     MD5 <digest>  bash-4.2.tar.gz
#     SHA1 <digest>  bash-4.2.tar.gz
#     SHA256 <digest>  bash-4.2.tar.gz
#     SHA384 <digest>  bash-4.2.tar.gz
#     SHA512 <digest>  bash-4.2.tar.gz
#
# The line that is output is based on the preferred digest algorithm
# currently in use in the tree (set in defs.mk).
#
# Really, all this does is run the coreutils commands to generate the
# digests for the new archive files.  It's just meant as a time saver
# for instances where we are creating a new package or updating an
# existing one.
#
# You can put your new archive names in the archives file, then run
#     make fetch
#     make makesum
# To download them and compute new digest hashes for them.
makesum: do-mkdir-wrkdir
	@env TOPSRC='$(TOPSRC)' \
	     WRKDIR='$(WRKDIR)' \
	     DIGEST_ALGORITHM='$(DIGEST_ALGORITHM)' \
	     PKGDISTFILES='$(PKGDISTFILES)' \
	$(makesum) || exit $$?

# Reset the archives file to just be include lines and archive filenames.
cleansum: do-mkdir-wrkdir
	@env WRKDIR='$(WRKDIR)' $(cleansum) || exit $$?

# Walk $(WRKSRC) and generate patches for any files that have a matching
# file of the same name that ends with ".orig".
update-patches: do-mkdir-patches pre-update-patches do-update-patches post-update-patches
do-update-patches:
	@env WRKDIR='$(WRKDIR)' \
	     WRKSRC='$(WRKSRC)' \
	     PATCHES='$(PATCHES)' \
	$(updatepatches) || exit $$?

# Generate a packing list file for the package.  The packing list file
# is the same format as the dist-tools attrs(5) file, which shows the
# file or directory name, owner, group, and octal permissions.
update-plist: pre-update-plist do-update-plist post-update-plist
do-update-plist:
	@env VER='$(VER)' \
	     REL='$(REL)' \
	     WRKDIR='$(WRKDIR)' \
	     WRKINST='$(WRKINST)' \
	     PACKAGE='$(PACKAGE)' \
	     TARGET_NAME='$(TARGET_NAME)' \
	     TARGET_MACH='$(TARGET_MACH)' \
	$(updateplist) || exit $$?

# Shortcut to do standard package upgrade commits
ci: update-plist
	git add .
	if [ "$(PKG)" = "kernel" ]; then \
		git commit -m "[$(PKG)] Upgrade to the Linux kernel version $(VER)" ; \
	else \
		git commit -m "[$(PKG)] Upgrade to $(PKG)-$(VER)" ; \
	fi

# If $(GPGKEYS) is set, update the $(PKG).gpg file with the key pointed
# to by this variable.
update-keyring:
	@env GPG='$(GPG)' \
	     GPGKEYS='$(GPGKEYS)' \
	     KEYRING='$(KEYRING)' \
	     TRUSTDB='$(TRUSTDB)' \
	$(updatekeyring) || exit $$?

# List the files and directories in $WRKINST, if it exists
list-wrkinst:
	@if [ -d "$(WRKINST)" ]; then \
		find $(WRKINST) | sed -e "s|$(WRKINST)||g" | grep -v -E "^$$" ; \
	else \
		$(ECHO) "$(WRKINST) does not exist." ; \
	fi

# STAGE: fetch

# Set this to 'y' to tell the fetch target to not download from mirrors.
ifdef FORCE_REFRESH
FORCE_REFRESH := $(shell $(parsebool) $(FORCE_REFRESH))
else
FORCE_REFRESH ?= n
endif

# Default to removing .git subdirectories
KEEPDOTGIT ?= n

fetch: do-pp-archives do-mkdir-distfiles pre-fetch do-fetch post-fetch
do-fetch:
	@env TOPSRC='$(TOPSRC)' \
	     MAKE='$(MAKE)' \
	     WRKDIR='$(WRKDIR)' \
	     OUR_MIRRORS='$(OUR_MIRRORS)' \
	     FORCE_REFRESH='$(FORCE_REFRESH)' \
	     URLS='$(URLS)' \
	     PKGDISTFILES='$(PKGDISTFILES)' \
	     DISTFILES='$(DISTFILES)' \
	     PKG='$(PKG)' \
	     VER='$(VER)' \
	     GITREPO='$(GITREPO)' \
	     GITBRANCH='$(GITBRANCH)' \
	     GITCOMMIT='$(GITCOMMIT)' \
	     GITSUBMODULES='$(GITSUBMODULES)' \
	     GITSCRIPT='$(GITSCRIPT)' \
	     GITPKG='$(GITPKG)' \
	     GITVER='$(GITVER)' \
	     KEEPDOTGIT='$(KEEPDOTGIT)' \
	$(fetch) || exit $$?

checksum: fetch
	@env CWD='$(CWD)' \
	     WRKDIR='$(WRKDIR)' \
	     PKGDISTFILES='$(PKGDISTFILES)' \
	$(checksum) || exit $$?

# STAGE: extract

ifdef GPGKEYS
KEYRING ?= $(CWD)/keyring.gpg
TRUSTDB ?= $(CWD)/trustdb.gpg

# gpg command line to run in the different build targets.  By putting
# it here, we only have to update the common options in one place.
GPG ?= gpg --no-sig-cache --no-default-keyring --allow-non-selfsigned-uid --trust-model pgp

ifdef GPG_KEYSERVER
GPG += --keyserver $(GPG_KEYSERVER)
endif

ifeq ($(GPG_SIGNATURE_ENDINGS),)
GPG_SIGNATURE_ENDINGS ?= .sig .asc .sign
endif
endif

extract: checksum pre-extract do-extract post-extract
do-extract:
	@env TOPSRC='$(TOPSRC)' \
	     PKGDISTFILES='$(PKGDISTFILES)' \
	     WRKDIR='$(WRKDIR)' \
	     WRKDIST='$(WRKDIST)' \
	     GPG='$(GPG)' \
	     KEYRING='$(KEYRING)' \
	     TRUSTDB='$(TRUSTDB)' \
	     GPG_SIGNATURE_ENDINGS='$(GPG_SIGNATURE_ENDINGS)' \
	     GITPKG='$(GITPKG)' \
	     GITVER='$(GITVER)' \
	     GITCOMMIT='$(GITCOMMIT)' \
	     OTHER_GIT_PKGS='$(OTHER_GIT_PKGS)' \
	$(extract) || exit $$?

# STAGE: patch
# All patches are assumed to have been generated by 'update-patches'.  We
# apply with -p1 to trim the first directory level, -F0 for no fuzz factor,
# and we back up the orignal to a file of the same name ending with ".orig"
# (so that 'update-patches' can be run later too).

# Default directory trim level (-p option on patch)
PATCHLEVEL ?= 1

patch: extract do-mkdir-wrkdir pre-patch do-patch post-patch
do-patch:
	@env WRKDIR='$(WRKDIR)' \
	     WRKSRC='$(WRKSRC)' \
	     PATCHES='$(PATCHES)' \
	     PATCHLEVEL='$(PATCHLEVEL)' \
	$(patch) || exit $$?

# STAGE: build

ifdef X11
X11 := $(shell $(parsebool) $(X11))
else
X11 ?= n
endif

ifdef WAYLAND
WAYLAND := $(shell $(parsebool) $(WAYLAND))
else
WAYLAND ?= n
endif

ifdef SKIP_BUILD
SKIP_BUILD := $(shell $(parsebool) $(SKIP_BUILD))
else
SKIP_BUILD ?= n
endif

# Default configure style
CONFIGURE_STYLE ?= gnu

ifeq ($(CONFIGURE_STYLE),gnu)
# Default ./configure options (for CONFIGURE_STYLE=gnu)
CONFIGURE_OPTS += --sysconfdir=$(SYSCONFDIR) \
                  --localstatedir=$(LOCALSTATEDIR) \

ifeq ($(X11),y)
CONFIGURE_OPTS += --prefix=$(X11_PREFIX) \
                  --bindir=$(X11_BINDIR) \
                  --sbindir=$(X11_SBINDIR) \
                  --libexecdir=$(X11_LIBEXECDIR) \
                  --libdir=$(X11_LIBDIR) \
                  --includedir=$(X11_INCLUDEDIR) \
                  --datadir=$(X11_DATADIR) \
                  --mandir=$(X11_MANDIR) \
                  --infodir=$(X11_INFODIR)
else
ifeq ($(WAYLAND),y)
CONFIGURE_OPTS += --prefix=$(WAYLAND_PREFIX) \
                  --bindir=$(WAYLAND_BINDIR) \
                  --sbindir=$(WAYLAND_SBINDIR) \
                  --libexecdir=$(WAYLAND_LIBEXECDIR) \
                  --libdir=$(WAYLAND_LIBDIR) \
                  --includedir=$(WAYLAND_INCLUDEDIR) \
                  --datadir=$(WAYLAND_DATADIR) \
                  --mandir=$(WAYLAND_MANDIR) \
                  --infodir=$(WAYLAND_INFODIR)
else
CONFIGURE_OPTS += --prefix=$(PREFIX) \
                  --bindir=$(BINDIR) \
                  --sbindir=$(SBINDIR) \
                  --libexecdir=$(LIBEXECDIR) \
                  --libdir=$(LIBDIR) \
                  --includedir=$(INCLUDEDIR) \
                  --datadir=$(DATADIR) \
                  --mandir=$(MANDIR) \
                  --infodir=$(INFODIR)
endif
endif
endif

ifeq ($(CONFIGURE_STYLE),haskell)
# Conditionally set the location of helper programs based on what is
# available in sysroot.

define find_tool
$(shell [ -x $(WRKSYSROOT)$(BINDIR)/$(1) ] && echo $(WRKSYSROOT)$(BINDIR)/$(1) || echo $(BINDIR)/$(1))
endef

ALEXDATA  := $(shell [ -x $(WRKSYSROOT)$(DATADIR)/haskell/alex ] && echo $(WRKSYSROOT)$(DATADIR)/haskell/alex || echo $(WRKSYSROOT)$(DATADIR)/haskell/alex)
HAPPYDATA := $(shell [ -x $(WRKSYSROOT)$(DATADIR)/haskell/happy ] && echo $(WRKSYSROOT)$(DATADIR)/haskell/happy || echo $(WRKSYSROOT)$(DATADIR)/haskell/happy)

CONFIGURE_OPTS += --enable-shared \
                  --enable-executable-dynamic \
                  --datadir=$(DATADIR)/haskell \
                  --extra-include-dirs=$(WRKSYSROOT)$(INCLUDEDIR)\
                  --extra-include-dirs=$(WRKSYSROOT)$(X11_INCLUDEDIR) \
                  --extra-include-dirs=$(WRKSYSROOT)$(WAYLAND_INCLUDEDIR) \
                  --extra-lib-dirs=$(WRKSYSROOT)$(LIBDIR) \
                  --extra-lib-dirs=$(WRKSYSROOT)$(X11_LIBDIR) \
                  --extra-lib-dirs=$(WRKSYSROOT)$(WAYLAND_LIBDIR) \
                  --with-alex=$(call find_tool,alex) \
                  --with-ar=$(call find_tool,ar) \
                  --with-c2hs=$(call find_tool,c2hs) \
                  --with-ld=$(call find_tool,ld) \
                  --with-gcc=$(call find_tool,gcc) \
                  --with-ghc=$(call find_tool,ghc) \
                  --with-happy=$(call find_tool,happy) \
                  --with-hscolour=$(call find_tool,HsColour) \
                  --with-pkg-config=$(call find_tool,pkg-config) \
                  --with-strip=$(call find_tool,strip) \
                  --with-tar=$(call find_tool,tar) \
                  --alex-options=--template=$(ALEXDATA) \
                  --happy-options=--template=$(HAPPYDATA) \
                  --gcc-options=-Wl,--sysroot=$(WRKSYSROOT)

ifeq ($(X11),y)
CONFIGURE_OPTS += --prefix=$(X11_PREFIX) \
                  --libdir=$(X11_LIBDIR) \
                  --libexecdir=$(X11_LIBEXECDIR) \
                  --libexecsubdir=$(PKG)
else
ifeq ($(WAYLAND),y)
CONFIGURE_OPTS += --prefix=$(WAYLAND_PREFIX) \
                  --libdir=$(WAYLAND_LIBDIR) \
                  --libexecdir=$(WAYLAND_LIBEXECDIR) \
                  --libexecsubdir=$(PKG)
else
CONFIGURE_OPTS += --prefix=$(PREFIX) \
                  --libdir=$(LIBDIR) \
                  --libexecdir=$(LIBEXECDIR) \
                  --libexecsubdir=$(PKG)
endif
endif
endif

ifeq ($(CONFIGURE_STYLE),cmake)
CONFIGURE_OPTS += -DCMAKE_SYSROOT:PATH=$(WRKSYSROOT) \
                  -DCMAKE_MODULE_PATH:PATH=$(WRKSYSROOT)$(DATADIR)/cmake/Modules:$(WRKSYSROOT)$(X11_DATADIR)/cmake/Modules:$(WRKSYSROOT)$(WAYLAND_DATADIR)/cmake/Modules \
                  -DCMAKE_VERBOSE_MAKEFILE=ON \
                  -DBUILD_SHARED_LIBS:BOOL=ON \
                  -DCMAKE_POLICY_DEFAULT_CMP0017=OLD \
                  -DSYSCONF_INSTALL_DIR:PATH=$(SYSCONFDIR)

CONFIGURE_OPTS += -DCMAKE_INSTALL_INCLUDEDIR:PATH=include \
                  -DCMAKE_INSTALL_LIBDIR:PATH=lib \
                  -DCMAKE_INSTALL_DATADIR:PATH=share

ifeq ($(X11),y)
CONFIGURE_OPTS += -DINCLUDE_INSTALL_DIR:PATH=$(X11_INCLUDEDIR) \
                  -DLIB_INSTALL_DIR:PATH=$(X11_LIBDIR) \
                  -DSHARE_INSTALL_PREFIX:PATH=$(X11_DATADIR) \
                  -DCMAKE_INSTALL_PREFIX:PATH=$(X11_PREFIX)
else
ifeq ($(WAYLAND),y)
CONFIGURE_OPTS += -DINCLUDE_INSTALL_DIR:PATH=$(WAYLAND_INCLUDEDIR) \
                  -DLIB_INSTALL_DIR:PATH=$(WAYLAND_LIBDIR) \
                  -DSHARE_INSTALL_PREFIX:PATH=$(WAYLAND_DATADIR) \
                  -DCMAKE_INSTALL_PREFIX:PATH=$(WAYLAND_PREFIX)
else
CONFIGURE_OPTS += -DINCLUDE_INSTALL_DIR:PATH=$(INCLUDEDIR) \
                  -DLIB_INSTALL_DIR:PATH=$(LIBDIR) \
                  -DSHARE_INSTALL_PREFIX:PATH=$(DATADIR) \
                  -DCMAKE_INSTALL_PREFIX:PATH=$(PREFIX)
endif
endif
endif

ifeq ($(CONFIGURE_STYLE),meson)
CONFIGURE_OPTS += --sysconfdir $(SYSCONFDIR) \
                  --localstatedir $(LOCALSTATEDIR) \
                  --backend ninja \
                  --default-library shared

ifeq ($(X11),y)
CONFIGURE_OPTS += --prefix $(X11_PREFIX) \
                  --libdir $(X11_LIBDIR) \
                  --libexecdir $(X11_LIBEXECDIR) \
                  --bindir $(X11_BINDIR) \
                  --sbindir $(X11_SBINDIR) \
                  --includedir $(X11_INCLUDEDIR) \
                  --datadir $(X11_DATADIR) \
                  --mandir $(X11_MANDIR) \
                  --infodir $(X11_INFODIR)
else
ifeq ($(WAYLAND),y)
CONFIGURE_OPTS += --prefix $(WAYLAND_PREFIX) \
                  --libdir $(WAYLAND_LIBDIR) \
                  --libexecdir $(WAYLAND_LIBEXECDIR) \
                  --bindir $(WAYLAND_BINDIR) \
                  --sbindir $(WAYLAND_SBINDIR) \
                  --includedir $(WAYLAND_INCLUDEDIR) \
                  --datadir $(WAYLAND_DATADIR) \
                  --mandir $(WAYLAND_MANDIR) \
                  --infodir $(WAYLAND_INFODIR)
else
CONFIGURE_OPTS += --prefix $(PREFIX) \
                  --libdir $(LIBDIR) \
                  --libexecdir $(LIBEXECDIR) \
                  --bindir $(BINDIR) \
                  --sbindir $(SBINDIR) \
                  --includedir $(INCLUDEDIR) \
                  --datadir $(DATADIR) \
                  --mandir $(MANDIR) \
                  --infodir $(INFODIR)
endif
endif
endif

ifeq ($(CONFIGURE_STYLE),python)
CONFIGURE_OPTS += --verbose
endif

ifeq ($(CONFIGURE_STYLE),pyproject)
CONFIGURE_OPTS += --no-deps \
                  --use-pep517 \
                  --no-build-isolation \
                  --disable-pip-version-check \
                  --no-clean \
                  --progress-bar off \
                  --verbose
BUILD_TARGET = wheel
SKIP_TEST ?= y
FAKEINST_ARGS += --no-deps \
                 --disable-pip-version-check \
                 --progress-bar off \
                 --verbose \
                 --ignore-installed \
                 --no-warn-script-location \
                 --no-index \
                 --no-cache-dir
endif

# Disable certain things by default for GNU style builds.
ifeq ($(CONFIGURE_STYLE),gnu)
CONFIGURE_ENV += am_cv_langinfo_codeset=no

# Pass --disable-dependency-tracking to ./configure script?
ifdef DISABLE_DEPENDENCY_TRACKING
DISABLE_DEPENDENCY_TRACKING := $(shell $(parsebool) $(DISABLE_DEPENDENCY_TRACKING))
else
DISABLE_DEPENDENCY_TRACKING ?= y
endif

ifeq ($(DISABLE_DEPENDENCY_TRACKING),y)
CONFIGURE_OPTS += --disable-dependency-tracking
endif

# Pass --disable-silent-rules to ./configure script?
ifdef DISABLE_SILENT_RULES
DISABLE_SILENT_RULES := $(shell $(parsebool) $(DISABLE_SILENT_RULES))
else
DISABLE_SILENT_RULES ?= y
endif

ifeq ($(DISABLE_SILENT_RULES),y)
CONFIGURE_OPTS += --disable-silent-rules
endif
endif

# Separate build directory from the source directory (recommended)
ifeq ($(CONFIGURE_STYLE),haskell)
SEPARATE_BUILD ?= n
endif

ifeq ($(CONFIGURE_STYLE),python)
SEPARATE_BUILD ?= n
endif

ifeq ($(CONFIGURE_STYLE),pyproject)
SEPARATE_BUILD ?= n
endif

ifeq ($(CONFIGURE_STYLE),qmake)
SEPARATE_BUILD ?= n
endif

ifeq ($(CONFIGURE_STYLE),cargo)
SEPARATE_BUILD ?= n
endif

ifeq ($(CONFIGURE_STYLE),perl)
SEPARATE_BUILD ?= n
endif

ifdef SEPARATE_BUILD
SEPARATE_BUILD := $(shell $(parsebool) $(SEPARATE_BUILD))
else
SEPARATE_BUILD ?= y
endif

ifeq ($(SEPARATE_BUILD),n)
WRKBLD := $(WRKSRC)
endif

# Make sure we use the correct perl executable
ifeq ($(CONFIGURE_STYLE),perl)
PERL ?= $(WRKSYSROOT)$(BINDIR)/perl
endif

# Default target
ifeq ($(CONFIGURE_STYLE),python)
BUILD_TARGET ?= build
endif

ifeq ($(CONFIGURE_STYLE),pyproject)
BUILD_TARGET ?= build
endif

# Default command
BUILD_COMMAND ?= $(MAKE)

BUILD_TARGET ?= all

build: patch do-mkdir-wrkdir pre-build do-build post-build
do-build:
	@if [ "$(SKIP)" = "y" ]; then \
		echo "Skipping $(PACKAGE) $(SKIP_REASON), set SKIP=n to override." ; \
	else \
		env WRKDIR='$(WRKDIR)' \
		    WRKSRC='$(WRKSRC)' \
		    WRKBLD='$(WRKBLD)' \
		    WRKSYSROOT='$(WRKSYSROOT)' \
		    PREFIX='$(PREFIX)' \
		    SYSCONFDIR='$(SYSCONFDIR)' \
		    BINDIR='$(BINDIR)' \
		    LIBDIR='$(LIBDIR)' \
		    LIBEXECDIR='$(LIBEXECDIR)' \
		    DATADIR='$(DATADIR)' \
		    INCLUDEDIR='$(INCLUDEDIR)' \
		    SEPARATE_BUILD='$(SEPARATE_BUILD)' \
		    CONFIGURE_STYLE='$(CONFIGURE_STYLE)' \
		    CONFIGURE_ENV='$(CONFIGURE_ENV)' \
		    CONFIGURE_OPTS='$(CONFIGURE_OPTS)' \
		    AUTORECONF='$(AUTORECONF)' \
		    GETTEXTIZE='$(GETTEXTIZE)' \
		    PKG='$(PKG)' \
		    VER='$(VER)' \
		    PACKAGE='$(PACKAGE)' \
		    CFLAGS='$(CFLAGS)' \
		    CXXFLAGS='$(CXXFLAGS)' \
		    LDFLAGS='$(LDFLAGS)' \
		    PYTHON='$(PYTHON)' \
		    BUILD_ENV='$(BUILD_ENV)' \
		    BUILD_TARGET='$(BUILD_TARGET)' \
		    BUILD_ARGS='$(BUILD_ARGS)' \
		    BUILD_COMMAND='$(BUILD_COMMAND)' \
		    PARALLEL_JOBS='$(PARALLEL_JOBS)' \
		    TOPSRC='$(TOPSRC)' \
		    SKIP_BUILD='$(SKIP_BUILD)' \
		    PERL='$(PERL)' \
		    HELPERS='$(HELPERS)' \
		$(build) || exit $$? ; \
	fi

# STAGE: test

ifeq ($(CONFIGURE_STYLE),haskell)
SKIP_TEST ?= y
endif

ifdef SKIP_TEST
SKIP_TEST := $(shell $(parsebool) $(SKIP_TEST))
else
SKIP_TEST ?= n
endif

ifeq ($(SKIP_TEST),n)
ifeq ($(CONFIGURE_STYLE),meson)
TEST_TARGET ?= test
endif

ifeq ($(CONFIGURE_STYLE),perl)
TEST_TARGET ?= test
endif

TEST_TARGET ?= check
endif

TEST_COMMAND ?= $(MAKE)

test: build do-mkdir-wrkdir pre-test do-test post-test
do-test:
	@if [ "$(SKIP)" = "y" ]; then \
		echo "Skipping $(PACKAGE) $(SKIP_REASON), set SKIP=n to override." ; \
	else \
		env WRKDIR='$(WRKDIR)' \
		    WRKSRC='$(WRKSRC)' \
		    WRKBLD='$(WRKBLD)' \
		    PACKAGE='$(PACKAGE)' \
		    CONFIGURE_STYLE='$(CONFIGURE_STYLE)' \
		    PYTHON='$(PYTHON)' \
		    SEPARATE_BUILD='$(SEPARATE_BUILD)' \
		    TEST_ENV='$(TEST_ENV)' \
		    TEST_COMMAND='$(TEST_COMMAND)' \
		    TEST_TARGET='$(TEST_TARGET)' \
		    TEST_ARGS='$(TEST_ARGS)' \
		    SKIP_TEST='$(SKIP_TEST)' \
		    HELPERS='$(HELPERS)' \
		$(test) || exit $$? ; \
	fi

# STAGE: fakeinst

# Default target
ifeq ($(CONFIGURE_STYLE),perl)
FAKEINST_TARGET ?= pure_install
FAKEINST_ENV += PERL_INSTALL_ROOT=$(WRKINST)
endif

ifeq ($(CONFIGURE_STYLE),cargo)
FAKEINST_TARGET ?= .
endif

FAKEINST_TARGET ?= install
FAKEINST_COMMAND ?= $(MAKE)

# Disable the debug package creation?
NO_DEBUG ?= n

# Strip DT_RPATH values unless instructed otherwise
ifdef STRIP_DT_RPATH
STRIP_DT_RPATH := $(shell $(parsebool) $(STRIP_DT_RPATH))
else
STRIP_DT_RPATH ?= y
endif

ifdef SKIP_FAKEINST
SKIP_FAKEINST := $(shell $(parsebool) $(SKIP_FAKEINST))
else
SKIP_FAKEINST ?= n
endif

fakeinst: test do-mkdir-wrkinst pre-fakeinst do-fakeinst post-fakeinst
do-fakeinst:
	@if [ "$(SKIP)" = "y" ]; then \
		echo "Skipping $(PACKAGE) $(SKIP_REASON), set SKIP=n to override." ; \
	else \
		env PKG='$(PKG)' \
		    VER='$(VER)' \
		    WRKDIR='$(WRKDIR)' \
		    WRKDIST='$(WRKDIST)' \
		    WRKSRC='$(WRKSRC)' \
		    WRKBLD='$(WRKBLD)' \
		    PREFIX='$(PREFIX)' \
		    BINDIR='$(BINDIR)' \
		    LIBDIR='$(LIBDIR)' \
		    SEPARATE_BUILD='$(SEPARATE_BUILD)' \
		    WRKSYSROOT='$(WRKSYSROOT)' \
		    WRKINST='$(WRKINST)' \
		    WRKDEBUG='$(WRKDEBUG)' \
		    DEBUGDIR='$(DEBUGDIR)' \
		    NO_DEBUG='$(NO_DEBUG)' \
		    CONFIGURE_STYLE='$(CONFIGURE_STYLE)' \
		    PYTHON='$(PYTHON)' \
		    PKG='$(PKG)' \
		    VER='$(VER)' \
		    PACKAGE='$(PACKAGE)' \
		    FAKEINST_ENV='$(FAKEINST_ENV)' \
		    FAKEINST_TARGET='$(FAKEINST_TARGET)' \
		    FAKEINST_ARGS='$(FAKEINST_ARGS)' \
		    DATADIR='$(DATADIR)' \
		    MANDIR='$(MANDIR)' \
		    INFODIR='$(INFODIR)' \
		    DOCDIR='$(DOCDIR)' \
		    LICENSEDIR='$(LICENSEDIR)' \
		    FAKEINST_COMMAND='$(FAKEINST_COMMAND)' \
		    MAKE='$(MAKE)' \
		    TOPSRC='$(TOPSRC)' \
		    TARGET_NAME='$(TARGET_NAME)' \
		    TARGET_MACH='$(TARGET_MACH)' \
		    INCLUDEDIR='$(INCLUDEDIR)' \
		    TARGET_MACH='$(TARGET_MACH)' \
		    STRIP_DT_RPATH='$(STRIP_DT_RPATH)' \
		    DOCS='$(DOCS)' \
		    LICENSES='$(LICENSES)' \
		    IGNORE_REQS='$(IGNORE_REQS)' \
		    SKIP_FAKEINST='$(SKIP_FAKEINST)' \
		    HELPERS='$(HELPERS)' \
		$(fakeinst) || exit $$? ; \
	fi

# STAGE: package
package: fakeinst do-mkdir-release pre-package do-package post-package
do-package:
	@if [ "$(SKIP)" = "y" ]; then \
		echo "Skipping $(PACKAGE) $(SKIP_REASON), set SKIP=n to override." ; \
	else \
		env WRKDIR='$(WRKDIR)' \
		    WRKINST='$(WRKINST)' \
		    PKG='$(PKG)' \
		    VER='$(VER)' \
		    REL='$(REL)' \
		    SUBPACKAGES='$(SUBPACKAGES)' \
		    TARGET_NAME='$(TARGET_NAME)' \
		    TARGET_MACH='$(TARGET_MACH)' \
		    TOPSRC='$(TOPSRC)' \
		    RELEASE_DIST='$(RELEASE_DIST)' \
		    WRKDEBUG='$(WRKDEBUG)' \
		    PKGCOLLECTION='$(PKGCOLLECTION)' \
		    NAME_PREFIX_CONFLICT='$(NAME_PREFIX_CONFLICT)' \
		    DIGEST_ALGORITHM='$(DIGEST_ALGORITHM)' \
		    DIST_URL='$(DIST_URL)' \
		    DIST_DOWNLOAD='$(DIST_DOWNLOAD)' \
		    PACKAGE='$(PACKAGE)' \
		    COMPRESSION='$(COMPRESSION)' \
		    DESCRIPTION='$(DESCRIPTION)' \
		    MAINTAINER='$(MAINTAINER)' \
		    LICENSE='$(LICENSE)' \
		    DEPSPATH='$(DEPSPATH)' \
		    HELPERSDB='$(HELPERSDB)' \
		$(package) || exit $$? ; \
	fi

# STAGE: clean
clean: pre-clean do-clean post-clean
do-clean:
ifeq ($(TOP_LEVEL),y)
	-find $(OBJ)/ -maxdepth 1 -type d -path $(OBJ)/ -prune -o -exec $(SUDO) find {} -delete \;
	-$(SUDO) rm -rf $(OBJ)/*
else
	-rm -rf $(WRKDIR)
endif

ispkg:
	@echo "yes"

# Show build system configuration information
show-config:
	@$(ECHO) "Peachtree Linux Build System version $(DISTRIBUTION_VERSION)"
	@echo
	@$(ECHO) "Target arch:   $(TARGET_MACH)"
	@$(ECHO) "Target name:   $(TARGET_NAME)"
	@$(ECHO) "Object dir:    $(OBJ)"
	@$(ECHO) "Release dir:   $(RELEASE)"
	@$(ECHO) "Distfiles dir: $(DISTFILES)"

help:
ifeq ($(TOP_LEVEL),y)
	$(MAKE) _world_help
else
	@echo
	@$(ECHO) "Peachtree Linux package building system"
	@echo
	@$(ECHO) "Package building consists of multiple stages, which can be supplemented with"
	@$(ECHO) "leaf stages.  The following table shows what stages there are and which you"
	@$(ECHO) "can override:"
	@echo
	@$(ECHO) "    +---------------+------------+-----------+"
	@$(ECHO) "    |               |   LEAFS    |           |"
	@$(ECHO) "    | STAGE         | PRE | POST | OVERRIDE? |"
	@$(ECHO) "    +---------------+-----+------+-----------+"
	@$(ECHO) "    | fetch         | yes | yes  | yes       |"
	@$(ECHO) "    | checksum      | no  | no   | no        |"
	@$(ECHO) "    | extract       | yes | yes  | yes       |"
	@$(ECHO) "    | patch         | yes | yes  | no        |"
	@$(ECHO) "    | build         | yes | yes  | yes       |"
	@$(ECHO) "    | test          | yes | yes  | yes       |"
	@$(ECHO) "    | fakeinst      | yes | yes  | yes       |"
	@$(ECHO) "    | package       | yes | yes  | no        |"
	@$(ECHO) "    | clean         | yes | yes  | yes       |"
	@$(ECHO) "    | update-plist  | yes | yes  | no        |"
	@$(ECHO) "    | take-upstream | yes | yes  | no        |"
	@$(ECHO) "    +---------------+-----+------+-----------+"
	@echo
	@$(ECHO) "The different stages have a dependent order, specifying one stage will"
	@$(ECHO) "invoke dependent stages in order.  The general order is as they are shown"
	@$(ECHO) "in the table.  Each stage with a pre or post leaf will have that target run"
	@$(ECHO) "either before or after the main stage target, respectively."
	@echo
	@$(ECHO) "If you override one of the stages, the leaf stages are disabled unless you"
	@$(ECHO) "explicitly set them up in your package Makefile."
	@echo
	@$(ECHO) "Typing 'make' with no target specified will run through the build target."
	@echo
	@$(ECHO) "Stage targets:"
	@$(ECHO) "    fetch            Fetch the source files listed in 'archives' or specified"
	@$(ECHO) "                     using GITREPO and GITCOMMIT variables; store in"
	@$(ECHO) "                     $(DISTFILES).  Run as 'env FORCE_REFRESH=y make fetch'"
	@$(ECHO) "                     to skip any mirrors and download from the upstream"
	@$(ECHO) "                     location."
	@$(ECHO) "    checksum         Verify checksums of downloaded source files when applicable."
	@$(ECHO) "    extract          Extract source files to $(WRKDIR)."
	@$(ECHO) "    patch            Apply patches to source tree."
	@$(ECHO) "    build            Configure and build software."
	@$(ECHO) "    test             Run test suite included with $(PACKAGE)."
	@$(ECHO) "    fakeinst         Install software to $(WRKINST)."
	@$(ECHO) "    package          Generate an installable package for $(PACKAGE)."
	@$(ECHO) "    clean            Remove $(WRKDIR)."
	@echo
	@$(ECHO) "User-provided leaf targets:"
	@$(ECHO) "    pre-fetch        Runs before 'fetch'."
	@$(ECHO) "    post-fetch       Runs after 'fetch'."
	@$(ECHO) "    pre-extract      Runs before 'extract'."
	@$(ECHO) "    post-extract     Runs after 'extract'."
	@$(ECHO) "    pre-patch        Runs before 'patch'."
	@$(ECHO) "    post-patch       Runs after 'patch'."
	@$(ECHO) "    pre-build        Runs before 'build'."
	@$(ECHO) "    post-build       Runs after 'build'."
	@$(ECHO) "    pre-test         Runs before 'test'."
	@$(ECHO) "    post-test        Runs after 'test'."
	@$(ECHO) "    pre-fakeinst     Runs before 'fakeinst'."
	@$(ECHO) "    post-fakeinst    Runs after 'fakeinst'."
	@$(ECHO) "    pre-package      Runs before 'package'."
	@$(ECHO) "    post-package     Runs after 'package'."
	@$(ECHO) "    pre-clean        Runs before 'clean'."
	@$(ECHO) "    post-clean       Runs after 'clean'."
	@echo
	@$(ECHO) "Helper targets for packagers:"
	@$(ECHO) "    makesum          Generate new checksums in 'archives' for new source"
	@$(ECHO) "                     files."
	@$(ECHO) "    cleansum         Clean out the 'archives' file and leave just filenames"
	@$(ECHO) "                     and #include lines.  This is a helper target when updating"
	@$(ECHO) "                     a package to a new upstream version."
	@$(ECHO) "    update-patches   Walk source tree and generate any new patches for files"
	@$(ECHO) "                     ending with .orig."
	@$(ECHO) "    update-keyring   Update the GPG keyring file if GPGKEYS is set."
	@$(ECHO) "    update-plist     Update the packing list file."
	@$(ECHO) "    ci               Do update-plist and then git commit."
	@$(ECHO) "    list-wrkinst     Show contents of $(WRKINST)"
	@$(ECHO) "    show var=VAR     Show the value of variable VAR."
	@$(ECHO) "                     e.g., make show var=WRKDIR"
	@$(ECHO) "    bumprel          Increment value of REL by 1 (requires REL be an integer)"
	@$(ECHO) "    verrel           Show the package triple PKG-VER-REL."
	@$(ECHO) "    license          Show licensing information for this package."
	@$(ECHO) "    prune-distfiles  Remove old distfiles, keep current ones."
	@$(ECHO) "    check-upstream   Check for a new upstream release."
	@$(ECHO) "    take-upstream    Apply new upstream info to package source files."
	@$(ECHO) "    show-buildreqs   Show the packages required to build this package."
	@$(ECHO) "    reduce-buildreqs Shrink existing buildreqs file to minimum required."
	@$(ECHO) "    missingreqs      Recursively run 'make world' for any missing buildreqs."
	@$(ECHO) "    bundle           Run 'make world' and create an update bundle."
	@$(ECHO) "    showprevpkg      Show the previous PKG-VER-REL.TARGET_MACH string."
	@echo
	@$(ECHO) "The business end of package maintenance:"
	@$(ECHO) "    world            Peform all steps through 'make package' but in the style"
	@$(ECHO) "                     of a top level world build.  Requires a populated release"
	@$(ECHO) "                     tree with built packages to satisfy the buildreqs."
	@$(ECHO) "    quickworld       Exactly like world, but does not wipe and reinstall the"
	@$(ECHO) "                     sysroot tree.  Just uses the one in place already."
endif
