/* Check the integrity of an installed dist package. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdlib.h>
#include <stdint.h>
#include <assert.h>
#include <string.h>
#include <err.h>
#include "dist-tools.h"
#include "db.h"

/*
 * Perform integrity checks on the specified package,
 * optionally repair database.
 */
int checkpkg(char *dbpath, struct dist *package, char *prepend)
{
    int ret = 0;
    char *query = NULL;
    char *errmsg = NULL;
    sqlite3 *db = NULL;
    sqlite3_stmt *stmt = NULL;
    struct finfo *fi = NULL;

    assert(dbpath != NULL);
    assert(package != NULL);

    /* open a connection to our database */
    db = open_db(dbpath);

    if (db == NULL) {
        return -1;
    }

    /* iterate and check contents */
    xasprintf(&query, PACKAGE_CONTENTS, package->id);

    if (sqlite3_prepare(db, query, -1, &stmt, 0) != SQLITE_OK) {
        warnx("sqlite3_prepare: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -2;
    }

    while (sqlite3_step(stmt) == SQLITE_ROW) {
        /* allocate a new struct finfo for each entry */
        fi = calloc(1, sizeof(*fi));
        assert(fi != NULL);

        /* build a struct finfo for each package entry */
        if (read_finfo(package->id, package->sigver, stmt, fi, prepend)) {
            warn("read_finfo");
            ret = -3;
        }

        /* check this package entry */
        if (checkfile(fi, prepend, true)) {
            ret = -4;
        }

        /* reset the struct finfo */
        free_finfo(fi);
    }

    if (sqlite3_finalize(stmt) != SQLITE_OK) {
        warnx("sqlite3_finalize: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -5;
    }

    sqlite3_free(errmsg);
    sqlite3_close(db);
    free(query);
    return ret;
}
