/*
 * Calculate an MD5, SHA-1, SHA-256, SHA-384, or SHA-512 checksum for
 * a file.
 */

/*
 * Copyright David Shea <david@reallylongword.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * Copyright David Cantrell <dcantrell@burdell.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <err.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include "dist-tools.h"

/*
 * Take in a file, return a checksum.
 * NOTE: The caller is responsible for freeing the string returned by
 *       this function.
 */
char *checksum(const char *filename, enum checksum type)
{
    struct stat sb;
    unsigned char buf[BUFSIZ];
    int input = 0;
    int len = 0;
    int i = 0;
    char *r = NULL;
    unsigned char digest[BUFSIZ];
    MD5_CTX md5c;
    SHA_CTX sha1c;
    SHA256_CTX sha256c;
    SHA512_CTX sha512c;

    /* don't calculate the checksum of a device node */
    if (lstat(filename, &sb) != 0) {
        return NULL;
    }

    if (S_ISCHR(sb.st_mode) || S_ISBLK(sb.st_mode) || S_ISFIFO(sb.st_mode) || S_ISSOCK(sb.st_mode)) {
        warnx("cannot calculate checksum on devices or FIFOs");
        return NULL;
    }

    /*
     * This holds all possible digest type computations.  if any
     * future digest is larger than BUFSIZ, hopefully we update this
     * and get faster systems.
     */
    memset(&buf, '\0', BUFSIZ);
    memset(&digest, '\0', BUFSIZ);

    /* Initialize the correct context based on the checksum type. */
    if (type == MD5SUM) {
        MD5_Init(&md5c);
    } else if (type == SHA1SUM) {
        SHA1_Init(&sha1c);
    } else if (type == SHA256SUM) {
        SHA256_Init(&sha256c);
    } else if (type == SHA384SUM) {
        SHA384_Init(&sha512c);
    } else if (type == SHA512SUM) {
        SHA512_Init(&sha512c);
    }

    /* read in the file to generate the requested checksum */
    if ((input = open(filename, O_RDONLY)) == -1) {
        warn("open");
        return NULL;
    }

    len = read(input, buf, sizeof(buf));

    if (len == -1) {
        warn("read");
        return NULL;
    }

    while (len > 0) {
        /* update the correct context based on the checksum type */
        if (type == MD5SUM) {
            MD5_Update(&md5c, buf, len);
        } else if (type == SHA1SUM) {
            SHA1_Update(&sha1c, buf, len);
        } else if (type == SHA256SUM) {
            SHA256_Update(&sha256c, buf, len);
        } else if (type == SHA384SUM) {
            SHA384_Update(&sha512c, buf, len);
        } else if (type == SHA512SUM) {
            SHA512_Update(&sha512c, buf, len);
        }

        len = read(input, buf, sizeof(buf));

        if (len == -1) {
            warn("read");
            return NULL;
        }
    }

    if (close(input) == -1) {
        warn("close");
        return NULL;
    }

    /* finalize the correct context type and determine the loop length */
    if (type == MD5SUM) {
        MD5_Final(digest, &md5c);
        len = MD5_DIGEST_LENGTH;
    } else if (type == SHA1SUM) {
        SHA1_Final(digest, &sha1c);
        len = SHA_DIGEST_LENGTH;
    } else if (type == SHA256SUM) {
        SHA256_Final(digest, &sha256c);
        len = SHA256_DIGEST_LENGTH;
    } else if (type == SHA384SUM) {
        SHA384_Final(digest, &sha512c);
        len = SHA384_DIGEST_LENGTH;
    } else if (type == SHA512SUM) {
        SHA512_Final(digest, &sha512c);
        len = SHA512_DIGEST_LENGTH;
    }

    /* this is our human readable digest, caller must free */
    r = calloc(1, (sizeof(char) * len * 2) + 1);

    if (r == NULL) {
        warn("calloc");
        return NULL;
    }

    for (i = 0; i < len; ++i) {
        sprintf(&r[i*2], "%02x", (unsigned int) digest[i]);
    }

    return r;
}
