/* Generic copy file function. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright David Shea <david@reallylongword.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <libgen.h>
#include <limits.h>
#include <err.h>
#include "dist-tools.h"

int copyfile(const char *src, const char *dest, bool force, bool verbose)
{
    FILE *in = NULL;
    FILE *out = NULL;
    int out_fd;                 /* use open() to test if file exists, then
                                 * fdopen for out */
    char buf[BUFSIZ];
    size_t i;
    ssize_t l;
    int success = 0;
    char *destpath = NULL;
    char *destdir = NULL;
    char linkdest[PATH_MAX];
    int oflags = O_WRONLY | O_CREAT | O_EXCL;
    mode_t mode = (S_IRWXU | S_IRGRP | S_IROTH) ^ S_IXUSR;
    struct stat sb;

    assert(src != NULL);
    assert(dest != NULL);

    /* stat the source */
    if (lstat(src, &sb) == -1) {
        warn("lstat");
        return -1;
    }

    /* create destination directory if it doesn't exist */
    destpath = strdup(dest);
    destdir = dirname(destpath);

    if (mkdirp(destdir, S_IRWXU) == -1) {
        warn("mkdirp");
        return -1;
    }

    free(destpath);

    /* if src is a symlink, handle it here */
    if (S_ISLNK(sb.st_mode)) {
        l = readlink(src, linkdest, PATH_MAX);

        if (l == -1) {
            warn("readlink");
            return -1;
        }

        /* make sure our readlink value is NULL terminated */
        linkdest[l >= PATH_MAX ? PATH_MAX - 1 : l] = '\0';

        if (symlink(linkdest, dest) == -1) {
            warn("symlink");
            return -1;
        }

        return 0;
    }

    /* copy src to dest */
    in = fopen(src, "r");

    if (in == NULL) {
        warn("fopen");
        return -1;
    }

    out_fd = open(dest, oflags, mode);

    if (out_fd == -1) {
        if (errno == EEXIST) {
            if (verbose) {
                fprintf(stderr, "*** %s already exists", dest);
            }

            if (force) {
                if (verbose) {
                    fprintf(stderr, ", overwriting\n");
                }

                if (remove(dest)) {
                    warn("remove");
                    return -1;
                } else {
                    out_fd = open(dest, oflags, mode);

                    if (out_fd == -1) {
                        warn("open");
                    }
                }
            } else {
                putc('\n', stderr);
                return -1;
            }
        } else {
            warn("open");
            return -1;
        }
    }

    out = fdopen(out_fd, "wb");

    if (out == NULL) {
        warn("fdopen");
        return -1;
    }

    while ((i = fread(buf, sizeof(char), BUFSIZ, in)) > 0) {
        if (fwrite(buf, sizeof(char), i, out) != i) {
            warn("fwrite");
            success = -1;
            break;
        }
    }

    if (fflush(out) != 0) {
        warn("fflush");
        success = -1;
    }

    if (fclose(out) != 0) {
        warn("fclose");
        success = -1;
    }

    if (fclose(in) != 0) {
        warn("fclose");
        success = -1;
    }

    if (success != 0) {
        remove(dest);
    }

    /* set ownerships and permissions */
    if (geteuid() == 0 && (chown(dest, sb.st_uid, sb.st_gid) == -1)) {
        warn("chown");
        success = -1;
    }

    if (chmod(dest, (sb.st_mode & S_INTER)) == -1) {
        warn("chmod");
        success = -1;
    }

    return success;
}

/* vim: set shiftwidth=3 softtabstop=3: */
