/* sqlite helper functions */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <err.h>
#include <sqlite3.h>

#include "dist-tools.h"

/*
 * Given a string representing a file path to an sqlite database, open
 * it and set the options for dist-tools.  The caller is responsible
 * for closing the database handle returned.
 */
sqlite3 *open_db(const char *dbpath)
{
    sqlite3 *db = NULL;
    int flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE | SQLITE_OPEN_NOMUTEX;
    char *errmsg = NULL;

    if (dbpath == NULL) {
        return NULL;
    }

    /* open a connection to our database */
    if (sqlite3_open_v2(dbpath, &db, flags, NULL) != SQLITE_OK) {
        warn("sqlite3_open: %d: %s", sqlite3_errcode(db), sqlite3_errmsg(db));
        return NULL;
    }

    /*
     * Various runtime options for libsqlite:
     * https://www.sqlite.org/pragma.html
     *
     * cache_size      The default is 2M of sqlite pages retained in
     *                 memory.  A negative value of -N means abs(N)*1024.  The
     *                 default is -2000.  A positive number is the number of
     *                 pages to use.
     * synchronous     EXTRA, FULL, NORMAL, or OFF.  How much libsqlite should
     *                 care about sync(2) operations.  OFF is not at all.
     * journal_mode    Several options about how libsqlite should handle the
     *                 transaction journal.  MEMORY means just do stuff in RAM
     *                 to favor speed over database integrity.  A crash mid-op
     *                 may leave the database file corrupt.  DELETE is the
     *                 default mode.
     */
    if (sqlite3_exec(db, "PRAGMA cache_size = 0; PRAGMA synchronous = FULL; PRAGMA journal_mode = DELETE;", NULL, NULL, &errmsg) != SQLITE_OK) {
        warn(errmsg);
        sqlite3_free(errmsg);
        sqlite3_close(db);
        return NULL;
    }

    sqlite3_free(errmsg);

    return db;
}
