/* struct finfo helper functions */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdlib.h>
#include <assert.h>
#include <err.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <endian.h>
#include "dist-tools.h"

/*
 * Build a struct finfo from the current row in an sqlite3_stmt.
 * If 'id' is <0, read the id from the sqlite3 row.
 * If 'id' is >=0, use this value for the package id.
 */
int read_finfo(int id, sigver_t sigver, sqlite3_stmt *stmt, struct finfo *fi, char *prepend)
{
    mode_t mode = 0;

    assert(stmt != NULL);
    assert(fi != NULL);

    /* finfo needs a struct stat, so build one sort of */
    if ((fi->attr = calloc(1, sizeof(struct stat))) == NULL) {
        warn("calloc");
        return -1;
    }

    if (id < 0) {
        fi->packageid = sqlite3_column_int(stmt, 0);
    } else {
        fi->packageid = id;
    }

    fi->filename = strdup((char *) sqlite3_column_text(stmt, 1));
    fi->owner = strdup((char *) sqlite3_column_text(stmt, 2));
    fi->attr->st_uid = (uid_t) sqlite3_column_int(stmt, 3);
    fi->group = strdup((char *) sqlite3_column_text(stmt, 4));
    fi->attr->st_gid = (gid_t) sqlite3_column_int(stmt, 5);

    memset(&(fi->attr->st_mode), 0, sizeof(fi->attr->st_mode));

    if (sigver == SIGVER_21) {
        /*
         * This is a known bug in v2.1 packages.  If you build a noarch
         * package, the host you install it on must match the endianness
         * of the build host.  That's why for v2.2 packages, we store the
         * mode in big endian format in the packages.
         */
        memmove(&(fi->attr->st_mode), sqlite3_column_blob(stmt, 6), sqlite3_column_bytes(stmt, 6));
    } else if (sigver >= SIGVER_22) {
        memmove(&mode, sqlite3_column_blob(stmt, 6), sqlite3_column_bytes(stmt, 6));
        mode = be32toh(mode);
        memmove(&(fi->attr->st_mode), &mode, sizeof(mode));
    } else {
        errx(EXIT_FAILURE, "unknown package signature %d", sigver);
    }

    fi->attr->st_size = (off_t) sqlite3_column_int(stmt, 7);
    fi->attr->st_mtime = sqlite3_column_int(stmt, 8);
    fi->type = sqlite3_column_int(stmt, 9);

    if (fi->type == TYPE_HARD || fi->type == TYPE_SOFT) {
        fi->linkname = strdup((char *) sqlite3_column_text(stmt, 10));
    } else {
        fi->linkname = NULL;
    }

    if (fi->type == TYPE_REG) {
        fi->checksum = strdup((char *) sqlite3_column_text(stmt, 11));
    } else {
        fi->checksum = NULL;
    }

    fi->fullpath = get_path(prepend, fi->filename);
    fi->refcnt = 0;

    return 0;
}

/*
 * Helper to make sure all of a struct finfo is properly freed.
 * Useful as a callback for glib data structures.
 */
void free_finfo(struct finfo *info)
{
    if (info == NULL) {
        return;
    }

    free(info->filename);
    free(info->fullpath);
    free(info->checksum);
    free(info->attr);
    free(info->linkname);
    free(info->owner);
    free(info->group);
    free(info);

    return;
}
