/* finfolist_t helper functions */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <assert.h>
#include <string.h>
#include "dist-tools.h"
#include "finfolist.h"

/*
 * Allocate memory for and initialize a new finfolist_t.
 */
finfolist_t *init_finfolist(void)
{
    finfolist_t *list = NULL;

    list = calloc(1, sizeof(*list));
    assert(list != NULL);
    TAILQ_INIT(list);
    return list;
}

/*
 * Given an allocated finfolist_t, free all the memory associated with it.
 */
void free_finfolist(finfolist_t *list)
{
    finfolist_entry_t *entry = NULL;

    if (list == NULL) {
        return;
    }

    while (!TAILQ_EMPTY(list)) {
        entry = TAILQ_FIRST(list);
        TAILQ_REMOVE(list, entry, items);

        if (entry->info) {
            entry->info->refcnt--;

            if (entry->info->refcnt == 0) {
                free_finfo(entry->info);
                entry->info = NULL;
            }
        }

        free(entry);
    }

    free(list);

    return;
}

/*
 * Return the entry count for the given finfolist.
 */
int length_finfolist(finfolist_t *list)
{
    int c = 0;
    finfolist_entry_t *entry = NULL;

    if (list == NULL) {
        return 0;
    }

    TAILQ_FOREACH(entry, list, items) {
        c++;
    }

    return c;
}

/*
 * Returns the entry specified by the filename if it is found in
 * the finfolist_t.  Return NULL if not found.
 */
finfolist_entry_t *find_finfolist_filename(finfolist_t *list, const char *filename)
{
    finfolist_entry_t *entry = NULL;

    assert(list != NULL);
    assert(filename != NULL);

    TAILQ_FOREACH(entry, list, items) {
        if (!strcmp(entry->info->filename, filename)) {
            return entry;
        }
    }

    return NULL;
}

/*
 * Remove the finfolist_entry_t from the list that matches the filename.
 */
finfolist_t *remove_finfolist_filename(finfolist_t *list, const char *filename)
{
    finfolist_entry_t *entry = NULL;

    assert(list != NULL);
    assert(filename != NULL);

    if (find_finfolist_filename(list, filename) != NULL) {
        return list;
    }

    TAILQ_FOREACH(entry, list, items) {
        if (!strcmp(entry->info->filename, filename)) {
            TAILQ_REMOVE(list, entry, items);
            entry->info->refcnt--;

            if (entry->info->refcnt == 0) {
                free(entry->info);
                entry->info = NULL;
            }

            free(entry);
            return list;
        }
    }

    return list;
}

/*
 * Insert an finfolist entry in sorted order.  Used to build finfolists
 * sorted by depth order of the pathname.  The caller must take care to
 * not free the struct finfo passed to this function because it will be
 * added to the list specified.
 */
finfolist_t *sorted_finfolist_insert(finfolist_t *list, struct finfo *info)
{
    int i = 0;
    finfolist_entry_t *iterator = NULL;
    finfolist_entry_t *entry = NULL;

    assert(info != NULL);

    /* do not insert if we already have it in the list */
    if (find_finfolist_filename(list, info->filename) != NULL) {
        return list;
    }

    if (list == NULL) {
        list = init_finfolist();
        assert(list != NULL);
    }

    entry = calloc(1, sizeof(*entry));
    assert(entry != NULL);
    entry->info = info;
    entry->info->refcnt++;

    TAILQ_FOREACH(iterator, list, items) {
        i = strcmp(iterator->info->filename, entry->info->filename);

        if (i < 0) {
            TAILQ_INSERT_AFTER(list, iterator, entry, items);
            return list;
        } else if (i > 0) {
            TAILQ_INSERT_BEFORE(iterator, entry, items);
            return list;
        } else if (i == 0) {
            /* duplicates probably shouldn't happen, but just in case */
            return list;
        }
    }

    /* Handle the empty list state */
    TAILQ_INSERT_TAIL(list, entry, items);

    return list;
}
