/*
 * Walk a directory tree and read attributes and generate checksums.
 * Also generate package dependencies.
 */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright David Shea <david@reallylongword.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <err.h>
#include <ftw.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <libelf.h>
#include <gelf.h>
#include "dist-tools.h"
#include "plist.h"
#include "strlist.h"
#include "signature.h"
#include "db.h"

extern char **environ;

/* We use this struct to pass in more info to the nftw() callback. */
struct attrsinfo {
    sqlite3_stmt *stmt;
    plist_t *plist;
    sigver_t sigver;
    enum checksum sumtype;
    bool debug;
};

static strlist_t *deplist = NULL;
static strlist_t *provlist = NULL;

/* pointers to caller-provided lists, do not free */
static strlist_t **manifest = NULL;
static strlist_t **symlinks = NULL;

static int trimlen = -1;
static struct attrsinfo ai;
static char *prepend_path = NULL;

/* Local prototypes */
static void insert_name(const char *, sqlite3_stmt *);
static bool elf_reader(const char *, const int, strlist_t **);
static int _add_attrs_entry(const char *, const struct stat *, int, struct FTW *);

/*
 * Insert an SONAME element in to an SQLite table.  The first
 * parameter is the string to insert in the database.  The second
 * parameter is the prepared statement ready to bind values to and
 * execute.  This function will bind the string to the statement, run
 * it, then unbind and reset the prepared statement for another call.
 * The caller needs to set up the statement.
 */
static void insert_name(const char *soname, sqlite3_stmt *stmt)
{
    assert(soname != NULL);
    assert(stmt != NULL);

    if (sqlite3_bind_text(stmt, 1, soname, strlen(soname), SQLITE_STATIC) != SQLITE_OK) {
        errx(EXIT_FAILURE, "sqlite3_bind_text");
    }

    if (sqlite3_step(stmt) != SQLITE_DONE) {
        errx(EXIT_FAILURE, "sqlite3_step");
    }

    if (sqlite3_reset(stmt) != SQLITE_OK) {
        errx(EXIT_FAILURE, "sqlite3_reset");
    }

    if (sqlite3_clear_bindings(stmt) != SQLITE_OK) {
        errx(EXIT_FAILURE, "sqlite3_clear_bindings");
    }

    return;
}

/*
 * ELF object reading support function.  This function should
 * be called when the current working directory is the location
 * of 'filename'.  This function does not change directories.
 */
static bool elf_reader(const char *filename, const int type, strlist_t **list)
{
    int fd = -1;
    int elfret = 0;
    Elf *elf = NULL;
    GElf_Ehdr ehdr;
    Elf_Scn *scn = NULL;
    strlist_entry_t *entry = NULL;

    if (filename == NULL) {
        return false;
    }

    /* skip non-ELF files */
    if (!is_elf(filename)) {
        return true;
    }

    elf_version(EV_CURRENT);

    if ((fd = open(filename, O_RDONLY)) == -1) {
        warn("open");
        return false;
    }

    elf = elf_begin(fd, ELF_C_READ_MMAP, NULL);

    if (gelf_getehdr(elf, &ehdr) == NULL) {
        warn("gelf_getehdr");
        return false;
    }

    /* only collect DT_NEEDED from ELF dynamic shared objects */
    if (ehdr.e_type != ET_EXEC && ehdr.e_type != ET_DYN) {
        elfret = elf_end(elf);
        assert(elfret == 0);

        if (close(fd) == -1) {
            warn("close");
            return false;
        }

        return true;
    }

    while ((scn = elf_nextscn(elf, scn)) != NULL) {
        char *name = NULL;
        int ndx, maxndx;
        GElf_Shdr shdr;
        Elf_Data *data = NULL;
        GElf_Dyn dyn;

        if (gelf_getshdr(scn, &shdr) == NULL) {
            continue;
        }

        if (shdr.sh_type != SHT_DYNAMIC) {
            continue;
        }

        if ((data = elf_getdata(scn, data)) == NULL || data->d_size == 0) {
            /* no data to read, go to next section */
            continue;
        }

        maxndx = data->d_size / shdr.sh_entsize;

        for (ndx = 0; ndx < maxndx; ++ndx) {
            if (gelf_getdyn(data, ndx, &dyn) == NULL) {
                continue;
            }

            if (dyn.d_tag == type) {
                /*
                 * This took a while to figure out since the elfutils
                 * documentation doesn't really explain anything.  The
                 * important thing to know is that all of the sections
                 * in an ELF file just contain offsets and addresses to
                 * useful data like strings such as "libc.so.6".
                 *
                 * Here we iterate over each section and get the GElf_Dyn
                 * object from the section.  If it's a DT_NEEDED dyn
                 * object, then it represents a shared library we need
                 * at runtime.  If it's DT_SONAME, it's the name of this
                 * library (there should really only be one SONAME per
                 * ELF object, but we won't assume that).
                 *
                 * To get the library name, we use elf_strptr() here to
                 * extract the string from the big string table.  The
                 * string table is pointed to by shdr.sh_link and the
                 * offset to the library name is given by dyn.d_un.d_ptr.
                 *
                 * And this is a standard?
                 *
                 * Also, we strdup() here because we need a copy of the
                 * string in the strlist_t.  It will be free()'ed later.
                 */
                name = elf_strptr(elf, shdr.sh_link, (size_t) (dyn.d_un.d_ptr));

                /* we need to initialize the list on first use */
                if (*list == NULL) {
                    *list = init_strlist();
                }

                if (!is_str_in_list(*list, name)) {
                    entry = calloc(1, sizeof(*entry));
                    assert(entry != NULL);
                    entry->val = strdup(name);
                    TAILQ_INSERT_TAIL(*list, entry, items);
                }
            }
        }
    }

    if (elf_end(elf) != 0) {
        warn("elf_end");
        return false;
    }

    if (close(fd) == -1) {
        warn("close");
        return false;
    }

    return true;
}

/*
 * Adds an individual entry to the manifest and symlinks lists.
 */
static int _add_lists_entry(const char *fpath, const struct stat *sb, __attribute__((unused)) int tflag, __attribute__((unused)) struct FTW *ftwbuf)
{
    strlist_t *list = NULL;
    strlist_entry_t *entry = NULL;

    if (S_ISREG(sb->st_mode)) {
        list = *manifest;
    } else if (S_ISLNK(sb->st_mode)) {
        list = *symlinks;
    } else {
        return 0;
    }

    if (!is_str_in_list(list, fpath)) {
        entry = calloc(1, sizeof(*entry));
        entry->val = strdup(fpath);
        TAILQ_INSERT_TAIL(list, entry, items);
    }

    return 0;
}

/*
 * Compute attribute values for the specified entry.
 * NOTE: Relies on the global 'ai' struct, initialized by genattrs().
 */
static int _add_attrs_entry(const char *fpath, const struct stat *sb, __attribute__((unused)) int tflag, __attribute__((unused)) struct FTW *ftwbuf)
{
    int r = 0;
    const char *vpath = NULL;
    struct finfo *info = NULL;
    plist_entry_t *plistentry = NULL;
    plist_entry_t *iterator = NULL;
    strlist_entry_t *strentry = NULL;

    assert(manifest != NULL);
    assert(symlinks != NULL);

    if (trimlen == -1) {
        trimlen = strlen(fpath);
        return 0;
    } else {
        vpath = fpath + trimlen;

        while (*vpath == '/') {
            vpath++;
        }
    }

    /* initialize a new finfo */
    info = calloc(1, sizeof(*info));
    assert(info != NULL);

    /* initialize finfo as other to avoid hard link checks */
    info->type = TYPE_OTHER;
    info->checksum = NULL;
    info->linkname = NULL;
    info->filename = strdup(vpath);
    info->fullpath = strdup(fpath);
    info->owner = NULL;
    info->group = NULL;
    info->has_checksum = false;

    info->attr = calloc(1, sizeof(struct stat));
    assert(info->attr != NULL);
    memmove(info->attr, sb, sizeof(struct stat));

    /* grab the attributes */
    r = gen_contents(info, ai.sumtype, prepend_path);

    /* override attributes from stat(2) with plist values */
    TAILQ_FOREACH(iterator, ai.plist, items) {
        if (!strcmp(iterator->filename, vpath)) {
            plistentry = iterator;
            break;
        }
    }

    if (plistentry != NULL) {
        if (!plistentry->is_symlink) {
            info->attr->st_mode = plistentry->mode;
        }

        info->owner = strdup(plistentry->owner);
        info->attr->st_uid = plistentry->uid;

        info->group = strdup(plistentry->group);
        info->attr->st_gid = plistentry->gid;

        /* we could have a checksum already computed by gen_contents() */
        if (!plistentry->has_checksum) {
            free(info->checksum);
            info->checksum = strdup("*");
        }

        info->has_checksum = plistentry->has_checksum;
    }

    /* gather dependency information */
    if (info->type == TYPE_REG && !ai.debug) {
        if (!elf_reader(vpath, DT_NEEDED, &deplist)) {
            warn("error checking file dependencies");
            r = -1;
        }

        /* remove dependencies provided by this package */
        if (length_strlist(deplist) > 0) {
            TAILQ_FOREACH(strentry, deplist, items) {
                if (is_str_in_list(*manifest, strentry->val) || is_str_in_list(*symlinks, strentry->val)) {
                    TAILQ_REMOVE(deplist, strentry, items);
                    free(strentry->val);
                    strentry->val = NULL;
                    free(strentry);
                }
            }
        }

        /* collect SONAMEs provided by this package */
        if (!elf_reader(vpath, DT_SONAME, &provlist)) {
            warn("error checking file provides");
            r = -1;
        }
    }

    /* write out infos */
    if (write_one_entry(-1, info, ai.stmt, ai.sigver)) {
        warn("unable to write info for %s", vpath);
        r = -1;
    }

    free_finfo(info);
    return r;
}

/*
 * Exposed call for programs that want to build the manifest and
 * symlinks lists for a package staging root.
 */
int makelists(const char *dir, strlist_t **user_manifest, strlist_t **user_symlinks)
{
    int flags = FTW_MOUNT | FTW_PHYS;

    assert(dir != NULL);

    manifest = user_manifest;
    symlinks = user_symlinks;

    if (nftw(dir, _add_lists_entry, 25, flags) == -1) {
        warn("nftw");
        return -1;
    }

    return 0;
}

/* Helper called by writedeps() and writeprovs() */
static int _write_soname_helper(struct dist *package, sqlite3 *bom, strlist_t *list, const char *sqlschema, const char *sqlinsert)
{
    char *errmsg = NULL;
    char *query = NULL;
    strlist_entry_t *entry = NULL;
    sqlite3_stmt *stmt = NULL;

    if (package->debug || length_strlist(list) == 0) {
        return 0;
    }

    assert(bom != NULL);
    assert(sqlschema != NULL);
    assert(sqlinsert != NULL);

    xasprintf(&query, "%s; BEGIN TRANSACTION;", sqlschema);
    assert(query != NULL);

    if (sqlite3_exec(bom, query, NULL, NULL, &errmsg) != SQLITE_OK) {
        warnx(errmsg);
        sqlite3_free(errmsg);
        return -1;
    }

    sqlite3_free(errmsg);
    free(query);

    if (sqlite3_prepare(bom, sqlinsert, -1, &stmt, 0) != SQLITE_OK) {
        warnx("sqlite3_prepare: %d: %s", sqlite3_errcode(bom), sqlite3_errmsg(bom));
        sqlite3_free(errmsg);
        return -1;
    }

    if (list != NULL) {
        TAILQ_FOREACH(entry, list, items) {
            insert_name(entry->val, stmt);
        }
    }

    if (sqlite3_finalize(stmt) != SQLITE_OK) {
        warnx("sqlite3_finalize: %d: %s", sqlite3_errcode(bom), sqlite3_errmsg(bom));
        sqlite3_free(errmsg);
        return -1;
    }

    if (sqlite3_exec(bom, "END TRANSACTION", NULL, NULL, &errmsg) != SQLITE_OK) {
        warnx(errmsg);
        sqlite3_free(errmsg);
        return -1;
    }

    sqlite3_free(errmsg);
    return 0;
}

/* Write the list of SONAME dependencies to the package */
int writedeps(struct dist *package, sqlite3 *bom)
{
    return _write_soname_helper(package, bom, deplist, BOM_DEPS_SCHEMA, BOM_DEPS_INSERT);
}

/* Write the list of SONAMEs provided by this package */
int writeprovs(struct dist *package, sqlite3 *bom)
{
    return _write_soname_helper(package, bom, provlist, BOM_PROVS_SCHEMA, BOM_PROVS_INSERT);
}

/*
 * Walk the tree and generate attributes for all entries.
 * Since we use nftw(), we take the input we are given and put it in
 * the global attrsinfo struct.
 */
int genattrs(char *tree, sqlite3 *bom, plist_t *plist, sigver_t sigver, enum checksum sumtype, bool debug, const char *prepend)
{
    char *errmsg = NULL;
    char *query = NULL;
    int flags = FTW_MOUNT | FTW_PHYS;
    sqlite3_stmt *stmt = NULL;

    assert(tree != NULL);
    assert(bom != NULL);

    /* set up the schema */
    xasprintf(&query, BOM_CONTENTS_SCHEMA"; BEGIN TRANSACTION;");
    assert(query != NULL);

    if (sqlite3_exec(bom, query, NULL, NULL, &errmsg) != SQLITE_OK) {
        warnx(errmsg);
        sqlite3_free(errmsg);
        return -1;
    }

    sqlite3_free(errmsg);
    free(query);

    /* create a prepared statement for the nftw() loop */
    if (sqlite3_prepare(bom, BOM_CONTENTS_INSERT, -1, &stmt, 0) != SQLITE_OK) {
        warnx("sqlite3_prepare: %d: %s", sqlite3_errcode(bom), sqlite3_errmsg(bom));
        sqlite3_free(errmsg);
        return -1;
    }

    /* pass info to the nftw() callback this way */
    ai.stmt = stmt;
    ai.plist = plist;
    ai.sigver = sigver;
    ai.sumtype = sumtype;
    ai.debug = debug;
    trimlen = -1;

    prepend_path = strdup(prepend);

    /* add entry entry to the table */
    if (nftw(tree, _add_attrs_entry, 25, flags) == -1) {
        warn("nftw");
        free(prepend_path);
        sqlite3_finalize(stmt);
        sqlite3_free(errmsg);
        return -1;
    }

    /* clean up */
    free(prepend_path);

    if (sqlite3_finalize(stmt) != SQLITE_OK) {
        warnx("sqlite3_finalize: %d: %s", sqlite3_errcode(bom), sqlite3_errmsg(bom));
        sqlite3_free(errmsg);
        return -1;
    }

    if (sqlite3_exec(bom, "END TRANSACTION", NULL, NULL, &errmsg) != SQLITE_OK) {
        warnx(errmsg);
        sqlite3_free(errmsg);
        return -1;
    }

    sqlite3_free(errmsg);
    return 0;
}
