/* Read a package file and collect the pkginfo values. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright David Shea <david@reallylongword.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <assert.h>
#include <string.h>
#include <sys/stat.h>
#include <err.h>
#include <archive.h>
#include "dist-tools.h"
#include "db.h"
#include "strlist.h"

/* Prototypes */
static void _get_basic_info(struct dist *, sqlite3_stmt *, int);
static strlist_t *_get_sonames(sqlite3 *, const char *);

/*
 * Helper to extract basic package information from an SQLite row.
 * NOTE:  The first parameter must be an allocated struct dist.
 */
static void _get_basic_info(struct dist *package, sqlite3_stmt *row, int type)
{
    int offset = 0;

    /* reading from the installed package database gives us IDs */
    if (type == PKG_DBENTRY) {
        package->id = sqlite3_column_int(row, 0);
        offset = 1;
    } else {
        package->id = -1;
    }

    package->name = strdup((char *) sqlite3_column_text(row, 0 + offset));
    package->ver = strdup((char *) sqlite3_column_text(row, 1 + offset));
    package->rel = sqlite3_column_int(row, 2 + offset);
    package->arch = strdup((char *) sqlite3_column_text(row, 3 + offset));

    if (sqlite3_column_text(row, 4 + offset) != NULL) {
        package->repo = strdup((char *) sqlite3_column_text(row, 4 + offset));
    }

    package->size = sqlite3_column_int(row, 5 + offset);

    if (sqlite3_column_int(row, 6 + offset)) {
        package->debug = true;
    } else {
        package->debug = false;
    }

    if (sqlite3_column_text(row, 7 + offset) != NULL) {
        package->desc = strdup((char *) sqlite3_column_text(row, 7 + offset));
    }

    if (sqlite3_column_text(row, 8 + offset) != NULL) {
        package->maintainer = strdup((char *) sqlite3_column_text(row, 8 + offset));
    }

    if (sqlite3_column_text(row, 9 + offset) != NULL) {
        package->license = strdup((char *) sqlite3_column_text(row, 9 + offset));
    }

    if (sqlite3_column_text(row, 10 + offset) != NULL) {
        package->depspath = strdup((char *) sqlite3_column_text(row, 10 + offset));
    }

    if (type == PKG_DBENTRY) {
        if (sqlite3_column_text(row, 11 + offset) != NULL) {
            package->pkgdigest = strdup((char *) sqlite3_column_text(row, 11 + offset));
        }
    }

    return;
}

/* Helper function to read ELF soname lists from bom.db */
static strlist_t *_get_sonames(sqlite3 *db, const char *query)
{
    strlist_t *sonames = NULL;
    strlist_entry_t *entry = NULL;
    sqlite3_stmt *stmt = NULL;

    assert(db != NULL);
    assert(query != NULL);

    if (sqlite3_prepare(db, query, -1, &stmt, 0) != SQLITE_OK) {
        return NULL;
    }

    while (sqlite3_step(stmt) == SQLITE_ROW) {
        if ((entry = calloc(1, sizeof(strlist_entry_t))) == NULL) {
            warn("calloc");
            return NULL;
        }

        if (sonames == NULL) {
            /* initialize a new strlist */
            sonames = init_strlist();
        }

        entry->val = strdup((char *) sqlite3_column_text(stmt, 0));
        TAILQ_INSERT_TAIL(sonames, entry, items);
    }

    if (sqlite3_finalize(stmt) != SQLITE_OK) {
        return NULL;
    }

    return sonames;
}

/*
 * Reads package information from one of two places:
 *    For PKG_DISTFILE, extracts bom.db and reads the info table.
 *    For PKG_DBENTRY, reads the packages table from installed.db.
 *
 * If type==PKG_DISTFILE, archive and where cannot be NULL, row must be NULL.
 * If type==PKG_DBENTRY, row cannot be NULL, archive and where must be NULL.
 */
int getpkginfo(struct dist *package, const int type, const char *archive, const char *where, sqlite3_stmt *row)
{
    int i = 0;
    char *buf = NULL;
    char *dbfilename = NULL;
    struct stat sb;
    sqlite3 *db = NULL;
    sqlite3_stmt *stmt = NULL;
    char *errmsg = NULL;

    assert(package != NULL);

    /* blank out the package structure */
    memset(package, '\0', sizeof(*package));

    /* determine how we get our pkginfo file and read in the data */
    switch (type) {
        case PKG_DISTFILE:     /* a .dist pkg */
            assert(archive != NULL);
            assert(where != NULL);

            /*
             * For packages, extract the bom.db file so we can read the info
             * table and collect all the values.
             */
            package->filename = strdup(archive);

            /*
             * Get the signature.
             */
            buf = calloc(1, BUFSIZ);
            assert(buf != NULL);

            if (unpack_metadata2str(archive, DIST_SIGNATURE, buf, true)) {
                warnx("error determining package signature");
                return -1;
            }

            buf[strcspn(buf, "\r\n")] = '\0';

            if (!strncmp(buf+strlen(SIGNATURE), SIGNATURE_21, strlen(SIGNATURE_21))) {
                package->sigver = SIGVER_21;
            } else if (!strncmp(buf+strlen(SIGNATURE), SIGNATURE_22, strlen(SIGNATURE_22))) {
                package->sigver = SIGVER_22;
            } else {
                warnx("unexpected package signature: %s", buf);
                return -1;
            }

            free(buf);

            if (unpack_metadata2file(archive, DIST_BOM, where, true)) {
                warnx("error reading %s from %s", DIST_BOM, archive);
                return -1;
            }

            /* Open bom.db */
            xasprintf(&dbfilename, "%s/%s", where, DIST_BOM);

            if ((package->bom = realpath(dbfilename, package->bom)) == NULL) {
                free(dbfilename);
                warn("realpath");
                return -1;
            }

            free(dbfilename);
            db = open_db(package->bom);

            if (db == NULL) {
                return -1;
            }

            /* Gather package information */
            if (sqlite3_prepare(db, INFO_SELECT, -1, &stmt, 0) != SQLITE_OK) {
                warnx("%s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
                return -1;
            }

            i = 0;
            while (sqlite3_step(stmt) == SQLITE_ROW) {
                if (i > 1) {
                    warnx("more than one row in the bom.db info table");
                    return -1;
                }

                _get_basic_info(package, stmt, type);
                i++;
            }

            if (sqlite3_finalize(stmt) != SQLITE_OK) {
                warnx("%s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
                return -1;
            }

            /* Gather ELF dependencies */
            package->deps = _get_sonames(db, DEPS_SELECT);

            /* Gather ELF provides */
            package->provs = _get_sonames(db, PROVS_SELECT);

            /* Finished reading the bom */
            sqlite3_free(errmsg);
            sqlite3_close(db);

            /* collect the size of the package file if applicable */
            if (stat(archive, &sb) == 0) {
                package->pkgsize = sb.st_size;
            }

            if (package->pkgsize == 0) {
                warn("package size of %s is 0 bytes", archive);
                return -1;
            }

            /* compute a digest for the package file itself (used by distadd) */
            package->sumtype = DIGEST_ALGORITHM;
            package->pkgdigest = checksum(archive, package->sumtype);

            break;
        case PKG_DBENTRY:      /* installed pkg */
            assert(row != NULL);

            /* we have received a path to a package database entry */

            /* entries do not record information about the package they were delivered in */
            package->tree = NULL;
            package->meta = NULL;
            package->plist = NULL;
            package->pkgsize = -1;
            package->filename = NULL;
            package->bom = NULL;
            package->reloc = NULL;
            package->etc = NULL;
            package->distname = NULL;
            package->depspath = NULL;
            package->pkgdigest = NULL;
            package->compression = ARCHIVE_FILTER_NONE;
            package->sigver = SIGVER_22;

            /* our row contains this information */
            _get_basic_info(package, row, type);

            /* digest of the package we installed from */
            package->sumtype = DIGEST_ALGORITHM;

            /* legal file is actually on disk -- only for backwards compatibility */
            xasprintf(&package->legal, "%s/%s/%s", DB_PATH, package->pkgdigest, DIST_LEGAL);
            memset(&sb, 0, sizeof(sb));

            if (stat(package->legal, &sb) != 0) {
                free(package->legal);
                package->legal = NULL;
            }

            break;
    }

    /* do we have a missing repo? */
    if (package->repo == NULL) {
        package->repo = strdup("?");
    }

    if (package->name == NULL || package->ver == NULL || package->arch == NULL || package->desc == NULL || package->maintainer == NULL || package->size == 0) {
        fprintf(stderr, "Missing packageinfo data: ");

        if (package->name == NULL) {
            fprintf(stderr, "name\n");
        } else if (package->ver == NULL) {
            fprintf(stderr, "ver\n");
        } else if (package->arch == NULL) {
            fprintf(stderr, "arch\n");
        } else if (package->desc == NULL) {
            fprintf(stderr, "desc\n");
        } else if (package->maintainer == NULL) {
            fprintf(stderr, "maintainer\n");
        } else if (package->size < 0) {
            fprintf(stderr, "size\n");
        }

        return -1;
    }

    return 0;
}
