/*
 * Given a list of path components as separate strings, join them in to a
 * correct (but unverified) Unix path.  Extra slashes are removed.  Spaces
 * and other special characters are not escaped.  This function allocates
 * memory for the returned value.  The caller must free this memory when
 * done.
 *
 * Usage: path = join(a, b, c, ..., NULL);
 * (where a, b, and c are char *)
 */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <limits.h>
#include "dist-tools.h"

char *joinpath(const char *path, ...)
{
    va_list ap;
    int i = 0;
    int extralen = 0;
    bool needsep = false;
    char *element = NULL;
    char *tail = NULL;
    char *built = NULL;
    char *tmp = NULL;
    char *near = NULL;
    char *far = NULL;

    assert(path != NULL);

    /* Allocate a large buffer to use for building the path. */
    built = calloc(1, PATH_MAX + 1);
    assert(built != NULL);

    /* Make sure the full path starts with a slash. */
    if (*path == '/') {
        /* this for loop trims multiple leading slashes down to just one */
        while (*(path + 1) == '/') path++;
    }

    /* begin our joined path */
    tail = stpcpy(built, path);

    /* the remaining elements come in this way */
    va_start(ap, path);

    while ((element = va_arg(ap, char *)) != NULL) {
        /* for the trailing NUL */
        extralen = 1;
        needsep = false;

        /* trim any extra trailing slashes */
        i = strlen(built);

        while (*tail == '/') {
            *tail = '\0';
            tail--;
        }

        /*
         * This loop trims multiple leading slashes down to just one.
         * If 'i' is 1, it will preserve 1 leading slash.  Actually, set
         * this to the number of slashes to preserve.
         */
        if (*element == '/' && *tail != '/') {
            i = 1;
        } else {
            i = 0;
        }

        while (*(element + i) == '/') element++;

        /* make sure we have at least one slash in case there are none */
        if (*element != '/' && *tail != '/') {
            extralen++;
            needsep = true;
        }

        /* perform the concatenations */
        if (needsep) {
            tail = stpcpy(tail, "/");
        }

        tail = stpcpy(tail, element);
    }

    va_end(ap);

    /* it's possible there are repeating slashes, eliminate them */
    near = built;
    far = built;

    while ((*near = *far)) {
        if (*near == '/') {
            while (*far == '/') {
                far++;
            }

            near++;
        } else {
            near++;
            far++;
        }
    }

    /* shrink memory allocation */
    tmp = realloc(built, strlen(built) + 1);
    built = tmp;

    return built;
}
