/* Read plist files. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <assert.h>
#include <string.h>
#include <err.h>
#include <glob.h>
#include <sys/queue.h>
#include <limits.h>
#include "dist-tools.h"
#include "plist.h"

/* Local structs */
struct ownership {
    char *owner;
    uint64_t uid;
    char *group;
    uint64_t gid;
};

/* Local prototypes */
static plist_entry_t *init_plist_entry(void);
static int create_plist_entry(plist_t *, char *, struct ownership *, char *, bool);

/*
 * Allocate memory for and initialize a plist_entry_t.
 * Caller must use free the plist_entry_t when done.
 */
static plist_entry_t *init_plist_entry(void)
{
    plist_entry_t *entry = NULL;

    if ((entry = calloc(1, sizeof(plist_entry_t))) == NULL) {
        warn("calloc");
        return NULL;
    }

    entry->has_checksum = true;

    return entry;
}

/* Create and add a new plist entry. */
static int create_plist_entry(plist_t *plist, char *filename, struct ownership *owners, char *mode, bool has_checksum)
{
    plist_entry_t *entry = NULL;

    assert(filename != NULL);
    assert(mode != NULL);

    /* we will not tolerate duplicate plist entries */
    TAILQ_FOREACH(entry, plist, items) {
        if (!strcmp(entry->filename, filename)) {
            warn("duplicate entries in plist for: %s", filename);
            return -1;
        }
    }

    if ((entry = init_plist_entry()) == NULL) {
        return -1;
    }

    /* collect the plist entry fields */
    entry->filename = strdup(filename);

    /* set the ownership */
    entry->owner = strdup(owners->owner);
    entry->uid = owners->uid;
    entry->group = strdup(owners->group);
    entry->gid = owners->gid;
    entry->has_checksum = has_checksum;

    /* set the mode */
    if (!strncmp(mode, "symlink", 7)) {
        entry->is_symlink = true;
        entry->mode = -1;
    } else {
        errno = 0;
        entry->is_symlink = false;
        entry->mode = strtoull(mode, NULL, 8);

        if (errno == ERANGE || (errno == EINVAL && entry->mode == 0)) {
            warn("strtoull");
            return -1;
        }
    }

    /* record this plist entry */
    TAILQ_INSERT_TAIL(plist, entry, items);

    return 0;
}

/*
 * Read a plist file and return a list of plist_entry_t structs.
 * Caller is responsible for freeing the returned list.
 */
plist_t *read_plist(char *filename, char *root)
{
    int ret = 0;
    int extrafields = 0;
    int foundfields = 0;
    size_t len = 0;
    unsigned int j = 0;
    int gflags = GLOB_ERR|GLOB_NOCHECK|GLOB_PERIOD|GLOB_TILDE|GLOB_BRACE;
    char *globpath = NULL;
    char *globsub = NULL;
    char *fname = NULL;
    char *pos = NULL;
    char *buf = NULL;
    char *fpath = NULL;
    char **fields = NULL;
    glob_t found;
    struct ownership owners;
    bool has_checksum = true;
    unsigned int fieldcount = 0;
    plist_t *plist = NULL;
    FILE *fp = NULL;

    /* initialize a new plist */
    plist = init_plist();

    owners.owner = NULL;
    owners.uid = 0;
    owners.group = NULL;
    owners.gid = 0;

    fpath = realpath(filename, NULL);
    assert(fpath != NULL);

    if ((fp = fopen(fpath, "r")) == NULL) {
        warn("fopen");
        free(fpath);
        return NULL;
    }

    free(fpath);

    while (getline(&buf, &len, fp) != -1) {
        has_checksum = true;

        /* grab the read line, less the line ending characters */
        len = strcspn(buf, "\r\n");
        fname = calloc(len + 1, sizeof(*fname));
        assert(fname != NULL);
        fname = strncpy(fname, buf, len);
        assert(fname != NULL);

        /* determine how many fields are after the filename */
        pos = strrchr(fname, ' ');
        assert(pos != NULL);
        pos++;

        if (!strcmp(pos, "nochecksum")) {
            extrafields = 5;
        } else {
            extrafields = 4;
        }

        /* break the line in to two parts: the filename and the fields */
        pos = fname;
        pos += strlen(fname);

        for (foundfields = 1; foundfields < extrafields; pos--) {
            if (*pos == ' ') {
               foundfields++;
            }
        }

        pos++;
        *pos = '\0';
        pos++;
        fields = strsplit(pos, " ");
        fieldcount = strvlen(fields);

        free(buf);
        buf = NULL;

        if (fieldcount < 3 || fieldcount > 4) {
            strvfree(fields);
            continue;
        }

        if (split_name_id(fields[0], &owners.owner, &owners.uid) == -1) {
            warn("unable to read owner information");
            strvfree(fields);
            break;
        }

        if (split_name_id(fields[1], &owners.group, &owners.gid) == -1) {
            warn("unable to read group information");
            strvfree(fields);
            break;
        }

        /* Adjust the filename based on actual location. */
        if (root == NULL) {
            globpath = strdup(fname);
        } else {
            /* skip staging root and trailing slash */
            len = strlen(root) + 1;
            globpath = NULL;
            xasprintf(&globpath, "%s/%s", root, fname);
        }

        /* An optional fifth field can disable checksums for these entries. */
        if ((fieldcount == 4) && !strcmp(fields[3], "nochecksum")) {
            has_checksum = false;
        }

        /* Each line may contain wildcards. */
        if ((ret = glob(globpath, gflags, NULL, &found)) != 0) {
            if (create_plist_entry(plist, fname, &owners, fields[2], has_checksum)) {
                strvfree(fields);
                err(EXIT_FAILURE, "failed expanding %s in plist", fname);
            }
        } else {
            j = 0;

            while (j < found.gl_pathc) {
                globsub = found.gl_pathv[j] + len;

                if (create_plist_entry(plist, globsub, &owners, fields[2], has_checksum)) {
                    strvfree(fields);
                    globfree(&found);
                    err(EXIT_FAILURE, "unable to create plist entry for %s", globsub);
                }

                j++;
            }
        }

        strvfree(fields);

        free(owners.owner);
        owners.owner = NULL;
        owners.uid = 0;
        free(owners.group);
        owners.group = NULL;
        owners.gid = 0;

        free(globpath);
        globpath = NULL;

        globfree(&found);

        free(fname);
        fname = NULL;
    }

    return plist;
}

/*
 * Return the entry count for the given plist.
 */
int plist_length(plist_t *plist)
{
    int c = 0;
    plist_entry_t *entry = NULL;

    if (plist == NULL) {
        return 0;
    }

    TAILQ_FOREACH(entry, plist, items) {
        c++;
    }

    return c;
}

/*
 * Allocate memory for and initialize a new plist_t.
 */
plist_t *init_plist(void)
{
    plist_t *list = NULL;

    list = calloc(1, sizeof(*list));
    assert(list != NULL);
    TAILQ_INIT(list);
    return list;
}

/*
 * Given an allocated plist_t, free all the memory associated with it.
 */
void free_plist(plist_t *plist)
{
    plist_entry_t *entry = NULL;

    if (plist == NULL) {
        return;
    }

    while (!TAILQ_EMPTY(plist)) {
        entry = TAILQ_FIRST(plist);
        TAILQ_REMOVE(plist, entry, items);
        free(entry->filename);
        entry->filename = NULL;
        free(entry->owner);
        entry->owner = NULL;
        free(entry->group);
        entry->group = NULL;
        free(entry);
    }

    free(plist);

    return;
}
