/* SQLite info gathering functions. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdlib.h>
#include <stdint.h>
#include <assert.h>
#include <err.h>
#include <string.h>
#include "dist-tools.h"
#include "db.h"

/*
 * Put together a prepared statement to insert package info in to the db.
 * Return rowid of inserted data, or negative value on failure.
 */
int insert_pkg_info(const struct dist *package, sqlite3 *db, const char *schema, const int column_offset, const bool upgrade)
{
    sqlite3_stmt *stmt = NULL;
    char *err = NULL;

    assert(package != NULL);
    assert(db != NULL);
    assert(schema != NULL);

    if (sqlite3_prepare(db, schema, -1, &stmt, 0) != SQLITE_OK) {
        warnx("sqlite3_prepare: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -1;
    }

    if (sqlite3_exec(db, "BEGIN TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        return -1;
    }

    /*
     * recording an installed package?  set first column to null for the
     * package id primary key to be automatically set.
     */
    if (column_offset > 0) {
        if (sqlite3_bind_null(stmt, 1) != SQLITE_OK) {
            warnx("sqlite3_bind_null: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            return -1;
        }
    }

    if (sqlite3_bind_text(stmt, 1 + column_offset, package->name, strlen(package->name), SQLITE_TRANSIENT) != SQLITE_OK) {
        warnx("sqlite3_bind_text: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -1;
    }

    if (sqlite3_bind_text(stmt, 2 + column_offset, package->ver, strlen(package->ver), SQLITE_TRANSIENT) != SQLITE_OK) {
        warnx("sqlite3_bind_text: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -1;
    }

    if (sqlite3_bind_int(stmt, 3 + column_offset, package->rel) != SQLITE_OK) {
        warnx("sqlite3_bind_int: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -1;
    }

    if (sqlite3_bind_text(stmt, 4 + column_offset, package->arch, strlen(package->arch), SQLITE_TRANSIENT) != SQLITE_OK) {
        warnx("sqlite3_bind_text: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -1;
    }

    if (package->repo == NULL) {
        if (sqlite3_bind_null(stmt, 5 + column_offset) != SQLITE_OK) {
            warnx("sqlite3_bind_null: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            return -1;
        }
    } else {
        if (sqlite3_bind_text(stmt, 5 + column_offset, package->repo, strlen(package->repo), SQLITE_TRANSIENT) != SQLITE_OK) {
            warnx("sqlite3_bind_text: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            return -1;
        }
    }

    if (sqlite3_bind_int64(stmt, 6 + column_offset, package->size) != SQLITE_OK) {
        warnx("sqlite3_bind_int64: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -1;
    }

    if (sqlite3_bind_int(stmt, 7 + column_offset, package->debug) != SQLITE_OK) {
        warnx("sqlite3_bind_int: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -1;
    }

    if (package->desc == NULL) {
        if (sqlite3_bind_null(stmt, 8 + column_offset) != SQLITE_OK) {
            warnx("sqlite3_bind_null: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            return -1;
        }
    } else {
        if (sqlite3_bind_text(stmt, 8 + column_offset, package->desc, strlen(package->desc), SQLITE_TRANSIENT) != SQLITE_OK) {
            warnx("sqlite3_bind_text: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            return -1;
        }
    }

    if (package->maintainer == NULL) {
        if (sqlite3_bind_null(stmt, 9 + column_offset) != SQLITE_OK) {
            warnx("sqlite3_bind_null: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            return -1;
        }
    } else {
        if (sqlite3_bind_text(stmt, 9 + column_offset, package->maintainer, strlen(package->maintainer), SQLITE_TRANSIENT) != SQLITE_OK) {
            warnx("sqlite3_bind_text: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            return -1;
        }
    }

    if (package->license == NULL) {
        if (sqlite3_bind_null(stmt, 10 + column_offset) != SQLITE_OK) {
            warnx("sqlite3_bind_null: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            return -1;
        }
    } else {
        if (sqlite3_bind_text(stmt, 10 + column_offset, package->license, strlen(package->license), SQLITE_TRANSIENT) != SQLITE_OK) {
            warnx("sqlite3_bind_text: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            return -1;
        }
    }

    /*
     * An offset of 0 indicates we are writing to a .dist file's bom.db,
     * so only record the depspath.  Any other offset (which should only
     * be 1) indicates we are writing to the packages table in installed.db
     * so record both depspath and pkgdigest.
     *
     * Upgrading a package records both the depspath and pkgdigest, but the
     * offset is zero because of the query used.  The first position is not
     * the id.
     */
    if (column_offset == 0 && !upgrade) {
        if (package->depspath != NULL) {
            if (sqlite3_bind_text(stmt, 11 + column_offset, package->depspath, strlen(package->depspath), SQLITE_TRANSIENT) != SQLITE_OK) {
                warnx("sqlite3_bind_text: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
                return -1;
           }
        }
    } else {
        if (package->depspath != NULL) {
            if (sqlite3_bind_text(stmt, 11 + column_offset, package->depspath, strlen(package->depspath), SQLITE_TRANSIENT) != SQLITE_OK) {
                warnx("sqlite3_bind_text: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
                return -1;
           }
        }

        if (package->pkgdigest != NULL) {
            if (sqlite3_bind_text(stmt, 12 + column_offset, package->pkgdigest, strlen(package->pkgdigest), SQLITE_TRANSIENT) != SQLITE_OK) {
                warnx("sqlite3_bind_text: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
                return -1;
            }
        }
    }

    if (sqlite3_step(stmt) != SQLITE_DONE) {
        warnx("sqlite3_step: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -1;
    }

    if (sqlite3_exec(db, "END TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        return -1;
    }

    sqlite3_finalize(stmt);

    if ((package->id < 0) && !upgrade) {
        return sqlite3_last_insert_rowid(db);
    } else {
        return package->id;
    }
}

/*
 * Copy the contents from bom.db to installed.db for the specified package id.
 */
int insert_contents(int id, const char *bomdb, struct sqlite3 *db)
{
    char *err = NULL;
    char *query = NULL;

    assert(id >= 0);
    assert(bomdb != NULL);
    assert(db != NULL);

    /* attach bom.db */
    xasprintf(&query, BOM_CONTENTS_ATTACH, bomdb);

    if (sqlite3_exec(db, query, NULL, 0, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        return -1;
    }

    free(query);
    query = NULL;

    /* insert all new rows, update existing rows */
    xasprintf(&query, DB_CONTENTS_INSERT, id);

    if (sqlite3_exec(db, "BEGIN TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        return -1;
    }

    if (sqlite3_exec(db, query, NULL, 0, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        return -1;
    }

    if (sqlite3_exec(db, "END TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        return -1;
    }

    free(query);
    query = NULL;

    /* detach bom */
    if (sqlite3_exec(db, BOM_CONTENTS_DETACH, NULL, 0, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        return -1;
    }

    return 0;
}

/*
 * Given an SQL query, return the number of rows in the result.
 * On error, return < 0.
 */
int result_count(sqlite3 *db, const char *query)
{
    int count = -1;
    sqlite3_stmt *stmt = NULL;

    assert(db != NULL);
    assert(query != NULL);

    if (sqlite3_prepare(db, query, -1, &stmt, 0) != SQLITE_OK) {
        warnx("sqlite3_prepare: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -1;
    }

    /* this will only ever return one row, but whatever */
    while (sqlite3_step(stmt) == SQLITE_ROW) {
        count += sqlite3_column_int(stmt, 0);
    }

    sqlite3_finalize(stmt);
    return count;
}
