/* File and directory removal helper functions. */

/*
 * Copyright David Shea <david@reallylongword.org>
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdint.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <err.h>
#include "dist-tools.h"
#include "opts.h"

/*
 * Callback when iterating a TAILQ of struct finfo objects.  Used to
 * safely remove files and directories.  Note, the caller needs to
 * ensure that directory removal happens in the correct order.  This
 * function cannot remove an empty tree of subdirectories, only empty
 * directories.
 */
int checkunlink(struct finfo *info, const char *prepend, const uint64_t opts)
{
    int r = 0;
    int err = 0;
    int bad = 0;
    char *efile = NULL;

    assert(info != NULL);

    bad = checkfile(info, prepend, (opts & OPTS_STRICT));
    efile = joinpath(prepend, info->filename, NULL);

    if ((opts & OPTS_QUICK) || (opts & OPTS_FORCE) || !bad) {
        if (info->type == TYPE_DIR) {
            r = rmdir(efile);

            if (r == -1) {
                if (errno == ENOTEMPTY && (opts & OPTS_VERBOSE)) {
                    printf("%s not empty, not removing\n", efile);
                } else if (errno != ENOTEMPTY) {
                    warn("rmdir: %s", efile);
                    err = -1;
                }
            }
        } else {
            r = unlink(efile);

            if (r == -1) {
                warn("unlink: %s", efile);
                err = -1;
            }
        }
    }

    free(efile);
    return err;

}
