/* Run a package script. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdlib.h>
#include <assert.h>
#include <stdbool.h>
#include <unistd.h>
#include <string.h>
#include <err.h>
#include "dist-tools.h"

/*
 * Generic function to handle running package scripts in the correct
 * environment and directory.
 *
 * The 'tmppath' string is only used when running preinstall scripts because
 * we do not yet have a package database entry so we have to run the script
 * from the temporary unpacking directory.  When calling runscript for any
 * other script types, it is ok to pass NULL for tmppath.
 */
int runscript(const struct dist *package, const struct dist *oldpackage, const unsigned int type, const bool upgrade, char *root, char *tmppath)
{
    int r = 0;
    int status = 0;
    size_t len = 0;
    char *oldval = NULL;
    char *newval = NULL;
    char *basescript = NULL;
    char *script = NULL;
    char *cmd = NULL;
    char *buf = NULL;
    strlist_t *scriptoutput = NULL;
    strlist_entry_t *entry = NULL;
    struct stat sb;
    FILE *fp = NULL;

    assert(package != NULL);
    assert(type > 0);

    umask(0022);

    /* Location of script. */
    if (type == SCRIPT_PREINSTALL) {
        basescript = DIST_PREINSTALL;
    } else if (type == SCRIPT_POSTINSTALL) {
        basescript = DIST_POSTINSTALL;
    } else if (type == SCRIPT_PREREMOVE) {
        basescript = DIST_PREREMOVE;
    } else if (type == SCRIPT_POSTREMOVE) {
        basescript = DIST_POSTREMOVE;
    } else {
        warnx("unknown script type");
        return -1;
    }

    if (!upgrade && type == SCRIPT_PREINSTALL) {
        assert(tmppath != NULL);
    }

    if (upgrade && type == SCRIPT_PREREMOVE) {
        assert(oldpackage != NULL);
    }

    if (type == SCRIPT_PREINSTALL) {
        /* run preinstall script from temporary package unpacking area */
        script = joinpath(tmppath, basescript, NULL);
    } else if (root == NULL) {
        if (upgrade && (type == SCRIPT_PREREMOVE || type == SCRIPT_POSTREMOVE)) {
            script = joinpath(DB_PATH, oldpackage->pkgdigest, basescript, NULL);
        } else {
            script = joinpath(DB_PATH, package->pkgdigest, basescript, NULL);
        }
    } else {
        if (upgrade && (type == SCRIPT_PREREMOVE || type == SCRIPT_POSTREMOVE)) {
            script = joinpath(root, DB_PATH, oldpackage->pkgdigest, basescript, NULL);
        } else {
            script = joinpath(root, DB_PATH, package->pkgdigest, basescript, NULL);
        }
    }

    if (stat(script, &sb) == -1) {
        if (errno != ENOENT) {
            /* script not existing is ok */
            warn("stat");
            status = -1;
        }

        free(script);
        return status;
    }

    /* ensure script is executable */
    if (((sb.st_mode) & S_IXUSR) != S_IXUSR) {
        if (chmod(script, (sb.st_mode & S_INTER) | S_IXUSR) == -1) {
            warn("chmod");
            return -1;
        }
    }

    /* paths that point in to the staging root, if specified */
    if (root != NULL) {
        oldval = getenv("PATH");

        if (oldval == NULL) {
            xasprintf(&newval, "%s/usr/bin", root);
        } else {
            xasprintf(&newval, "%s/usr/bin:%s", root, oldval);
        }

        if (setenv("PATH", newval, 1) == -1) {
            warn("setenv");
            return -1;
        }

        free(newval);
        oldval = getenv("LD_LIBRARY_PATH");

        if (oldval == NULL) {
            xasprintf(&newval, "%s/usr/lib", root);
        } else {
            xasprintf(&newval, "%s/usr/lib:%s", root, oldval);
        }

        if (setenv("LD_LIBRARY_PATH", newval, 1) == -1) {
            warn("setenv");
            return -1;
        }

        free(newval);
    }

    /* package meta data that is offered to scripts */
    if (setenv("PKG", package->name, 1) == -1) {
        warn("setenv");
        return -1;
    }

    if (upgrade && type == SCRIPT_PREREMOVE) {
        newval = oldpackage->ver;
    } else {
        newval = package->ver;
    }

    assert(newval != NULL);

    if (setenv("VER", newval, 1) == -1) {
        warn("setenv");
        return -1;
    }

    if (upgrade && type == SCRIPT_PREREMOVE) {
        xasprintf(&newval, "%d", oldpackage->rel);
    } else {
        xasprintf(&newval, "%d", package->rel);
    }

    if (setenv("REL", newval, 1) == -1) {
        warn("setenv");
        return -1;
    }

    free(newval);

    /* STAGING is only set if the user run distadd with '-r' */
    if (root == NULL) {
        if (unsetenv("STAGING") == -1) {
            warn("unsetenv");
            return -1;
        }
    } else {
        if (setenv("STAGING", root, 1) == -1) {
            warn("setenv");
            return -1;
        }
    }

    /* set the script action */
    if (upgrade && (type == SCRIPT_PREINSTALL || type == SCRIPT_POSTINSTALL || type == SCRIPT_PREREMOVE || type == SCRIPT_POSTREMOVE)) {
        if (setenv("ACTION", "upgrade", 1) == -1) {
            warn("setenv");
            return -1;
        }
    } else if (!upgrade && (type == SCRIPT_PREINSTALL || type == SCRIPT_POSTINSTALL)) {
        if (setenv("ACTION", "install", 1) == -1) {
            warn("setenv");
            return -1;
        }
    } else if (!upgrade && (type == SCRIPT_PREREMOVE || type == SCRIPT_POSTREMOVE)) {
        if (setenv("ACTION", "remove", 1) == -1) {
            warn("setenv");
            return -1;
        }
    } else {
        /* unknown? */
        if (unsetenv("ACTION") == -1) {
            warn("unsetenv");
            return -1;
        }
    }

    /* upgrades need the version number in preremove */
    if (upgrade && type == SCRIPT_PREREMOVE) {
        if (setenv("UPGRADEVER", package->ver, 1) == -1) {
            warn("setenv");
            return -1;
        }
    }

    /* run the script */
    xasprintf(&cmd, "%s 2>&1", script);
    free(script);

    if (root == NULL) {
        root = "/";
    }

    if (chdir(root) == -1) {
        warn("chdir");
        return -1;
    }

    if ((fp = popen(cmd, "r")) == NULL) {
        warn("popen");
        return -1;
    }

    while (getline(&buf, &len, fp) != -1) {
        /* capture the line as a new list entry */
        entry = calloc(1, sizeof(*entry));
        assert(entry != NULL);

        if (buf != NULL) {
            entry->val = strdup(buf);
        }

        /* initialize the string list if needed */
        if (scriptoutput == NULL) {
            scriptoutput = init_strlist();
            assert(scriptoutput != NULL);
        }

        /* add the output line */
        TAILQ_INSERT_TAIL(scriptoutput, entry, items);
    }

    free(buf);
    status = pclose(fp);

    if (!WIFEXITED(status) || WEXITSTATUS(status) != EXIT_SUCCESS) {
        r = -1;
    }

    if (scriptoutput != NULL) {
        TAILQ_FOREACH(entry, scriptoutput, items) {
            printf("%s\n", entry->val);
        }
    }

    free(cmd);
    free_strlist(scriptoutput);
    return r;
}
