/*
 * Takes a package format string of the form:
 *     NAME
 *     NAME.ARCH
 *     NAME.debug
 *     NAME.debug.ARCH
 *     NAME%VER
 *     NAME-VER
 *     NAME-VER.ARCH
 *     NAME-VER.debug
 *     NAME-VER.debug.ARCH
 *     NAME-VER-REL
 *     NAME-VER-REL.ARCH
 *     NAME-VER-REL.debug
 *     NAME-VER-REL.debug.ARCH
 * finds all versions of an installed package, and fills a list with struct
 * dists.  Returns the list.
 */

/*
 * Copyright David Shea <david@reallylongword.org>
 * Copyright David Cantrell <dcantrell@burdell.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>
#include <err.h>
#include "dist-tools.h"
#include "db.h"
#include "pkglist.h"

/*
 * Data types handled by append() below.  Kinda lazy here and pass in
 * a void * for what to append to the given string and then handle it
 * accordingly.
 */
enum APPEND_TYPE {
    STRING,
    INT
};

/*
 * Append 'new' to 'dest' taking in to account the data type of 'new'.
 * If dest is NULL, the function returns a newly allocated copy of
 * 'new' as a string.  Otherwise it concatenates dest and new together
 * as a newly allocated string and frees the existing string.
 *
 * The type of 'new' must be a supported APPEND_TYPE as listed above.
 */
static char *append(char *dest, enum APPEND_TYPE type, void *new)
{
    char *s = NULL;
    char *r = NULL;

    if (type == STRING) {
        s = strdup((char *) new);
    } else if (type == INT) {
        xasprintf(&s, "%i", *((int *) new));
    }

    assert(s != NULL);

    if (dest == NULL) {
        return s;
    } else {
        xasprintf(&r, "%s%s", dest, s);
        free(dest);
        free(s);
        return r;
    }
}

static bool check_match(const char *pkgstr, const char *compare, bool *first, char **query, int id)
{
    assert(pkgstr != NULL);
    assert(compare != NULL);

    if (!strcmp(pkgstr, compare)) {
        if (*first) {
            *first = false;
        } else {
            *query = append(*query, STRING, ",");
        }

        *query = append(*query, INT, &id);
        return true;
    }

    return false;
}

/*
 * This is the main function that finds package database entries.  Users
 * and callers can specify packages using this format:
 *     NAME
 *     NAME.ARCH
 *     NAME.debug
 *     NAME.debug.ARCH
 *     NAME%VER
 *     NAME-VER
 *     NAME-VER.ARCH
 *     NAME-VER.debug
 *     NAME-VER.debug.ARCH
 *     NAME-VER-REL
 *     NAME-VER-REL.ARCH
 *     NAME-VER-REL.debug
 *     NAME-VER-REL.debug.ARCH
 *
 * The db_installed variable is the path to installed.db, which is generally
 * set by main() in whatever program is running.
 *
 * The pkglist variable is what will be filled with the returned packages.
 * Caller must free this list, but they should know that since they own it.
 *
 * Returns the list of matching packages, which could be NULL.  In which
 * case there were no matches.  Only zuul.
 */
pkglist_t *scanpkg(const char *db_installed, const char *pkgstr)
{
    char *query = NULL;
    char *compare = NULL;
    sqlite3 *db = NULL;
    sqlite3_stmt *stmt = NULL;
    struct dist *tmppkg = NULL;
    bool first = false;
    pkglist_entry_t *entry = NULL;
    pkglist_t *pkglist = NULL;

    /* open the installed package database */
    db = open_db(db_installed);

    if (db == NULL) {
        return NULL;
    }

    /* handle input in various forms */
    if (pkgstr != NULL) {
        if (index(pkgstr, '%') || index(pkgstr, '-') || index(pkgstr, '.')) {
            /*
             * support these formats:
             * NAME.ARCH
             * NAME.debug
             * NAME.debug.ARCH
             * NAME%VER
             * NAME-VER
             * NAME-VER.ARCH
             * NAME-VER.debug
             * NAME-VER.debug.ARCH
             * NAME-VER-REL
             * NAME-VER-REL.ARCH
             * NAME-VER-REL.debug
             * NAME-VER-REL.debug.ARCH
             *
             * This is complicated, but let me explain.  Rather than parse the
             * input string, iterate over each package and build candidate input
             * strings and compare those with what we were actually given.  If
             * a match is found, add that package to the list and continue.  The
             * idea here is to avoid parsing errors on the input string at the
             * expense of a slightly longer run time.
             */
            first = true;
            query = strdup(SCANPKG_SELECT);

            if (sqlite3_prepare(db, "SELECT * FROM packages", -1, &stmt, 0) != SQLITE_OK) {
                warnx("sqlite3_prepare: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
                sqlite3_close(db);
                return NULL;
            }

            while (sqlite3_step(stmt) == SQLITE_ROW) {
                /* start by getting this package's information */
                freepkg(tmppkg);
                tmppkg = initpkg();
                assert(tmppkg != NULL);

                if (getpkginfo(tmppkg, PKG_DBENTRY, NULL, NULL, stmt)) {
                    sqlite3_close(db);
                    return NULL;
                }

                if (!tmppkg->debug) {
                    /* NAME-VER-REL.ARCH */
                    xasprintf(&compare, "%s-%s-%d.%s", tmppkg->name, tmppkg->ver, tmppkg->rel, tmppkg->arch);

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);

                    /* NAME-VER-REL */
                    xasprintf(&compare, "%s-%s-%d", tmppkg->name, tmppkg->ver, tmppkg->rel);

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);

                    /* NAME-VER.ARCH */
                    xasprintf(&compare, "%s-%s.%s", tmppkg->name, tmppkg->ver, tmppkg->arch);

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);

                    /* NAME-VER */
                    xasprintf(&compare, "%s-%s", tmppkg->name, tmppkg->ver);

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);

                    /* NAME%VER */
                    xasprintf(&compare, "%s%%%s", tmppkg->name, tmppkg->ver);

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);

                    /* NAME.ARCH */
                    xasprintf(&compare, "%s.%s", tmppkg->name, tmppkg->arch);

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);

                    /* NAME */
                    xasprintf(&compare, "%s", tmppkg->name);

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);
                } else if (tmppkg->debug) {
                    /* NAME-VER-REL.debug.ARCH */
                    xasprintf(&compare, "%s-%s-%d.debug.%s", tmppkg->name, tmppkg->ver, tmppkg->rel, tmppkg->arch);

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);

                    /* NAME-VER-REL.debug */
                    xasprintf(&compare, "%s-%s-%d.debug", tmppkg->name, tmppkg->ver, tmppkg->rel);

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);

                    /* NAME-VER.debug.ARCH */
                    xasprintf(&compare, "%s-%s%s.%s", tmppkg->name, tmppkg->ver, (tmppkg->debug ? ".debug" : ""), tmppkg->arch);

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);

                    /* NAME-VER.debug */
                    xasprintf(&compare, "%s-%s%s", tmppkg->name, tmppkg->ver, (tmppkg->debug ? ".debug" : ""));

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);

                    /* NAME.debug.ARCH */
                    xasprintf(&compare, "%s%s.%s", tmppkg->name, (tmppkg->debug ? ".debug" : ""), tmppkg->arch);

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);

                    /* NAME.debug */
                    xasprintf(&compare, "%s%s", tmppkg->name, (tmppkg->debug ? ".debug" : ""));

                    if (check_match(pkgstr, compare, &first, &query, tmppkg->id)) {
                        free(compare);
                        continue;
                    }

                    free(compare);
                }
            }

            freepkg(tmppkg);

            if (sqlite3_finalize(stmt) != SQLITE_OK) {
                warnx("sqlite3_finalize: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
                sqlite3_close(db);
                return NULL;
            }

            query = append(query, STRING, ")");
        } else {
            /* assume package name as a last resort */
            xasprintf(&query, "SELECT * FROM packages WHERE name='%s'", pkgstr);
        }
    } else {
        /* user specified no package, gather entire database */
        xasprintf(&query, "SELECT * FROM packages");
    }

    /* run the query and for each row, build a struct dist */
    if (sqlite3_prepare(db, query, -1, &stmt, 0) != SQLITE_OK) {
        free(query);
        warnx("sqlite3_prepare: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_close(db);
        return NULL;
    }

    while (sqlite3_step(stmt) == SQLITE_ROW) {
        /* the pkglist may need initializing */
        if (pkglist == NULL) {
            pkglist = init_pkglist();
        }

        /* init pkglist entry */
        entry = calloc(1, sizeof(*entry));
        assert(entry != NULL);

        /* init the struct dist for this entry */
        entry->pkg = calloc(1, sizeof(*entry->pkg));
        assert(entry->pkg != NULL);

        if (getpkginfo(entry->pkg, PKG_DBENTRY, NULL, NULL, stmt)) {
            sqlite3_close(db);
            return NULL;
        }

        TAILQ_INSERT_TAIL(pkglist, entry, items);
    }

    if (sqlite3_finalize(stmt) != SQLITE_OK) {
        warnx("sqlite3_finalize: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
    }

    free(query);
    sqlite3_close(db);
    return pkglist;
}
