/* Generic string list functions. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <err.h>
#include <assert.h>
#include <sys/queue.h>
#include "strlist.h"

/*
 * Allocate memory for and initialize a new strlist_t.
 */
strlist_t *init_strlist(void)
{
    strlist_t *list = NULL;

    list = calloc(1, sizeof(*list));
    assert(list != NULL);
    TAILQ_INIT(list);
    return list;
}

/*
 * Given an allocated strlist_t, free all the memory associated with it.
 */
void free_strlist(strlist_t *list) {
    strlist_entry_t *entry = NULL;

    if (list == NULL) {
        return;
    }

    while (!TAILQ_EMPTY(list)) {
        entry = TAILQ_FIRST(list);
        TAILQ_REMOVE(list, entry, items);
        free(entry->val);
        free(entry);
    }

    free(list);

    return;
}

/*
 * Returns true if the specified string is in the given list.
 */
bool is_str_in_list(strlist_t *list, const char *str) {
    strlist_entry_t *entry = NULL;

    assert(str != NULL);

    if (list == NULL) {
        return false;
    }

    TAILQ_FOREACH(entry, list, items) {
        if (!strcmp(entry->val, str)) {
            return true;
        }
    }

    return false;
}

/*
 * Return the entry count for the given strlist.
 */
int length_strlist(strlist_t *list) {
    int c = 0;
    strlist_entry_t *entry = NULL;

    if (list == NULL || TAILQ_EMPTY(list)) {
        return 0;
    }

    TAILQ_FOREACH(entry, list, items) {
        c++;
    }

    return c;
}

/*
 * Given a string and a delimiter string, tokenize it and store
 * the resulting strings in a strlist.  Return the strlist.
 */
strlist_t *split_strlist(char *in, char *delim) {
    char *str = NULL;
    char *end = NULL;
    char *tok = NULL;
    strlist_t *list = NULL;
    strlist_entry_t *entry = NULL;

    assert(in != NULL);

    str = strdup(in);
    end = str;
    list = init_strlist();

    while ((tok = strsep(&end, delim)) != NULL) {
        if ((entry = calloc(1, sizeof(strlist_entry_t))) == NULL) {
            warn("calloc");
            free(str);
            free_strlist(list);
            return NULL;
        }

        entry->val = strdup(tok);
        TAILQ_INSERT_TAIL(list, entry, items);
    }

    free(str);
    return list;
}
