/* Unpack a file from a .dist package. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <stdbool.h>
#include <string.h>
#include <limits.h>
#include <err.h>
#include <archive.h>
#include <archive_entry.h>
#include "dist-tools.h"

/* Local prototypes */
static int copy_data(struct archive *, struct archive *);
static int extract_entry(struct archive *, struct archive *, struct archive_entry *);

/*
 * From:
 * https://github.com/libarchive/libarchive/wiki/Examples
 * Reads data from the input archive stream and writes it to the output
 * archive stream.
 */
static int copy_data(struct archive *ar, struct archive *aw)
{
    int r;
    const void *buf = NULL;
    size_t s;
    off_t o;

    while ((r = archive_read_data_block(ar, &buf, &s, &o)) != ARCHIVE_EOF) {
        if (r != ARCHIVE_OK) {
            return r;
        }

        r = archive_write_data_block(aw, buf, s, o);

        if (r != ARCHIVE_OK) {
            warnx("archive_write_data_block: %s", archive_error_string(aw));
            return r;
        }
    }

    return ARCHIVE_OK;
}

/*
 * Unpack metadata from package and put the contents in buf.  None of
 * the parameters can be NULL.  buf will be grown as necessary, so make
 * sure it is allocated using calloc() or malloc() in the caller.
 *
 * NOTE:  buf is allocated or reallocated by this function, but must be
 * freed by the caller.
 */
int unpack_metadata2str(const char *package, const char *metadata, char *buf, bool fatal)
{
    la_ssize_t s = ARCHIVE_OK;
    int r = 0;
    bool found = false;
    char *reader = buf;
    struct archive *input = NULL;
    struct archive_entry *entry = NULL;

    assert(package != NULL);
    assert(metadata != NULL);
    assert(buf != NULL);

    /* open the package */
    input = archive_read_new();
#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_read_support_compression_all(input);
#else
    archive_read_support_filter_all(input);
#endif
    archive_read_support_format_all(input);
    r = archive_read_open_filename(input, package, 10240);

    if (r != ARCHIVE_OK) {
        warnx("archive_read_open_filename");
        return -1;
    }

    /* look for and read in the requested metadata file */
    while (archive_read_next_header(input, &entry) == ARCHIVE_OK) {
        if (!strcmp(archive_entry_pathname(entry), metadata)) {
            while ((s = archive_read_data(input, reader, BUFSIZ)) == ARCHIVE_OK) {
                buf = realloc(buf, strlen(buf) + 1 + BUFSIZ);
                assert(buf != NULL);
                reader += strlen(buf) + 1;
            }

            if (s == ARCHIVE_WARN || s == ARCHIVE_FAILED || s == ARCHIVE_FATAL) {
                warnx("archive_read_data");
                archive_read_close(input);
                return 1;
            }

            if (strlen(buf) > 0) {
                found = true;
            }

            break;
        }
    }

    /* close the package */
    r = archive_read_free(input);

    if (r != ARCHIVE_OK) {
        warnx("archive_read_free");
        return -1;
    }

    /* if we didn't find the metadata file, error */
    if (!found && fatal) {
        return -11;
    }

    return 0;
}

/* Extract a single entry from an archive and write it to disk. */
static int extract_entry(struct archive *input, struct archive *output, struct archive_entry *entry)
{
    int r = 0;
    int ret = 0;

    r = archive_write_header(output, entry);

    if (r != ARCHIVE_OK) {
        warnx("archive_write_header: %s", archive_error_string(output));
        ret = -1;
    } else if (archive_entry_size(entry) > 0) {
        if (copy_data(input, output) != ARCHIVE_OK) {
            ret = -1;
        }

        if (r != ARCHIVE_OK) {
            warnx("archive_write_header: %s", archive_error_string(output));
        } else if (r < ARCHIVE_WARN) {
            ret = -1;
        }
    }

    r = archive_write_finish_entry(output);

    if (r != ARCHIVE_OK) {
        warnx("archive_write_finish_entry: %s", archive_error_string(output));
    } else if (r < ARCHIVE_WARN) {
        ret = -1;
    }

    return ret;
}

/*
 * Unpack metadata from package and write it to the dir specified by
 * 'where'.  None of the parameters can be NULL.
 */
int unpack_metadata2file(const char *package, const char *metadata, const char *where, bool fatal)
{
    int ret = 0;
    int r = 0;
    char cwd[PATH_MAX + 1];
    struct archive *input = NULL;
    struct archive *output = NULL;
    struct archive_entry *entry = NULL;

    assert(package != NULL);
    assert(metadata != NULL);
    assert(where != NULL);
    memset(cwd, '\0', sizeof(cwd));

    /* open the package */
    input = archive_read_new();
#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_read_support_compression_all(input);
#else
    archive_read_support_filter_all(input);
#endif
    archive_read_support_format_all(input);

    if (archive_read_open_filename(input, package, 10240) != ARCHIVE_OK) {
        warn("archive_read_open_filename");
        return -1;
    }

    /* handle to write .dist members to disk */
    output = archive_write_disk_new();
    archive_write_disk_set_options(output, 0);
    archive_write_disk_set_standard_lookup(output);

    /* change to where */
    if (getcwd(cwd, PATH_MAX) == NULL) {
        err(EXIT_FAILURE, "getcwd");
    }

    if (chdir(where) != 0) {
        warn("chdir");
        return -1;
    }

    /* look for and read in the requested metadata file */
    while ((r = archive_read_next_header(input, &entry)) != ARCHIVE_EOF) {
        if (r != ARCHIVE_OK) {
            warnx("archive_read_next_header: %s", archive_error_string(input));
        } else if (r < ARCHIVE_WARN) {
            ret = -1;
        }

        if (!strcmp(archive_entry_pathname(entry), metadata)) {
            if (extract_entry(input, output, entry)) {
                ret = -1;
            }
        }
    }

    /* close the package */
    if (archive_read_free(input) != ARCHIVE_OK) {
        warnx("archive_read_free");
        return -1;
    }

#if ARCHIVE_VERSION_NUMBER < 3000000
    if (archive_write_finish(output) != ARCHIVE_OK) {
        warnx("archive_write_finish");
#else
    if (archive_write_free(output) != ARCHIVE_OK) {
        warnx("archive_write_free");
#endif
        return -1;
    }

    /* change back to original directory */
    if (chdir(cwd) != 0) {
        warn("chdir");
        return -1;
    }

    /* if we didn't find the metadata file, error */
    return (fatal ? ret : 0);
}

/*
 * Unpack an archive file to a destination directory.  The directory must
 * exist before calling this function.
 */
int unpack_archive(const char *archive, const char *dest, const bool force)
{
    int flags, r, ret = 0;
    char *rfilename = NULL;
    char cwd[PATH_MAX + 1];
    struct archive *input = NULL;
    struct archive *output = NULL;
    struct archive_entry *entry = NULL;

    assert(archive != NULL);
    assert(dest != NULL);
    memset(cwd, '\0', sizeof(cwd));

    /* attributes to restore */
    flags = ARCHIVE_EXTRACT_TIME;
    flags |= ARCHIVE_EXTRACT_PERM;
    flags |= ARCHIVE_EXTRACT_ACL;
    flags |= ARCHIVE_EXTRACT_FFLAGS;

    if (force) {
        /* user passed a -f flag, so try to force past errors unpacking */
        flags |= ARCHIVE_EXTRACT_UNLINK;
    }

    /* full location to the .dist */
    if ((rfilename = realpath(archive, rfilename)) == NULL) {
        warn("realpath");
        return -1;
    }

    /* .dist reader */
    input = archive_read_new();
#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_read_support_compression_all(input);
#else
    archive_read_support_filter_all(input);
#endif
    archive_read_support_format_all(input);
    r = archive_read_open_filename(input, archive, 10240);

    if (r != ARCHIVE_OK) {
        warnx("archive_read_open_filename");
        return -1;
    }

    /* change to dest */
    if (getcwd(cwd, PATH_MAX) == NULL) {
        err(EXIT_FAILURE, "getcwd");
    }

    if (chdir(dest) != 0) {
        warn("chdir");
        return -1;
    }

    /* handler to write .dist members to disk */
    output = archive_write_disk_new();
    archive_write_disk_set_options(output, flags);
    archive_write_disk_set_standard_lookup(output);

    /* extract each .dist member */
    while ((r = archive_read_next_header(input, &entry)) != ARCHIVE_EOF) {
        if (r != ARCHIVE_OK) {
            warnx("archive_read_next_header: %s", archive_error_string(input));
        } else if (r < ARCHIVE_WARN) {
            ret = -1;
        }

        if (extract_entry(input, output, entry)) {
            ret = -1;
        }
    }

    archive_read_free(input);
#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_write_finish(output);
#else
    archive_write_free(output);
#endif

    /* change back to original directory */
    if (chdir(cwd) != 0) {
        warn("chdir");
        return -1;
    }

    free(rfilename);
    return ret;
}
