/* Unquote a string like /bin/sh would. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <assert.h>
#include <errno.h>
#include <err.h>

/* Local prototypes */
static bool unquote_string_inplace(char *, char **);

/*
 * Single quotes preserve the literal string exactly.  Escape sequences
 * are not allowed; not even \' - if you want a ' in the quoted text,
 * you have to do something like 'foo'\''bar'
 *
 * Double quotes allow $ ' " \ and newline to be escaped with backslash.
 * Otherwise double quotes preserve things literally.
 *
 * More information
 * https://github.com/GNOME/glib/blob/master/glib/gshell.c
 */
static bool unquote_string_inplace(char *in, char **end)
{
    char *dest = NULL;
    char *s = NULL;
    char quote_char;

    assert(in != NULL);
    assert(end != NULL);

    dest = s = in;
    quote_char = *s;

    if (!(*s == '"' || *s == '\'')) {
        warnx("Quoted text doesn't begin with a quotation mark: %s", in);
        *end = in;
        return false;
    }

    /* Skip the initial quote mark */
    ++s;

    if (quote_char == '"') {
        while (*s) {
            assert(s > dest);

            switch (*s) {
                case '"':
                    /* End of string, return now */
                    *dest = '\0';
                    ++s;
                    *end = s;
                    return true;
                    break;

                case '\\':
                    /* Possible escaped quote or \ */
                    ++s;

                    switch (*s) {
                        case '"':
                        case '\\':
                        case '`':
                        case '$':
                        case '\n':
                            *dest = *s;
                            ++s;
                            ++dest;
                            break;
                        default:
                            /* not an escaped char */
                            *dest = '\\';
                            ++dest;
                            /* ++s already done */
                            break;
                    }

                    break;

                default:
                    *dest = *s;
                    ++dest;
                    ++s;
                    break;
            }

            assert(s > dest); /* loop invariant */
        }
    } else {
        while (*s) {
            assert(s > dest); /* loop invariant */

            if (*s == '\'') {
                /* End of string, return now */
                *dest = '\0';
                ++s;
                *end = s;
                return true;
            } else {
                *dest = *s;
                ++dest;
                ++s;
            }

            assert(s > dest); /* loop invariant */
        }
    }

    /* If we reach here this means the close quote was never encountered */
    *dest = '\0';
    warnx("Unmatched quotation mark in command line or other shell-quoted text: %s", in);
    *end = s;
    return false;
}

/*
 * This is a simplified version of g_shell_unquote() from glib,
 * which is what we were using before dropping out dependency on
 * glib.  There's no clearly defined behavior for shell unquoting,
 * so we need to maintain compatibility with what glib had been
 * doing for us.
 *
 * For more information on the glib version:
 * https://github.com/GNOME/glib/blob/master/glib/gshell.c
 *
 * The caller is responsible for freeing the returned string.
 */
char *unquote(const char *quoted)
{
    char *unquoted = NULL;
    char *end = NULL;
    char *start = NULL;
    char *retval = NULL;
    char *retwalk = NULL;

    assert(quoted != NULL);

    /* Create a working buffer to build the unquoted string */
    retval = calloc(1, strlen(quoted) + 1);
    assert(retval != NULL);

    unquoted = strdup(quoted);
    start = unquoted;
    end = unquoted;
    retwalk = retval;

    /*
     * The loop allows cases such as
     * "foo"blah blah'bar'woo foo"baz"la la la\'\''foo'
     */
    while (*start) {
        /* Append all non-quoted chaers, honoring backslash escape */
        while (*start && !(*start == '"' || *start == '\'')) {
            if (*start == '\\') {
                /*
                 * all characters can get escaped by backslash,
                 * except newline, which is removed if it follows
                 * a backslash outside of quotes
                 */
                ++start;

                if (*start) {
                    if (*start != '\n') {
                        *retwalk++ = *start;
                    }

                    ++start;
                }
            } else {
                *retwalk++ = *start;
                ++start;
            }
        }

        if (*start) {
            if (!unquote_string_inplace(start, &end)) {
                free(unquoted);
                free(retval);
                return NULL;
            } else {
                retwalk = stpcpy(retwalk, start);
                start = end;
            }
        }
    }

    free(unquoted);

    /* Shrink the unquoted string to only the size required */
    retval = realloc(retval, strlen(retval) + 1);
    assert(retval != NULL);
    return retval;
}
