#!/bin/sh
#
# build.sh: Implement build stage.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# Copyright Chris Lumens <chris@bangmoney.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $WRKDIR               The top level work directory.
#    $WRKSRC               The extracted source directory.
#    $WRKBLD               The source build directory.
#    $WRKSYSROOT           The --sysroot tree for this build.
#    $PREFIX               default: /usr
#    $SYSCONFDIR           default: /etc
#    $BINDIR               default: /usr/bin
#    $LIBDIR               default: /usr/lib
#    $LIBEXECDIR           default: /usr/libexec
#    $DATADIR              default: /usr/share
#    $INCLUDEDIR           default: /usr/include
#    $SEPARATE_BUILD       (y/n) Use source build directories?
#    $CONFIGURE_STYLE      The method used to configure $WRKSRC for builds.
#    $PKG                  The package name.
#    $VER                  The package version.
#    $PACKAGE              The full package name.
#    $CFLAGS               C compiler flags.
#    $CXXFLAGS             C++ compiler flags.
#    $LDFLAGS              Linker flags.
#    $PYTHON               Python interpreter to use.
#    $BUILD_ENV            Environment variables passed to $(BUILD_COMMAND).
#    $BUILD_TARGET         The name of the build target (default=all).
#    $BUILD_ARGS           Extra arguments to pass to $(BUILD_COMMAND).
#    $BUILD_COMMAND        The make command.
#    $PARALLEL_JOBS        Number of make(1) jobs to run (the -j option).
#    $TOPSRC               The top level distribution source directory.
#    $HELPERS              Path to the helpers.sh script.
#
# For CONFIGURE_STYLE=gnu
#    $CONFIGURE_ENV        Environment variables passed to ./configure
#    $CONFIGURE_OPTS       Extra options passed to ./configure
#    $AUTORECONF           Run 'autoreconf' on the package (y/n).
#    $GETTEXTIZE           Run 'gettextize' on the package (y/n).
# This script is called by the build system scripts and should not be called
# directly.

[ "${SKIP_BUILD}" = "y" ] && exit 0

if [ -z "${HELPERS}" ] || [ ! -f "${HELPERS}" ]; then
    exit 1
fi

# Helper functions
# shellcheck disable=SC1090
. "${HELPERS}"

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 2
fi

[ -z "${PACKAGE}" ] && exit 3

if [ ! -d "${WRKSRC}" ] && [ ! -d "${WRKDIR}"/src ]; then
    exit 4
fi

LOG="${WRKDIR}/build.log"
ERR="${WRKDIR}/build.err"
JOPT=

# Setup logging via a couple fifo's connected to tee
outlog="${TMPDIR:-/tmp}/out.$$"
errlog="${TMPDIR:-/tmp}/err.$$"
mkfifo "${outlog}" "${errlog}"
trap 'rm "${outlog}" "${errlog}"' EXIT

if [ -n "${PARALLEL_JOBS}" ] && [ "${PARALLEL_JOBS}" -gt 1 ]; then
    JOPT="-j ${PARALLEL_JOBS}"
fi

if [ -d "${WRKSRC}" ]; then
    for p in ${PACKAGE} ; do
        echo ">>> Building ${p}..."
    done

    if [ "${SEPARATE_BUILD}" = "y" ]; then
        [ -d "${WRKBLD}" ] || mkdir -p "${WRKBLD}"
        cd "${WRKBLD}" || exit 5
    else
        cd "${WRKSRC}" || exit 6
    fi

    case "${CONFIGURE_STYLE}" in
        gnu)
            # Regenerate autoconf files if requested
            if [ "${AUTORECONF}" = "y" ]; then
                if [ "${SEPARATE_BUILD}" = "y" ]; then
                    cd "${WRKSRC}" || exit 7
                fi

                [ -d m4 ] || mkdir -p m4

                run_tee
                # shellcheck disable=SC2086
                ( env ${CONFIGURE_ENV} ${WRKSYSROOT}${BINDIR}/autoreconf -v -f -i -B "${WRKSYSROOT}""${DATADIR}"/aclocal >"${outlog}" 2>"${errlog}" ) || exit $?

                # update any gettext files
                if [ "${GETTEXTIZE}" = "y" ]; then
                    if [ -d "${WRKSRC}"/po ] && [ -f "${WRKSRC}"/po/Makefile.in.in ]; then
                        run_tee
                        # shellcheck disable=SC2086
                        ( env ${CONFIGURE_ENV} ${WRKSYSROOT}${BINDIR}/gettextize --force --no-changelog >"${outlog}" 2>"${errlog}" ) || exit $?
                    fi
                fi

                # If there are resulting po/Makefile.in.in files, check to
                # see if it has datarootdir and if it doesn't, add it.  This
                # suppresses a harmless warning message, but makes our log
                # files nicer.
                find "${WRKSRC}" -name "Makefile.in.in" | grep "po/Makefile.in.in" | while read -r mfile ; do
                    if ! grep -q ^datarootdir "${mfile}" >/dev/null 2>&1 ; then
                        sed -i -e '/^datadir/a datarootdir = @datarootdir@' "${mfile}"
                    fi
                done

                if [ "${SEPARATE_BUILD}" = "y" ]; then
                    cd "${WRKBLD}" || exit 8
                fi
            fi

            # Run ./configure
            run_tee
            # shellcheck disable=SC2086
            ( env CFLAGS="${CFLAGS}" \
                  CPPFLAGS="${CFLAGS}" \
                  CXXFLAGS="${CXXFLAGS}" \
                  LDFLAGS="${LDFLAGS}" \
                  ${CONFIGURE_ENV} \
              "${WRKSRC}"/configure ${CONFIGURE_OPTS} \
              >"${outlog}" 2>"${errlog}"
            ) || exit $?

            # Run make
            run_tee
            # shellcheck disable=SC2086
            ( env ${BUILD_ENV} ${BUILD_COMMAND} ${JOPT} ${BUILD_TARGET} ${BUILD_ARGS} >"${outlog}" 2>"${errlog}" ) || exit $?
            ;;
        noconfigure|manual)
            if [ "${CONFIGURE_STYLE}" = "manual" ]; then
                echo "*** WARNING: CONFIGURE_STYLE=manual is deprecated, use CONFIGURE_STYLE=noconfigure"
            fi

            run_tee
            # shellcheck disable=SC2086
            ( env ${BUILD_ENV} ${BUILD_COMMAND} ${JOPT} ${BUILD_TARGET} ${BUILD_ARGS} >"${outlog}" 2>"${errlog}" ) || exit $?
            ;;
        haskell)
            cp "${TOPSRC}"/base/haskell/haskell.mk "${WRKSRC}"/Makefile
            cd "${WRKSRC}" || exit 9

            run_tee
            # shellcheck disable=SC2086
            ( env ${BUILD_ENV} \
                  PREFIX="${PREFIX}" \
                  BINDIR="${BINDIR}" \
                  LIBDIR="${LIBDIR}" \
                  DATADIR="${DATADIR}" \
                  PACKAGE="${PACKAGE}" \
                  CONFIGURE_OPTS="${CONFIGURE_OPTS}" \
              ${BUILD_COMMAND} ${JOPT} build ${BUILD_ARGS} \
              >"${outlog}" 2>"${errlog}"
            ) || exit $?

            RET=$?

            # Haskell stuff can silently build to completion but incorrectly,
            # check for the rogue errors that are really important.
            BADSTR="Setup: Encountered missing dependencies:"
            if grep -q "${BADSTR}" "${ERR}" >/dev/null 2>&1 ; then
                echo "*** HACKAGE: Haskell build failed with missing dependencies:" | tee -a "${ERR}"

                # Errors found, extract the blocks
                grep -n "${BADSTR}" "${ERR}" | cut -d ':' -f 1 | while read -r badline ; do
                    # Give the module name
                    modline=$((badline - 1))
                    line="$(sed -n "${modline} p" "${ERR}")"
                    echo "*** HACKAGE: ${line}" | tee -a "${ERR}"

                    # Find the bad line starting point
                    badline=$((badline + 1))
                    line="$(sed -n "${badline} p" "${ERR}")"

                    while [ -n "${line}" ]; do
                        echo "*** HACKAGE:     ${line}" | tee -a "${ERR}"
                        badline=$((badline + 1))
                        line="$(sed -n "${badline} p" "${ERR}")"
                    done
                done
            fi

            cp -p "${ERR}" "${LOG}"
            [ ${RET} -ne 0 ] && exit ${RET}
            ;;
        cmake)
            run_tee
            # shellcheck disable=SC2086
            ( env ${BUILD_ENV} \
                  CMAKE_C_FLAGS="${CFLAGS}" \
                  CMAKE_CXX_FLAGS="${CXXFLAGS}" \
                  CMAKE_C_LINK_FLAGS="${LDFLAGS}" \
                  CMAKE_CXX_LINK_FLAGS="${LDFLAGS}" \
              cmake ${CONFIGURE_OPTS} "${WRKSRC}"
              # shellcheck disable=SC2086
              env ${BUILD_ENV} ${BUILD_COMMAND} ${JOPT} ${BUILD_TARGET} ${BUILD_ARGS} >"${outlog}" 2>"${errlog}"
            ) || exit $?
            ;;
        python)
            run_tee
            # shellcheck disable=SC2086
            ( env ${BUILD_ENV} ${PYTHON} setup.py --no-user-cfg ${BUILD_TARGET} ${CONFIGURE_OPTS} ${BUILD_ARGS} >"${outlog}" 2>"${errlog}" ) || exit $?
            ;;
        pyproject)
            run_tee
            # shellcheck disable=SC2086
            ( env ${BUILD_ENV} ${PYTHON} -m pip ${BUILD_TARGET} ${CONFIGURE_OPTS} ${BUILD_ARGS} . >"${outlog}" 2>"${errlog}" ) || exit $?
            ;;
        qmake)
            # qmake is a terrible program when it comes to integrating it
            # with build systems.  We need to write a qt.conf file that
            # defines the Qt prefix to the sysroot location.  This file
            # needs to live in the same directory as qmake, which is the
            # /usr/bin directory in the sysroot.  Dumb, right?
            #
            # Figuring out these variable names was not trivial.  I finally
            # came across this page:
            # https://doc.qt.io/qt-5/qt-conf.html

            qtconf="${WRKSYSROOT}${BINDIR}/qt.conf"
            echo "[Paths]" > "${qtconf}"
            echo "Sysroot = ${WRKSYSROOT}" >> "${qtconf}"
            echo "Prefix = ${WRKSYSROOT}${X11_LIBDIR}/qt" >> "${qtconf}"
            echo "Headers = ${WRKSYSROOT}${X11_INCLUDEDIR}/qt" >> "${qtconf}"
            echo "Documentation = ${WRKSYSROOT}${DATADIR}/doc/qt" >> "${qtconf}"
            echo "Libraries = ${WRKSYSROOT}${X11_LIBDIR}" >> "${qtconf}"
            echo "LibraryExecutables = ${WRKSYSROOT}${X11_LIBEXECDIR}/qt" >> "${qtconf}"
            echo "Binaries = ${WRKSYSROOT}${X11_BINDIR}" >> "${qtconf}"
            echo "Settings = ${WRKSYSROOT}${SYSCONFDIR}/qt" >> "${qtconf}"

            [ -f "${WRKSYSROOT}${X11_BINDIR}/qmake" ] && QMAKE="${WRKSYSROOT}${X11_BINDIR}/qmake" || QMAKE="qmake"

            # Run qmake, which is sort of like configure.
            run_tee
            # shellcheck disable=SC2086
            ( env ${BUILD_ENV} ${CONFIGURE_ENV} ${QMAKE} ${CONFIGURE_OPTS} >"${outlog}" 2>"${errlog}" ) || exit $?

            # Now run 'make'
            run_tee
            # shellcheck disable=SC2086
            ( env ${BUILD_ENV} ${BUILD_COMMAND} ${JOPT} ${BUILD_TARGET} ${BUILD_ARGS} >"${outlog}" 2>"${errlog}" ) || exit $?
            ;;
        meson)
            run_tee
            ( cd "${WRKSRC}"
              # shellcheck disable=SC2086
              env ${CONFIGURE_ENV} meson setup "${WRKBLD}" ${CONFIGURE_OPTS}
              # shellcheck disable=SC2086
              env ${BUILD_ENV} ninja -C "${WRKBLD}" ${BUILD_TARGET} ${BUILD_ARGS} >"${outlog}" 2>"${errlog}"
            ) || exit $?
            ;;
        cargo)
            run_tee
            ( cd "${WRKBLD}"
              mkdir -p .cargo
              # shellcheck disable=SC2086
              env ${CONFIGURE_ENV} cargo vendor -v > .cargo/config.toml
              # shellcheck disable=SC2086
              env ${BUILD_ENV} cargo build -v --offline --release --all-targets ${BUILD_ARGS} >"${outlog}" 2>"${errlog}"
            ) || exit $?
            ;;
        perl)
            run_tee
            ( cd "${WRKBLD}"
              # shellcheck disable=SC2086
              env ${CONFIGURE_ENV} ${PERL} Makefile.PL INSTALLDIRS=vendor ${CONFIGURE_OPTS} < /dev/null
              # shellcheck disable=SC2086
              env ${BUILD_ENV} ${BUILD_COMMAND} -C "${WRKBLD}" ${BUILD_TARGET} ${BUILD_ARGS} >"${outlog}" 2>"${errlog}"
            ) || exit $?
            ;;
        *)
            echo "*** Do not understand CONFIGURE_STYLE=${CONFIGURE_STYLE}"
            exit 10
            ;;
    esac
fi

[ -s "${LOG}" ] || rm -f "${LOG}"
[ -s "${ERR}" ] || rm -f "${ERR}"
