#!/usr/bin/python3 -E
#
# buildreqs.py: A new-style way of getting BUILD_DEPENDS for a package.
#               This script gets the direct build requirements for the
#               package being built, and then recursively checks the
#               build requirements of all those, building up a complete
#               set of everything needed.  The results are printed in
#               sorted order on stdout.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

import os
import sys
import subprocess

def valid_mach(pkg, target_mach):
    proc = subprocess.Popen(["make", "--no-print-directory", "-C", pkg, "show", "var=TARGET_MACH"],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (mach, err) = proc.communicate()
    mach = mach.decode("utf-8").strip()
    return mach == target_mach or mach == "noarch"

if __name__ == "__main__":
    cwd = os.getcwd()
    topsrc = os.getenv("TOPSRC")
    buildreqs_paths = os.getenv("BUILDREQS_PATHS").split()
    wrkdir = os.getenv("WRKDIR")
    target_mach = os.getenv("TARGET_MACH")
    reduce = os.getenv("REDUCE")

    if not os.path.isfile(os.path.join(cwd, "buildreqs")):
        # not an error, but nothing to do
        sys.exit(0)

    if topsrc is None or not os.path.isdir(topsrc):
        sys.exit(1)

    if buildreqs_paths is None:
        sys.exit(2)
    else:
        for brpath in buildreqs_paths:
            if not os.path.isdir(brpath):
                sys.exit(3)

    if wrkdir is None or not os.path.isdir(wrkdir):
        sys.exit(4)

    if target_mach is None:
        sys.exit(5)

    # start by reading in the current buildreqs file
    f = open(os.path.join(cwd, "buildreqs"), "r")
    buildreqs = set(f.read().split())
    f.close()

    # initialize the worklist as a copy of buildreqs
    worklist = buildreqs.copy()

    # packages will contain ALL packages installed for buildreqs, even
    # the indirect ones
    packages = buildreqs.copy()

    # go through the worklist until there is nothing left
    while len(worklist) != 0:
        # Pick a worklist entry to work with
        entry = list(worklist)[0]

        # Handle any blank entries
        if entry == "" or entry is None:
            worklist.remove(entry)
            continue

        # Skip reqs that do not apply to this architecture nor noarch
        for brpath in buildreqs_paths:
            # package does not exist in this buildreqs tree
            if not os.path.isdir(os.path.join(brpath, entry)):
                continue

            # check the target architecture
            if not valid_mach(os.path.join(brpath, entry), target_mach):
                worklist.remove(entry)
                entry = None
                break

        if entry is None:
            continue

        # Add the first package in the work list to the packages list
        packages.add(entry)

        # Then add all the build dependencies of that package to the end of the
        # work list.  Note that not all packages have build dependencies.  eglibc,
        # for example...
        package_deps = set()
        for brpath in buildreqs_paths:
            brfile = os.path.join(brpath, entry, "buildreqs")

            if not os.path.isfile(brfile):
                continue

            f = open(brfile, "r")
            deplist = f.read().split()
            f.close()

            wl = list(packages)
            for dep in deplist:
                # see if we already have the dep
                try:
                    i = wl.index(dep)
                    if dep in buildreqs:
                        buildreqs.remove(dep)
                    continue
                except ValueError:
                    pass

                deppath = os.path.join(brpath, dep)

                if cwd == deppath:
                    sys.stderr.write("\n*** Circular dependency found!\n")
                    sys.stderr.write("*** %s requires %s\n" % (entry, cwd[len(brpath)+1:]))
                    sys.exit(6)

                if not os.path.isdir(deppath):
                    continue

                if valid_mach(deppath, target_mach):
                    if dep in buildreqs:
                        buildreqs.remove(dep)
                    package_deps.add(dep)

        # determine new deps that we have not yet looked at
        new_deps = package_deps.difference(packages)
        worklist = worklist.union(new_deps)

        # add these package deps to the main list of expanded deps
        packages = packages.union(package_deps)

        # remove the dependency we were working with
        worklist.remove(entry)

    # Our packages list now probably has lots of dupes in it.  Let's
    # filter those out so we don't try to install the same packages
    # over and over again.  Once again, it's probably glibc that's
    # going to be here a dozen times.
    if reduce == "y":
        f = open(os.path.join(cwd, "buildreqs"), "w")
        for entry in buildreqs:
            f.write("%s\n" % entry)
        f.close()
    else:
        output = list(packages)
        output.sort()

        for entry in output:
            print(entry)
