#!/bin/sh
#
# bundle.sh: Runs 'make world' then creates a bundle with the newly built
#            packages and integration script to merge it with an existing
#            built tree.  Bundles can be uploaded to the download server
#            and integrated in to the tree.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC                The top level of the source tree.
#    $WRKDIR                The package work directory.
#    $MAKE                  The make command.
#    $PKG                   The name of the package.
#    $FULLPKGPATH           Path to the built package.
#    $FULLDEBUGPKGPATH      Path to the built debug package.
#    $PACKAGE               The basename of the built package.
#    $DEBUGPACKAGE          The basename of the built debug package.
#    $DISTRIBUTION_VERSION  The version number of the entire distribution.
#    $PKGCOLLECTION         The collection name for this package.
#    $TARGET_MACH           The build architecture or 'noarch'.
#    $RELEASE               Path to the main release tree.
#    $HELPERSDB             Path to the helpers-db.sh file.
#    $MERGE_SH_IN           Path to the merge.sh.in template.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

[ -z "${WRKDIR}" ] && exit 2
[ -z "${MAKE}" ] && exit 3
[ -z "${PKG}" ] && exit 4
[ -z "${FULLPKGPATH}" ] && exit 5
[ -z "${DISTRIBUTION_VERSION}" ] && exit 6
[ -z "${PKGCOLLECTION}" ] && exit 7
[ -z "${TARGET_MACH}" ] && exit 8
[ -z "${RELEASE}" ] && exit 9
[ -z "${MERGE_SH_IN}" ] && exit 10

TMPDIR="$(mktemp -d)"
makecmd="${MAKE} --no-print-directory"

trap 'rm -rf "${TMPDIR}"' EXIT

# Is this a stable branch update or not?
if git rev-parse --abbrev-ref HEAD | grep -q -E '-stable$' >/dev/null 2>&1 ; then
    STABLE="y"
else
    STABLE="n"
fi

# Build the package.  Exit if it fails
${MAKE} world || exit 11

# Create a name for this bundle directory
BUNDLE_NAME="${PKG}-$(date +%s)"
BUNDLE="${WRKDIR}/${BUNDLE_NAME}"
rm -f "${BUNDLE}"
mkdir -p "${BUNDLE}"

# Copy in new packages to the bundle
for subpkg in ${FULLPKGPATH} ; do
    cp -p "${subpkg}" "${BUNDLE}"/"$(basename "${subpkg}")"

    if [ -n "${FULLDEBUGPKGPATH}" ]; then
        for subdebug in ${FULLDEBUGPKGPATH} ; do
            cp -p "${subdebug}" "${BUNDLE}"/"$(basename "${DEBUGPACKAGE}")" || exit 11
        done
    fi
done

# Generate a possible list of old packages to remove
PREV="$(${makecmd} showprevpkg)"
line="${DISTRIBUTION_VERSION}/${PKGCOLLECTION}/${TARGET_MACH}/${PREV}.dist"
[ "${STABLE}" = "n" ] && line="dist/${line}"
echo "${line}" > "${BUNDLE}"/old-packages.txt

# Debug package only on non-noarch
if [ ! "${TARGET_MACH}" = "noarch" ]; then
    PREVDEBUG="$(echo "${PREV}" | sed 's/\(.*\)\./\1.debug./')"
    line="${DISTRIBUTION_VERSION}/${PKGCOLLECTION}/${TARGET_MACH}/${PREVDEBUG}.dist"
    [ "${STABLE}" = "n" ] && line="dist/${line}"
    echo "${line}" >> "${BUNDLE}"/old-packages.txt
fi

# Create a merge script based on the template
cp "${MERGE_SH_IN}" "${BUNDLE}"/merge.sh
sed -i -e "/%%HELPERS_DB%%/ r ${HELPERSDB}" "${BUNDLE}"/merge.sh
sed -i -e "/%%HELPERS_DB%%/d" "${BUNDLE}"/merge.sh
sed -i -e "s|%%DIGEST_ALGORITHM%%|${DIGEST_ALGORITHM}|g" "${BUNDLE}"/merge.sh
sed -i -e "s|%%PACKAGE%%|${PACKAGE}|g" "${BUNDLE}"/merge.sh
sed -i -e "s|%%DEBUGPACKAGE%%|${DEBUGPACKAGE}|g" "${BUNDLE}"/merge.sh

if [ "${STABLE}" = "y" ]; then
    sed -i -e "s|%%PACKAGEPATH%%|${DISTRIBUTION_VERSION}/${PKGCOLLECTION}/${TARGET_MACH}|g" "${BUNDLE}"/merge.sh
    sed -i -e "s|%%DEBUGPACKAGEPATH%%|${DISTRIBUTION_VERSION}/debug/${PKGCOLLECTION}/${TARGET_MACH}|g" "${BUNDLE}"/merge.sh
else
    sed -i -e "s|%%PACKAGEPATH%%|dist/${DISTRIBUTION_VERSION}/${PKGCOLLECTION}/${TARGET_MACH}|g" "${BUNDLE}"/merge.sh
    sed -i -e "s|%%DEBUGPACKAGEPATH%%|dist/${DISTRIBUTION_VERSION}/debug/${PKGCOLLECTION}/${TARGET_MACH}|g" "${BUNDLE}"/merge.sh
fi

chmod 0755 "${BUNDLE}"/merge.sh

# Tar up the bundle and report the info
cd "${WRKDIR}" || exit 12
rm -f ./*.bundle
printf "Creating update bundle %s.bundle..." "${BUNDLE}"
tar -cvf "${BUNDLE_NAME}".bundle "${BUNDLE_NAME}" >/dev/null 2>&1
echo "DONE."

# Clean up and exit
exit 0
