#!/bin/sh
#
# check-upstream.sh: Implement check-upstream target.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level project directory.
# Variables provided for optional purposes when an upstream file is
# not present in the source tree or is missing values in order to take
# the build system defaults:
#    $PKG                  The package name.
#    $VER                  The package version.
#    $URLS                 The list of download URLs.
#    $UPSTREAM_URLS        The list of download URLs.
#    $UPSTREAM_GITREPO     The upstream project's git source repository.
#    $UPSTREAM_REGEXP      The version checking regular expression.
#    $UPSTREAM_PREFIX      The prefix trimming regular expression.
#    $UPSTREAM_SUFFIX      The suffix trimming regular expression.
#    $UPSTREAM_FILTER      Optional filter script to get latest version.
#    $HTML_OUTPUT          (y/n) Set to y to output an HTML page.
#    $HASKELL              (Y/n) Whether or not to include Haskell packages.
#    $UPDATE               (y/n) Set to y to let the script update the
#                          version in source files.
# This script is called by the build system scripts and should not be called
# directly.

# Strip all HTML tags from a string
html_filter() {
    sed -E 's/<[^>]*>//g' "$*" | sed -E 's/<[^>]*>$//g' | tr -s ' ' | grep -vE '^[ \t]*$' | sed 's/\r$//'
}

# Gather all version information.
find_latest() {
    ret="${UPSTREAMS}"/ret
    tmpfile="$(mktemp -p "${UPSTREAMS}")"
    curl_opts="-A Mozilla -s -f -L --retry 3 --ftp-pasv --stderr /dev/null"

    [ -f "${ret}" ] && rm -f "${ret}"

    # The default is always just checking the download URLs, so check any
    # optional methods first.
    if [ -n "${gitrepo}" ]; then
        # Check a git source repo tags for latest release
        git ls-remote --tags "${gitrepo}" | grep -E "${regexp}" > "${tmpfile}"
    elif [ -n "${hackage}" ]; then
        # Get list of preferred versions from hackage.haskell.org
        # shellcheck disable=SC2086
        curl ${curl_opts} -X GET -H "Content-type: application/json" -H "Accept: application/json" \
            https://hackage.haskell.org/package/"${hackage}"/preferred | \
            tr ',' '\n' | grep -v deprecated | cut -d '[' -f 2 | cut -d ']' -f 1 | \
            cut -d '"' -f 2 | grep -E '^[0-9\.]+$' > "${tmpfile}"
    elif [ -n "${stackage}" ]; then
        # Get latest version of the package from the LTS Stackage list
        # for the version of ghc in the source tree.
        LTS_VER="$(make --no-print-directory -C "${TOPSRC}"/base/haskell/stackage show var=VER)"

        # Find the Stackage LTS release for this version of ghc
        curlout="$(mktemp -p "${UPSTREAMS}")"
        # shellcheck disable=SC2086
        curl ${curl_opts} -o "${curlout}" "https://www.stackage.org/lts-${LTS_VER}" 2>/dev/null

        grep "href=" "${curlout}" | sed -e 's|href=|\nhref=|g' | grep "lts-${LTS_VER}" | \
            cut -d '"' -f 2 | rev | cut -d '/' -f 1 | rev | \
            grep -E "^${stackage}-[0-9\.]+$" | sed -e "s|^${stackage}-||g" > "${tmpfile}"

        rm -f "${curlout}"
    elif [ -n "${urls}" ]; then
        # Check one or more HTTP, HTTPS, or FTP sites for the latest version.
        for site in ${urls} ; do
            protocol="$(echo "${site}" | cut -d ':' -f 1)"

            if [ "${protocol}" = "ftp" ]; then
                # make sure the url ends with a slash, otherwise curl gets confused
                trail="$(echo "${site}" | rev | cut -c1)"
                [ "${trail}" = "/" ] || site="${site}/"

                # force an FTP NLST command to get filenames only
                curl_opts="${curl_opts} -l"
            elif [ "${protocol}" = "http" ] || [ "${protocol}" = "https" ]; then
                # perform an insecure download with curl (may want to change this
                # in the future?)
                curl_opts="${curl_opts} -k"
            else
                echo "*** ERROR: Unsupported upstream URL protocol for ${PKGPATH}: ${protocol}" >&2
                continue
            fi

            # collect upstream info and strip HTML tags
            # shellcheck disable=SC2086
            curlout="$(mktemp -p "${UPSTREAMS}")"
            # shellcheck disable=SC2086
            curl ${curl_opts} -o "${curlout}" "${site}" 2>/dev/null
            t="$(file -b --mime-type "${curlout}")"

            if [ "${t}" = "application/x-gzip" ] || [ "${t}" = "application/gzip" ]; then
                # some web servers give you gzip compressed web pages
                mv "${curlout}" "${curlout}.gz"
                gzip -d "${curlout}.gz"
            fi

            # grab all lines that match the upstream regular expression or use the filter
            filtercmd=
            if [ -n "${filter}" ]; then
                filtercmd="$(realpath "${filter}" 2>/dev/null)"
            fi

            if [ -x "${filtercmd}" ]; then
                html_filter "${curlout}" | ${filtercmd} > "${tmpfile}"
            else
                html_filter "${curlout}" | grep -E "${regexp}" > "${tmpfile}"
            fi

            rm -f "${curlout}"
        done
    fi

    # if we have exclude expression, filter out those lines
    if [ -n "${exclude}" ]; then
        grep -vE "${exclude}" "${tmpfile}" > "${tmpfile}.new"
        mv "${tmpfile}.new" "${tmpfile}"
    fi

    # trim the prefix and suffix
    if [ -n "${prefix}" ]; then
        if echo "${prefix}" | grep -q '/' >/dev/null 2>&1 ; then
            while read -r line ; do
                echo "${line}" | sed -E "s#^${prefix}##g"
            done < "${tmpfile}" > "${tmpfile}.new"
        else
            while read -r line ; do
                echo "${line}" | sed -E "s/^${prefix}//g"
            done < "${tmpfile}" > "${tmpfile}.new"
        fi

        mv "${tmpfile}.new" "${tmpfile}"
    fi

    if [ -n "${suffix}" ]; then
        if echo "${suffix}" | grep -q '/' >/dev/null 2>&1 ; then
            while read -r line ; do
                echo "${line}" | sed -E "s#${suffix}\$##g"
            done < "${tmpfile}" > "${tmpfile}.new"
        else
            while read -r line ; do
                echo "${line}" | sed -E "s/${suffix}$//g"
            done < "${tmpfile}" > "${tmpfile}.new"
        fi

        mv "${tmpfile}.new" "${tmpfile}"
    fi

    # Do a final trim of any surrounding whitespace
    awk '{$1=$1};1' < "${tmpfile}" > "${tmpfile}.new"
    mv "${tmpfile}.new" "${tmpfile}"

    # version filtering
    # Many projects use odd numbered major or minor versions to
    # indicate an unstable or development release.
    while read -r line ; do
        parts=$(echo "${line}" | tr '.' '\n' | wc -l)

        if [ "${parts}" -le 1 ]; then
            echo "${line}"
            continue
        fi

        filter=

        if [ -n "${skip_odd}" ]; then
            if [ "${skip_odd}" = "major" ]; then
                filter="$(echo "${line}" | cut -d '.' -f 1)"
            elif [ "${skip_odd}" = "minor" ]; then
                filter="$(echo "${line}" | cut -d '.' -f 2)"
            fi
        fi

        if [ -n "${filter}" ]; then
            num="$(echo "${filter}" | sed -r 's/^0+//g')"
            isodd=$((num % 2))
            [ "${isodd}" -eq 0 ] && echo "${line}"
        else
            echo "${line}"
        fi
    done < "${tmpfile}" > "${ret}"

    # remove duplicates, natural sort, and get latest
    latest="$(sort -uV "${ret}" | tail -n 1)"

    # clean up
    rm -f "${ret}"
    rm -f "${tmpfile}"

    echo "${latest}"
}

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

CWD="$(pwd)"
UPSTREAMS="$(mktemp -d)"
UPSTREAM_FILE="${CWD}"/upstream
PKGPATH="$(echo "${CWD}" | sed -e "s|${TOPSRC}/||g")"

trap 'rm -rf "${UPSTREAMS}"' EXIT

# create a directory to hold transformed upstream file fields
[ -d "${UPSTREAMS}" ] || mkdir -p "${UPSTREAMS}"

if [ -f "${UPSTREAM_FILE}" ]; then
    # transform upstream file to shell scripts
    output=
    have_ver=0
    have_urls=0
    have_gitrepo=0
    have_hackage=0
    have_stackage=0
    have_regexp=0
    have_prefix=0
    have_suffix=0
    have_filter=0

    while read -r line ; do
        # skip blank lines and comments
        [ -z "${line}" ] && continue
        [ "$(echo "${line}" | cut -c1)" = "#" ] && continue

        # do we have a new block to start?
        start="$(echo "${line}" | cut -c1)"
        end="$(echo "${line}" | rev | cut -c1)"

        if [ "${start}" = "[" ] && [ "${end}" = "]" ]; then
            # write out any missing values for the previous block
            if [ -n "${output}" ]; then
                if [ ${have_ver} -eq 0 ] && [ -n "${VER}" ]; then
                    printf "ver=\"%s\"\n" "${VER}" >> "${output}"
                fi

                if [ ${have_urls} -eq 0 ] && [ -n "${UPSTREAM_URLS}" ]; then
                    printf "urls=\"%s\"\n" "${UPSTREAM_URLS}" >> "${output}"
                fi

                if [ ${have_gitrepo} -eq 0 ] && [ -n "${UPSTREAM_GITREPO}" ]; then
                    printf "gitrepo=\"%s\"\n" "${UPSTREAM_GITREPO}" >> "${output}"
                fi

                if [ ${have_hackage} -eq 0 ] && [ -n "${UPSTREAM_HACKAGE}" ]; then
                    printf "hackage=\"%s\"\n" "${UPSTREAM_HACKAGE}" >> "${output}"
                fi

                if [ ${have_stackage} -eq 0 ] && [ -n "${UPSTREAM_STACKAGE}" ]; then
                    printf "stackage=\"%s\"\n" "${UPSTREAM_STACKAGE}" >> "${output}"
                fi

                if [ ${have_regexp} -eq 0 ] && [ -n "${UPSTREAM_REGEXP}" ]; then
                    printf "regexp=\"%s\"\n" "${UPSTREAM_REGEXP}" >> "${output}"
                fi

                if [ ${have_prefix} -eq 0 ] && [ -n "${UPSTREAM_PREFIX}" ]; then
                    printf "prefix=\"%s\"\n" "${UPSTREAM_PREFIX}" >> "${output}"
                fi

                if [ ${have_suffix} -eq 0 ] && [ -n "${UPSTREAM_SUFFIX}" ]; then
                    printf "suffix=\"%s\"\n" "${UPSTREAM_SUFFIX}" >> "${output}"
                fi

                if [ ${have_filter} -eq 0 ] && [ -n "${UPSTREAM_FILTER}" ]; then
                    printf "filter=\"%s\"\n" "${UPSTREAM_FILTER}" >> "${output}"
                fi
            fi

            # reset checks for default overrides
            have_ver=0
            have_urls=0
            have_gitrepo=0
            have_hackage=0
            have_stackage=0
            have_regexp=0
            have_prefix=0
            have_suffix=0
            have_filter=0

            # new block in the upstream file
            label="$(echo "${line}" | cut -d '[' -f 2 | cut -d ']' -f 1)"
            output="${UPSTREAMS}/$(echo "${label}" | tr ' ' '_').sh"
            echo "label=\"${label}\"" > "${output}"
            continue
        fi

        # lines in each block are in 'key: value' format
        key="$(echo "${line}" | cut -d ' ' -f 1 | cut -d ':' -f 1)"
        len=$((${#key} + 2))
        value="$(echo "${line}" | cut -c${len}- | awk '{$1=$1};1')"

        # convert booleans to ints for later use
        if [ "${key}" = "no_upstream" ]; then
            case ${value} in
                [Yy][Ee][Ss]|Y|y|[Tt][Rr][Uu][Ee]|T|t)
                    value="1" ;;
                [Nn][Oo]|[Ff][Aa][Ll][Ss][Ee]|N|n|F|f)
                    value="0" ;;
            esac
        fi

        # write out the line
        echo "$key=\"$value\"" >> "${output}"

        # note any default overrides we had
        [ "${key}" = "ver" ] && have_ver=1
        [ "${key}" = "urls" ] && have_urls=1
        [ "${key}" = "gitrepo" ] && have_gitrepo=1
        [ "${key}" = "hackage" ] && have_hackage=1
        [ "${key}" = "stackage" ] && have_stackage=1
        [ "${key}" = "regexp" ] && have_regexp=1
        [ "${key}" = "prefix" ] && have_prefix=1
        [ "${key}" = "suffix" ] && have_suffix=1
        [ "${key}" = "filter" ] && have_filter=1
    done < "${UPSTREAM_FILE}"

    # write out any missing values for the previous block
    # this catches upstream files with one block, the loop
    # above handles additional blocks in the upstream file
    if [ -n "${output}" ]; then
        if [ ${have_ver} -eq 0 ] && [ -n "${VER}" ]; then
            printf "ver=\"%s\"\n" "${VER}" >> "${output}"
        fi

        if [ ${have_urls} -eq 0 ] && [ -n "${UPSTREAM_URLS}" ]; then
            printf "urls=\"%s\"\n" "${UPSTREAM_URLS}" >> "${output}"
        fi

        if [ ${have_gitrepo} -eq 0 ] && [ -n "${UPSTREAM_GITREPO}" ]; then
            printf "gitrepo=\"%s\"\n" "${UPSTREAM_GITREPO}" >> "${output}"
        fi

        if [ ${have_hackage} -eq 0 ] && [ -n "${UPSTREAM_HACKAGE}" ]; then
            printf "hackage=\"%s\"\n" "${UPSTREAM_HACKAGE}" >> "${output}"
        fi

        if [ ${have_stackage} -eq 0 ] && [ -n "${UPSTREAM_STACKAGE}" ]; then
            printf "stackage=\"%s\"\n" "${UPSTREAM_STACKAGE}" >> "${output}"
        fi

        if [ ${have_regexp} -eq 0 ] && [ -n "${UPSTREAM_REGEXP}" ]; then
            printf "regexp=\"%s\"\n" "${UPSTREAM_REGEXP}" >> "${output}"
        fi

        if [ ${have_prefix} -eq 0 ] && [ -n "${UPSTREAM_PREFIX}" ]; then
            printf "prefix=\"%s\"\n" "${UPSTREAM_PREFIX}" >> "${output}"
        fi

        if [ ${have_suffix} -eq 0 ] && [ -n "${UPSTREAM_SUFFIX}" ]; then
            printf "suffix=\"%s\"\n" "${UPSTREAM_SUFFIX}" >> "${output}"
        fi

        if [ ${have_filter} -eq 0 ] && [ -n "${UPSTREAM_FILTER}" ]; then
            printf "filter=\"%s\"\n" "${UPSTREAM_FILTER}" >> "${output}"
        fi
    fi
else
    # default case for packages without an 'upstream' file
    output="${UPSTREAMS}"/"${PKG}".sh
    { printf "label=\"%s\"\n" "${PKG}"
      printf "ver=\"%s\"\n" "${VER}"
      # shellcheck disable=SC2153
      printf "urls=\"%s\"\n" "${URLS}"
      printf "gitrepo=\"%s\"\n" "${UPSTREAM_GITREPO}"
      printf "hackage=\"%s\"\n" "${UPSTREAM_HACKAGE}"
      printf "stackage=\"%s\"\n" "${UPSTREAM_STACKAGE}"
      printf "regexp=\"%s\"\n" "${UPSTREAM_REGEXP}"
      printf "prefix=\"%s\"\n" "${UPSTREAM_PREFIX}"
      printf "suffix=\"%s\"\n" "${UPSTREAM_SUFFIX}"
      printf "filter=\"%s\"\n" "${UPSTREAM_FILTER}"
    } >> "${output}"
fi

# iterate over each upstream source
for upstream in "${UPSTREAMS}"/*.sh ; do
    # Skip Haskell packages or not?
    if [ -n "${HASKELL}" ]; then
        echo "${PKGPATH}" | grep -q "^base/haskell" >/dev/null 2>&1
        haskell=$?

        if [ "${HASKELL}" = "y" ] && [ ${haskell} -eq 1 ]; then
            continue
        fi

        if [ "${HASKELL}" = "n" ] && [ ${haskell} -eq 0 ]; then
            continue
        fi
    fi

    # optional booleans that may be overridden by the upstream data
    no_upstream=0

    # Reset variables before each section runs
    label=
    ver=
    urls=
    gitrepo=
    hackage=
    stackage=
    regexp=
    prefix=
    suffix=
    filter=
    exclude=
    skip_odd=

    # source in this block's values
    # shellcheck disable=SC1090
    . "${upstream}"

    # some packages have no upstream source, we are it
    if [ -n "${no_upstream}" ] && [ "${no_upstream}" -eq 1 ]; then
        continue
    fi

    # pick a URL to use
    [ -z "${urls}" ] || linkdest="${urls}"
    [ -z "${gitrepo}" ] || linkdest="${gitrepo}"

    # look for the latest upstream version
    latest="$(find_latest)"
    if [ -z "${latest}" ] && [ ! "${UPDATE}" = "y" ]; then
        if [ "${HTML_OUTPUT}" = "y" ]; then
            echo "<tr><td>${PKGPATH}</td><td style=\"background-color: red;\"><p style=\"color:white\">UNKNOWN</p></td><td><a href=\"${linkdest}\">${label}</a></td></tr>"
        else
            echo "Unable to determine the latest version of ${label} for ${PKGPATH}"
        fi
    elif [ ! "${latest}" = "${ver}" ]; then
        if [ "${UPDATE}" = "y" ]; then
            # update the version number in the source files
            if [ -f "${CWD}"/upstream ]; then
                sed -i -r "s|^ver:\s*${ver}$|ver: ${latest}|g" "${CWD}"/upstream
            fi

            if [ -f "${CWD}"/Makefile ]; then
                sed -i -r "s|^VER\s*=\s*${ver}$|VER = ${latest}|g" "${CWD}"/Makefile
                sed -i -r "s|^REL\s*=.*$|REL = 1|g" "${CWD}"/Makefile
            fi

            if [ -f "${CWD}"/archives ]; then
                ( cd "${CWD}" || exit 2 ; make cleansum )
                sed -i -e "s|${ver}|${latest}|g" "${CWD}"/archives
            fi
        else
            # just report the new version information
            if [ "${HTML_OUTPUT}" = "y" ]; then
                echo "<tr><td>${PKGPATH}</td><td>${ver} < ${latest}</td><td><a href=\"${linkdest}\">${label}</a></td></tr>"
            else
                echo "${PKGPATH} is at ${latest} (${label})"
            fi
        fi
    fi
done

if [ "${UPDATE}" = "y" ]; then
    if [ -f "${CWD}"/archives ]; then
        ( cd "${CWD}" || exit 4 ; make fetch makesum )
    fi

    if grep -q "^post-take-upstream:" "${CWD}"/Makefile >/dev/null 2>&1 ; then
        ( cd "${CWD}" || exit 5 ; env NEW_VER="${latest}" make post-take-upstream )
    fi

    ( cd "${CWD}" || exit 6 ; git diff . )
fi
