#!/bin/sh
#
# checksum.sh: Validate checksums of $DISTFILES
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $WRKDIR               The top level work directory.
#    $PKGDISTFILES         The package specific distfiles subdirectory.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    echo "*** WRKDIR is empty in $(basename "${0}")." >&2
    exit 1
fi

if [ ! -f "${CWD}"/archives ]; then
    exit 0
fi

if [ -f "${WRKDIR}"/archives ]; then
    if [ -z "${PKGDISTFILES}" ] || [ ! -d "${PKGDISTFILES}" ]; then
        echo "*** Missing package distfiles, verify the 'fetch' target worked." >&2
        exit 2
    fi

    grep -E -v "^#" "${WRKDIR}"/archives | grep -E -v "^$" | while read -r first second third ; do
        if [ -z "${third}" ]; then
            # archives file is in openssl(1) dgst format
            checkcmd="openssl"
            checkopts="dgst -$(echo "${first}" | cut -d '(' -f 1 | tr '[:upper:]' '[:lower:]')"
        else
            # archives file is in legacy Peachtree Linux format, use GNU coreutils
            checkcmd="$(echo "${first}" | tr '[:upper:]' '[:lower:]')sum"
        fi

        ${checkcmd} --version >/dev/null 2>&1
        if [ $? -eq 127 ]; then
            echo "*** Unable to find ${checkcmd} program." >&2
            exit 3
        fi

        if [ -z "${third}" ]; then
            # archives file is in openssl(1) dgst format
            filename="$(echo "${first}" | cut -d '(' -f 2 | cut -d ')' -f 1)"
            # shellcheck disable=SC2086
            computed="$(${checkcmd} ${checkopts} "${PKGDISTFILES}"/"${filename}" | cut -d ' ' -f 2)"
        else
            # archives file is in legacy Peachtree Linux format
            filename="${third}"
            computed="$(${checkcmd} "${PKGDISTFILES}"/"${third}" | cut -d ' ' -f 1)"
        fi

        if [ -f "${PKGDISTFILES}"/"${filename}" ]; then
            printf ">>> Validating %s..." "${filename}"

            if [ "${computed}" = "${second}" ]; then
                echo "ok."
            else
                echo "FAILED."
                rm -f "${PKGDISTFILES}"/"${filename}"
            fi
        fi
    done
fi
