#!/bin/sh
#
# fakeinst.sh: Implement fakeinst stage.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# Copyright Chris Lumens <chris@bangmoney.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $PKG                  Package name.
#    $VER                  Package version.
#    $UPSTREAM_PKG         Upstream version of the package name.
#    $WRKDIR               The top level work directory.
#    $WRKDIST              The name of the extracted source directory.
#    $WRKINST              The staging root directory.
#    $WRKSYSROOT           The --sysroot tree for this build.
#    $WRKDEBUG             A copy of $WRKINST with non-stripped ELF files.
#    $DEBUGDIR             The directory in $WRKDEBUG.
#    $NO_DEBUG             Create a debug package (y/n)?
#    $WRKSRC               The extracted source directory.
#    $WRKBLD               The source build directory.
#    $SEPARATE_BUILD       (y/n) Use source build directories?
#    $CONFIGURE_STYLE      The method used to configure $WRKSRC for builds.
#    $PYTHON               Python interpreter to use.
#    $PACKAGE              The full package name.
#    $FAKEINST_ENV         Environment variables passed to fakeinst target.
#    $FAKEINST_TARGET      Name of fakeinst target (default=install).
#    $FAKEINST_ARGS        Extra arguments to $(FAKEINST_COMMAND).
#    $FAKEINST_COMMAND     The make command, usually.
#    $MAKE                 The make command driving the build system.
#    $TOPSRC               The top level distribution source directory.
#    $IGNORE_REQS          List of package build deps to ignore (optional).
#    $TARGET_NAME          Target arch name (e.g., vax-peachtree-linux).
#    $TARGET_MACH          arch(1) or 'uname -m' value (or 'noarch').
#    $BINDIR               Install path for binaries (e.g., /usr/bin).
#    $DATADIR              Directory for architecture-independent data files.
#    $MANDIR               Directory where man pages are installed.
#    $INFODIR              Directory where GNU info pages are installed.
#    $DOCDIR               Directory where documentation files are installed.
#    $LICENSEDIR           Directory where license files are installed.
#    $INCLUDEDIR           Install path for headers (e.g., /usr/include).
#    $LIBDIR               Install path for libraries (e.g., /usr/lib).
#    $STRIP_DT_RPATH       Remove DT_RPATH and DT_RUNPATH from ELF files (Y/n).
#    $DOCS                 List of doc files in $(WRKSRC) to install to $(DOCDIR).
#    $LICENSES             List of license files in $(WRKSRC) to install to
#                          $(LICENSEDIR).
#    $HELPERS              Path to the helpers.sh script.
# This script is called by the build system scripts and should not be called
# directly.

CWD="$(pwd)"
TOPSRC="$(readlink -q -f "${TOPSRC}")"
LOG="${WRKDIR}/fakeinst.log"
ERR="${WRKDIR}/fakeinst.err"
DT_NEEDED="${WRKDIR}/DT_NEEDED"
LISTED_DEPS="${WRKDIR}/listed.deps"
NEEDED_DEPS="${WRKDIR}/needed.deps"
HAVE_DEPS="${WRKDIR}/have.deps"
ALL_PLISTS="${WRKDIR}/expanded-plists"
STATUS=0

if [ -z "${HELPERS}" ] || [ ! -f "${HELPERS}" ]; then
    exit 1
fi

# Helper functions
# shellcheck disable=SC1090
. "${HELPERS}"

find_empty() {
    find "${1}" -type f -empty | while read -r pathname ; do
        echo "*** WARNING: $(echo "${pathname}" | sed -e "s|^${WRKINST}||g") is empty"
    done
}

# Install any documentation files
if [ -n "${DOCS}" ]; then
    mkdir -p "${WRKINST}""${DOCDIR}"

    for doc in ${DOCS} ; do
        # doc make contain a wilcard which we want to expand
        # shellcheck disable=SC2086
        if [ -d "${WRKSRC}" ] && [ -n "$(ls -1 "${WRKSRC}"/${doc} 2>/dev/null)" ]; then
            # doc make contain a wilcard which we want to expand
            # shellcheck disable=SC2086
            ( cd "${WRKSRC}" || exit 14 ; cp -prv --parents ${doc} "${WRKINST}""${DOCDIR}" )
            continue
        fi

        # doc make contain a wilcard which we want to expand
        # shellcheck disable=SC2086
        if [ -n "$(ls -1 "${CWD}"/${doc} 2>/dev/null)" ]; then
            # doc make contain a wilcard which we want to expand
            # shellcheck disable=SC2086
            ( cd "${CWD}" || exit 15 ; cp -prv --parents ${doc} "${WRKINST}""${DOCDIR}" )
            continue
        fi

        echo "WARNING: Unable to find ${doc} for ${DOCDIR}" >&2
    done

    find "${WRKINST}""${DOCDIR}" -type d | while read -r dname ; do
        chmod ugo+rx "${dname}"
    done

    find "${WRKINST}""${DOCDIR}" -type f | while read -r fname ; do
        chmod ugo+r "${fname}"
    done

    find_empty "${WRKINST}""${DOCDIR}"
fi

# Install any license files
if [ -n "${LICENSES}" ]; then
    mkdir -p "${WRKINST}""${LICENSEDIR}"

    for lic in ${LICENSES} ; do
        # lic make contain a wilcard which we want to expand
        # shellcheck disable=SC2086
        if [ -d "${WRKSRC}" ] && [ -n "$(ls -1 "${WRKSRC}"/${lic} 2>/dev/null)" ]; then
            # lic make contain a wilcard which we want to expand
            # shellcheck disable=SC2086
            ( cd "${WRKSRC}" || exit 16 ; cp -prv --parents ${lic} "${WRKINST}""${LICENSEDIR}" )
            continue
        fi

        # lic make contain a wilcard which we want to expand
        # shellcheck disable=SC2086
        if [ -n "$(ls -1 "${CWD}"/${lic} 2>/dev/null)" ]; then
            # lic make contain a wilcard which we want to expand
            # shellcheck disable=SC2086
            ( cd "${CWD}" || exit 17 ; cp -prv --parents ${lic} "${WRKINST}""${LICENSEDIR}" )
            continue
        fi

        echo "WARNING: Unable to find ${lic} for ${LICENSEDIR}" >&2
    done

    find "${WRKINST}""${LICENSEDIR}" -type d | while read -r dname ; do
        chmod ugo+rx "${dname}"
    done

    find "${WRKINST}""${LICENSEDIR}" -type f | while read -r fname ; do
        chmod ugo+r "${fname}"
    done

    find_empty "${WRKINST}""${LICENSEDIR}"
fi

[ "${SKIP_FAKEINST}" = "y" ] && exit 0

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 1
fi

if [ -z "${WRKSRC}" ] || [ ! -d "${WRKSRC}" ]; then
    exit 2
fi

if [ -z "${WRKINST}" ] || [ ! -d "${WRKINST}" ]; then
    exit 4
fi

[ -z "${WRKDEBUG}" ] && exit 5
[ -z "${PACKAGE}" ] && exit 6

if [ "${SEPARATE_BUILD}" = "y" ]; then
    if [ -z "${WRKBLD}" ] || [ ! -d "${WRKBLD}" ]; then
        exit 7
    fi
fi

[ -z "${FAKEINST_TARGET}" ] && exit 9

unique() {
    [ -f "${1}" ] || return
    sort -u "${1}" > "${1}".new
    mv "${1}".new "${1}"
}

# Prepare a plist string for use by grep.  That is:
# 1) Escape the '+' character that may be in a filename.
# 2) Convert any '*' wildcards to the '.*' regex syntax.
escape_string() {
    echo "${1}" | sed -e 's|+|\\+|g' | sed -e 's|\*|.*|g'
}

# Setup logging via a couple fifo's connected to tee
outlog="${TMPDIR:-/tmp}/out.$$"
errlog="${TMPDIR:-/tmp}/err.$$"
mkfifo "${outlog}" "${errlog}"
trap 'rm "${outlog}" "${errlog}"' EXIT

if [ -d "${WRKSRC}" ]; then
    for p in ${PACKAGE} ; do
        echo ">>> Installing ${p} in fake root..."
    done

    if [ "${SEPARATE_BUILD}" = "y" ]; then
        cd "${WRKBLD}" || exit 13
    else
        cd "${WRKSRC}" || exit 12
    fi

    case "${CONFIGURE_STYLE}" in
        gnu|noconfigure|manual|cmake|perl)
            if [ "${CONFIGURE_STYLE}" = "manual" ]; then
                echo "*** WARNING: CONFIGURE_STYLE=manual is deprecated, use CONFIGURE_STYLE=noconfigure" >&2
            fi

            run_tee
            # shellcheck disable=SC2086
            ( env ${FAKEINST_ENV} DESTDIR="${WRKINST}" ${FAKEINST_COMMAND} ${FAKEINST_TARGET} ${FAKEINST_ARGS} \
              >"${outlog}" 2>"${errlog}"
            ) || exit $?

            ;;
        haskell)
            run_tee
            # shellcheck disable=SC2086
            ( env ${FAKEINST_ENV} LIBDIR="${LIBDIR}" PACKAGE="${PACKAGE}" DESTDIR="${WRKINST}" ${FAKEINST_COMMAND} ${FAKEINST_TARGET} ${FAKEINST_ARGS} \
              >"${outlog}" 2>"${errlog}"
            ) || exit $?

            # Special handling for Haskell package docs
            if [ -d "${WRKINST}""${DATADIR}"/doc/"${UPSTREAM_PKG}"-"${VER}" ]; then
                rm -rf "${WRKINST}""${DATADIR}"/doc/"${UPSTREAM_PKG}"-"${VER}" || exit $?
            fi

            # license files are common in directories by themselves
            for legalfile in LICENSE COPYING ; do
                find "${WRKINST}${DATADIR}" -type f -name ${legalfile} 2>/dev/null | while read -r legalfile ; do
                    if [ -f "${legalfile}" ]; then
                        dn="$(dirname "${legalfile}")"

                        if echo "${dn}" | grep -q "^${WRKINST}${LICENSEDIR}" >/dev/null 2>&1 ; then
                            continue
                        fi

                        rm "${legalfile}"
                        rmdir -p --ignore-fail-on-non-empty "${dn}"
                    fi
                done
            done

            ;;
        python)
            run_tee
            # shellcheck disable=SC2086
            ( env ${FAKEINST_ENV} ${PYTHON} setup.py --no-user-cfg ${FAKEINST_TARGET} --verbose -O1 --skip-build --root "${WRKINST}" ${FAKEINST_ARGS} \
              >"${outlog}" 2>"${errlog}"
            ) || exit $?

            # Special handling for Python module installation
            # (this may not exist in all things that build with setup.py, but
            # we have seen it in more than one package)
            if [ -d "${WRKINST}""${WRKSYSROOT}" ]; then
                mkdir -p "${WRKDIR}"/tmp
                mv "${WRKINST}""${WRKSYSROOT}"/* "${WRKDIR}"/tmp
                rmdir -p --ignore-fail-on-non-empty "${WRKINST}""${WRKSYSROOT}"
                cp -pr "${WRKDIR}"/tmp/* "${WRKINST}"
                rmdir "${WRKDIR}"/tmp
            fi

            ;;
        pyproject)
            specifier="$(find "${WRKBLD}"/*.whl -print0 | xargs -0 basename --multiple | sed -E 's/([^-]+)-([^-]+)-.+\\\.whl/\\\1==\\\2/')"

            run_tee
            # shellcheck disable=SC2086
            ( env ${FAKEINST_ENV} ${PYTHON} -m pip ${FAKEINST_TARGET} --root ${WRKINST} ${FAKEINST_ARGS} ${specifier} \
              >"${outlog}" 2>"${errlog}"
            ) || exit $?

            # Special handling for Python module installation
            # This happened with setuptools and is also happening with pip.
            if [ -d "${WRKINST}""${WRKSYSROOT}" ]; then
                mkdir -p "${WRKDIR}"/tmp
                mv "${WRKINST}""${WRKSYSROOT}"/* "${WRKDIR}"/tmp
                rmdir -p --ignore-fail-on-non-empty "${WRKINST}""${WRKSYSROOT}"
                cp -pr "${WRKDIR}"/tmp/* "${WRKINST}"
                rmdir "${WRKDIR}"/tmp
            fi

            ;;
        qmake)
            if [ -f "${WRKSYSROOT}""${BINDIR}"/qt.conf ]; then
                rm -f "${WRKSYSROOT}""${BINDIR}"/qt.conf
            fi

            run_tee
            # shellcheck disable=SC2086
            ( env ${FAKEINST_ENV} INSTALL_ROOT="${WRKINST}" ${FAKEINST_COMMAND} ${FAKEINST_TARGET} ${FAKEINST_ARGS} >"${outlog}" 2>"${errlog}" ) || exit $?

            # everything will be installed to WRKSYSROOT under WRKINST, so
            # manually move things to they are root in WRKINST
            if [ -d "${WRKINST}""${WRKSYSROOT}" ]; then
                cp -pr "${WRKINST}""${WRKSYSROOT}"/* "${WRKINST}"
                rm -rf "${WRKINST:?}""${WRKSYSROOT:?}"/*
                rmdir -p --ignore-fail-on-non-empty "${WRKINST}""${WRKSYSROOT}"
            fi

            ;;
        meson)
            run_tee
            # shellcheck disable=SC2086
            ( env ${FAKEINST_ENV} DESTDIR="${WRKINST}" ninja -C "${WRKBLD}" ${FAKEINST_TARGET} ${FAKEINST_ARGS} >"${outlog}" 2>"${errlog}" ) || exit $?

            ;;
        cargo)
            cd "${WRKBLD}" || exit 11

            run_tee
            # shellcheck disable=SC2086
            ( env ${FAKEINST_ENV} cargo install -v --offline --path ${FAKEINST_TARGET} ${FAKEINST_ARGS} >"${outlog}" 2>"${errlog}" ) || exit $?

            find "${WRKINST}" -type f -name ".crates.toml" | while read -r fname ; do
                rm -f "${fname}"
            done

            find "${WRKINST}" -type f -name ".crates2.json" | while read -r fname ; do
                rm -f "${fname}"
            done

            ;;
        *)
            echo "*** Do not understand CONFIGURE_STYLE=${CONFIGURE_STYLE}"
            exit 10
            ;;
    esac
fi

# Remove __pycache__ files
find "${WRKINST}" -type d -name __pycache__ | while read -r dname ; do
    rm -rf "${dname}"
done

# Remove $WRKSRC and $WRKSYSROOT references from installed files.
# NOTE:  Put the longest patterns before shorter patterns.
if [ -n "${WRKSYSROOT}" ] && [ -n "${WRKSRC}" ]; then
    find "${WRKINST}" -type f | while read -r f ; do
        if file -b --mime-type "${f}" 2>/dev/null | grep -q -E "^text/" >/dev/null 2>&1 ; then
            continue
        fi

        if grep -q -E "(${WRKSYSROOT}|${WRKSRC})" "${f}" >/dev/null 2>&1 ; then
            continue
        fi

        sed -i -e "s|'LDFLAGS=-L${WRKSYSROOT}${LIBDIR}\ -Wl,--sysroot=${WRKSYSROOT}'|'LDFLAGS=-L${LIBDIR}'|g" "${f}"
        sed -i -e "s|-Wl,--sysroot=${WRKSYSROOT}||g" "${f}"
        sed -i -e "s|--sysroot=${WRKSYSROOT}||g" "${f}"
        sed -i -e "s|'--with-sysroot=${WRKSYSROOT}'||g" "${f}"
        sed -i -e "s|-I${WRKSYSROOT}${INCLUDEDIR}|-I${INCLUDEDIR}|g" "${f}"
        sed -i -e "s|-L${WRKSYSROOT}${LIBDIR}|-L${LIBDIR}|g" "${f}"
        sed -i -e "s|${WRKSYSROOT}${INCLUDEDIR}|${INCLUDEDIR}|g" "${f}"
        sed -i -e "s|${WRKSYSROOT}${LIBDIR}|${LIBDIR}|g" "${f}"
        sed -i -e "s|${WRKSRC}|${WRKDIST}|g" "${f}"
    done
fi

# Clean up the man pages
if [ -d "${WRKINST}""${MANDIR}" ]; then
    # Create non symlink style man page references.
    find "${WRKINST}""${MANDIR}" -type l | while read -r manlink ; do
        section="$(basename "$(dirname "${manlink}")")"
        link="$(basename "$(readlink "${manlink}")")"
        rm -f "${manlink}"
        manlink="$(dirname "${manlink}")"/"$(basename "${manlink}")"
        echo ".so ${section}/${link}" > "${manlink}"
        chmod 0644 "${manlink}"
    done

    # Warn about compressed man pages
    find "${WRKINST}""${MANDIR}" -type f | while read -r manpage ; do
        fullmtype="$(file -b --mime-type "${manpage}")"
        mtype="$(echo "${fullmtype}" | cut -c-5)"
        # file(1) is not really consistent about man page types, we could
        # see any of the following:
        #     text/html
        #     text/troff
        #     text/x-c
        #     text/x-c++
        #     text/x-makefile
        #     text/plain
        #     application/javascript
        # and possibly others.  The objective of this check is to catch
        # compressed man pages or executable files in a man page subdir.
        if [ ! "${mtype}" = "text/" ] && [ ! "${fullmtype}" = "application/javascript" ]; then
            echo "*** ERROR: Unknown man page type for ${manpage}: $(file -b --mime-type "${manpage}")" >&2
        fi
    done
fi

# Clean up the info pages
if [ -d "${WRKINST}""${INFODIR}" ]; then
    [ -f "${WRKINST}""${INFODIR}"/dir ] && rm -f "${WRKINST}""${INFODIR}"/dir
fi

# Strip ELF files, add debugging symbols, collect ELF dependencies
if [ -d "${WRKINST}" ] && [ "${NO_DEBUG}" = "n" ] && [ ! "${TARGET_MACH}" = "noarch" ]; then
    find "${WRKINST}" -type f | while read -r filename ; do
        # Skip symlinks
        [ -L "${filename}" ] && continue

        # Gather MIME type of the file
        ftype="$(file -b --mime-type "${filename}" 2>/dev/null)"

        # Skip non-ELF files and likely ELF object (.o) files
        case "${ftype}" in
            application/x-executable|application/x-sharedlib|application/x-pie-executable|application/x-pie-sharedlib)
                if [ "$(echo "${filename}" | rev | cut -c2)" = "o." ]; then
                    echo "${filename} ends with '.o', skipping ELF debug processing"
                    continue
                else
                    echo "${filename} is ${ftype}, proceeding with ELF debug processing"
                fi
                ;;
            *)
                continue
                ;;
        esac

        # Skip unknown architecture files, which are basically things that come
        # with emulator software (e.g., firmware images)
        elfarch="$(objdump -x "${filename}" 2>/dev/null | grep ^architecture: | cut -d ' ' -f 2)"
        [ "${elfarch}" = "UNKNOWN!," ] && continue

        # Strip any DT_RPATH values unless told otherwise.
        if [ "${STRIP_DT_RPATH}" = "y" ]; then
            patchelf --remove-rpath "${filename}"
        fi

        # Find all ELF shared objects, copy them to $WRKDEBUG,
        # and strip them in $WRKINST.
        dn="$(dirname "$(echo "${filename}" | sed -e "s|^${WRKINST}||g")")"
        debugfilename=${WRKDEBUG}${DEBUGDIR}${dn}/$(basename "${filename}").debug

        [ -d "${WRKDEBUG}""${DEBUGDIR}""${dn}" ] || mkdir -p "${WRKDEBUG}""${DEBUGDIR}""${dn}"

        owner="$(stat -c %U "${filename}")"
        group="$(stat -c %G "${filename}")"
        perms="$(stat -c %a "${filename}")"

        run_tee
        ( cp -av "${filename}" "${debugfilename}" >"${outlog}" 2>"${errlog}" ) || exit $?

        # Only perform objcopy step on dynamic executables and libraries
        if ldd "${filename}" >/dev/null 2>&1 ; then
            run_tee
            ( objcopy -v --remove-section=.gnu_debuglink "${filename}" >"${outlog}" 2>"${errlog}" ) || :
            run_tee
            ( objcopy -v --add-gnu-debuglink="${debugfilename}" "${filename}" >"${outlog}" 2>"${errlog}" ) || exit $?
        fi

        # If this is a Haskell package, we need to fix up the rpath in
        # it to no longer point at $WRKSYSROOT.
        if [ "${CONFIGURE_STYLE}" = "haskell" ]; then
            if patchelf --print-rpath "${filename}" | grep -q "${WRKSYSROOT}" >/dev/null 2>&1 ; then
                OLDPATH="$(patchelf --print-rpath "${filename}" | sed -e "s|^${filename}: RPATH=||")"
                NEWPATH="$(echo "${OLDPATH}" | sed -e "s|${WRKSYSROOT}||g")"
                patchelf --set-rpath "${NEWPATH}" "${filename}"
            fi
        fi

        # remove debugging symbols
        strip "${filename}"

        chown "${owner}":"${group}" "${filename}"
        chmod "${perms}" "${filename}"

        # Collect DT_NEEDED names to verify the build dependencies.
        readelf -d "${filename}" | grep "(NEEDED)" | cut -d '[' -f 2 | cut -d ']' -f 1 >> "${DT_NEEDED}"
    done

    # Remove all .la files.
    find "${WRKINST}" -type f -name '*.la' -exec rm {} \;
    find "${WRKINST}" -type l -name '*.la' -exec rm {} \;
fi

# Verify the build dependencies are correct on this package.
if [ -f "${DT_NEEDED}" ]; then
    unique "${DT_NEEDED}"

    rm -f "${LISTED_DEPS}" "${NEEDED_DEPS}" "${HAVE_DEPS}"

    if [ -z "${BUILD_DEPENDS}" ] && [ -f "${CWD}"/buildreqs ]; then
        BUILD_DEPENDS="$(${MAKE} -C "${CWD}" show-buildreqs)" || exit $?
    fi

    for dep in ${BUILD_DEPENDS} ; do
        echo "${dep}" >> "${LISTED_DEPS}"
    done

    # Expand all the plist files first so we can check them
    for plist in $(find "${TOPSRC}"/ -type f -name "plist*" 2>/dev/null | grep -v -E "(^${TOPSRC}/obj/|plist\.[cho5])" 2>/dev/null | grep -E "(plist|plist.${TARGET_MACH})$" 2>/dev/null) ; do
        p="$(dirname "${plist}")"

        # this is not a package if it lacks a Makefile
        [ -f "${p}"/Makefile ] || continue

        pkgpath="$(echo "${p}" | sed -e "s|^${TOPSRC}/||g")"
        sed -e "s|\${VER}|$(${MAKE} -C "${p}" show var=VER)|g" "${plist}" | \
        sed -e "s|\${REL}|$(${MAKE} -C "${p}" show var=REL)|g" | \
        sed -e "s|\${TARGET_NAME}|$(${MAKE} -C "${p}" show var=TARGET_NAME)|g" | \
        sed -e "s|\${TARGET_MACH}|$(${MAKE} -C "${p}" show var=TARGET_MACH)|g" | \
        sed -e "s|^|${pkgpath}:|g" >> "${ALL_PLISTS}"
    done

    # First, check to see if the dependency exists in an existing plist file
    while read -r needed ; do
        # we know nothing yet.  absolutely nothing.
        owned=0

        # check all the files!  enhance!
        if grep -q "${needed} " "${ALL_PLISTS}" >/dev/null 2>&1 ; then
            owned=1
        fi

        # grep could be stupid
        if [ ${owned} -eq 0 ]; then
            if find "${WRKINST}"/ -type f | grep -q "${needed}" >/dev/null 2>&1 ; then
                # This dep is owned by a package, so note that
                owned=1
            fi
        fi

        # oh it's been owned
        if [ ${owned} -eq 1 ]; then
            echo "${needed}" >> "${DT_NEEDED}".owned
        fi

        # Figure out what packages provide these dependencies and list
        # them in the needed deps list, except for anything we want to
        # ignore.
        grep "${needed} " "${ALL_PLISTS}" | cut -d ':' -f 1 | while read -r p ; do
            if [ ! "$(basename "${p}")" = "$(basename "${CWD}")" ]; then
                if [ -z "${IGNORE_REQS}" ]; then
                    echo "${p}" >> "${NEEDED_DEPS}"
                else
                    ignored=0
                    for req in ${IGNORE_REQS} ; do
                        if [ "${req}" = "${p}" ]; then
                            ignored=1
                            break
                        fi
                    done
                    [ "${ignored}" -eq 0 ] && echo "${p}" >> "${NEEDED_DEPS}"
                fi
            fi
        done
    done < "${DT_NEEDED}"

    # Second, if the dependency didn't exist as-is, see if a wildcard
    # line covers it.  We do this by gathering any wildcard lines in
    # plist file and seeing if any of those match something installed
    # in the sysroot or the pkgroot.  If there is a match, that means
    # either a dependent package owns it or this package owns it, so
    # then we can say this dependency is owned.
    while read -r needed ; do
        if grep -q -E "(\?|\*)" "${ALL_PLISTS}" >/dev/null 2>&1 ; then
            # This plist file actually has wildcard lines, so iterate
            # over those.
            grep -E "(\?|\*)" "${ALL_PLISTS}" | while read -r wild ; do
                # Filename may contain spaces, so parse accordingly
                RL="$(echo "${wild}" | rev)"
                LEN=${#wild}

                field=1
                limit=1
                end=""
                while [ "${field}" -le "${limit}" ]; do
                    val="$(echo "${RL}" | cut -d ' ' -f ${field})"

                    if [ "${field}" -eq 1 ]; then
                        rval="$(echo "${val}" | rev)"
                        [ "${rval}" = "nochecksum" ] && limit=4 || limit=3
                    fi

                    field=$((field + 1))
                    end="${end}${val} "
                done

                filepath="$(echo "${wild}" | cut -c-$((LEN - ${#end})) | cut -d ':' -f 2)"

                # If this package owns it, note that and move on.
                if ls "${WRKINST}"/"${filepath}" >/dev/null 2>&1 ; then
                    find "${WRKINST}" -type f -name /"${filepath}" | while read -r line ; do
                        filename="$(basename "${line}")"
                        if [ "${filename}" = "${needed}" ]; then
                            echo "${needed}" >> "${DT_NEEDED}".owned
                            break
                        fi
                    done
                fi

                # If we find it in the sysroot, assume another package
                # owns it and move on.
                if ls "${WRKSYSROOT}"/"${filepath}" >/dev/null 2>&1 ; then
                    find "${WRKSYSROOT}"/ -type f -name "${filepath}" | while read -r line ; do
                        filename="$(basename "${line}")"
                        if [ "${filename}" = "${needed}" ]; then
                            echo "${needed}" >> "${DT_NEEDED}".owned
                            break
                        fi
                    done
                fi
            done
        fi
    done < "${DT_NEEDED}"
fi

if [ -s "${DT_NEEDED}".owned ]; then
    # NOTE: this command must use uniq(1) because we want to print
    # lines in DT_NEEDED that are missing from DT_NEEDED.owned, which
    # will be the unique lines.  The sort -u command is not the same
    # in this case.
    sort "${DT_NEEDED}" "${DT_NEEDED}".owned | uniq -u > "${DT_NEEDED}".unowned
fi

rm -f "${DT_NEEDED}".owned
[ -s "${NEEDED_DEPS}" ] && unique "${NEEDED_DEPS}"

if [ -s "${NEEDED_DEPS}" ]; then
    while read -r needed ; do
        needed="$(escape_string "${needed}")"
        if grep -s -q -E "^${needed}$" "${LISTED_DEPS}" ; then
            grep -v -E "^${needed}$" "${LISTED_DEPS}" > "${LISTED_DEPS}".new
            mv "${LISTED_DEPS}".new "${LISTED_DEPS}"
            echo "${needed}" >> "${HAVE_DEPS}"
        fi
    done < "${NEEDED_DEPS}"
fi

if [ -s "${HAVE_DEPS}" ]; then
    while read -r have ; do
        have="$(escape_string "${have}")"
        grep -v -E "^${have}$" "${NEEDED_DEPS}" > "${NEEDED_DEPS}".new
        mv "${NEEDED_DEPS}".new "${NEEDED_DEPS}"
    done < "${HAVE_DEPS}"
fi

if [ -s "${NEEDED_DEPS}" ]; then
    echo "*** WARNING: The following build dependencies are missing:" >&2
    while read -r needed ; do
        echo "***     ${needed}" >&2
    done < "${NEEDED_DEPS}"
fi

if [ -s "${DT_NEEDED}".unowned ]; then
    unique "${DT_NEEDED}".unowned

    echo "*** WARNING: These build dependencies are not owned by any package:" >&2
    while read -r unowned ; do
        echo "***     ${unowned}" >&2
    done < "${DT_NEEDED}".unowned
fi

if [ ! -f "${DT_NEEDED}" ] && [ ! "${TARGET_MACH}" = "noarch" ] && [ ! "${PKG}" = "kernel" ] && [ ! "${PKG}" = "ovmf" ]; then
    echo "*** WARNING: No platform-specific files, but missing TARGET_MACH=noarch" >&2
fi

rm -f "${HAVE_DEPS}" "${NEEDED_DEPS}" "${LISTED_DEPS}" "${DT_NEEDED}".unowned "${ALL_PLISTS}"

[ -s "${LOG}" ] || rm -f "${LOG}"
[ -s "${ERR}" ] || rm -f "${ERR}"

exit ${STATUS}
