#!/bin/sh
#
# fetch.sh: Implement 'fetch' stage.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level project directory.
#    $MAKE                 The make command.
#    $WRKDIR               The top level work directory.
#    $PKG                  The package name.
#    $VER                  The package version.
#    $OUR_MIRRORS          The list of our mirror download sites.
#    $FORCE_REFRESH        Skip download attempts from mirrors (y/n).
#    $URLS                 The URLS list from the package Makefile.
#    $GITREPO              Upstream git repo to clone from.
#    $GITBRANCH            Optional git branch to change to first.
#    $GITCOMMIT            Upstream git repo commit ID to build from.
#    $GITSUBMODULES        Set to 'y' to fetch submodules too (y/n).
#    $GITSCRIPT            Optional shell command to run in the repo
#                          clone after cloning.
#    $GITPKG               The name of the git repo project (default: $PKG)
#    $GITVER               The version of the git repo project (default: $VER)
#    $KEEPDOTGIT           (y/n) Keep the .git subdirs in the archive?
#    $DISTFILES            The top level distfiles directory.
#    $PKGDISTFILES         The package specific distfiles subdirectory.
# This script is called by the build system scripts and should not be called
# directly.

CURL="curl -k -f -L --retry 3 --ftp-pasv"
CWD="$(pwd)"

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 1
fi

if [ -z "${DISTFILES}" ] || [ ! -d "${DISTFILES}" ]; then
    exit 2
fi

if [ -z "${URLS}" ] && [ -z "${GITREPO}" ]; then
    [ -d "${CWD}"/src ] && exit 0
    [ -f "${CWD}"/archives ] || exit 0

    # URLS can be absent if archives contains nothing local
    remainder="$(grep -v "^#include" "${CWD}"/archives | grep -v "^$")"
    if [ -n "${remainder}" ]; then
        echo "*** Missing URLS variable in the Makefile, but 'archives' file present." >&2
        exit 3
    fi
fi

[ -z "${PKGDISTFILES}" ] && exit 4

[ -z "${FORCE_DOWNLOAD}" ] || FORCE_DOWNLOAD="$(echo "${FORCE_DOWNLOAD}" | tr '[:upper:]' '[:lower:]')"

fetch_file() {
    archive="${1}"
    force_download="${2}"

    # If the download is forced, do not try to fetch from our mirrors.
    if [ "${force_download}" -eq 1 ]; then
        urls="${URLS}"
    else
        urls="${OUR_MIRRORS} ${URLS}"
    fi

    for url in ${urls} ; do
        # Skip if we already have it.
        if [ -d "${PKGDISTFILES}" ] && [ -f "${PKGDISTFILES}/${archive}" ] && [ "${force_download}" -ne 1 ]; then
            if grep "${archive}" "${WRKDIR}"/fetch.log >/dev/null 2>&1 ; then
                echo ">>> ${archive} found, skipping." >> "${WRKDIR}"/.msg
            else
                echo "${archive}" >> "${WRKDIR}"/fetch.log
            fi

            continue
        fi

        # Some URLS may end with a slash, some may not.  Be careful not
        # to add an extra slash on the end.
        if echo "${url}" | grep -q "\/$" >/dev/null 2>&1 ; then
            slash=""
        else
            slash="/"
        fi

        # If a URL ends with '@', we want to use it as-is without adding the archive
        # name to the URL.
        redirect=0
        if echo "${url}" | grep -q -E '@$' >/dev/null 2>&1 ; then
            url="$(echo "${url}" | sed -e 's|@$||g')"
            redirect=1
        fi

        echo ">>> Fetching ${archive}..."
        [ -d "${PKGDISTFILES}" ] || mkdir -p "${PKGDISTFILES}"

        if echo "${url}" | grep -q "^pip:" >/dev/null 2>&1 ; then
            pippkg="$(echo "${url}" | cut -c5-)"
            pip download --no-deps --disable-pip-version-check --ignore-requires-python \
                         --no-binary=:all: --progress-bar ascii -d "${PKGDISTFILES}" "${pippkg}"=="${VER}"
        else
            if [ ${redirect} -eq 1 ]; then
                ${CURL} -o "${PKGDISTFILES}"/"${archive}" "${url}"
            else
                ${CURL} -o "${PKGDISTFILES}"/"${archive}" "${url}""${slash}""${archive}"
            fi

            RET=$?
        fi

        if [ "${RET}" -eq 0 ]; then
            echo "${archive}" >> "${WRKDIR}"/fetch.log
            force_download=0
        else
            rm -f "${PKGDISTFILES}"/"${archive}"
        fi

        # SourceForge check...*sigh*
        t="$(file -b --mime-type "${PKGDISTFILES}/${archive}" | cut -c-5)"
        if echo "${url}" | grep -q "sourceforge\.net" >/dev/null 2>&1 && [ "${t}" = "text/" ]; then
            echo >&2
            echo "*** SOURCEFORGE BREAKS DOWNLOADING ***" >&2
            echo "The build system downloaded an HTML page for your archive because" >&2
            echo "SourceForge uses a client-side JavaScript-based redirect function" >&2
            echo "to provide you the download link.  It is impossible to script." >&2
            echo >&2
            echo "You need to manually download the files and copy them to distfiles." >&2
            echo "Sorry, but SourceForge is terrible." >&2
            echo >&2
            echo "${url}${slash}${archive}" >&2
            echo >&2
            exit 5
        fi
    done
}

# Fetch archives not included from another package.
if [ -f "${CWD}"/archives ]; then
    rm -f "${WRKDIR}"/.msg

    grep -v "^#" "${CWD}"/archives | grep -v "^$" | while read -r first second third ; do
        force_download=0

        if [ -z "${second}" ] && [ -z "${third}" ]; then
            filename="${first}"
            force_download=1
        else
            if [ -z "${third}" ]; then
                # archives file is in the openssl(1) dgst format
                filename="$(echo "${first}" | cut -d '(' -f 2 | cut -d ')' -f 1)"
            else
                # archives file is in legacy Peachtree Linux format
                filename="${third}"
            fi
        fi

        [ "${FORCE_REFRESH}" = "y" ] && force_download=1

        # We may have the archive in an old location, move it
        if [ -d "${PKGDISTFILES}" ]; then
            if [ -f "${DISTFILES}"/"${filename}" ] && [ ! -f "${PKGDISTFILES}"/"${filename}" ]; then
                mv "${DISTFILES}"/"${filename}" "${PKGDISTFILES}"
            fi
        fi

        # Fetch this archive file
        fetch_file "${filename}" "${force_download}"
    done
fi

# Fetch any archives included from another package.
if [ -f "${CWD}"/archives ]; then
    # change to each included package directory and fetch
    grep -E "^#include" "${CWD}"/archives | while read -r _ includepath ; do
        pkgdir="${TOPSRC}"/"$(dirname "$(echo "${includepath}" | cut -d '<' -f 2 | cut -d '>' -f 1)")"
        # shellcheck disable=SC2086
        env -u GITREPO \
            -u GITBRANCH \
            -u GITCOMMIT \
            -u GITSUBMODULES \
            -u GITSCRIPT \
            -u GITPKG \
            -u GITVER \
            -u KEEPDOTGIT \
        ${MAKE} --no-print-directory -C "${pkgdir}" fetch
    done

    # symlink the included archives to our distfiles directory
    grep -E "^#include" "${CWD}"/archives | while read -r _ includepath ; do
        includepath="$(echo "${includepath}" | cut -d '<' -f 2 | cut -d '>' -f 1)"
        pkgdir="${TOPSRC}"/"$(dirname "${includepath}")"
        includedarchives="${TOPSRC}"/"${includepath}"
        includeddistfiles="$(${MAKE} --no-print-directory -C "${pkgdir}" show var=PKGDISTFILES)"

        grep -v "^#" "${includedarchives}" | grep -v "^$" | while read -r first second third ; do
            [ -d "${PKGDISTFILES}" ] || mkdir -p "${PKGDISTFILES}"

            if [ -z "${third}" ]; then
                # archives file is in openssl(1) dgst format
                filename="$(echo "${first}" | cut -d '(' -f 2 | cut -d ')' -f 1)"
            else
                # archives file is in legacy Peachtree Linux format
                filename="${third}"
            fi

            # create an absolute symlink, then convert it to relative
            ln -sf "${includeddistfiles}"/"${filename}" "${PKGDISTFILES}"/"${filename}"

            if symlinks --help >/dev/null 2>&1 ; then
                symlinks -c "${PKGDISTFILES}"
            fi
        done
    done
fi

# Clone a git repo and cache the archive if we need to
if [ -n "${GITREPO}" ] && [ -z "${GITCOMMIT}" ]; then
    echo "*** GITREPO specified, but missing GITCOMMIT" >&2
    exit 6
fi

if [ -z "${GITREPO}" ] && [ -n "${GITCOMMIT}" ]; then
    echo "*** GITCOMMIT specified, but missing GITREPO" >&2
    exit 7
fi

if [ -n "${GITREPO}" ] && [ -n "${GITCOMMIT}" ]; then
    # we must have PKG and VER
    if [ -z "${GITPKG}" ]; then
        echo "*** Missing PKG" >&2
        exit 8
    fi

    if [ -z "${GITVER}" ]; then
        echo "*** Missing VER" >&2
        exit 9
    fi

    # first, construct the archive name that we would cache in distfiles
    archive="${GITPKG}-${GITVER}-${GITCOMMIT}.tar.gz"

    # check to see if we have it
    if [ -d "${PKGDISTFILES}" ] && [ -f "${PKGDISTFILES}/${archive}" ] && [ ! "${FORCE_REFRESH}" = "y" ]; then
        echo ">>> ${archive} found, skipping." >> "${WRKDIR}"/.msg
    else
        # get the source and create a cached archive
        cd "${WRKDIR}" || exit 10
        git clone "${GITREPO}" "${GITPKG}-${GITVER}"
        cd "${GITPKG}-${GITVER}" || exit 11
        [ -z "${GITBRANCH}" ] || git checkout -b "${GITBRANCH}" "${GITBRANCH}"
        [ "${GITCOMMIT}" = "HEAD" ] || git reset --hard "${GITCOMMIT}"
        [ "${GITSUBMODULES}" = "y" ] && git submodule update --init
        [ -z "${GITSCRIPT}" ] || ${GITSCRIPT}
        [ "${KEEPDOTGIT}" = "n" ] && find . -type d -name .git -print0 | xargs -0 rm -rf
        cd "${WRKDIR}" || exit 12
        [ -d "${PKGDISTFILES}" ] || mkdir -p "${PKGDISTFILES}"
        printf "Generating %s/%s (this will take some time)..." "${PKGDISTFILES}" "${archive}"
        tar -cf - "${GITPKG}-${GITVER}" | gzip -9c > "${PKGDISTFILES}/${archive}"
        rm -rf "${GITPKG}-${GITVER}"
        printf "done\n"
    fi
fi

if [ -f "${WRKDIR}"/.msg ]; then
    sort -u "${WRKDIR}"/.msg
    rm -f "${WRKDIR}"/.msg
fi
