#!/bin/sh
#
# helpers.sh: Useful helper functions
#
# Copyright David Cantrell <dcantrell@burdell.org>
# Copyright Chris Lumens <chris@bangmoney.org>
# Copyright Brian C. Lane <bcl@brianlane.com>
# SPDX-License-Identifier: Apache-2.0

# These are all set in the scripts that source in helpers.sh, but
# having fallbacks here is just a safety measure.
[ -z "${CURL}" ] && CURL="curl -k -f -L --retry 3 --ftp-pasv"
[ -z "${outlog}" ] && outlog="${TMPDIR:-/tmp}/out.$$"
[ -z "${errlog}" ] && errlog="${TMPDIR:-/tmp}/err.$$"

# tee gets killed when the subshell closes and needs to be re-started before each use
run_tee() {
    tee -a "${LOG}" < "${outlog}" &
    tee -a "${ERR}" < "${errlog}" >&2 &
}

# Generic variable gathering function for packages.
get_var() {
    pkg="${1}"
    var="${2}"
    for brpath in ${BUILDREQS_PATHS} ; do
        if [ -d "${brpath}/${pkg}" ]; then
            env -u TARGET_MACH make --no-print-directory -C "${brpath}"/"${pkg}" show var="${var}"
            break
        fi
    done
}

# Called by download_helper() to perform digest verification.
verify_digest() {
    path="${1}"
    package="${2}"
    digest="${3}"

    [ -z "${digest}" ] && return 1

    # determine algorithm based on digest string length
    len=${#digest}

    if [ -z "${len}" ]; then
        echo "*** Unknown digest type for ${package}" >&2
        return 1
    fi

    # ensure we have openssl available
    openssl --version >/dev/null 2>&1
    [ ${?} -eq 127 ] && return

    # compute the appropriate message digest
    if [ "${len}" -eq 32 ]; then
        computed="$(openssl dgst -md5 "${path}"/"${package}" 2>/dev/null | cut -d ' ' -f 2)"
    elif [ "${len}" -eq 40 ]; then
        computed="$(openssl dgst -sha1 "${path}"/"${package}" 2>/dev/null | cut -d ' ' -f 2)"
    elif [ "${len}" -eq 64 ]; then
        computed="$(openssl dgst -sha256 "${path}"/"${package}" 2>/dev/null | cut -d ' ' -f 2)"
    elif [ "${len}" -eq 128 ]; then
        computed="$(openssl dgst -sha512 "${path}"/"${package}" 2>/dev/null | cut -d ' ' -f 2)"
    else
        echo "*** Unknown digest type for ${package}" >&2
        return 1
    fi

    [ "${computed}" = "${digest}" ] && return 0 || return 1
}

# Called by download_dist() to fetch regular and debug packages.
download_helper() {
    package="${1}"
    path="${2}"
    collection="${3}"
    digest=
    download_url="$(echo "${path}" | sed -e "s|^${RELEASE}|${DIST_URL}|g")"

    [ -d "${path}" ] || mkdir -p "${path}"

    echo ">>> Downloading repodata for ${collection}..."
    if [ -f "${path}"/repodata.db ]; then
        # last modification time of repodata, used to give curl a timestamp
        ${CURL} -z "${path}"/repodata.db -o "${path}"/repodata.db "${download_url}"/repodata.db
    else
        ${CURL} -o "${path}"/repodata.db "${download_url}"/repodata.db
    fi

    if [ -f "${path}"/repodata.db ]; then
        digest="$(sqlite3 "${path}"/repodata.db "SELECT digest FROM dists WHERE filename='${package}'")"
    fi

    if [ -z "${digest}" ]; then
        echo ">>> Downloading latest repodata for ${collection}..."
        ${CURL} -o "${path}"/repodata.db "${download_url}"/repodata.db

        if [ -f "${path}"/repodata.db ]; then
            digest="$(sqlite3 "${path}"/repodata.db "SELECT digest FROM dists WHERE filename='${package}'")"
        fi
    fi

    if [ -f "${path}"/"${package}" ]; then
        verify_digest "${path}" "${package}" "${digest}" && return
    else
        echo ">>> Downloading ${package}..."

        if ! ${CURL} -o "${path}"/"${package}" "${download_url}"/"${package}" ; then
            echo ">>> FAILED"

            if [ -f "${path}"/repodata.db ]; then
                sqlite3 "${path}"/repodata.db "DELETE FROM dists WHERE filename='${package}'"
            fi

            rm -f "${path}"/"${package}"
            return
        fi
    fi

    if verify_digest "${path}" "${package}" "${digest}" ; then
        return
    else
        if [ -f "${path}"/repodata.db ]; then
            sqlite3 "${path}"/repodata.db "DELETE FROM dists WHERE filename='${package}'"
        fi

        rm -f "${path}"/"${package}"
        return
    fi
}

# Attempt a download and cache of this dep's built packages.
download_dist() {
    [ "$(echo "${DIST_DOWNLOAD}" | tr '[:upper:]' '[:lower:]')" = "n" ] && return

    # many variables
    dep="${1}"
    packages="$(get_var "${dep}" PACKAGE)"
    debugpackages="$(get_var "${dep}" DEBUGPACKAGE)"
    collection="$(get_var "${dep}" PKGCOLLECTION)"
    path="$(get_var "${dep}" RELEASE_DIST)"/"$(get_var "${dep}" RELPKGPATH)"
    debugpath="$(get_var "${dep}" RELEASE_DIST)"/debug/"$(get_var "${dep}" RELPKGPATH)"

    # try to download the main dist
    for pkg in ${packages} ; do
        download_helper "${pkg}" "${path}" "${collection}"
    done

    # try to download the debug dist (this is not fatal if it's missing)
    for debugpkg in ${debugpackages} ; do
        download_helper "${debugpkg}" "${debugpath}" "${collection}"
    done
}
