#!/bin/sh
#
# makesum.sh: Generate additional data for the archives file.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level project directory.
#    $WRKDIR               The top level work directory.
#    $DIGEST_ALGORITHM     Message digest algorithm to use.
#    $PKGDISTFILES         The package specific distfiles subdirectory.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 2
fi

[ -z "${DIGEST_ALGORITHM}" ] && exit 3

if [ -z "${PKGDISTFILES}" ] || [ ! -d "${PKGDISTFILES}" ]; then
    exit 4
fi

CWD="$(pwd)"
INCLUDES="${WRKDIR}"/archives.includes
HANDLED="${WRKDIR}"/archives.handled

if [ -f "${CWD}"/archives ]; then
    cmd="openssl"
    cmdopts="dgst -$(echo "${DIGEST_ALGORITHM}" | tr '[:upper:]' '[:lower:]')"

    ${cmd} --version >/dev/null 2>&1
    [ $? -eq 127 ] && exit 5

    # collect list of included archives files
    included="$(grep "^#include" "${CWD}"/archives | cut -d '<' -f 2 | cut -d '>' -f 1)"
    if [ -n "${included}" ]; then
        rm -f "${HANDLED}"

        for ifile in ${included} ; do
            cat "${TOPSRC}"/"${ifile}" >> "${HANDLED}"
        done
    fi

    rm -f "${WRKDIR}"/archives

    grep -v "^#" "${CWD}"/archives | grep -v "^$" | while read -r a b ; do
        if [ -n "${a}" ] && [ -n "${b}" ]; then
            echo "${a} ${b}" >> "${WRKDIR}"/archives
            continue
        fi

        archive="$(echo "${a}" | cut -d '(' -f 2 | cut -d ')' -f 1)"
        if [ ! -f "${PKGDISTFILES}"/"${archive}" ]; then
            echo "*** Missing ${PKGDISTFILES}/${archive}, try fetch again." >&2
            exit 6
        fi

        if [ -f "${HANDLED}" ] && [ -n "${b}" ]; then
            if grep -q -E "${b}$" "${HANDLED}" >/dev/null 2>&1 ; then
                continue
            fi
        fi

        # shellcheck disable=SC2086
        ( cd "${PKGDISTFILES}" || exit 7 ; ${cmd} ${cmdopts} "${archive}" ) >> "${WRKDIR}"/archives
    done

    # keep any include lines
    grep "^#include" "${CWD}"/archives > "${INCLUDES}"

    if [ -f "${WRKDIR}"/archives ]; then
        if [ -s "${INCLUDES}" ]; then
            # remove expanded lines that we are including
            while read -r x ; do
                oa="$(echo "${x}" | cut -d '<' -f 2 | cut -d '>' -f 1)"
                [ -f "${TOPSRC}"/"${oa}" ] || continue

                while read -r y ; do
                    grep -v "^${y}" "${WRKDIR}"/archives > "${WRKDIR}"/a.new
                    [ -f "${WRKDIR}/a.new" ] && mv "${WRKDIR}"/a.new "${WRKDIR}"/archives
                done < "${TOPSRC}"/"${oa}"
            done < "${INCLUDES}"

            cat "${INCLUDES}" > "${CWD}"/archives
            rm -f "${INCLUDES}"
        else
            rm -f "${CWD}"/archives
        fi

        # only take the fully expanded lines beginning with an algorithm
        grep -E "\(.+\)=" "${WRKDIR}"/archives | sort -u >> "${CWD}"/archives
        rm -rf "${WRKDIR}"/archives
    fi

    rm -rf "${HANDLED}"
fi

# Clear out blank lines
sed -i -e '/^$/d' "${CWD}"/archives
