#!/bin/sh
#
# missingreqs.sh: Recursively build missing buildreqs
#
# Copyright David Cantrell <dcantrell@burdell.org>
# Copyright Chris Lumens <chris@bangmoney.org>
# Copyright Brian C. Lane <bcl@brianlane.com>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level distribution source directory.
#    $WRKDIR               The package work directory.
#    $WRKSYSROOT           The directory passed to --sysroot on gcc(1).
#    $MAKE                 The make(1) command.
#    $DIST_DOWNLOAD        Try to download build dependencies (y/n).
#    $DIST_URL             Site to download build dependencies from.
#    $RELEASE              The release directory containing built packages.
#    $DISTADD_TMPDIR       Temp directory for distadd(8) to use.
#    $HELPERS              Path to the helpers.sh script.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

# Helper functions
# shellcheck disable=SC1090
[ -r "${HELPERS}" ] && . "${HELPERS}"

[ -z "${WRKDIR}" ] && exit 2
[ -z "${WRKSYSROOT}" ] && exit 3
[ -z "${MAKE}" ] && exit 4
[ -z "${DISTADD_TMPDIR}" ] && exit 5

[ -d "${DISTADD_TMPDIR}" ] || mkdir -p "${DISTADD_TMPDIR}"
CWD="$(pwd)"

# Some URLs may end with a slash, some may not.  Be careful not
# to add an extra slash on the end.
if echo "${DIST_URL}" | grep -q "\/$" >/dev/null 2>&1 ; then
    slash=""
else
    # shellcheck disable=SC2034
    slash="/"
fi

# Gather list of buildreqs
if [ -f "${CWD}/buildreqs" ]; then
    BUILDREQS="$(${MAKE} show-buildreqs)" || exit ${?}
fi

[ -z "${BUILDREQS}" ] && exit 0

# We use a single sysroot, so wipe it and start a new one.
rm -rf "${WRKSYSROOT}"
mkdir -p "${WRKSYSROOT}"

# distadd (standard Peachtree Linux location is /usr/sbin)
# shellcheck disable=SC2034
distadd="/usr/sbin/distadd"

# Find curl, or disable downloading
CURL="curl -k -f -L --retry 3 --ftp-pasv"
${CURL} >/dev/null 2>&1
# shellcheck disable=SC2034
[ ${?} -eq 127 ] && DIST_DOWNLOAD=n

[ -d "${WRKDIR}" ] || mkdir -p "${WRKDIR}"
LOG="${WRKDIR}/usereqs.log"
ERR="${WRKDIR}/usereqs.err"

# Setup logging via a couple fifo's connected to tee
outlog="${TMPDIR:-/tmp}/out.$$"
errlog="${TMPDIR:-/tmp}/err.$$"
mkfifo "${outlog}" "${errlog}"
trap 'rm "${outlog}" "${errlog}"' EXIT

# Step one:  Verify all build dependencies exist.
run_tee
( for dep in ${BUILDREQS} ; do
      fullpkgpath="$(get_var "${dep}" FULLPKGPATH)"

      if [ -z "${fullpkgpath}" ]; then
          echo "Build requirement ${dep} does not exist, cannot continue." >&2
          exit 4
      fi

      [ -f "${fullpkgpath}" ] || download_dist "${dep}"

      if [ ! -f "${fullpkgpath}" ]; then
          echo "*** Missing ${fullpkgpath}, running 'make missingreqs' for ${dep}" >&2
          ${MAKE} -C "${TOPSRC}"/"${dep}" missingreqs
          if [ ! -f "${fullpkgpath}" ]; then
              echo "*** make world failed for ${dep}, exiting."
              exit 5
          fi
      fi
  done \
  >"${outlog}" 2>"${errlog}"
) || exit ${?}

# All missing reqs are built, do a make world now
echo "Running 'make world'"
run_tee
( ${MAKE} world  >"${outlog}" 2>"${errlog}" ) || exit ${?}

[ -s "${LOG}" ] || rm -f "${LOG}"
[ -s "${ERR}" ] || rm -f "${ERR}"
