#!/bin/sh
#
# package.sh: Implement package stage.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $WRKDIR               The top level work directory.
#    $WRKINST              The staging root directory.
#    $WRKDEBUG             The debug directory.
#    $PKG                  The package name.
#    $VER                  The package version.
#    $REL                  The package release.
#    $SUBPACKAGES          Any optional subpackages.
#    $TARGET_NAME          The target arch name (e.g., vax-peachtree-linux).
#    $TARGET_MACH          arch(1) or 'uname -m' value (or 'noarch').
#    $TOPSRC               The top level directory of the project.
#    $RELEASE_DIST         The release tree directory of the project.
#    $PKGCOLLECTION        The name of the package collection.
#    $NAME_PREFIX_CONFLICT Set to a list of package names that name $PKG as
#                          their package name prefix.
#    $MULTIPLE_VERSIONS    (y/n), y=There a multiple versions of this $PKG
#                          in this $PKGCOLLECTION (e.g., gtk+)
#    $DIGEST_ALGORITHM     The digest algorithm to use (e.g., sha1) when
#                          generating message digests for built packages.
#    $COMPRESSION          Compression algorithm to use in dist payloads.  It
#                          is zstd by default, but that's expensive on slower
#                          processors.  Packages can override it by setting
#                          this variable to an accepted distmk(1) -z setting.
#    $DESCRIPTION          Short package description (optional).
#    $MAINTAINER           Package maintainer name and email (optional).
#    $LICENSE              License information string (optional).
#    $DEPSPATH             Additional shared library search dirs (optional).
#    $HELPERSDB            Path to the helpers-db.sh script.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 1
fi

if [ -z "${WRKINST}" ] || [ ! -d "${WRKINST}" ]; then
    exit 2
fi

[ -z "${PKG}" ] && exit 3
[ -z "${VER}" ] && exit 4
[ -z "${REL}" ] && exit 5
[ -z "${TARGET_MACH}" ] && exit 6

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 7
fi

if [ -z "${PKGCOLLECTION}" ] || [ ! -d "${TOPSRC}/${PKGCOLLECTION}" ]; then
    exit 8
fi

[ -z "${DIGEST_ALGORITHM}" ] && exit 9
[ -z "${RELEASE_DIST}" ] && exit 12
[ -z "${COMPRESSION}" ] && exit 13

CWD="$(pwd)"
TMPDIST="$(mktemp -d)"
LOG="${WRKDIR}/package.log"
ERR="${WRKDIR}/package.err"
DEBUGLOG="${WRKDIR}/package-debug.log"
DEBUGERR="${WRKDIR}/package-debug.err"

DISTDIR="${RELEASE_DIST}/${PKGCOLLECTION}/${TARGET_MACH}"
DEBUGDIR="${RELEASE_DIST}/debug/${PKGCOLLECTION}/${TARGET_MACH}"

trap 'rm -rf "${TMPDIST}"' EXIT

# shellcheck disable=SC1090
. "${HELPERSDB}"

escape_pkgname() {
    echo "${1}" | sed -e 's|\+|\\\+|g'
}

is_subpackage() {
    pkgname="${1}"

    [ -z "${SUBPACKAGES}" ] && return 1

    for subpkg in ${SUBPACKAGES} ; do
        subpkgname="${PKG}-${subpkg}-${VER}-${REL}.${TARGET_MACH}.dist"

        if [ "${pkgname}" = "${subpkgname}" ]; then
            return 0
        fi
    done

    return 1
}

remove_old_pkg() {
    _pkgdir="${1}"
    _curpkg="${2}"
    _pkg="${3}"
    _msg="${4}"

    find "${_pkgdir}"/"${_pkg}"-*.dist -prune -print 2>&1 | while read -r line ; do
        pkgname="$(basename "${line}")"

        # Check for the lead of the package distfile matching the package name
        pkgnamelen=$(echo -n "${_pkg}-" | wc -c)
        beginning="$(echo "${pkgname}" | cut -c-${pkgnamelen})"

        # Special handling for package names that confuse the wildcards
        # or that otherwise need special handling.
        if [ ! "${_pkg}-" = "${beginning}" ]; then
            continune
        elif [ -n "${NAME_PREFIX_CONFLICT}" ]; then
            conflict=0
            for c in ${NAME_PREFIX_CONFLICT} ; do
                c="$(escape_pkgname "${c}")"
                if echo "${pkgname}" | grep -E -s -q "^${c}" ; then
                    conflict=1
                    break
                fi
            done
            [ ${conflict} -eq 1 ] && continue
        elif [ "$(echo "${MULTIPLE_VERSIONS}" | tr '[:upper:]' '[:lower:]')" = "y" ]; then
            c="$(escape_pkgname "${_pkg}-")"
            if echo "${pkgname}" | grep -E -s -q "^${c}" ; then
                majorver="$(echo "${VER}" | cut -d '.' -f 1)"
                echo "${pkgname}" | grep -E -s -q "^${c}${majorver}"
                [ $? -eq 1 ] && continue
            fi
        fi

        if [ ! "${pkgname}" = "${_curpkg}" ]; then
            printf "Removing old %s %s..." "${_msg}" "${pkgname}"
            rm -f "${line}"

            if [ -f "${_pkgdir}"/repodata.db ]; then
                sqlite3 "${_pkgdir}"/repodata.db "DELETE FROM dists WHERE INSTR(filename, '${pkgname}') > 0"
            fi

            if [ -f "${_pkgdir}"/00-ORDER ]; then
                sed -i -e "s|^${pkgname}$|${_curpkg}|g" "${_pkgdir}"/00-ORDER
            fi

            echo "done."
        fi
    done
}

# Run distmk with specified arguments.
# $1  stdout log
# $2  stderr log
# $3  pkgroot dir
# $4  plist path
# $5  output dir
# $6  is a debug pkg? (y/n)
# $7  subpackage name
run_distmk() {
    DLOG="$1"
    DERR="$2"
    PKGROOT_DIR="$3"
    PLIST_PATH="$4"
    OUTPUT_DIR="$5"
    ISDEBUG="$6"
    SUBPKG="$7"

    # These are killed after distmk is done
    tee -a "${DLOG}" < "${outlog}" &
    tee -a "${DERR}" < "${errlog}" >&2 &

    # Standard arguments
    set -- --tree="${PKGROOT_DIR}" "$@"
    set -- --plist="${PLIST_PATH}" "$@"
    set -- --arch="${TARGET_MACH}" "$@"
    set -- --outdir="${OUTPUT_DIR}" "$@"
    set -- --repo="${PKGCOLLECTION}" "$@"
    set -- --compression="${COMPRESSION}" "$@"

    if [ -z "${SUBPKG}" ]; then
        # Creating the main package
        set -- --meta="${CWD}" "$@"

        # Optional arguments
        [ -z "${DESCRIPTION}" ] || set -- --description="${DESCRIPTION}" "$@"
        [ -z "${MAINTAINER}" ] || set -- --maintainer="${MAINTAINER}" "$@"
        [ -z "${LICENSE}" ] || set -- --license="${LICENSE}" "$@"
        [ -z "${DEPSPATH}" ] || set -- --depspath="${DEPSPATH}" "$@"

        PKGNAME="${PKG}"
    else
        # Creating a subpackage
        set -- --meta="${CWD}"/metadata/"${SUBPKG}" "$@"

        # Optional arguments
        SUBPKG_DESCRIPTION=
        SUBPKG_MAINTAINER=
        SUBPKG_LICENSE=
        SUBPKG_DEPSPATH=

        # shellcheck disable=SC1090
        [ -f "${CWD}"/metadata/"${SUBPKG}"/pkginfo ] && . "${CWD}"/metadata/"${SUBPKG}"/pkginfo

        [ -z "${SUBPKG_DESCRIPTION}" ] || set -- --description="${SUBPKG_DESCRIPTION}" "$@"
        [ -z "${SUBPKG_MAINTAINER}" ] || set -- --maintainer="${SUBPKG_MAINTAINER}" "$@"
        [ -z "${SUBPKG_LICENSE}" ] || set -- --license="${SUBPKG_LICENSE}" "$@"
        [ -z "${SUBPKG_DEPSPATH}" ] || set -- --depspath="${SUBPKG_DEPSPATH}" "$@"

        PKGNAME="${PKG}-${SUBPKG}"
    fi

    [ "${ISDEBUG}" = "y" ] && set -- -d "$@"

    distmk -p "${PKGNAME}" -v "${VER}" -r "${REL}" -c "$@" >"${outlog}" 2>"${errlog}"
}

# Run to combine plist template files from the package source
# directory in to a plist file the distmk command can use.  The
# function takes one argument and if it's empty, it assumes we are
# working on the main package.  If the argument has a value, it is a
# subpackage name.  The generated plist for distmk is always written
# to ${TMPDIST}/plist
# $1    Subpackage name or ""
combine_plists() {
    SUBPKG="${1}"
    PLISTS=

    # If an argument is given to this function, it identifies a
    # subpackage we are working on.  Otherwise it is the main package.
    if [ -z "${SUBPKG}" ]; then
        # Combine multiple plist files for the package and replace variables
        [ -f "${CWD}/plist" ] && PLISTS="${CWD}/plist"
        [ -f "${CWD}/plist.${TARGET_MACH}" ] && PLISTS="${PLISTS} ${CWD}/plist.${TARGET_MACH}"
    else
        # Combine multiple plist files for the package and replace variables
        [ -f "${CWD}/metadata/${SUBPKG}/plist" ] && PLISTS="${CWD}/metadata/${SUBPKG}/plist"
        [ -f "${CWD}/metadata/${SUBPKG}/plist.${TARGET_MACH}" ] && PLISTS="${PLISTS} ${CWD}/metadata/${SUBPKG}/plist.${TARGET_MACH}"
    fi

    if [ -z "${PLISTS}" ]; then
        # For new packages, we may not have a plist file yet, so just use an
        # empty file here to move things along for the packager.
        touch "${TMPDIST}"/emptyplist
        PLISTS="${TMPDIST}/emptyplist"
    fi

    # shellcheck disable=SC2086
    sort -u ${PLISTS} | \
    sed -e "s|\${VER}|${VER}|g" | \
    sed -e "s|\${REL}|${REL}|g" | \
    sed -e "s|\${TARGET_NAME}|${TARGET_NAME}|g" > "${TMPDIST}"/plist

    if [ ! "${TARGET_MACH}" = "noarch" ]; then
        sed -i -e "s|\${TARGET_MACH}|${TARGET_MACH}|g" "${TMPDIST}"/plist
    fi
}

# Force packages without ELF content to have TARGET_MACH=noarch
if [ "${TARGET_MACH}" = "noarch" ]; then
    if find "${WRKINST}" -type f -exec file -b {} \; | grep -s ELF >/dev/null 2>&1 ; then
        echo "*** ELF files found in WRKINST but TARGET_MACH = noarch." >&2
        exit 11
    fi
fi

[ -d "${DISTDIR}" ] || mkdir -p "${DISTDIR}"
[ -f "${DISTDIR}"/repodata.db ] || init_repodata_db "${DISTDIR}"

# Fail if we have arch-specific plists for 'noarch'
if [ "${TARGET_MACH}" = "noarch" ] && [ "$(find "${CWD}"/plist* -type f -prune -print | wc -l)" -gt 1 ]; then
    echo "*** Architecture-specific plist files found, but TARGET_MACH=noarch" 2>&1
    exit 1
fi

# Setup logging via a couple fifo's connected to tee
outlog="${TMPDIR:-/tmp}/out.$$"
errlog="${TMPDIR:-/tmp}/err.$$"
mkfifo "${outlog}" "${errlog}"
trap 'rm "${outlog}" "${errlog}"' EXIT

# Create the dist package
combine_plists
run_distmk "${LOG}" "${ERR}" "${WRKINST}" "${TMPDIST}"/plist "${DISTDIR}" "n" ""
curpkg="${PKG}-${VER}-${REL}.${TARGET_MACH}.dist"
add_repodata "${DISTDIR}" "${curpkg}"
remove_old_pkg "${DISTDIR}" "${curpkg}" "${PKG}" "package"

# Any subpackages?
if [ -n "${SUBPACKAGES}" ]; then
    for subpkg in ${SUBPACKAGES} ; do
        combine_plists "${subpkg}"
        run_distmk "${LOG}" "${ERR}" "${WRKINST}"."${subpkg}" "${TMPDIST}"/plist "${DISTDIR}" "n" "${subpkg}"
        curpkg="${PKG}-${subpkg}-${VER}-${REL}.${TARGET_MACH}.dist"
        add_repodata "${DISTDIR}" "${curpkg}"
        remove_old_pkg "${DISTDIR}" "${curpkg}" "${PKG}-${subpkg}" "package"
    done
fi

# Do we need a debug package?
if [ -d "${WRKDEBUG}" ]; then
    [ -d "${DEBUGDIR}" ] || mkdir -p "${DEBUGDIR}"
    [ -f "${DEBUGDIR}"/repodata.db ] || init_repodata_db "${DEBUGDIR}"

    distplist -t "${WRKDEBUG}" -u root -g root > "${TMPDIST}"/plist.debug
    run_distmk "${DEBUGLOG}" "${DEBUGERR}" "${WRKDEBUG}" "${TMPDIST}"/plist.debug "${DEBUGDIR}" "y" ""
    curpkg="${PKG}-${VER}-${REL}.debug.${TARGET_MACH}.dist"
    add_repodata "${DEBUGDIR}" "${curpkg}"

    # Remove old debug packages
    remove_old_pkg "${DEBUGDIR}" "${curpkg}" "${PKG}" "debug package"

    # Any subpackages?
    if [ -n "${SUBPACKAGES}" ]; then
        for subpkg in ${SUBPACKAGES} ; do
            if [ -d "${WRKDEBUG}"."${subpkg}" ]; then
                distplist -t "${WRKDEBUG}"."${subpkg}" -u root -g root > "${TMPDIST}"/plist.debug
                run_distmk "${DEBUGLOG}" "${DEBUGERR}" "${WRKDEBUG}"."${subpkg}" "${TMPDIST}"/plist.debug "${DEBUGDIR}" "y" "${subpkg}"
                curpkg="${PKG}-${subpkg}-${VER}-${REL}.debug.${TARGET_MACH}.dist"
                add_repodata "${DEBUGDIR}" "${curpkg}"
                remove_old_pkg "${DEBUGDIR}" "${curpkg}" "${PKG}-${subpkg}" "debug package"
            fi
        done
    fi
fi

[ -s "${LOG}" ] || rm -f "${LOG}"
[ -s "${DEBUGLOG}" ] || rm -f "${DEBUGLOG}"
[ -s "${ERR}" ] || rm -f "${ERR}"
[ -s "${DEBUGERR}" ] || rm -f "${DEBUGERR}"
rm -rf "${TMPDIST}"
