#!/bin/sh
#
# prune-distfiles.sh: Implement prune-distfiles target.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $WRKDIR               The top level work directory.
#    $DISTFILES            The top level distfiles directory.
#    $PKGDISTFILES         The package specifc distfiles subdirectory.
#    $TOPSRC               The top level of the source tree.
#    $GITCOMMIT            The git commit ID to build from.
#    $GITPKG               The name of the git repo project (default: $PKG)
#    $GITVER               The version of the git repo project (default: $VER)
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 1
fi

if [ -z "${DISTFILES}" ] || [ ! -d "${DISTFILES}" ]; then
    exit 2
fi

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 3
fi

# Detect format of the archives file and filter it down to just
# a list of the archive filenames.
filter_distfiles() {
    if grep -q -E "\(.+\)=" "${1}" >/dev/null 2>&1 ; then
        # archives file is in openssl(1) dgst format
        grep -vE "(^#|^$)" "${1}" | cut -d '(' -f 2 | cut -d ')' -f 1 | sort -u
    else
        # archives file is in legacy Peachtree Linux format
        grep -vE "(^#|^$)" "${1}" | rev | cut -d ' ' -f 1 | rev | sort -u
    fi
}

if [ ! -d "${PKGDISTFILES}" ]; then
    exit 0
fi

numfiles=$(find "${PKGDISTFILES}" -not -type d | wc -l)

if [ "${numfiles}" -eq 0 ]; then
    rmdir "${PKGDISTFILES}"
    exit 0
fi

keep="$(mktemp -d -p "${PKGDISTFILES}")"
basekeep="$(basename "${keep}")"

trap 'rm -rf "${keep}"' EXIT

if [ -f "${WRKDIR}/archives" ]; then
    if [ -z "${PKGDISTFILES}" ] || [ ! -d "${PKGDISTFILES}" ]; then
        exit 4
    fi

    filter_distfiles "${WRKDIR}"/archives | while read -r archive ; do
        echo "Keeping ${archive}"
        [ -f "${PKGDISTFILES}"/"${archive}" ] && mv "${PKGDISTFILES}"/"${archive}" "${keep}"
    done

fi

if [ -n "${GITCOMMIT}" ]; then
    if [ -z "${PKGDISTFILES}" ] || [ ! -d "${PKGDISTFILES}" ]; then
        exit 5
    fi

    archive="${GITPKG}-${GITVER}-${GITCOMMIT}.tar.gz"

    if [ -f "${PKGDISTFILES}"/"${archive}" ]; then
        echo "Keeping ${archive}"
        mv "${PKGDISTFILES}"/"${archive}" "${keep}"
    fi
fi

if [ -n "${keep}" ] && [ -n "${basekeep}" ]; then
    cd "${PKGDISTFILES}" || exit 6

    for f in * ; do
        bf="$(basename "${f}")"

        if [ "${bf}" = "${basekeep}" ] || [ "${bf}" = "SHA512" ]; then
            continue
        fi

        if [ -f "${f}" ]; then
            echo "Removing ${f}"
            rm "${f}"
        fi
    done

    if [ -d "${keep}" ] && [ -d "${PKGDISTFILES}" ]; then
        mv "${keep}"/* "${PKGDISTFILES}"
    fi
fi
