#!/bin/sh
#
# showprevpkg.sh: Show the previous package name as PKG-VER-REL.TARGET_MACH
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level project directory.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

TMPDIR="$(mktemp -d)"
trap 'rm -rf "${TMPDIR}"' EXIT

PREVPKG=
PREVVER=
PREVREL=
PREVTARGET_MACH=

PKGDIR="$(pwd | sed -e "s|${TOPSRC}/||g")"

# See if there is a pending commit on the package Makefile
if git status --porcelain Makefile | grep -q " M " >/dev/null 2>&1 ; then
    COMMIT=HEAD
else
    COMMIT="$(git log -n 2 --pretty=format:%H -- Makefile | tail -n 1)"
fi

git show "${COMMIT}":"${PKGDIR}"/Makefile > "${TMPDIR}"/Makefile

# Pick up the previous values for these variables
PREVPKG="$(make -f "${TMPDIR}"/Makefile show var=PKG)"
PREVREL="$(make -f "${TMPDIR}"/Makefile show var=REL)"
PREVTARGET_MACH="$(make -f "${TMPDIR}"/Makefile show var=TARGET_MACH)"

# If the package VER is the same as DISTRIBUTION_VERSION, we need
# figure out if the previous package change was before or after
# the most recent release tag.  If it was before, we need to get
# the previous DISTRIBUTION_VERSION.
# shellcheck disable=SC2016
if grep ^VER "${TMPDIR}"/Makefile | grep -q '$(DISTRIBUTION_VERSION)' >/dev/null 2>&1 ; then
    # timestamp of the package commit
    PKG_COMMIT_DATE=$(git show -s --format=%ct "${COMMIT}")

    # most recent release tag
    TAG="$(git describe --abbrev=0)"

    # timestamp of the tag
    TAG_COMMIT_DATE=$(git show -s --format=%ct "${TAG}" | tail -n 1)

    if [ "${PKG_COMMIT_DATE}" -lt "${TAG_COMMIT_DATE}" ]; then
        # need previous DISTRIBUTION_VERSION value
        DISTRIBUTION_VERSION="$(git show "${TAG}":mk/defs.mk | grep ^DISTRIBUTION_VERSION | cut -d ' ' -f 3)"
        PREVVER="$(env DISTRIBUTION_VERSION="${DISTRIBUTION_VERSION}" make -f "${TMPDIR}"/Makefile show var=VER)"
    fi
fi

if [ -z "${PREVVER}" ]; then
    PREVVER="$(make -f "${TMPDIR}"/Makefile show var=VER)"
fi

# Previous package string
echo "${PREVPKG}-${PREVVER}-${PREVREL}.${PREVTARGET_MACH}"

exit 0
