#!/bin/sh
#
# test.sh: Implement test stage.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# Copyright Chris Lumens <chris@bangmoney.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $WRKDIR               The top level work directory.
#    $WRKSRC               The extracted source directory.
#    $WRKBLD               The source build directory.
#    $PACKAGE              The full package name.
#    $CONFIGURE_STYLE      The method used to configure $WRKSRC for builds.
#    $PYTHON               Python interpreter to use.
#    $SEPARATE_BUILD       (y/n) Use source build directories?
#    $TEST_ENV             Environment variables passed to $(TEST_COMMAND).
#    $TEST_COMMAND         The make command.
#    $TEST_TARGET          The name of the test target in the Makefile.
#    $TEST_ARGS            Extra arguments passed to $(TEST_COMMAND).
#    $SKIP_TEST            (y/n) Run tests?
#    $TEST_COMMAND         The make command.
#    $HELPERS              Path to the helpers.sh script.
# This script is called by the build system scripts and should not be called
# directly.

[ "${SKIP_TEST}" = "y" ] && exit 0

if [ -z "${HELPERS}" ] || [ ! -f "${HELPERS}" ]; then
    exit 1
fi

# Helper functions
# shellcheck disable=SC1090
. "${HELPERS}"

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 2
fi

if [ "${SEPARATE_BUILD}" = "y" ]; then
    if [ -z "${WRKBLD}" ] || [ ! -d "${WRKBLD}" ]; then
        exit 3
    fi
else
    if [ -z "${WRKSRC}" ] || [ ! -d "${WRKSRC}" ]; then
        exit 4
    fi
fi

LOG="${WRKDIR}/test.log"
ERR="${WRKDIR}/test.err"

# Setup logging via a couple fifo's connected to tee
outlog="${TMPDIR:-/tmp}/out.$$"
errlog="${TMPDIR:-/tmp}/err.$$"
mkfifo "${outlog}" "${errlog}"
trap 'rm "${outlog}" "${errlog}"' EXIT

for p in ${PACKAGE} ; do
    echo ">>> Testing ${p}..."
done

if [ "${CONFIGURE_STYLE}" = "python" ]; then
    cd "${WRKSRC}" || exit 5

    run_tee
    # shellcheck disable=SC2086
    ( env ${TEST_ENV} ${PYTHON} setup.py --no-user-cfg ${TEST_TARGET} ${TEST_ARGS} >"${outlog}" 2>"${errlog}" ) || exit $?
elif [ "${CONFIGURE_STYLE}" = "meson" ]; then
    cd "${WRKSRC}" || exit 6

    run_tee
    # shellcheck disable=SC2086
    ( env ${TEST_ENV} ninja -C "${WRKBLD}" ${TEST_TARGET} ${TEST_ARGS} >"${outlog}" 2>"${errlog}" ) || exit $?
elif [ "${CONFIGURE_STYLE}" = "cargo" ]; then
    cd "${WRKBLD}" || exit 7

    run_tee
    # shellcheck disable=SC2086
    ( env ${TEST_ENV} cargo test -v --offline --release ${TEST_ARGS} >"${outlog}" 2>"${errlog}" ) || exit $?
else
    if [ -d "${WRKBLD}" ] || [ -d "${WRKSRC}" ]; then
        [ "${SEPARATE_BUILD}" = "y" ] && D=${WRKBLD} || D=${WRKSRC}

        run_tee
        # shellcheck disable=SC2086
        ( env ${TEST_ENV} ${TEST_COMMAND} -C "${D}" ${TEST_TARGET} ${TEST_ARGS} >"${outlog}" 2>"${errlog}" ) || exit $?
    fi
fi

[ -s "${LOG}" ] || rm -f "${LOG}"
[ -s "${ERR}" ] || rm -f "${ERR}"
