#!/bin/sh
#
# update-patches.sh: Generate new or update existing source patches
#
# Copyright David Cantrell <dcantrell@burdell.org>
# Copyright Chris Lumens <chris@bangmoney.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $WRKDIR               The top level work directory.
#    $WRKSRC               The source subdirectory in $WRKDIR.
#    $PATCHES              The directory where patches are stored.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${WRKDIR}" ] || [ ! -d "${WRKDIR}" ]; then
    exit 1
fi

[ -z "${WRKSRC}" ] && exit 2
[ -z "${PATCHES}" ] && exit 3

generate_patches() {
    d="${1}"
    find "${d}" -type f -name "*.orig" | while read -r original ; do
        modified="$(dirname "${original}")/$(basename "${original}" .orig)"
        patchfile="$(echo "${modified}" | sed -e "s|${d}/||g" -e 's/\//_/g' -e 's/\ /_/g').patch"

        if [ ! -f "${modified}" ]; then
            echo ">>> WARNING: found ${original} but not ${modified}"
            continue
        fi

        original="$(echo "${original}" | sed -e "s|^${WRKDIR}/||g")"
        modified="$(echo "${modified}" | sed -e "s|^${WRKDIR}/||g")"
        diff -u "${original}" "${modified}" > "${PATCHES}/${patchfile}.NEW" 2>/dev/null || :

        if [ -f "${PATCHES}/${patchfile}" ]; then
            sed -e '1,2d' "${PATCHES}/${patchfile}" > "${WRKDIR}"/orig
            sed -e '1,2d' "${PATCHES}/${patchfile}.NEW" > "${WRKDIR}"/new

            if diff -u "${WRKDIR}/orig" "${WRKDIR}"/new >/dev/null 2>&1 ; then
                echo ">>> No changes detected in ${modified}."
                rm -f "${PATCHES}/${patchfile}.NEW"
            else
                echo ">>> Updating patch for ${modified}..."
                cat "${PATCHES}/${patchfile}.NEW" > "${PATCHES}/${patchfile}" && rm -f "${PATCHES}/${patchfile}.NEW"
            fi

            rm -f "${WRKDIR}"/orig "${WRKDIR}"/new
        elif [ -s "${PATCHES}/${patchfile}.NEW" ]; then
            echo ">>> Generating patch for ${modified}..."
            mv -f "${PATCHES}/${patchfile}.NEW" "${PATCHES}/${patchfile}"
        else
            rm -f "${PATCHES}/${patchfile}.NEW"
        fi
    done
}

[ -d "${PATCHES}" ] || mkdir -p "${PATCHES}"

cd "${WRKDIR}" || exit 4
[ -d "${WRKSRC}" ] && generate_patches "${WRKSRC}"

# we may have no patches, silently try to remove directory
rmdir --ignore-fail-on-non-empty "${PATCHES}" >/dev/null 2>&1

exit 0
