#!/bin/sh
#
# usereqs.sh: Install packages listed in buildreqs to the sysroot tree
#             specified.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# Copyright Chris Lumens <chris@bangmoney.org>
# Copyright Brian C. Lane <bcl@brianlane.com>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level distribution source directory.
#    $WRKDIR               The package work directory.
#    $WRKSYSROOT           The directory passed to --sysroot on gcc(1).
#    $MAKE                 The make(1) command.
#    $RELEASE              The release directory containing built packages.
#    $DISTADD_TMPDIR       Temp directory for distadd(8) to use.
#    $DIST_URL             Site to download build dependencies from.
#    $DIST_DOWNLOAD        Try to download build dependencies (y/n).
#    $HELPERS              Path to the helpers.sh script.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${HELPERS}" ] || [ ! -f "${HELPERS}" ]; then
    exit 1
fi

# Helper functions
# shellcheck disable=SC1090
. "${HELPERS}"

[ -z "${WRKDIR}" ] && exit 2
[ -z "${WRKSYSROOT}" ] && exit 3
[ -z "${MAKE}" ] && exit 4
[ -z "${DISTADD_TMPDIR}" ] && exit 5

[ -d "${DISTADD_TMPDIR}" ] || mkdir -p "${DISTADD_TMPDIR}"
CWD="$(pwd)"

# Gather list of buildreqs
if [ -f "${CWD}"/buildreqs ]; then
    BUILDREQS="$(${MAKE} show-buildreqs)" || exit ${?}
fi

[ -z "${BUILDREQS}" ] && exit 0

# We use a single sysroot, so wipe it and start a new one.
rm -rf "${WRKSYSROOT}"
mkdir -p "${WRKSYSROOT}"

# distadd (standard Peachtree Linux location is /usr/sbin)
distadd="/usr/sbin/distadd"

[ -d "${WRKDIR}" ] || mkdir -p "${WRKDIR}"
LOG="${WRKDIR}/usereqs.log"
ERR="${WRKDIR}/usereqs.err"

# Setup logging via a couple fifo's connected to tee
outlog="${TMPDIR:-/tmp}/out.$$"
errlog="${TMPDIR:-/tmp}/err.$$"
mkfifo "${outlog}" "${errlog}"
trap 'rm "${outlog}" "${errlog}"' EXIT

# Step one:  Verify all build dependencies exist.
run_tee
( for dep in ${BUILDREQS} ; do
      fullpkgpath="$(get_var "${dep}" FULLPKGPATH)"

      if [ -z "${fullpkgpath}" ]; then
          echo "Build requirement ${dep} does not exist, cannot continue." >&2
          exit 4
      fi

      for pkgpath in ${fullpkgpath} ; do
          [ -f "${pkgpath}" ] || download_dist "${dep}"

          if [ ! -f "${pkgpath}" ]; then
              echo "*** Missing ${pkgpath}, cannot continue." >&2
              exit 5
          fi
      done
  done \
  >"${outlog}" 2>"${errlog}"
) || exit ${?}

# Step two:  Install build dependencies.
run_tee
( for dep in ${BUILDREQS} ; do
      fullpkgpath="$(get_var "${dep}" FULLPKGPATH)"

      if [ -z "${fullpkgpath}" ]; then
          echo "Build requirement ${dep} does not exist, cannot continue." >&2
          exit 6
      fi

      for pkgpath in ${fullpkgpath} ; do
          ${distadd} --tmpdir="${DISTADD_TMPDIR}" -S -d -r "${WRKSYSROOT}" "${pkgpath}"
      done
  done \
  >"${outlog}" 2>"${errlog}"
) || exit ${?}

[ -s "${LOG}" ] || rm -f "${LOG}"
[ -s "${ERR}" ] || rm -f "${ERR}"
