/* Install or upgrade a .dist package. */

/*
 * Copyright David Shea <david@reallylongword.org>
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <assert.h>
#include <libgen.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <utime.h>
#include <errno.h>
#include <err.h>
#include <glob.h>
#include <paths.h>
#include <limits.h>
#include "dist-tools.h"
#include "db.h"
#include "opts.h"

extern char *optarg;
extern int optind, opterr, optopt;

/* ld.so cache structure local to distadd */
typedef struct _ldso_entry_t {
    char *soname;
    strlist_t *libs;
    TAILQ_ENTRY(_ldso_entry_t) items;
} ldso_entry_t;

typedef TAILQ_HEAD(ldso_s, _ldso_entry_t) ldso_t;

/* Global variables */
char *db_installed = NULL;      /* path to installed.db to use */
char *prepend = NULL;           /* root path to use (-r) */
uint64_t opts = 0;

/* Local prototypes */
static void usage(void);
static ldso_t *get_ldso_cache(void);
static ldso_t *ldso_insert(ldso_t *, char *, char *);
static bool check_depspath(const struct dist *, const char *);
static bool check_deps(const struct dist *);
static int instreloc(struct dist *, struct dist *, bool);
static int copymeta(const char *, char *, bool);
static int remove_obsolete(int, sqlite3 *, const char *);
static int mkdistentry(const struct dist *, const struct dist *, bool);
static int gendistentry(const struct dist *);
static int initialetc(const struct dist *);
static bool is_downgrade(const struct dist *, const struct dist *);
static char *format_msg(const struct dist *, const struct dist *, bool);

/* Local prototypes - instreloc support functions */
static int installdir(struct finfo *);
static int installnew(struct dist *, struct finfo *, bool);
static int installhard(struct finfo *);
static int fixdir(struct finfo *);
static int setattr(const char *, const char *, const struct stat *);
static int _read_finfo_row(sqlite3_stmt *, finfolist_t *, finfolist_t *);
static int installfile(struct finfo *, struct dist *, bool);

/* Local global variables */
static char *prog = NULL;
static char *dbpath = NULL;             /* package database location */
static char *tmppath = NULL;            /* temporary directory */
static finfolist_t *errfiles = NULL;    /* old files that failed to check */
static finfolist_t *newerrfiles = NULL; /* new files that failed to inst */
static finfolist_t *newmanifest = NULL; /* new manifest list */
static finfolist_t *newhards = NULL;    /* new hardlink list */
static finfolist_t *newdirs = NULL;     /* new directory list */
static finfolist_t *oldfiles = NULL;    /* old files to remove */
static finfolist_t *olddirs = NULL;     /* old directories to remove */

/* Command line options */
static struct option long_options[] = {
    { "noarch", no_argument, NULL, 'a' },
    { "nodeps", no_argument, NULL, 'd' },
    { "force", no_argument, NULL, 'f' },
    { "force-script", no_argument, NULL, 'F' },
    { "genattrs", no_argument, NULL, 'g' },
    { "multi", no_argument, NULL, 'm' },
    { "quick", no_argument, NULL, 'q' },
    { "strict", no_argument, NULL, 's' },
    { "noscripts", no_argument, NULL, 'S' },
    { "upgrade", no_argument, NULL, 'u' },
    { "dry-run", no_argument, NULL, 'y' },
    { "root", required_argument, NULL, 'r' },
    { "upgrade-version", required_argument, NULL, 'v' },
    { "etcmerge", no_argument, NULL, 'e' },
    { "version", no_argument, NULL, 'V' },
    { "help", no_argument, NULL, 'h' },
    { "verbose", no_argument, NULL, 0 },
    { "tmpdir", required_argument, NULL, 0 },
    { "no-x11", no_argument, NULL, 0 },
    { "no-wayland", no_argument, NULL, 0 },
    { "no-man", no_argument, NULL, 0 },
    { "no-info", no_argument, NULL, 0 },
    { "no-doc", no_argument, NULL, 0 },
    { "no-licenses", no_argument, NULL, 0 },
    { NULL, 0, NULL, 0 }
};

/* Displays messages only when --verbose is passed. */
#define verbose_error(...) \
    if (opts & OPTS_VERBOSE) { \
        warnx(__VA_ARGS__); \
    }

/* Usage screen displayed when the user passes --help as an argument. */
static void usage(void)
{
    printf("Usage: %s [OPTION]... archive...\n", prog);
    puts("Install a Peachtree Linux dist package\n"
         "\n"
         "  -a, --noarch               Skip architecture check\n"
         "  -d, --nodeps               Skip library dependency check\n"
         "  -f, --force                Force replacement of all files\n"
         "  -F, --force-script         Continue install if any package script fails\n"
         "  -g, --genattrs             Regenerate metadata after package is installed\n"
         "  -m, --multi                Install additional version of a package\n"
         "  -q, --quick                Skip all file checks\n"
         "  -s, --strict               Perform attribute checks\n"
         "  -S, --noscripts            Do not run pre/post-install scripts\n"
         "  -u, --upgrade              Upgrade a package\n"
         "  -y, --dry-run              Stop before installing or upgrading anything\n"
         "  -r PATH, --root=PATH       Install package to directory PATH\n"
         "  -v VER, --upgrade-version=VER\n"
         "                             Choose VER for upgrading if multiple\n"
         "                             versions of a package are installed\n"
         "  -P SPEC, --pkgspec=SPEC    Specify exact installed package to upgrade\n"
         "  -e, --etcmerge             Run etcmerge -i after install\n");
    printf("  -no-x11                    Do not install anything to %s\n", X11_PREFIX);
    printf("  -no-wayland                Do not install anything to %s\n", WAYLAND_PREFIX);
    puts("  -no-man                    Do not install any man pages\n");
    puts("  -no-info                   Do not install any info pages\n");
    printf("  -no-doc                    Do not install anything to %s\n", DOCDIR);
    printf("  -no-licenses               Do not install anything to %s\n", LICENSESDIR);
    printf("  --verbose                    Display more output\n");
    printf("  --tmpdir=PATH              Use PATH for temporary files (default: %s)\n", _PATH_VARTMP);
    puts("  -V, --version                Display version information\n"
         "  -h, --help                   This help message\n");
    return;
}

/*
 * Takes the output of /sbin/ldconfig -p, places it in a hash table.
 * sonames are keys, value is the path to the filename providing that
 * soname.
 *
 * Caller is responsible for freeing the hash table returned.
 */
static ldso_t *get_ldso_cache(void)
{
    char *cmd = NULL;
    char *line = NULL, *tmp = NULL, *soname = NULL, *filename = NULL;
    char **fields = NULL;;
    strlist_t *lines = NULL;
    strlist_entry_t *entry = NULL;
    char *sout = NULL;
    int status = 0;
    ldso_t *cache = NULL;
    size_t len = 0;
    char *buf = NULL;
    FILE *fp = NULL;

    /* Initialize the cache */
    cache = calloc(1, sizeof(*cache));
    assert(cache != NULL);
    TAILQ_INIT(cache);

    /* Prepare and run the ldconfig command */
    if (access(LDCONFIG, X_OK) == -1) {
        warnx("ERROR: unable to execute %s to check ELF dependencies", LDCONFIG);
        return NULL;
    }

    if (prepend != NULL) {
        xasprintf(&cmd, LDCONFIG" -p -r %s", prepend);
    } else {
        xasprintf(&cmd, LDCONFIG" -p");
    }

    if ((fp = popen(cmd, "r")) == NULL) {
        warn("popen");
        return NULL;
    }

    lines = init_strlist();

    while (getline(&buf, &len, fp) != -1) {
        entry = calloc(1, sizeof(*entry));
        assert(entry != NULL);
        entry->val = strdup(buf);
        TAILQ_INSERT_TAIL(lines, entry, items);
    }

    free(buf);
    status = pclose(fp);

    if (!WIFEXITED(status) || WEXITSTATUS(status) != EXIT_SUCCESS) {
        warnx("unable to execute %s", cmd);
        return NULL;
    }

    free(cmd);

    /*
     * Collect the sonames and actual file locations from the ldconfig
     * output.  Lines that do not have " => " in the string are ignored.
     * Valid cache output lines take this format:
     *     libGL.so.1 (libc6,x86-64) => /usr/lib/libGL.so.1
     * The soname followed by a space, then optional text of some length
     * (which can include spaces), followed by ' => ', followed by the
     * path to the actual file providing the symbol.  What we want in our
     * hash table is the soname and the filename.  Strip everything else.
     */
    TAILQ_FOREACH(entry, lines, items) {
        line = strtrim(entry->val);

        if (strstr(line, " => ") == NULL) {
            free(line);
            continue;
        }

        fields = strsplit(line, "=>");

        if (strvlen(fields) == 3 && !strcmp(fields[1], "")) {
            soname = strtrim(fields[0]);
            filename = strtrim(fields[2]);
            tmp = index(soname, ' ');

            if (tmp != NULL) {
                *tmp = '\0';
            }

            cache = ldso_insert(cache, soname, filename);
            free(soname);
            free(filename);
        }

        strvfree(fields);
        free(line);
    }

    free_strlist(lines);
    free(sout);
    return cache;
}

/*
 * Insert an SONAME in to an ldso_t if it doesn't exist, otherwise
 * add the filename to the existing entry.
 */
static ldso_t *ldso_insert(ldso_t *list, char *soname, char *filename)
{
    ldso_entry_t *entry = NULL;
    strlist_entry_t *sentry = NULL;

    assert(list != NULL);
    assert(soname != NULL);
    assert(filename != NULL);

    /* First, try to find the SONAME in the list. */
    TAILQ_FOREACH(entry, list, items) {
        if (!strcmp(entry->soname, soname)) {
            if (!is_str_in_list(entry->libs, filename)) {
                sentry = calloc(1, sizeof(*sentry));
                assert(sentry != NULL);
                sentry->val = strdup(filename);
                TAILQ_INSERT_TAIL(entry->libs, sentry, items);
            }

            return list;
        }
    }

    /* Second, add the entry to the list with the filename to init it. */
    entry = calloc(1, sizeof(*entry));
    assert(entry != NULL);
    entry->soname = strdup(soname);
    entry->libs = init_strlist();
    sentry = calloc(1, sizeof(*sentry));
    assert(sentry != NULL);
    sentry->val = strdup(filename);
    TAILQ_INSERT_TAIL(entry->libs, sentry, items);
    TAILQ_INSERT_TAIL(list, entry, items);

    return list;
}

/*
 * Reads the depspath= line from pkginfo and looks in those directories for
 * the dependency specified.  The idea here is to cover libraries that are
 * not in the usual library search path.
 */
static bool check_depspath(const struct dist *package, const char *soname)
{
    bool ret = false;
    char *filename = NULL, *rfilename = NULL;
    strlist_t *dirs = NULL;
    strlist_entry_t *entry = NULL;
    struct stat sb;

    assert(package != NULL);
    assert(soname != NULL);

    /*
     * If depspath is empty and we are installing to a path other than /,
     * use a default library directory.
     */
    if (package->depspath == NULL) {
        dirs = split_strlist(LIBDIR, ":");
    } else {
        dirs = split_strlist(package->depspath, ":");
    }

    /* Iterate each directory in depspath. */
    TAILQ_FOREACH(entry, dirs, items) {
        /* Concatenate dir with soname. */
        filename = NULL;
        rfilename = NULL;
        xasprintf(&filename, "%s%s/%s", ((prepend == NULL) ? "" : prepend), entry->val, soname);

        if ((rfilename = realpath(filename, rfilename)) == NULL) {
            free(filename);
            continue;
        }

        if (stat(rfilename, &sb) == -1) {
            warn("stat");
            break;
        }

        /*
         * Assume the dep is met if we find the file we are looking for
         * and it's an ELF file.  This check could probably be more robust.
         */
        if (S_ISREG(sb.st_mode) && is_elf(rfilename)) {
            ret = true;
            break;
        }

        free(filename);
        free(rfilename);
    }

    free_strlist(dirs);
    return ret;
}

/*
 * Check the running system to see if the ELF dependencies required
 * by the package are available.
 */
static bool check_deps(const struct dist *package)
{
    bool ret = true;
    bool cached = false;
    strlist_entry_t *entry = NULL;
    strlist_entry_t *pentry = NULL;
    ldso_t *cache = NULL;
    ldso_entry_t *ldsoentry = NULL;

    /* if a package has no dependency data, just continue */
    if (package->deps && TAILQ_EMPTY(package->deps)) {
        return true;
    }

    /* what does the system provide? */
    cache = get_ldso_cache();

    /* look for each dep we expect */
    if (cache != NULL && package->deps != NULL) {
        TAILQ_FOREACH(entry, package->deps, items) {
            cached = false;

            if (package->provs && !TAILQ_EMPTY(package->provs)) {
                TAILQ_FOREACH(pentry, package->provs, items) {
                    if (!strcmp(entry->val, pentry->val)) {
                        cached = true;
                        break;
                    }
                }
            }

            if (!cached) {
                TAILQ_FOREACH(ldsoentry, cache, items) {
                    if (!strcmp(entry->val, ldsoentry->soname)) {
                        cached = true;
                        break;
                    }
                }
            }

            if (!cached && !check_depspath(package, entry->val)) {
                ret = false;
                warnx("*** %s not found.", entry->val);
            }
        }

        while (!TAILQ_EMPTY(cache)) {
            ldsoentry = TAILQ_FIRST(cache);
            TAILQ_REMOVE(cache, ldsoentry, items);
            free_strlist(ldsoentry->libs);
            free(ldsoentry->soname);
            free(ldsoentry);
        }

        free(cache);
    }

    return ret;
}

/*
 * Iterate a row of DB_CONTENTS_GONE results and add them to the
 * appropriate hash table.
 */
static int _read_finfo_row(sqlite3_stmt *stmt, finfolist_t *oldfiles, finfolist_t *olddirs)
{
    int type = -1;
    struct finfo *info = NULL;

    if ((info = calloc(1, sizeof(*info))) == NULL) {
        warn("calloc");
        return -1;
    }

    info->packageid = sqlite3_column_int(stmt, 0);
    info->type = sqlite3_column_int(stmt, 1);
    info->filename = strdup((const char *) sqlite3_column_text(stmt, 2));
    info->fullpath = joinpath(prepend, info->filename, NULL);

    if (info->type == TYPE_REG) {
        info->checksum = strdup((const char *) sqlite3_column_text(stmt, 3));
    } else {
        info->checksum = NULL;
    }

    /* we don't care about the attrs here, we're removing */
    info->attr = NULL;

    if (info->type == TYPE_HARD || info->type == TYPE_SOFT) {
        info->linkname = strdup((const char *) sqlite3_column_text(stmt, 4));
    } else {
        info->linkname = NULL;
    }

    /* add this to the list */
    if (type == TYPE_DIR) {
        olddirs = sorted_finfolist_insert(olddirs, info);
    } else {
        oldfiles = sorted_finfolist_insert(oldfiles, info);
    }

    return 0;
}

/*
 * Walk the extracted reloc.payload and install new files and directories
 * to the system.  Handle removal of obsolete files and directories for
 * upgrades too.  Also, if upgrading a package, run the old package's
 * preremove script here.
 */
static int instreloc(struct dist *package, struct dist *oldpackage, bool upgrade)
{
    int errs = 0;
    finfolist_entry_t *entry = NULL;
    char *tmp = NULL;
    char *dberr = NULL;
    char *query = NULL;
    char *err = NULL;
    sqlite3 *db = NULL;
    sqlite3_stmt *stmt = NULL;

    assert(package != NULL);

    umask(0077);

    newhards = init_finfolist();
    newdirs = init_finfolist();

    if (oldfiles == NULL) {
        oldfiles = init_finfolist();
    }

    if (olddirs == NULL) {
        olddirs = init_finfolist();
    }

    /*
     * If we are upgrading an existing package, run the preremove
     * script for the existing package.
     */
    if (oldpackage != NULL && !oldpackage->debug && !(opts & OPTS_NOSCRIPTS) && !(opts & OPTS_DRYRUN)) {
        if (runscript(package, oldpackage, SCRIPT_PREREMOVE, upgrade, prepend, NULL)) {
            printf("%s failed for %s, ", DIST_PREREMOVE, oldpackage->name);

            if (opts & OPTS_FORCESCRIPT) {
                printf("continuing anyway\n");
            } else {
                printf("skipping\n");
                return -2;
            }
        }
    }

    /* Read bom.db for the new package */
    if (sqlite3_open(package->bom, &db) != SQLITE_OK) {
        warnx("sqlite3_open: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -1;
    }

    if (sqlite3_exec(db, "PRAGMA synchronous = OFF", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_free(err);
        return -1;
    }

    sqlite3_free(err);

    if (sqlite3_exec(db, "PRAGMA journal_mode = MEMORY", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_free(err);
        return -1;
    }

    sqlite3_free(err);

    if (read_contents(db, package->sigver, &newmanifest, &newdirs, &newhards)) {
        warnx("read_contents");
        return -1;
    }

    sqlite3_close(db);
    db = NULL;

    /* For upgrades, generate list of files to remove */
    if (upgrade) {
        if (sqlite3_open(db_installed, &db) != SQLITE_OK) {
            warnx("sqlite3_open: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            return -1;
        }

        if (sqlite3_exec(db, "PRAGMA synchronous = OFF", NULL, NULL, &err) != SQLITE_OK) {
            warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            sqlite3_free(err);
            return -1;
        }

        sqlite3_free(err);

        if (sqlite3_exec(db, "PRAGMA journal_mode = MEMORY", NULL, NULL, &err) != SQLITE_OK) {
            warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            sqlite3_free(err);
            return -1;
        }

        sqlite3_free(err);

        /* attach bom.db from the new package */
        xasprintf(&query, BOM_CONTENTS_ATTACH, package->bom);

        if (sqlite3_exec(db, query, NULL, 0, &dberr) != SQLITE_OK) {
            warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            sqlite3_free(dberr);
            return -1;
        }

        sqlite3_free(dberr);
        free(query);
        query = NULL;

        /* generate a list of files and directories on the system not in bom.db */
        if (sqlite3_prepare(db, DB_CONTENTS_GONE, -1, &stmt, 0) != SQLITE_OK) {
            warnx("sqlite3_prepare: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            return -1;
        }

        if (sqlite3_bind_int(stmt, 1, oldpackage->id) != SQLITE_OK) {
            warnx("sqlite3_bind_int: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            return -1;
        }

        while (sqlite3_step(stmt) == SQLITE_ROW) {
            if (_read_finfo_row(stmt, oldfiles, olddirs)) {
                warnx("sqlite3_step: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
                return -1;
            }
        }

        sqlite3_finalize(stmt);
    }

    /* removing any old package metadata */
    if (oldpackage != NULL && oldpackage->pkgdigest != NULL && !(opts & OPTS_DRYRUN)) {
        tmp = joinpath(dbpath, oldpackage->pkgdigest, NULL);

        if (rmtree(tmp, true, false)) {
            warn("rmtree");
            return -1;
        }

        free(tmp);
    }

    if (!(opts & OPTS_DRYRUN)) {
        /* install new directories */
        TAILQ_FOREACH(entry, newdirs, items) {
            errs += installdir(entry->info);
        }

        /* install new files */
        TAILQ_FOREACH(entry, newmanifest, items) {
            errs += installnew(package, entry->info, upgrade);
        }

        /* Make new hard links */
        TAILQ_FOREACH_REVERSE(entry, newhards, finfolist_s, items) {
            errs += installhard(entry->info);
        }

        /* All that installing messed with the mtime/atime of directories, fix 
         * now.  Fortunately, thanks to laziness, the list is still in the
         * right order */
        TAILQ_FOREACH(entry, newdirs, items) {
            errs += fixdir(entry->info);
        }
    }

    return errs;
}

/* make all directories */
static int installdir(struct finfo *info)
{
    int err = 0;
    char *curfile = NULL;
    char *newfile = NULL;

    assert(info != NULL);

    curfile = joinpath(tmppath, PAYLOAD_SUBDIR, info->filename, NULL);
    newfile = joinpath(prepend, info->filename, NULL);
    umask(0077);

    if (mkdirp(newfile, _DIR_MODE)) {
        verbose_error("%s: Error making path %s: %s\n", __func__,
                      newfile, strerror(errno));
        err = 1;

        if (!(opts & OPTS_GENATTRS)) {
            newerrfiles = sorted_finfolist_insert(newerrfiles, info);
        }
    } else if (setattr(newfile, curfile, info->attr)) {
        err = 1;
    }

    free(curfile);
    free(newfile);
    return err;
}

/*
 * Install all files
 * old files that didn't check should be in errfiles
 * remove files in new and errfiles from errfiles for genattrs
 */
static int installnew(struct dist *package, struct finfo *info, bool upgrade)
{
    int errs = 0;
    bool go = true;
    char *curfile = NULL;
    char *distfile = NULL;
    char *reffile = NULL;
    struct stat sb;

    if (info->type == TYPE_DIR || info->type == TYPE_HARD) {
        return errs;
    }

    if (find_finfolist_filename(errfiles, info->filename) == NULL) {
        curfile = joinpath(prepend, info->filename, NULL);

        /* if not in the table, the file shouldn't exist */
        if (!(opts & OPTS_QUICK) && (!lstat(curfile, &sb) || errno != ENOENT)) {
            if (!upgrade) {
                verbose_error("%s already exists", curfile);
            }

            if (opts & OPTS_FORCE) {
                if (!upgrade) {
                    verbose_error(", overwriting\n");
                }
            } else {
                go = false;
                errs += 1;
                verbose_error("\n");

                if (!(opts & OPTS_GENATTRS)) {
                    newerrfiles = sorted_finfolist_insert(newerrfiles, info);
                }
            }
        }

        if (go) {
            if (installfile(info, package, upgrade)) {
                errs += 1;

                if (!(opts & OPTS_GENATTRS)) {
                    newerrfiles = sorted_finfolist_insert(newerrfiles, info);
                }
            }
        } else if (info->attr != NULL) {
            /* if not copying, reset attributes in place */
            distfile = joinpath(prepend, info->filename, NULL);
            reffile = joinpath(tmppath, PAYLOAD_SUBDIR, info->filename, NULL);

            if (info->type != TYPE_SOFT &&
                setattr(distfile, reffile, info->attr)) {
                errs += 1;
            }

            free(distfile);
            free(reffile);
        }

        free(curfile);
    } else {
        errfiles = remove_finfolist_filename(errfiles, info->filename);
    }

    return errs;
}

static int installhard(struct finfo *info)
{
    int errs = 0;
    char *oldpath = NULL;
    char *newpath = NULL;
    char *distpath = NULL;

    if (find_finfolist_filename(errfiles, info->filename) == NULL) {
        oldpath = joinpath(prepend, info->linkname, NULL);
        distpath = joinpath(prepend, info->filename, NULL);
        xasprintf(&newpath, "%s.distnew", distpath);

        /* step 0: must remove the target before a hard link will be made */
        if (!access(newpath, F_OK)) {
            if (unlink(newpath) == -1) {
                 verbose_error("%s: unable to unlink %s: (%d) %s\n", __func__, newpath, errno, strerror(errno));
                 errs = 1;
            }
        }

        /* step 1, create a hard link from the link to the new staging link */
        if (link(oldpath, newpath)) {
            verbose_error("%s: unable to link %s to %s: (%d) %s\n", __func__, oldpath, newpath, errno, strerror(errno));
            errs = 1;

            if (!(opts & OPTS_GENATTRS)) {
                newerrfiles = sorted_finfolist_insert(newerrfiles, info);
            }
        }

        /* step 2: remove the target name */
        if (!access(distpath, F_OK)) {
            if (unlink(distpath) == -1) {
                verbose_error("%s: unable to unlink %s: (%d) %s\n", __func__, distpath, errno, strerror(errno));
                errs = 1;
            }
        }

        /* step 3, rename the staging link to the target name */
        if (rename(newpath, distpath) == -1) {
            verbose_error("%s: unable to rename %s to %s: (%d) %s\n", __func__, newpath, distpath, errno, strerror(errno));
            errs = 1;
        }

        /* step 4, if staging link name exists, remove it */
        if (!access(newpath, F_OK)) {
            if (unlink(newpath) == -1) {
                verbose_error("%s: unable to unlink %s: (%d) %s\n", __func__, newpath, errno, strerror(errno));
                errs = 1;
            }
        }

        errno = 0;
        free(oldpath);
        free(newpath);
        free(distpath);
    } else {
        errfiles = remove_finfolist_filename(errfiles, info->filename);
    }

    return errs;
}

/* run setattr on directories again to fix times */
static int fixdir(struct finfo *info)
{
    int err = 0;
    char *curfile = NULL;
    char *newfile = NULL;

    assert(info != NULL);

    curfile = joinpath(tmppath, PAYLOAD_SUBDIR, info->filename, NULL);
    newfile = joinpath(prepend, info->filename, NULL);

    if (setattr(newfile, curfile, info->attr)) {
        err = 1;
    }

    free(curfile);
    free(newfile);
    return err;
}

/* Reffile used to set mtime/atime */
static int setattr(const char *file, const char *reffile, const struct stat *attr)
{
    int success = 0;
    struct utimbuf times;
    struct stat timestat;

    assert(file != NULL);
    assert(attr != NULL);
    assert(reffile != NULL);

    if (chown(file, attr->st_uid, attr->st_gid) == -1 ||
        chmod(file, (attr->st_mode & S_INTER)) == -1) {
        verbose_error("%s: Could not set mode for %s: %s\n", __func__, file, strerror(errno));
        errno = 0;
        success = -1;
    }

    if (lstat(reffile, &timestat) == -1) {
        verbose_error("%s: Could not stat %s: %s\n", __func__, reffile, strerror(errno));
        return -1;
    }

    times.actime = timestat.st_atime;
    times.modtime = timestat.st_mtime;

    if (utime(file, &times) == -1) {
        verbose_error("%s: Could not set times for %s: %s\n", __func__, file, strerror(errno));
        errno = 0;
        success = -1;
    }

    return success;
}

/* Handle installation of individual payload entries. */
static int installfile(struct finfo *info, struct dist *package, bool upgrade)
{
    char *curfile = NULL;
    char *newfile = NULL;
    char *distfile = NULL;
    int success = 0;

    assert(info != NULL);
    assert(package != NULL);

    curfile = joinpath(tmppath, PAYLOAD_SUBDIR, info->filename, NULL);
    distfile = joinpath(prepend, info->filename, NULL);
    xasprintf(&newfile, "%s.distnew", distfile);

    if (info->type == TYPE_SOFT) {
        if (!access(distfile, F_OK)) {
            if (unlink(distfile)) {
                verbose_error("%s: Unable to remove existing link %s: %s\n", __func__, distfile, strerror(errno));
                return -1;
            }
        }

        if (symlink(info->linkname, newfile)) {
            verbose_error("%s: Unable to create link %s: %s\n", __func__, newfile, strerror(errno));
            return -1;
        }
    } else if (info->type == TYPE_CHAR || info->type == TYPE_BLOCK) {
        if (mknod(newfile, info->attr->st_mode, info->attr->st_rdev)) {
            verbose_error("%s: Unable to create device node %s: %s\n", __func__, newfile, strerror(errno));
            return -1;
        }
    } else if (info->type == TYPE_SOCK) {
        /* XXX: should we support this?  technically it's not possible */
        verbose_error("%s: cannot install socket %s with distadd\n", __func__, newfile);
        return -1;
    } else if (info->type == TYPE_FIFO) {
        if (mkfifo(newfile, info->attr->st_mode)) {
            verbose_error("%s: Unable to create FIFO %s: %s\n", __func__, newfile, strerror(errno));
            return -1;
        }
    } else if (info->type == TYPE_REG) {
        if (copyfile(curfile, newfile, (opts & OPTS_FORCE) || upgrade, (opts & OPTS_VERBOSE) || !upgrade)) {
            return -1;
        }
    } else {
        verbose_error("%s: unable to handle file type %d\n", __func__, info->type);
        return -1;
    }

    if (info->attr != NULL && info->type != TYPE_SOFT &&
        setattr(newfile, curfile, info->attr)) {
        success = -1;
    }

    if (rename(newfile, distfile)) {
        verbose_error("%s: Unable to rename %s to %s: %s\n", __func__, newfile, distfile, strerror(errno));
        success = -1;

        if (unlink(newfile)) {
            verbose_error("%s: Unable to unlink %s: %m\n", __func__, newfile);
        }
    }

    free(curfile);
    free(newfile);
    free(distfile);
    return success;
}

/*
 * Little support function to prevent having to do fifty billion string
 * concatenations in mkdistentry.
 */
static int copymeta(const char *filename, char *libdir, bool upgrade)
{
    int success = 0;
    struct stat sb;
    char *src = NULL;
    char *dest = NULL;

    assert(filename != NULL);
    assert(libdir != NULL);

    src = joinpath(tmppath, filename, NULL);
    dest = joinpath(libdir, filename, NULL);

    if (lstat(src, &sb) == 0) {
        /* force copying if the user passed -f (force) or -u (upgrade) */
        success = copyfile(src, dest, (opts & OPTS_FORCE) || upgrade, (opts & OPTS_VERBOSE) || !upgrade);
    } else {
        errno = 0;
        success = -1;
    }

    free(src);
    free(dest);
    return success;
}

/*
 * Remove obsolete package contents on a package upgrade.
 * This function removes database entries in the contents table.
 */
static int remove_obsolete(int id, sqlite3 *db, const char *bomdb)
{
    char *err = NULL;
    char *query = NULL;

    assert(id >= 0);
    assert(db != NULL);
    assert(bomdb != NULL);

    /* attach bom.db */
    xasprintf(&query, BOM_CONTENTS_ATTACH, bomdb);

    if (sqlite3_exec(db, query, NULL, 0, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_free(err);
        return -1;
    }

    free(query);
    query = NULL;

    /* drop the existing entries from the contents table */
    xasprintf(&query, DB_CONTENTS_DELETE, id);

    if (sqlite3_exec(db, "BEGIN TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_free(err);
        return -1;
    }

    if (sqlite3_exec(db, query, NULL, 0, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_free(err);
        return -1;
    }

    if (sqlite3_exec(db, "END TRANSACTION", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_free(err);
        return -1;
    }

    free(query);

    /* detach bom.db */
    if (sqlite3_exec(db, BOM_CONTENTS_DETACH, NULL, 0, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_free(err);
        return -1;
    }

    return 0;
}

/*
 * Adds entry to installed.db.  oldpackage is NULL unless we are performing
 * and upgrade, in which case is holds the package info we are upgrading.
 * Returns < 0 on error, otherwise returns the package id recorded.
 */
static int mkdistentry(const struct dist *package, const struct dist *oldpackage, bool upgrade)
{
    int id = 0;
    int offset = 0;
    char *schema = NULL;
    struct stat sb;
    char *dbsubpath = NULL;
    char *payload = NULL;
    char *err = NULL;
    sqlite3 *db = NULL;

    /* open database */
    if (sqlite3_open(db_installed, &db) != SQLITE_OK) {
        warnx("sqlite3_open: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        return -1;
    }

    if (sqlite3_exec(db, "PRAGMA synchronous = OFF", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_free(err);
        return -1;
    }

    if (sqlite3_exec(db, "PRAGMA journal_mode = MEMORY", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_free(err);
        return -1;
    }

    /*
     * Record package information in the database.  New packages return
     * the ID of the row inserted.  Upgraded packages return the same ID.
     */
    if (upgrade) {
        offset = 0;
        xasprintf(&schema, DB_PACKAGES_UPDATE, oldpackage->id);
    } else {
        offset = 1;
        schema = strdup(DB_PACKAGES_INSERT);
    }

    id = insert_pkg_info(package, db, schema, offset, upgrade);

    if ((id < 0)) {
        warnx("insert_pkg_info");
        return -1;
    }

    free(schema);

    /* for upgrade, remove all old contents */
    if (upgrade) {
        if (remove_obsolete(id, db, package->bom)) {
            warn("remove_obsolete");
            return -1;
        }
    }

    /* insert rows in the contents table */
    if (!(opts & OPTS_GENATTRS)) {
        if (insert_contents(id, package->bom, db)) {
            warn("insert_contents");
            return -1;
        }
    }

    /* we are finished with the databases here */
    sqlite3_close(db);

    /* meta data goes in a subdirectory matching the digest of the package file */
    dbsubpath = joinpath(dbpath, package->pkgdigest, NULL);

    if (mkdirp(dbsubpath, _DIR_MODE)) {
        warn("mkdirp");
        return -1;
    }

    if (chmod(dbsubpath, _DIR_MODE)) {
        warn("chmod");
        return -1;
    }

    /* optional */
    if (!package->debug) {
        if (copymeta(DIST_PREINSTALL, dbsubpath, upgrade)) {
            verbose_error("%s (%d): Error copying meta data for %s: %s\n", __func__, __LINE__, package->name, strerror(errno));
        }

        if (copymeta(DIST_POSTINSTALL, dbsubpath, upgrade)) {
            verbose_error("%s (%d): Error copying meta data for %s: %s\n", __func__, __LINE__, package->name, strerror(errno));
        }

        if (copymeta(DIST_PREREMOVE, dbsubpath, upgrade)) {
            verbose_error("%s (%d): Error copying meta data for %s: %s\n", __func__, __LINE__, package->name, strerror(errno));
        }

        if (copymeta(DIST_POSTREMOVE, dbsubpath, upgrade)) {
            verbose_error("%s (%d): Error copying meta data for %s: %s\n", __func__, __LINE__, package->name, strerror(errno));
        }
    }

    /* unpack the etc payload if we have it */
    payload = joinpath(tmppath, DIST_ETC, NULL);

    if (lstat(payload, &sb) == 0 && unpack_archive(payload, dbsubpath, (opts & OPTS_FORCE))) {
        warnx("Unable to unpack %s to %s", payload, dbsubpath);
        return -1;
    }

    free(dbsubpath);
    free(payload);
    return id;
}

/* Replaces the dist files with the actual state of the system */
static int gendistentry(const struct dist *package)
{
    int success = 0;
    finfolist_t *merged = NULL;
    finfolist_entry_t *entry = NULL;
    sqlite3 *db = NULL;
    char *err = NULL;

    assert(package != NULL);

    /* add everything to the table */
    if (newmanifest != NULL) {
        TAILQ_FOREACH(entry, newmanifest, items) {
            merged = sorted_finfolist_insert(newmanifest, entry->info);
        }
    }

    /* take things that weren't installed back out */
    if (newerrfiles != NULL) {
        TAILQ_FOREACH(entry, newerrfiles, items) {
            merged = remove_finfolist_filename(merged, entry->info->filename);
        }
    }

    /* add in things that weren't removed */
    if (errfiles != NULL) {
        TAILQ_FOREACH(entry, errfiles, items) {
            merged = sorted_finfolist_insert(merged, entry->info);
        }
    }

    /* open the database for adding this entry */
    if (sqlite3_open(db_installed, &db) != SQLITE_OK) {
        warnx("sqlite3_open: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        free_finfolist(merged);
        return -1;
    }

    if (sqlite3_exec(db, "PRAGMA synchronous = OFF", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_free(err);
        free_finfolist(merged);
        return -1;
    }

    if (sqlite3_exec(db, "PRAGMA journal_mode = MEMORY", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_free(err);
        free_finfolist(merged);
        return -1;
    }

    /* make sure checksums are correct */
    if (merged != NULL) {
        TAILQ_FOREACH(entry, merged, items) {
            success = gen_contents(entry->info, package->sumtype, prepend);
        }
    }

    /* record this package */
    success = write_contents(package->id, package->sigver, db, merged);
    sqlite3_close(db);
    return success;
}

/* Install initial set of /etc configuration files. */
static int initialetc(const struct dist *package)
{
    int status = 0, ret = 0;
    char *cmd = NULL;
    char *etcmerge = NULL;
    char **etcmerge_argv = NULL;
    pid_t pid = 0;

    assert(package != NULL);

    /* if there is an ETCMERGE environment variable, use it */
    if (getenv("ETCMERGE") != NULL) {
        etcmerge = strdup(getenv("ETCMERGE"));
    } else {
        etcmerge = strdup(ETCMERGE);
    }

    /* build the command with an odd argument delimiter to split later */
    if (prepend == NULL) {
        xasprintf(&cmd, "%s|-i|%s-%s-%d.%s", etcmerge, package->name, package->ver, package->rel, package->arch);
    } else {
        xasprintf(&cmd, "%s|-i|-r|%s|%s-%s-%d.%s", etcmerge, prepend, package->name, package->ver, package->rel, package->arch);
    }

    etcmerge_argv = strsplit(cmd, "|");
    free(cmd);

    if ((pid = fork()) == 0) {
        if (execv(etcmerge, etcmerge_argv) == -1) {
            err(EXIT_FAILURE, "execv");
            exit(EXIT_FAILURE);
        }
    } else if (pid == -1) {
        warn("fork");
        strvfree(etcmerge_argv);
        return 0;
    } else {
        if (waitpid(pid, &status, 0) == -1) {
            warn("waitpid");
            strvfree(etcmerge_argv);
            return -1;
        }

        if (!WIFEXITED(status) || WEXITSTATUS(status) != EXIT_SUCCESS) {
            ret = -1;
        } else {
            ret = 0;
        }
    }

    free(etcmerge);
    strvfree(etcmerge_argv);
    return ret;
}

/* Return true if the second package specified is a downgrade from the first. */
static bool is_downgrade(const struct dist *oldpackage, const struct dist *package)
{
    assert(package != NULL);

    if (oldpackage == NULL) {
        return false;
    }

    if (opts & OPTS_MULTI) {
        return false;
    }

    if ((strverscmp(package->ver, oldpackage->ver) < 0) || (!strcmp(package->ver, oldpackage->ver) && (package->rel < oldpackage->rel))) {
        return true;
    }

    return false;
}

/* Format the upgrade/downgrade/installing messages. */
static char *format_msg(const struct dist *oldpackage, const struct dist *package, bool upgrade)
{
    bool downgrade = is_downgrade(oldpackage, package);

    if (opts & OPTS_FORCE) {
        if (upgrade) {
            return "Force upgrad";
        } else if (downgrade) {
            return "Force downgrad";
        } else {
            return "Force install";
        }
    } else {
        if (upgrade) {
            return "Upgrad";
        } else if (downgrade) {
            return "Downgrad";
        } else {
            return "Install";
        }
    }
}

/*
 * Find the package that provides the specified dependency.  Returns
 * NULL if not package provides it or if the system provides it.  If
 * the dependency cannot be found, the program displays an error
 * indicating a required dependency cannot be found and then exits.
 */
static pkglist_entry_t *find_dep_provider(pkglist_t *pkgs, const pkglist_entry_t *pkgentry, const char *dep)
{
    pkglist_entry_t *provider = NULL;
    pkglist_entry_t *pkgsentry = NULL;
    strlist_entry_t *sentry = NULL;
    sqlite3 *db = NULL;
    sqlite3_stmt *stmt = NULL;
    char *query = NULL;
    char *path = NULL;
    size_t deplen = 0;
    size_t pathlen = 0;
    bool sys = false;

    assert(pkgs != NULL);
    assert(pkgentry != NULL);
    assert(dep != NULL);

    /* check the package set given to the program */
    TAILQ_FOREACH(pkgsentry, pkgs, items) {
        if (pkgsentry->pkg->provs != NULL && !TAILQ_EMPTY(pkgsentry->pkg->provs)) {
            TAILQ_FOREACH(sentry, pkgsentry->pkg->provs, items) {
                if (!strcmp(sentry->val, dep)) {
                    /* found the provider */
                    provider = pkgsentry;
                    break;
                }
            }
        }

        if (provider != NULL) {
            if (provider == pkgentry) {
                /* the provider is the calling package */
                return NULL;
            } else {
                return provider;
            }
        }
    }

    /* not found in the package set, check the installed system db */
    db = open_db(db_installed);

    if (db == NULL) {
        return NULL;
    }

    xasprintf(&query, SYS_DEP_CHECK, dep);

    if (sqlite3_prepare(db, query, -1, &stmt, 0) != SQLITE_OK) {
        warnx("sqlite3_prepare: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_close(db);
        return NULL;
    }

    deplen = strlen(dep);

    while (sqlite3_step(stmt) == SQLITE_ROW) {
        path = strdup((char *) sqlite3_column_text(stmt, 1));

        /* verify the dep is at the end of the path string */
        pathlen = strlen(path);

        if ((pathlen >= deplen) && !strcmp(path + pathlen - deplen, dep)) {
            sys = true;
        }

        /* XXX: verify path exists on the system or do we trust our database? */

        free(path);
    }

    if (sqlite3_finalize(stmt) != SQLITE_OK) {
        warnx("sqlite3_finalize: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_close(db);
        free(query);
        return NULL;
    }

    sqlite3_close(db);
    free(query);

    if (sys) {
        /* the dep is provided by the installed system */
        return NULL;
    }

    /* not found on the system either, we are missing a dependency */
    if (provider == NULL) {
        fprintf(stderr, "*** Cannot find `%s' dependency for:\n", dep);
        fprintf(stderr, "***     %s-%s-%d.%s\n", pkgentry->pkg->name, pkgentry->pkg->ver, pkgentry->pkg->rel, pkgentry->pkg->arch);
        fprintf(stderr, "*** Neither the package set nor installed system provide the dependency.\n");
        exit(EXIT_FAILURE);
    }

    /* not reached */
    return NULL;
}

/*
 * Sorting helper for depsort_pkgs().  This is like depth first
 * searching across the deps list.  Or graph.
 */
static void visit_dep(pkglist_t *sorted, deplist_t *deps, deplist_entry_t *entry)
{
    deplist_entry_t *other = NULL;
    pkglist_entry_t *pkgsentry = NULL;

    assert(deps != NULL);
    assert(entry != NULL);

    if (entry->permanent) {
        return;
    }

    if (entry->temporary) {
        fprintf(stderr, "*** Unable to sort package set for installation.\n");
        fprintf(stderr, "***     %s-%s-%d.%s is the problem package\n", entry->pkg->name, entry->pkg->ver, entry->pkg->rel, entry->pkg->arch);
        exit(EXIT_FAILURE);
    }

    entry->temporary = true;

    TAILQ_FOREACH(other, deps, items) {
        if (other->dep == entry->pkg) {
            visit_dep(sorted, deps, other);
        }
    }

    entry->temporary = false;
    entry->permanent = true;

    if (!have_pkg(sorted, entry->dep)) {
        pkgsentry = calloc(1, sizeof(*pkgsentry));
        assert(pkgsentry != NULL);
        pkgsentry->pkg = entry->dep;
        TAILQ_INSERT_TAIL(sorted, pkgsentry, items);
    }

    return;
}

/*
 * Given a pkglist_t, order the packages based on deps and provs
 * NOTE: this destroys the pkglist_t passed in
 */
static pkglist_t *depsort_pkgs(pkglist_t *pkgs)
{
    pkglist_entry_t *pkgsentry = NULL;
    pkglist_entry_t *provider = NULL;
    pkglist_entry_t *leaf = NULL;
    strlist_entry_t *sentry = NULL;
    deplist_t *deps = NULL;
    deplist_entry_t *depsentry = NULL;
    pkglist_t *sorted = NULL;
    bool sort = true;

    if (pkgs == NULL) {
        return pkgs;
    }

    /* make a DAG of deps */
    /* this list just holds pointers to other data */
    deps = calloc(1, sizeof(*deps));
    assert(deps != NULL);
    TAILQ_INIT(deps);

    TAILQ_FOREACH(pkgsentry, pkgs, items) {
        /*
         * If this package has dependencies, figure out where each one comes
         * from right now.  Either it comes from a package in the install set
         * we have now or it comes from the system or it doesn't exist.  In
         * the case of a dep provided by another package in this set, that
         * goes in the DAG.  If the dependency comes from the system, that
         * returns NULL and we don't care.  If the dependency cannot be found
         * then we need to fail now and tell the user we are missing a dep.
         */
        if (pkgsentry->pkg->deps != NULL && !TAILQ_EMPTY(pkgsentry->pkg->deps)) {
            TAILQ_FOREACH(sentry, pkgsentry->pkg->deps, items) {
                provider = find_dep_provider(pkgs, pkgsentry, sentry->val);

                if (provider != NULL) {
                    depsentry = calloc(1, sizeof(*depsentry));
                    assert(depsentry != NULL);
                    depsentry->pkg = pkgsentry->pkg;
                    depsentry->dep = provider->pkg;
                    TAILQ_INSERT_TAIL(deps, depsentry, items);
                }
            }
        }
    }

    /* Depsort the packages in the install set with inter-dependencies */
    sorted = calloc(1, sizeof(*sorted));
    assert(sorted != NULL);
    TAILQ_INIT(sorted);

    sort = true;

    while (sort) {
        TAILQ_FOREACH(depsentry, deps, items) {
            if (depsentry->permanent == false) {
                sort = true;
                visit_dep(sorted, deps, depsentry);
            } else {
                sort = false;
            }
        }
    }

    /*
     * Now add in all of the packages in the install set that are not
     * inter-dependent
     */
    TAILQ_FOREACH(pkgsentry, pkgs, items) {
        if (!have_pkg(sorted, pkgsentry->pkg)) {
            leaf = calloc(1, sizeof(*leaf));
            assert(leaf != NULL);
            leaf->pkg = pkgsentry->pkg;
            TAILQ_INSERT_TAIL(sorted, leaf, items);
        }
    }

    /*
     * Clear out the unsorted list and reset it to the sorted one
     */
    while (!TAILQ_EMPTY(pkgs)) {
        pkgsentry = TAILQ_FIRST(pkgs);
        TAILQ_REMOVE(pkgs, pkgsentry, items);
        free(pkgsentry);
    }

    free(pkgs);
    pkgs = sorted;

    /* cleanup */
    if (deps != NULL) {
        while (!TAILQ_EMPTY(deps)) {
            depsentry = TAILQ_FIRST(deps);
            TAILQ_REMOVE(deps, depsentry, items);
            free(depsentry);
        }

        free(deps);
    }

    /* return sorted package list */
    return sorted;
}

/* MAIN */
int main(int argc, char **argv)
{
    int i = 0;
    int status = 0;
    int long_idx = 0;
    size_t j = 0;
    char *cmd = NULL;
    char *pkgspec = NULL;
    char *payload = NULL;
    char *relocpath = NULL;
    char *oldpath = NULL;
    char *targetarch = getdefaultarch();
    pkglist_t *pkgs = NULL;
    pkglist_entry_t *pkgsentry = NULL;
    finfolist_entry_t *finfoentry = NULL;
    struct dist *package = NULL;
    struct dist *oldpackage = NULL;
    FILE *fp = NULL;
    bool exiterr = false;
    bool locerr = false;            /* print "installed with errors"? */
    char *oldver = NULL;            /* -v */
    char cwd[PATH_MAX + 1];
    char *tmpdir = NULL;
    pkglist_t *pkgmatches = NULL;
    struct stat sb;
    glob_t found;
    bool upgrade = false;

    setlinebuf(stdout);

    memset(cwd, '\0', sizeof(cwd));

    if (getcwd(cwd, PATH_MAX) == NULL) {
        err(EXIT_FAILURE, "getcwd");
    }

    prog = basename(argv[0]);

    /* parse options */
    while ((i = getopt_long(argc, argv, "adfFgmqsSuyr:v:P:t:eVh?", long_options, &long_idx)) != -1) {
        switch (i) {
            case 'a':
                opts |= OPTS_NOARCHCHECK;
                break;
            case 'd':
                opts |= OPTS_NODEPCHECK;
                break;
            case 'f':
                opts |= OPTS_FORCE;
                break;
            case 'F':
                opts |= OPTS_FORCESCRIPT;
                break;
            case 'g':
                opts |= OPTS_GENATTRS;
                break;
            case 'm':
                opts |= OPTS_MULTI;
                break;
            case 'q':
                opts |= OPTS_QUICK;
                break;
            case 's':
                opts |= OPTS_STRICT;
                break;
            case 'S':
                opts |= OPTS_NOSCRIPTS;
                break;
            case 'r':
                if (*optarg != '/') {
                    printf("Root directory must be an absolute pathname\n");
                    return EXIT_FAILURE;
                }

                prepend = strdup(optarg);
                break;
            case 'v':
                oldver = strdup(optarg);
                break;
            case 'P':
                opts |= OPTS_PKGSPEC;
                pkgspec = strdup(optarg);
                break;
            case 'u':
                opts |= OPTS_UPGRADE;
                break;
            case 'y':
                opts |= OPTS_DRYRUN;
                break;
            case 'e':
                opts |= OPTS_INSTALLETC;
                break;
            case 'V':
                ver(prog);
                return EXIT_SUCCESS;
            case 'h':
            case '?':
                usage();
                return EXIT_SUCCESS;
            case 0:
                /* handle long options with no short option equivalent */
                if (!strcmp(long_options[long_idx].name, "verbose")) {
                    opts |= OPTS_VERBOSE;
                } else if (!strcmp(long_options[long_idx].name, "tmpdir")) {
                    /* if this directory already exists, do not remove it */
                    assert(optarg != NULL);

                    if (stat(optarg, &sb) == -1) {
                        if (errno == ENOENT) {
                            warnx("%s does not exist", optarg);
                        } else {
                            warn("stat");
                        }

                        return EXIT_FAILURE;
                    }

                    if (S_ISDIR(sb.st_mode)) {
                        opts |= OPTS_SET_TMPDIR;
                    }

                    tmpdir = optarg;
                }

                break;
            default:
                usage();
                return EXIT_FAILURE;
        }
    }

    /* exit and tell user if no packages specified */
    if (optind == argc) {
        errx(EXIT_FAILURE, "You must specify at least one package.");
    }

    /* -u -m requires -P */
    if ((opts & OPTS_UPGRADE) && (opts & OPTS_MULTI) && !((opts & OPTS_PKGSPEC) || (oldver != NULL))) {
        errx(EXIT_FAILURE, "The -P or -v option is required when specifying -u and -m.");
    }

    /* -P and -v are mutually exclusive */
    if ((opts & OPTS_PKGSPEC) && (oldver != NULL)) {
        errx(EXIT_FAILURE, "The -P and -v options cannot be used together.");
    }

    /* -P option requires a single package */
    if ((opts & OPTS_PKGSPEC) && ((optind + 1) != argc)) {
        errx(EXIT_FAILURE, "The -P option only works with a single package.");
    }

    /* use / if the user did not specify a root path */
    if (prepend == NULL) {
        prepend = strdup("/");
    }

    /* initialize the installed package database if we need to */
    if (initdb(prepend, DB_PATH, DB_FILE)) {
        err(EXIT_FAILURE, "initdb");
    }

    /* handle an alternative root path for installation */
    dbpath = get_path(prepend, DB_PATH);
    db_installed = get_path(prepend, DB_FILE);

    /* find a suitable tmp dir if one was not specified */
    if (tmppath == NULL) {
        xasprintf(&tmppath, "%s/%s.%s.XXXXXX", (tmpdir == NULL) ? _PATH_VARTMP : tmpdir, prog, targetarch);

        if ((tmppath = mkdtemp(tmppath)) == NULL) {
            err(EXIT_FAILURE, "Unable to find a suitable temporary directory, please use the --tmpdir option");
        }
    }

    /* initialize the list that will hold all packages to install */
    pkgs = init_pkglist();
    assert(pkgs != NULL);

    /* initialize glob struct at least once before the main loop */
    memset(&found, 0, sizeof(found));

    /* gather all of the package deps and provs */
    do {
        /* initialize the glob structure */
        globfree(&found);
        memset(&found, 0, sizeof(found));

        /* handle any potential wildcards */
        if (glob(argv[optind], GLOB_BRACE|GLOB_TILDE, NULL, &found) != 0) {
            warn("*** ERROR: failed expanding %s", argv[optind]);
            continue;
        }

        /* loop over expanded items */
        j = 0;

        while (j < found.gl_pathc) {
            /* verify the package exists */
            if (access(found.gl_pathv[j], F_OK) != 0) {
                warnx("package file '%s' does not exist.", found.gl_pathv[j]);
                j++;
                continue;
            }

            /* verify we have a valid dist package */
            if (!checksig(found.gl_pathv[j])) {
                warnx("%s is not a valid package.", found.gl_pathv[j]);
                j++;
                continue;
            }

            /* initialize some things */
            pkgsentry = calloc(1, sizeof(*pkgsentry));
            assert(pkgsentry != NULL);
            pkgsentry->pkg = initpkg();

            /* collect package information */
            if (getpkginfo(pkgsentry->pkg, PKG_DISTFILE, found.gl_pathv[j], tmppath, NULL)) {
                warnx("Could not extract package information from %s", found.gl_pathv[j]);
                exiterr = true;
                j++;
                freepkg(pkgsentry->pkg);
                free(pkgsentry);
                continue;
            }

            /* add the package to the list */
            TAILQ_INSERT_TAIL(pkgs, pkgsentry, items);

            j++;
        }

        globfree(&found);
    } while (optind++, optind < argc);

    /* order the packages to install based on dependencies */
    if ((length_pkglist(pkgs) > 1) && !(opts & OPTS_NODEPCHECK)) {
        pkgs = depsort_pkgs(pkgs);
    }

    /* clear out the temporary dir */
    rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));

    /* make the tmp path if it doesn't exist */
    if (mkdirp(tmppath, _DIR_MODE)) {
        err(EXIT_FAILURE, "mkdirp");
    }

    /* iterate over the sorted install set and install packages */
    TAILQ_FOREACH(pkgsentry, pkgs, items) {
        /* reset the status of the -u switch for each package */
        upgrade = opts & OPTS_UPGRADE;

        /* initialize some things for this loop iteration */
        free_pkglist(pkgmatches);
        free(oldpath);
        free(relocpath);
        free(payload);
        pkgmatches = NULL;
        oldpath = relocpath = payload = NULL;

        /* initialize these lists */
        free_finfolist(errfiles);
        errfiles = init_finfolist();

        free_finfolist(newerrfiles);
        newerrfiles = init_finfolist();

        free_finfolist(newmanifest);
        newmanifest = init_finfolist();

        /* convenience name */
        package = pkgsentry->pkg;

        /* check for existing packages on non-force and non-upgrade */
        if (!(opts & OPTS_FORCE) && !upgrade && !(opts & OPTS_MULTI)) {
            if (pkgspec == NULL) {
                if (oldver != NULL) {
                    xasprintf(&pkgspec, "%s-%s%s.%s", package->name, oldver, (package->debug ? ".debug" : ""), package->arch);
                } else {
                    xasprintf(&pkgspec, "%s%s.%s", package->name, (package->debug ? ".debug" : ""), package->arch);
                }
            }

            pkgmatches = scanpkg(db_installed, pkgspec);
            free(pkgspec);
            pkgspec = NULL;

            if (pkgmatches != NULL) {
                warnx("%s already installed, must upgrade (-u) or multi (-m) install", package->name);
                exiterr = true;
                j++;
                continue;
            }
        }

        /* cannot use -m if the package specified matches the installed package */
        if ((opts & OPTS_MULTI) && !upgrade) {
            if (pkgspec == NULL) {
                if (oldver != NULL) {
                    xasprintf(&pkgspec, "%s-%s%s.%s", package->name, oldver, (package->debug ? ".debug" : ""), package->arch);
                } else {
                    xasprintf(&pkgspec, "%s%s.%s", package->name, (package->debug ? ".debug" : ""), package->arch);
                }
            }

            pkgmatches = scanpkg(db_installed, pkgspec);
            free(pkgspec);
            pkgspec = NULL;

            if (pkgmatches != NULL && length_pkglist(pkgmatches) == 1) {
                oldpackage = (TAILQ_FIRST(pkgmatches))->pkg;

                if (!strcmp(package->name, oldpackage->name) && !strcmp(package->ver, oldpackage->ver) &&
                    package->rel == oldpackage->rel && package->debug == oldpackage->debug && !strcmp(package->arch, oldpackage->arch)) {
                    warnx("%s-%s-%d%s.%s is already installed, cannot multi install (-m).", package->name, package->ver, package->rel, (package->debug ? ".debug" : ""), package->arch);
                    exiterr = true;
                    j++;
                    continue;
                }
            }
        } else {
            /* figure out the current state of installed packages set things
             * up for a new install, an upgrade, a multi-version install, or
             * just fall over
             */
            if (pkgspec == NULL) {
                if (oldver != NULL) {
                    xasprintf(&pkgspec, "%s-%s%s.%s", package->name, oldver, (package->debug ? ".debug" : ""), package->arch);
                } else {
                    xasprintf(&pkgspec, "%s%s.%s", package->name, (package->debug ? ".debug" : ""), package->arch);
                }
            }

            pkgmatches = scanpkg(db_installed, pkgspec);
            i = length_pkglist(pkgmatches);
            free(pkgspec);
            pkgspec = NULL;

            if (i == 0) {
                /*
                 * user asked to upgrade, but no package by this name is
                 * installed, so just treat as a regular installation
                 */
                upgrade = false;
                oldpackage = NULL;
            } else if (i == 1 && !upgrade && !(opts & OPTS_MULTI) && !(opts & OPTS_FORCE)) {
                /*
                 * user is trying to install a package that we think is
                 * already installed
                 */
                printf("%sackage %s already installed, skipping\n", package->debug ? "Debug p" : "P", package->name);
                j++;
                continue;
            } else if ((i == 1 && upgrade) || (i == 1 && (opts & OPTS_FORCE) && !upgrade)) {
                /* user wants to upgrade and we have one version installed */
                oldpackage = (TAILQ_FIRST(pkgmatches))->pkg;

                /* do we already have it?  suggest force */
                if (!(opts & OPTS_FORCE) && (!strcmp(package->ver, oldpackage->ver) && (package->rel == oldpackage->rel))) {
                    printf("%sackage %s-%s-%d already installed, skipping.\n", package->debug ? "Debug p" : "P", package->name, package->ver, package->rel);
                    j++;
                    continue;
                }

                /* detect if this is a version downgrade */
                if (!(opts & OPTS_FORCE) && is_downgrade(oldpackage, package)) {
                    printf("%sackage %s-%s-%d is older than %s-%s-%d, skipping.\n", package->debug ? "Debug p" : "P", package->name,
                           package->ver, package->rel, oldpackage->name, oldpackage->ver, oldpackage->rel);
                    j++;
                    continue;
                }
            } else if (i > 1 && upgrade && !(opts & OPTS_MULTI)) {
                /* user wants to upgrade, but multiple versions are installed */
                printf("More than one version of %spackage %s already installed\n", package->debug ? "debug " : "", package->name);
                exiterr = true;
                j++;
                continue;
            } else if (i > 1 && upgrade && (opts & OPTS_MULTI)) {
                /* the user is trying to multi upgrade */
                printf("Unable to match %s to an installed package.\n", package->filename);
                printf("More package specification required for a multiversion upgrade.\n");
                exiterr = true;
                j++;
                continue;
            } else {
                /* not sure how we got here */
                printf("Unknown distadd state reached, congratulations!\n");
                exiterr = true;
                j++;
                continue;
            }
        }

        if (upgrade) {
            printf("%sing %spackage %s-%s-%d.%s to", format_msg(oldpackage, package, true), oldpackage->debug ? "debug " : "",
                   oldpackage->name, oldpackage->ver, oldpackage->rel, oldpackage->arch);
        } else {
            printf("%sing %spackage", format_msg(oldpackage, package, false), package->debug ? "debug " : "");
        }

        printf(" %s-%s-%d.%s\n", package->name, package->ver, package->rel, package->arch);

        /* for upgrades, run an integrity check on the installed package */
        if (upgrade && !(opts & OPTS_FORCE) && geteuid() == 0) {
            if (checkpkg(db_installed, oldpackage, prepend)) {
                printf("ERRORS with installed %spackage %s-%s-%d.%s, cannot upgrade.\n", oldpackage->debug ? "debug " : "",
                       oldpackage->name, oldpackage->ver, oldpackage->rel, oldpackage->arch);
                printf("Use distchk(8) to investigate and repair these errors.\n");
                exiterr = true;
                j++;
                continue;
            }
        }

        /* architecture check */
        if (!(opts & OPTS_NOARCHCHECK) && check_targetarch(package->arch)) {
            printf("Machine type (%s) does not match type of package %s %s(%s)\n", targetarch, package->name,
                   package->debug ? "(debug symbols) " : "", package->arch);
            printf("Skipping %s\n", package->name);
            exiterr = true;
            j++;
            continue;
        }

        /* dependency check */
        if (!package->debug && !(opts & OPTS_NODEPCHECK)) {
            if (!check_deps(package)) {
                printf("Dependency check for %s failed, skipping\n", package->name);
                exiterr = true;
                j++;
                continue;
            }
        }

        /* unpack the contents of the .dist to tmppath */
        if (unpack_archive(package->filename, tmppath, (opts & OPTS_FORCE))) {
            warnx("Unable to unpack %s to %s", package->filename, tmppath);
            exiterr = true;
            rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
            j++;
            continue;
        }

        /* run the preinstall script */
        if (!package->debug && !(opts & OPTS_NOSCRIPTS) && !(opts & OPTS_DRYRUN)) {
            if (runscript(package, NULL, SCRIPT_PREINSTALL, upgrade, prepend, tmppath)) {
                printf("%s failed for %s, ", DIST_PREINSTALL, package->name);

                if (opts & OPTS_FORCESCRIPT) {
                    printf("continuing anyway");
                } else {
                    printf("skipping package");
                    exiterr = true;
                    rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
                    j++;
                    continue;
                }

                printf("\n");
            }
        }

        /* unpack reloc.payload */
        relocpath = joinpath(tmppath, "/reloc", NULL);

        if (mkdirp(relocpath, _DIR_MODE)) {
            warn("mkdirp");
            exiterr = true;
            rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
            j++;
            continue;
        }

        xasprintf(&payload, "%s.payload", relocpath);

        if (unpack_archive(payload, relocpath, (opts & OPTS_FORCE))) {
            warnx("Unable to unpack %s to %s", payload, relocpath);
            exiterr = true;
            rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
            j++;
            continue;
        }

        /* instreloc prints out any appropriate error messages */
        i = instreloc(package, oldpackage, upgrade);

        if (i == -2) {          /* preremove failed, fall over */
            exiterr = true;
            rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
            j++;
            continue;
        } else if (i == -1) {   /* some files installed, replace metadata */
            printf("errors installing %s%s\n", package->name, package->debug ? " (debug symbols)" : "");
            locerr = true;
        }

        /* record the package */
        if (upgrade && !(opts & OPTS_DRYRUN)) {
            package->id = oldpackage->id;
        }

        if (!(opts & OPTS_DRYRUN)) {
            if ((package->id = mkdistentry(package, oldpackage, upgrade)) < 0) {
                printf("Unable to create database entry for %s%s\n", package->name, package->debug ? " (debug symbols" : "");
                locerr = 1;
            }
        }

        /* generate new attrs if requested */
        if ((opts & OPTS_GENATTRS) && !(opts & OPTS_DRYRUN)) {
            if (gendistentry(package)) {
                printf("Failed to re-generate dist info for %s%s\n", package->name, package->debug ? " (debug symbols" : "");
                locerr = 1;
            }
        }

        /* update the library cache if we provide anything */
        if ((!(opts & OPTS_NODEPCHECK)) && (package->provs && !TAILQ_EMPTY(package->provs))) {
            xasprintf(&cmd, LDCONFIG" -r %s", (prepend == NULL) ? "/" : prepend);

            if ((fp = popen(cmd, "r")) == NULL) {
                err(EXIT_FAILURE, "popen");
            }

            status = pclose(fp);

            if (!WIFEXITED(status) || WEXITSTATUS(status) != EXIT_SUCCESS) {
                errx(EXIT_FAILURE, "unable to execute %s", cmd);
            }

            free(cmd);
        }

        /* call etcmerge if requested */
        if (!package->debug) {
            if ((opts & OPTS_INSTALLETC) && !(opts & OPTS_DRYRUN) && initialetc(package)) {
                printf("Failed merging initial configuration files for %s\n", package->name);
                locerr = 1;
            }
        }

        /* run postinstall */
        if (!package->debug && !(opts & OPTS_NOSCRIPTS) && !(opts & OPTS_DRYRUN)) {
            if (runscript(package, NULL, SCRIPT_POSTINSTALL, upgrade, prepend, NULL)) {
                printf("%s failed for %s, ", DIST_POSTINSTALL, package->name);

                if (opts & OPTS_FORCESCRIPT) {
                    printf("continuing anyway\n");
                } else {
                    printf("skipping\n");
                    locerr = true;
                }
            }
        }

        /* clean up */
        free_finfolist(newhards);
        free_finfolist(newdirs);
        free_finfolist(newmanifest);
        free_finfolist(errfiles);
        free_finfolist(newerrfiles);
        newhards = newdirs = newmanifest = errfiles = newerrfiles = NULL;

        /* results */
        if (upgrade) {
            printf("%sed %spackage %s-%s-%d.%s to", format_msg(oldpackage, package, true), oldpackage->debug ? "debug " : "",
                   oldpackage->name, oldpackage->ver, oldpackage->rel, oldpackage->arch);
        } else {
            printf("%sed %spackage", format_msg(oldpackage, package, false), package->debug ? "debug " : "");
        }

        printf(" %s-%s-%d.%s ", package->name, package->ver, package->rel, package->arch);

        if (locerr) {
            printf("with errors\n");
            locerr = false;
            exiterr = true;
        } else {
            printf("successfully\n");
        }

        /* return to original directory */
        if (chdir(cwd) != 0) {
            warn("chdir");
            exiterr = true;
        }
    }

    /* remove all of the old files and directories */
    if (oldfiles != NULL) {
        TAILQ_FOREACH(finfoentry, oldfiles, items) {
            if (checkunlink(finfoentry->info, prepend, opts)) {
                exiterr = true;
            }
        }

        free_finfolist(oldfiles);
    }

    if (olddirs != NULL) {
        TAILQ_FOREACH_REVERSE(finfoentry, olddirs, finfolist_s, items) {
            if (checkunlink(finfoentry->info, prepend, opts)) {
                exiterr = true;
            }
        }

        free_finfolist(olddirs);
    }

    /* clean up the temporary directory */
    if (rmtree(tmppath, false, (opts & OPTS_SET_TMPDIR))) {
        warnx("*** error removing %s", tmppath);
        exiterr = true;
    }

    /* clean up */
    rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));

    free(relocpath);
    free(payload);
    free(prepend);
    free(targetarch);
    free(tmppath);
    free(dbpath);
    free(oldver);
    free_pkglist(pkgmatches);
    free_pkglist(pkgs);

    return (exiterr ? EXIT_FAILURE : EXIT_SUCCESS);
}
