/* Show information about installed packages or package files. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright David Shea <david@reallylongword.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <glob.h>
#include <paths.h>
#include <err.h>
#include "dist-tools.h"
#include "opts.h"
#include "db.h"

extern char *optarg;
extern int optind, opterr, optopt;

/* Local prototypes */
static void usage(void);
static void _querypkg(const char *);
static void _printpkg(struct dist *);
static void _print_bom_list(const char *, const char *);

/* Local global variables */
static char *prog = NULL;
static char *db_path = NULL;
static char *db_installed = NULL;
static char *tmppath = NULL;
static char *prepend = NULL;
static bool first = true;
static int success = EXIT_SUCCESS;

/* Command line options */
/* Use this global opts variable to hold boolean options. */
static uint64_t opts = 0;

/* Usage screen displayed when the user passes --help as an argument. */
static void usage(void)
{
    printf("Usage:\n");
    printf("  %s [info option]... [(packagespec... | archive...)]\n", prog);
    printf("  %s -q filespec\n", prog);
    puts("\n"
         "Information gathering options:\n"
         "  -v, --ver                  Display package version\n"
         "  -m, --maintainer           Display package maintainer\n"
         "  -d, --description          Display package description\n"
         "  -s, --size                 Display installed package size\n"
         "  -p, --pkgsize              Display size of package file\n"
         "  -a, --arch                 Display package architecture\n"
         "  -A, --all                  Display all package information\n"
         "  -f, --files                Display package file listing\n"
         "  -l, --license              Display license name (e.g., GPLv3+)\n"
         "  -R, --repo                 Display package repository name\n"
         "  -g, --isdebug              Indicate if debug package or not\n\n"
         "Query options:\n"
         "  -q, --query                Display package containing file\n\n"
         "Package file options:\n"
         "  --provs                    Display dependencies this package provides\n"
         "  --deps                     Display dependencies this package requires\n\n"
         "Miscellaneous options:\n"
         "  --tmpdir=PATH              Use PATH for temporary files\n"
         "  -S, --shell                Output shell variable syntax\n"
         "  -D, --nodebug              Skip debug packages\n"
         "  -r PATH, --root=PATH       Use PATH for the root (/) directory\n"
         "  -V, --version              Show distinfo version\n"
         "  -h, --help                 This help message\n");
    return;
}

/* Called by main() when program invoked in query mode (-q). */
static void _querypkg(const char *filename)
{
    char *rfilename = NULL;
    char *qfilename = NULL;
    char *cfgfile = NULL;
    char *err = NULL;
    int count = 0;
    sqlite3 *db = NULL;
    sqlite3_stmt *stmt = NULL;

    if (filename == NULL) {
        return;
    }

    /* normalize the path we've been asked to look at */
    if ((rfilename = realpath(filename, NULL)) == NULL) {
        warn("realpath: %s", filename);
        free(rfilename);
        return;
    }

    /* trim the leading prepend from rfilename if need be */
    if (prepend != NULL) {
        qfilename = rfilename + strlen(prepend);
    } else {
        qfilename = rfilename;
    }

    /* paths are stored in the database without a leading slash, so trim it */
    while ((strlen(qfilename) > 0) && *qfilename == '/') {
        qfilename++;
    }

    /* get a database handle */
    if (sqlite3_open(db_installed, &db) != SQLITE_OK) {
        warnx("sqlite3_open: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_close(db);
        free(rfilename);
        return;
    }

    if (sqlite3_exec(db, "PRAGMA synchronous = OFF", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        sqlite3_close(db);
        free(rfilename);
        return;
    }

    if (sqlite3_exec(db, "PRAGMA journal_mode = MEMORY", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        sqlite3_close(db);
        free(rfilename);
        return;
    }

    /* are we looking for a config file or not */
    if (!strncmp(qfilename, "etc/", 4)) {
        /* looking for a configuration file or directory */
        if (sqlite3_prepare(db, DB_INFO_ETC_QUERY, -1, &stmt, 0) != SQLITE_OK) {
            warnx("sqlite3_open: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            sqlite3_close(db);
            free(rfilename);
            return;
        }

        while (sqlite3_step(stmt) == SQLITE_ROW) {
            cfgfile = joinpath(db_path, (char *) sqlite3_column_text(stmt, 4), qfilename, NULL);

            if (!access(cfgfile, F_OK)) {
                printf("%s: %s-%s-%d.%s\n", rfilename, (char *) sqlite3_column_text(stmt, 0), (char *) sqlite3_column_text(stmt, 1),
                       sqlite3_column_int(stmt, 2), sqlite3_column_text(stmt, 3));
                count++;
            }

            free(cfgfile);
            cfgfile = NULL;
        }

        if (sqlite3_finalize(stmt) != SQLITE_OK) {
            warnx("sqlite3_finalize: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            sqlite3_close(db);
            free(rfilename);
            return;
        }
    } else {
        /* query the database for a non config file */
        if (sqlite3_prepare(db, DB_INFO_QUERY, -1, &stmt, 0) != SQLITE_OK) {
            warnx("sqlite3_prepare: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            sqlite3_close(db);
            free(rfilename);
            return;
        }

        if (sqlite3_bind_text(stmt, 1, qfilename, strlen(qfilename), SQLITE_TRANSIENT) != SQLITE_OK) {
            warnx("sqlite3_bind_text: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            sqlite3_close(db);
            free(rfilename);
            return;
        }

        while (sqlite3_step(stmt) == SQLITE_ROW) {
            printf("%s: %s-%s-%d", rfilename, (char *) sqlite3_column_text(stmt, 0), (char *) sqlite3_column_text(stmt, 1), sqlite3_column_int(stmt, 2));

            if (sqlite3_column_int(stmt, 4)) {
                printf(".debug");
            }

            printf(".%s\n", (char *) sqlite3_column_text(stmt, 3));
            count++;
        }

        if (sqlite3_finalize(stmt) != SQLITE_OK) {
            warnx("sqlite3_finalize: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            sqlite3_close(db);
            free(rfilename);
            return;
        }
    }

    sqlite3_close(db);

    if (count == 0) {
        printf("%s is not owned by any installed package.\n", rfilename);
    }

    free(rfilename);
    return;
}

/*
 * Helper function to print the contents of list tables in bom.db,
 * such as the provs or deps tables.
 */
static void _print_bom_list(const char *pkgfile, const char *list_query)
{
    char *bomfilename = NULL;
    sqlite3 *db = NULL;
    sqlite3_stmt *stmt = NULL;
    char *err = NULL;

    assert(pkgfile != NULL);
    assert(list_query != NULL);

    if (unpack_metadata2file(pkgfile, DIST_BOM, tmppath, true)) {
        return;
    }

    bomfilename = joinpath(tmppath, DIST_BOM, NULL);
    db = open_db(bomfilename);

    if (db == NULL) {
        return;
    }

    if (sqlite3_exec(db, "PRAGMA synchronous = OFF", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        sqlite3_close(db);
        return;
    }

    if (sqlite3_exec(db, "PRAGMA journal_mode = MEMORY", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        sqlite3_close(db);
        return;
    }

    if (sqlite3_prepare(db, list_query, -1, &stmt, 0) != SQLITE_OK) {
        warnx("sqlite3_prepare: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_close(db);
        return;
    }

    while (sqlite3_step(stmt) == SQLITE_ROW) {
        printf("%s\n", (char *) sqlite3_column_text(stmt, 0));
    }

    if (sqlite3_finalize(stmt) != SQLITE_OK) {
        warnx("sqlite3_finalize: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_close(db);
        return;
    }

    free(bomfilename);
    sqlite3_close(db);
    return;
}

/* Print information on all versions of a package. */
static void _printpkg(struct dist *package)
{
    sqlite3 *db = NULL;
    sqlite3_stmt *stmt = NULL;
    char *tmp = NULL;
    char *err = NULL;

    /* Skip debug packages if the user requested that */
    if (package->debug && (opts & OPTS_NO_DEBUG)) {
        return;
    }

    /* Skip a line after the first package's info is displayed. */
    if (opts != 0 && opts != OPTS_NO_DEBUG) {
        if (first) {
            first = false;
        } else {
            printf("\n");
        }
    }

    /* Enforce --provs and --deps requiring package files. */
    if (package->filename == NULL && ((opts & OPTS_PROVS) || (opts & OPTS_DEPS))) {
        rmtree(tmppath, true, false);
        errx(EXIT_FAILURE, "The --provs and --deps options are only valid on package files.");
    }

    if (opts == 0 || opts == OPTS_NO_DEBUG) {
        /* show single line output */
        if (package->debug && !(opts & OPTS_NO_DEBUG)) {
            xasprintf(&tmp, "%s-%s-%d.debug.%s", package->name, package->ver, package->rel, package->arch);
        } else {
            xasprintf(&tmp, "%s-%s-%d.%s", package->name, package->ver, package->rel, package->arch);
        }

        printf("%-15s %s\n", package->repo, tmp);
        free(tmp);
        tmp = NULL;
    } else if (opts & OPTS_SHELL) {
        /* output information in shell variable format */
        if (package->filename != NULL) {
            printf("FILENAME=\"%s\"\n", package->filename);
        }

        printf("PKG=\"%s\"\n", package->name);

        if (opts & OPTS_VER) {
            printf("VER=\"%s\"\n", package->ver);
            printf("REL=\"%d\"\n", package->rel);
        }

        if (opts & OPTS_LICENSE) {
            printf("LICENSE=\"%s\"\n", package->license);
        }

        if (package->repo != NULL) {
            printf("REPO=\"%s\"\n", package->repo);
        }

        if (opts & OPTS_ARCH) {
            printf("ARCH=\"%s\"\n", package->arch);
        }

        if (opts & OPTS_MAINTAINER) {
            printf("MAINTAINER=\"%s\"\n", package->maintainer);
        }

        if (opts & OPTS_DESC) {
            printf("DESC=\"%s\"\n", package->desc);
        }

        if ((opts & OPTS_PKGSIZE) && package->filename != NULL) {
            printf("PKGSIZE=\"%llu\"\n",
                   (unsigned long long int) package->pkgsize);
        }

        if (opts & OPTS_SIZE) {
            printf("SIZE=\"%lu\"\n", package->size);
        }

        printf("DEBUG=\"%s\"\n", (package->debug) ? "y" : "n");
    } else {
        /* normal invocation */
        if (!(opts & OPTS_PROVS) && !(opts & OPTS_DEPS)) {
            if (package->filename != NULL) {
                printf("Package:        %s\n", package->filename);
            }

            printf("Name:           %s\n", package->name);
        }

        if (opts & OPTS_VER) {
            printf("Version:        %s\n", package->ver);
            printf("Release:        %d\n", package->rel);
        }

        if (opts & OPTS_LICENSE) {
            printf("License:        %s\n", package->license);
        }

        if ((opts & OPTS_REPO) && package->repo != NULL) {
            printf("Repository:     %s\n", package->repo);
        }

        if (opts & OPTS_ARCH) {
            printf("Architecture:   %s\n", package->arch);
        }

        if (opts & OPTS_ISDEBUG) {
            printf("Debug package:  %s\n", (package->debug) ? "yes" : "no");
        }

        if (opts & OPTS_MAINTAINER) {
            printf("Maintainer:     %s\n", package->maintainer);
        }

        if (opts & OPTS_DESC) {
            printf("\n%s\n", package->desc);
        }

        if ((opts & OPTS_PKGSIZE) && package->filename != NULL) {
            if (opts & OPTS_DESC) {
                printf("\n");
            }

            printf("Package size:   %llu bytes\n",
                   (unsigned long long int) package->pkgsize);
        }

        if (opts & OPTS_SIZE) {
            printf("Installed size: %lu bytes\n", package->size);
        }

        if (opts & OPTS_FILES) {
            if (package->filename == NULL) {
                /* package is installed, gather file list */
                if (sqlite3_open(db_installed, &db) != SQLITE_OK) {
                    warnx("sqlite3_open: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
                    sqlite3_close(db);
                    return;
                }

                if (sqlite3_exec(db, "PRAGMA synchronous = OFF", NULL, NULL, &err) != SQLITE_OK) {
                    warnx("sqlite3_exec: %s", err);
                    sqlite3_free(err);
                    sqlite3_close(db);
                    return;
                }

                if (sqlite3_exec(db, "PRAGMA journal_mode = MEMORY", NULL, NULL, &err) != SQLITE_OK) {
                    warnx("sqlite3_exec: %s", err);
                    sqlite3_free(err);
                    sqlite3_close(db);
                    return;
                }

                if (sqlite3_prepare(db, DB_INFO_FILELIST, -1, &stmt, 0) != SQLITE_OK) {
                    warnx("sqlite3_prepare: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
                    sqlite3_close(db);
                    return;
                }

                if (sqlite3_bind_int(stmt, 1, package->id) != SQLITE_OK) {
                    warnx("sqlite3_bind_int: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
                    sqlite3_close(db);
                    return;
                }

                while (sqlite3_step(stmt) == SQLITE_ROW) {
                    printf("/%s\n", (char *) sqlite3_column_text(stmt, 0));
                }

                if (sqlite3_finalize(stmt) != SQLITE_OK) {
                    warnx("sqlite3_finalize: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
                    sqlite3_close(db);
                    return;
                }

                sqlite3_close(db);
            } else {
                /* package is not installed, unpack and read bom for file list */
                _print_bom_list(package->filename, BOM_INFO_FILELIST);
            }
        }

        /* package is not installed, unpack and read bom for provs */
        if ((package->filename != NULL) && (opts & OPTS_PROVS)) {
            _print_bom_list(package->filename, BOM_INFO_PROVS);
        }

        /* package is not installed, unpack and read bom for deps */
        if ((package->filename != NULL) && (opts & OPTS_DEPS)) {
            _print_bom_list(package->filename, BOM_INFO_DEPS);
        }
    }

    return;
}

/* MAIN */
int main(int argc, char **argv)
{
    int i = 0;
    size_t j = 0;
    int long_idx = 0;
    char *tmp = NULL, *rtmp = NULL;
    struct stat sb;
    struct dist *package = NULL;
    pkglist_t *pkglist = NULL;
    pkglist_entry_t *entry = NULL;
    glob_t found;
    struct option long_options[] = {
        { "ver", no_argument, NULL, 'v' },
        { "maintainer", no_argument, NULL, 'm' },
        { "description", no_argument, NULL, 'd' },
        { "size", no_argument, NULL, 's' },
        { "pkgsize", no_argument, NULL, 'p' },
        { "arch", no_argument, NULL, 'a' },
        { "all", no_argument, NULL, 'A' },
        { "files", no_argument, NULL, 'f' },
        { "license", no_argument, NULL, 'l' },
        { "repo", no_argument, NULL, 'R' },
        { "isdebug", no_argument, NULL, 'g' },
        { "query", no_argument, NULL, 'q' },
        { "shell", no_argument, NULL, 'S' },
        { "nodebug", no_argument, NULL, 'D' },
        { "root", required_argument, NULL, 'r' },
        { "tmpdir", required_argument, NULL, 0 },
        { "version", no_argument, NULL, 'V' },
        { "help", no_argument, NULL, 'h' },
        { "provs", no_argument, NULL, 0 },
        { "deps", no_argument, NULL, 0 },
        { NULL, 0, NULL, 0 }
    };

    setlinebuf(stdout);
    prog = basename(argv[0]);

    /* parse options */
    while ((i = getopt_long(argc, argv, "vmdDspaAflRgqTSr:Vh?", long_options, &long_idx)) != EOF) {
        switch (i) {
            case 'v':
                opts |= OPTS_VER;
                break;
            case 'm':
                opts |= OPTS_MAINTAINER;
                break;
            case 'd':
                opts |= OPTS_DESC;
                break;
            case 'D':
                opts |= OPTS_NO_DEBUG;
                break;
            case 's':
                opts |= OPTS_SIZE;
                break;
            case 'p':
                opts |= OPTS_PKGSIZE;
                break;
            case 'a':
                opts |= OPTS_ARCH;
                break;
            case 'l':
                opts |= OPTS_LICENSE;
                break;
            case 'R':
                opts |= OPTS_REPO;
                break;
            case 'g':
                opts |= OPTS_ISDEBUG;
                break;
            case 'A':
                opts |= OPTS_ARCH | OPTS_VER | OPTS_MAINTAINER | OPTS_DESC | OPTS_SIZE | OPTS_PKGSIZE | OPTS_LICENSE;
                break;
            case 'S':
                opts |= OPTS_SHELL;
                break;
            case 'f':
                opts |= OPTS_FILES;
                break;
            case 'q':
                /* we're querying for the package that owns the specified file */
                opts = OPTS_QUERY;
                break;
            case 'r':
                if (*optarg != '/') {
                    printf("Root directory must be an absolute pathname\n");
                    return EXIT_FAILURE;
                }

                prepend = strdup(optarg);
                break;
            case 'V':
                ver(prog);
                return EXIT_SUCCESS;
            case 'h':
            case '?':
                usage();
                return EXIT_SUCCESS;
            case 0:
                if (!strcmp(long_options[long_idx].name, "tmpdir")) {
                    tmppath = strdup(optarg);
                } else if (!strcmp(long_options[long_idx].name, "provs")) {
                    opts |= OPTS_PROVS;
                } else if (!strcmp(long_options[long_idx].name, "deps")) {
                    opts |= OPTS_DEPS;
                }

                break;
            default:
                usage();
                return EXIT_FAILURE;
        }
    }

    /* use / if the user did not specify a root path */
    if (prepend == NULL) {
        prepend = strdup("/");
    }

    /* handle an alternative root path for installation */
    db_path = get_path(prepend, DB_PATH);
    db_installed = get_path(prepend, DB_FILE);

    /* create a temporary directory if the user did not specify one */
    if (tmppath == NULL) {
        xasprintf(&tmppath, "%s/%s.XXXXXX", _PATH_VARTMP, prog);

        if ((tmppath = mkdtemp(tmppath)) == NULL) {
            err(EXIT_FAILURE, "mkdtemp");
        }
    }

    /*
     * Iterate over everything passed to us.  The algorithm is to check
     * if the argument is a file that we can see.  If it is, perform the
     * distinfo operations on a package file.  If we don't see a file by
     * the argument name, assume the user is requesting info on an
     * installed package, and go about collecting information that way.
     *
     * If no arguments were given, iterate over every installed package
     * on the system.
     */

    /* no packages specified on the command line, iterate over everything */
    if (argc == optind || opts == OPTS_QUERY) {
        if (opts & OPTS_PROVS) {
            rmtree(tmppath, true, false);
            err(EXIT_FAILURE, "The --provs option is only valid for package files.");
        }

        if (opts & OPTS_DEPS) {
            rmtree(tmppath, true, false);
            err(EXIT_FAILURE, "The --deps option is only valid for package files.");
        }

        if (stat(db_installed, &sb) == -1 || !S_ISREG(sb.st_mode)) {
            rmtree(tmppath, true, false);
            err(EXIT_FAILURE, "This system lacks a Peachtree Linux package database.");
        }

        pkglist = scanpkg(db_installed, NULL);

        if (pkglist == NULL) {
            rmtree(tmppath, true, false);
            free_pkglist(pkglist);
            return EXIT_FAILURE;
        }

        /* sort the list alphabetically */
        pkglist = sort_pkglist(pkglist);

        /* either print information or run an ownership query */
        if (opts == OPTS_QUERY) {
            /* search for the package that owns each argument */
            for (i = optind; i < argc; i++) {
                tmp = joinpath(prepend, argv[i], NULL);
                rtmp = realpath(tmp, NULL);

                if (rtmp == NULL) {
                    warn("realpath: %s", tmp);
                    continue;
                }

                free(tmp);

                if (stat(rtmp, &sb) == -1) {
                    warn("stat");
                    free(rtmp);
                    continue;
                }

                _querypkg(rtmp);
                free(rtmp);
            }
        } else {
            /* normal information gathering mode */
            if (pkglist != NULL && !TAILQ_EMPTY(pkglist)) {
                TAILQ_FOREACH(entry, pkglist, items) {
                    _printpkg(entry->pkg);
                }
            }
        }

        /* free the list of packages */
        free_pkglist(pkglist);
        pkglist = NULL;
    } else {
        for (i = optind; i < argc; i++) {
            /* handle any potential wildcards */
            j = glob(argv[i], GLOB_BRACE|GLOB_TILDE, NULL, &found);

            if (j == 0) {
                /* matches found */
                while (j < found.gl_pathc) {
                    /* try to read a package dist file */
                    if ((stat(found.gl_pathv[j], &sb) == 0) && S_ISREG(sb.st_mode)) {
                        /* if not really a package, fall through and try the regular way */
                        if (checksig(found.gl_pathv[j])) {
                            package = initpkg();
                            assert(package != NULL);

                            if (getpkginfo(package, PKG_DISTFILE, found.gl_pathv[j], tmppath, NULL)) {
                                success = EXIT_FAILURE;
                                continue;
                            }

                            _printpkg(package);
                            freepkg(package);
                        } else {
                            warnx("Warning: file %s not a package", found.gl_pathv[j]);
                            success = EXIT_FAILURE;
                        }
                    }

                    j++;
                }
            } else if (j == GLOB_NOMATCH) {
                /* only check the package database if we have one */
                if ((stat(db_installed, &sb) == 0) && S_ISREG(sb.st_mode)) {
                    /* make a list of possible installed packages */
                    pkglist = scanpkg(db_installed, argv[i]);

                    if (pkglist == NULL) {
                        continue;
                    }

                    /* display information on the installed packages we found */
                    TAILQ_FOREACH(entry, pkglist, items) {
                        _printpkg(entry->pkg);
                    }

                    /* free the list of packages */
                    free_pkglist(pkglist);
                    pkglist = NULL;
                }
            } else {
                warn("glob");
                continue;
            }

            globfree(&found);
        }
    }

    rmtree(tmppath, true, false);
    free(db_path);
    free(db_installed);
    free(tmppath);
    return success;
}
