/* Make a new dist package. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright David Shea <david@reallylongword.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdlib.h>
#include <assert.h>
#include <stdbool.h>
#include <string.h>
#include <strings.h>
#include <ftw.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <paths.h>
#include <limits.h>
#include <err.h>
#include <archive.h>
#include <archive_entry.h>
#include "dist-tools.h"
#include "plist.h"
#include "db.h"
#include "opts.h"
#include "signature.h"

extern char *optarg;
extern int optind, opterr, optopt;

/* Global variables */
/*
 * prepend needs to be set to "/" for use by gendist.c
 * It's really only necessary for -g on distadd, but since distmk calls
 * the same functions, we need to define it in distmk too.
 */
char *prepend = "/";

/* Local prototypes */
static void usage(void);
static void set_archive_compression(unsigned int);
static int _add_payload_entry(const char *, const struct stat *, int, struct FTW *);
static int create_reloc(struct dist *);
static int create_etc(struct dist *);
static int create_dist(struct dist *);
static int mkstagedir(struct dist *);
static int genstrings(struct dist *, char *, char *);
static void cpmeta(struct dist *, char *, bool);
static int _tree_entry_size(const char *, const struct stat *, int, struct FTW *);
static int get_tree_size(const char *);
static int dometa(struct dist *);

/* Global variables */
static char *targetarch = NULL;
static uint64_t opts = 0;
static char *prog = NULL;
static char *tmppath = NULL;
static char *staging = NULL;
static int verbose = 0;
static struct archive *output = NULL;
static unsigned int treesize = 0;
static int payload = PAYLOAD_NULL;
static int trimlen = -1;
static int plist_checked = 0;
static plist_t *plist = NULL;

static struct option long_options[] = {
    { "tree", required_argument, NULL, 't' },
    { "meta", required_argument, NULL, 'm' },
    { "pkg", required_argument, NULL, 'p' },
    { "ver", required_argument, NULL, 'v' },
    { "rel", required_argument, NULL, 'r' },
    { "plist", required_argument, NULL, 'l' },
    { "debug", no_argument, NULL, 'd' },
    { "repo", required_argument, NULL, 'R' },
    { "arch", required_argument, NULL, 'a' },
    { "compression", required_argument, NULL, 'z' },
    { "checksum", required_argument, NULL, 's' },
    { "clean", no_argument, NULL, 'c' },
    { "version", no_argument, NULL, 'V' },
    { "help", no_argument, NULL, 'h' },
    { "verbose", no_argument, &verbose, -1 },
    { "tmpdir", required_argument, NULL, 0 },
    { "description", required_argument, NULL, 'D' },
    { "maintainer", required_argument, NULL, 'M' },
    { "license", required_argument, NULL, 'L' },
    { "depspath", required_argument, NULL, 'E' },
    { "outdir", required_argument, NULL, 'o' },
    { NULL, 0, NULL, 0 }
};

#define REASON_TREE "Missing package root tree (-t)"
#define REASON_PKGNAME "Missing package name (-p)"
#define REASON_PKGVERSION "Missing package version (-v)"
#define REASON_PKGRELEASE "Missing package release (-r)"
#define REASON_PLIST "Missing path to plist file (-l)"

/*
 * Usage screen.
 */
static void usage(void)
{
    printf("Usage: %s [OPTION]...\n", prog);
    puts("Make a Peachtree Linux dist package\n"
         "\nRequired:\n"
         "  -t PATH, --tree=PATH         Source path to package root tree\n"
         "  -p NAME, --pkg=NAME          Package name (e.g., dist-tools)\n"
         "  -v VER, --ver=VER            Package version (e.g., 2.0)\n"
         "  -r REL, --rel=REL            Package release (e.g., 1)\n"
         "  -l FILE, --plist=FILE        Path to the packing list file\n"
         "\nOptional:\n"
         "  -D TEXT, --description=TEXT  Short package description string\n"
         "  -M NAME, --maintainer=NAME   Name and email of maintainer\n"
         "  -L TAG, --license=TAG        Short license tag (e.g., GPL-2.0-or-later)\n"
         "  -E PATH, --depspath=PATH     Additional shared library directories\n"
         "  -d, --debug                  Create a debug package\n"
         "  -m PATH, --meta=PATH         Source path to package meta data\n"
         "  -R REPO, --repo=REPO         Package repository name\n"
         "  -a ARCH, --arch=ARCH         Target architecture\n"
         "  -o PATH, --outdir=PATH       Output directory for final package\n"
         "  -z TYPE, --compression=TYPE  Set compression type (see below)\n"
         "  -s TYPE, --checksum=TYPE     Set the checksum type (see below)\n"
         "  -c, --clean                  Clean up temporary packaging files\n"
         "  --verbose                    Display more output\n"
         "  --tmpdir=PATH                Use PATH for temporary files");
    printf("                               (default: %s)\n", _PATH_VARTMP);
    puts("  -V, --version                Display version information\n"
         "  -h, --help                   This help message\n");
    printf("Target architecture is '%s' by default, ", targetarch);
    printf("but can be explicitly\nset to 'noarch'.\n\n");
    puts("Compression type is Zstd by default, but you may specify any of\n"
         "the following types:\n"
         "    bzip2, bz2               bzip2(1) compression\n"
         "    gzip, gz                 gzip(1) compression\n"
         "    compress, Z              compress(1) compression\n"
         "    zstd                     Zstandard compression (default)\n"
         "    none                     Do not compress payload\n\n"
         "Checksum type is SHA-256 by default, but you may specify any of the\n"
         "following types:\n"
         "    md5                      MD5 128-bit, as described in RFC 1321\n"
         "    sha1                     SHA-1 160-bit, as described in FIPS-180-1\n"
         "    sha256                   SHA-256 256-bit, as described in FIPS-180-2\n"
         "    sha384                   SHA-384 384-bit, as described in FIPS-180-4\n"
         "    sha512                   SHA-512 512-bit, as described in FIPS-180-4\n\n"
         "Both compression type and checksum type are compared as case\n"
         "insensitive strings.");
    return;
}

/*
 * Set the compression type on the payload based on what the user
 * specified.  The default is ZSTD set in initpkg().
 */
static void set_archive_compression(unsigned int compression)
{
    switch (compression) {
        case ARCHIVE_FILTER_BZIP2:
#if ARCHIVE_VERSION_NUMBER < 3000000
            archive_write_set_compression_bzip2(output);
#else
            archive_write_add_filter_bzip2(output);
#endif
            break;
        case ARCHIVE_FILTER_COMPRESS:
#if ARCHIVE_VERSION_NUMBER < 3000000
            archive_write_set_compression_compress(output);
#else
            archive_write_add_filter_compress(output);
#endif
            break;
        case ARCHIVE_FILTER_GZIP:
#if ARCHIVE_VERSION_NUMBER < 3000000
            archive_write_set_compression_gzip(output);
#else
            archive_write_add_filter_gzip(output);
#endif
            break;
        case ARCHIVE_FILTER_ZSTD:
#if ARCHIVE_VERSION_NUMBER < 3000000
            archive_write_set_compression_zstd(output);
#else
            archive_write_add_filter_zstd(output);
#endif
            break;
        case ARCHIVE_FILTER_NONE:
#if ARCHIVE_VERSION_NUMBER < 3000000
            archive_write_set_compression_none(output);
#else
            archive_write_add_filter_none(output);
#endif
            break;
    }

    return;
}

/*
 * Adds an individual entry to a payload
 * (called by create_reloc() and create_etc())
 *
 * Callers need to set up 'output' and reset 'trimlen' to -1 before calling
 * this function.
 */
static int _add_payload_entry(const char *fpath, const struct stat *sb, __attribute__((unused)) int tflag, __attribute__((unused)) struct FTW *ftwbuf)
{
    int fd;
    ssize_t len;
    char buf[BUFSIZ];
    char *hardlink = NULL;
    bool is_hardlink = false;
    const char *vpath = NULL;
    struct archive_entry *entry = NULL;
    plist_entry_t *plistentry = NULL;
    plist_entry_t *iterator = NULL;

    if (trimlen == -1) {
        trimlen = strlen(fpath);
        return 0;
    } else {
        vpath = fpath + trimlen;

        while (*vpath == '/') {
            vpath++;
        }
    }

    /* Skip parts of the tree that do not match this payload type. */
    if ((strncmp(vpath, CONFDIR_PREFIX, strlen(CONFDIR_PREFIX)) && payload == PAYLOAD_ETC) ||
        (!strncmp(vpath, CONFDIR_PREFIX, strlen(CONFDIR_PREFIX)) && payload == PAYLOAD_RELOC)) {
        return 0;
    }

    entry = archive_entry_new();

    /* Is this file a hard link? */
    hardlink = lookup_inode(sb->st_ino);
    if (hardlink != NULL && strcmp(vpath, hardlink)) {
        is_hardlink = true;
    }

    if (S_ISREG(sb->st_mode)) {
        if (is_hardlink) {
            archive_entry_set_hardlink(entry, hardlink);
        } else {
            archive_entry_set_filetype(entry, AE_IFREG);
        }
    } else if (S_ISLNK(sb->st_mode) && (tflag == FTW_SL || tflag == FTW_SLN)) {
        archive_entry_set_filetype(entry, AE_IFLNK);
        memset(buf, '\0', sizeof(buf));

        if (readlink(fpath, buf, sizeof(buf)-1) == -1) {
            warn("readlink");
            return -1;
        }

        archive_entry_set_symlink(entry, buf);
    } else if (S_ISSOCK(sb->st_mode)) {
        archive_entry_set_filetype(entry, AE_IFSOCK);
    } else if (S_ISCHR(sb->st_mode)) {
        archive_entry_set_filetype(entry, AE_IFCHR);
    } else if (S_ISBLK(sb->st_mode)) {
        archive_entry_set_filetype(entry, AE_IFBLK);
    } else if (S_ISDIR(sb->st_mode)) {
        archive_entry_set_filetype(entry, AE_IFDIR);
    } else if (S_ISFIFO(sb->st_mode)) {
        archive_entry_set_filetype(entry, AE_IFIFO);
    } else {
        warn("stat");
        return -1;
    }

    archive_entry_set_pathname(entry, vpath);
    archive_entry_set_atime(entry, sb->st_atime, 0);
    archive_entry_set_mtime(entry, sb->st_mtime, 0);
    archive_entry_set_ctime(entry, sb->st_ctime, 0);
    archive_entry_set_perm(entry, sb->st_mode);

    if (payload == PAYLOAD_DIST) {
        archive_entry_set_gid(entry, 0);
        archive_entry_set_uid(entry, 0);
    } else {
        archive_entry_set_gid(entry, sb->st_uid);
        archive_entry_set_uid(entry, sb->st_gid);
    }

    if (!is_hardlink) {
        archive_entry_set_size(entry, sb->st_size);
    }

    if (payload != PAYLOAD_DIST) {
        if (verbose) {
            printf(">>> ");

            if (payload == PAYLOAD_RELOC) {
                printf("reloc");
            } else if (payload == PAYLOAD_ETC) {
                printf("  "CONFDIR_PREFIX);
            }

            printf(": /%s\n", vpath);
        }

        /*
         * We get the mode, owner, and group from the plist file.  We also
         * check and ensure there are no entries in the package root tree
         * that are not specified in the plist file.
         */
        TAILQ_FOREACH(iterator, plist, items) {
            if (!strcmp(iterator->filename, vpath)) {
                plistentry = iterator;
                break;
            }
        }

        if (plistentry == NULL) {
            warnx("*** WARNING: %s not specified in plist file.", vpath);
        } else {
            if (!plistentry->is_symlink && S_ISLNK(sb->st_mode)) {
                warnx("*** WARNING: %s not a 'symlink' in plist file.", vpath);
            }

            if (!plistentry->is_symlink) {
                archive_entry_set_perm(entry, plistentry->mode);
            }

            archive_entry_set_gid(entry, plistentry->uid);
            archive_entry_set_uid(entry, plistentry->gid);
            plist_checked++;
        }
    }

    /* write the entry header */
    archive_write_header(output, entry);

    /* write the data to the payload for regular files only */
    if (S_ISREG(sb->st_mode) && !is_hardlink) {
        if ((fd = open(fpath, O_RDONLY)) == -1) {
            warn("open");
            return -1;
        } else {
            memset(buf, 0, sizeof(buf));
        }

        if ((len = read(fd, buf, sizeof(buf))) == -1) {
            warn("read");
            return -1;
        }

        while (len > 0) {
            archive_write_data(output, buf, len);

            if ((len = read(fd, buf, sizeof(buf))) == -1) {
                warn("read");
                return -1;
            }
        }

        if (close(fd) == -1) {
            warn("close");
            return -1;
        }
    }

    archive_entry_free(entry);
    return 0;
}

/*
 * Create the DIST_RELOC file of package->tree, excluding /etc.
 */
static int create_reloc(struct dist *package)
{
    int flags = FTW_MOUNT | FTW_PHYS;

    output = archive_write_new();
    set_archive_compression(package->compression);
    archive_write_set_format_pax_restricted(output);
    archive_write_open_filename(output, package->reloc);

    /* add all tree directories except /etc to the payload */
    payload = PAYLOAD_RELOC;
    trimlen = -1;

    if (nftw(package->tree, _add_payload_entry, 25, flags) == -1) {
        warn("nftw");
#if ARCHIVE_VERSION_NUMBER < 3000000
        archive_write_finish(output);
#else
        archive_write_free(output);
#endif
        return -1;
    }

#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_write_close(output);
    archive_write_finish(output);
#else
    archive_write_free(output);
#endif

    return 0;
}

/*
 * Pack up /etc tree if it exists.
 */
static int create_etc(struct dist *package)
{
    int flags = FTW_MOUNT | FTW_PHYS;
    struct stat sb;

    /* debug packages do not have configuration files */
    if (package->debug) {
        return 0;
    }

    memset(&sb, 0, sizeof(sb));

    /* do we even have an /etc directory? */
    if (chdir(package->tree)) {
        warn("chdir");
        return -1;
    } else if (lstat(CONFDIR_PREFIX, &sb) || !S_ISDIR(sb.st_mode)) {
        return 0;
    }

    /* create an /etc payload */
    output = archive_write_new();
    set_archive_compression(package->compression);
    archive_write_set_format_pax_restricted(output);
    archive_write_open_filename(output, package->etc);

    payload = PAYLOAD_ETC;
    trimlen = -1;

    if (nftw(package->tree, _add_payload_entry, 25, flags) == -1) {
        warn("nftw");
#if ARCHIVE_VERSION_NUMBER < 3000000
        archive_write_finish(output);
#else
        archive_write_free(output);
#endif
        return -1;
    }

#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_write_close(output);
    archive_write_finish(output);
#else
    archive_write_free(output);
#endif

    return 0;
}

/*
 * Create the package file.
 */
static int create_dist(struct dist *package)
{
    int flags = FTW_MOUNT | FTW_PHYS;
    char cwd[PATH_MAX + 1];

    memset(cwd, '\0', sizeof(cwd));

    if (getcwd(cwd, PATH_MAX) == NULL) {
        err(EXIT_FAILURE, "getcwd");
    }

    if (chdir(staging) != 0) {
        warn("chdir");
        return -1;
    }

    /* create dist package */
    output = archive_write_new();
#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_write_set_compression_none(output);
#else
    archive_write_add_filter_none(output);
#endif
    archive_write_set_format_pax_restricted(output);
    archive_write_open_filename(output, package->distname);

    payload = PAYLOAD_DIST;
    trimlen = -1;

    if (nftw(staging, _add_payload_entry, 25, flags) == -1) {
        warn("nftw");
#if ARCHIVE_VERSION_NUMBER < 3000000
        archive_write_finish(output);
#else
        archive_write_free(output);
#endif
        return -1;
    }

#if ARCHIVE_VERSION_NUMBER < 3000000
    archive_write_close(output);
    archive_write_finish(output);
#else
    archive_write_free(output);
#endif

    if (chdir(cwd) != 0) {
        warn("chdir");
        return -1;
    }

    return 0;
}

/*
 * Creating the package staging directory.
 */
static int mkstagedir(struct dist *package)
{
    int mode = S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH;
    struct stat statbuf;

    assert(staging != NULL);

    if (stat(staging, &statbuf)) {
        /* directory doesn't exist, go ahead and make it */
        if (mkdirp(staging, mode)) {
            warn("mkdirp");
            return -1;
        }
    } else {
        /* we fail if the staging directory already exists */
        if (S_ISDIR(statbuf.st_mode)) {
            warnx("Staging directory for %s already exists", package->filename);
        } else {
            warnx("%s already exists", package->filename);
        }

        return -1;
    }

    return 0;
}

/*
 * Generate string values in the dist structure.
 */
static int genstrings(struct dist *package, char *targetarch, char *outdir)
{
    package->arch = targetarch;

    if (outdir == NULL) {
        outdir = calloc(1, PATH_MAX);
        assert(outdir != NULL);
        outdir = getcwd(outdir, PATH_MAX);

        if (outdir == NULL) {
            err(EXIT_FAILURE, "getcwd");
        }
    }

    if (package->debug) {
        xasprintf(&(package->filename), "%s-%s-%d.debug.%s", package->name, package->ver, package->rel, package->arch);
    } else {
        xasprintf(&(package->filename), "%s-%s-%d.%s", package->name, package->ver, package->rel, package->arch);
    }

    /* package bill of materials database */
    xasprintf(&(package->bom), "%s/%s", staging, DIST_BOM);

    /* package reloc file full path */
    xasprintf(&(package->reloc), "%s/%s", staging, DIST_RELOC);

    /* package etc file full path */
    xasprintf(&(package->etc), "%s/%s", staging, DIST_ETC);

    /* package distname file full path */
    xasprintf(&(package->distname), "%s/%s.dist", outdir, package->filename);

    return 0;
}

/*
 * Copy a meta file to the package staging directory.
 */
static void cpmeta(struct dist *package, char *filename, bool create)
{
    char *src = NULL;
    char *dest = NULL;
    struct stat sb;
    FILE *fp = NULL;

    if (package->meta == NULL) {
        return;
    }

    /* set the source and dest filenames */
    xasprintf(&src, "%s/%s", package->meta, filename);
    xasprintf(&dest, "%s/%s", staging, filename);

    /* exit if the source doesn't exist */
    /* or create empty file if create is True */
    if (stat(src, &sb)) {
        if (create == true) {
            if ((fp = fopen(dest, "w")) == NULL) {
                warn("fopen");
                free(src);
                free(dest);
                return;
            }

            fprintf(fp, "Missing %s file.\n", src);

            if (fclose(fp) == EOF) {
                warn("fclose");
                free(src);
                free(dest);
                return;
            }
        }

        free(src);
        free(dest);
        return;
    }

    /* copy the file */
    if (copyfile(src, dest, true, true)) {
        warn("copyfile");
    }

    free(src);
    free(dest);
    return;
}

/* Called by get_tree_size() for each directory entry. */
static int _tree_entry_size(__attribute__((unused)) const char *fpath, const struct stat *sb, int tflag, __attribute__((unused)) struct FTW *ftwbuf)
{
    if (tflag == FTW_DNR || tflag == FTW_NS) {
        return -1;
    }

    treesize += sb->st_size;
    return 0;
}

/* Computes the size of the given tree */
static int get_tree_size(const char *tree)
{
    int flags = FTW_MOUNT | FTW_PHYS;

    if (tree == NULL) {
        return -1;
    }

    treesize = 0;

    if (nftw(tree, _tree_entry_size, 25, flags) == -1) {
        warn("nftw");
        return -1;
    }

    return treesize;
}

/*
 * Add package meta data to the staging directory.
 * NOTE:  When this function is entered, we are already in the pkgroot
 * directory.  Any chdir() out of the current directory needs to take
 * in to account the genattrs() call, which expects us to be in the
 * pkgroot directory.  Just make sure you are chdir()'ed in to pkgroot
 * before genattrs() is called.
 */
static int dometa(struct dist *package)
{
    char pkgroot[PATH_MAX + 1];
    FILE *fd = NULL;
    char *errmsg = NULL;
    sqlite3 *bom = NULL;
    strlist_t *manifest = NULL;
    strlist_t *symlinks = NULL;

    memset(pkgroot, '\0', sizeof(pkgroot));

    if (getcwd(pkgroot, PATH_MAX) == NULL) {
        err(EXIT_FAILURE, "getcwd");
    }

    /* initialize manifest and symlinks lists before calling genattrs() */
    manifest = init_strlist();
    symlinks = init_strlist();

    /* populates the manifest list, to avoid self-dependencies */
    if (makelists(package->tree, &manifest, &symlinks)) {
        warn("makelists");
        return -1;
    }

    /* initialize bill of materials */
    bom = open_db(package->bom);

    if (bom == NULL) {
        return -1;
    }

    /* add the Peachtree Linux package identifier */
    if (chdir(staging)) {
        warn("chdir");
        return -1;
    }

    if ((fd = fopen("signature", "w")) == NULL) {
        warn("fopen");
        return -1;
    }

    /* write out the current package signature we are using */
    fprintf(fd, "%s%s\n", SIGNATURE, SIGNATURE_IN_USE);

    if (fclose(fd)) {
        warn("fclose");
        return -1;
    }

    /* BOM(contents): write the contents of the reloc payload to the contents table */
    if (chdir(pkgroot)) {
        warn("chdir");
        return -1;
    }

    if (genattrs(package->tree, bom, plist, package->sigver, package->sumtype, package->debug, prepend)) {
        warn("genattrs");
        return -1;
    }

    /* Free manifest strlists */
    free_strlist(manifest);
    free_strlist(symlinks);

    /* BOM(deps): write the collected dependencies to the deps table */
    if (writedeps(package, bom)) {
        warnx("writedeps");
        sqlite3_close(bom);
        return -1;
    }

    /* BOM(provs): write the SONAMEs in this package to the provs table */
    if (writeprovs(package, bom)) {
        warnx("writeprovs");
        sqlite3_close(bom);
        return -1;
    }

    /* add the size of the package root tree */
    if ((package->size = get_tree_size(package->tree)) == -1) {
        warn("get_tree_size");
        return -1;
    }

    /* BOM(info): create the bill of materials info table */
    if (sqlite3_exec(bom, BOM_INFO_SCHEMA, NULL, NULL, &errmsg) != SQLITE_OK) {
        warn(errmsg);
        sqlite3_free(errmsg);
        sqlite3_close(bom);
        return -1;
    }

    sqlite3_free(errmsg);

    if (insert_pkg_info(package, bom, BOM_INFO_INSERT, 0, 0) < 0) {
        warnx("insert_pkg_info: %d: %s", sqlite3_errcode(bom), sqlite3_errmsg(bom));
        sqlite3_close(bom);
        return -1;
    }

    /* close out the bill of materials */
    sqlite3_close(bom);

    /* copy in any meta information that's present */
    if (!package->debug) {
        cpmeta(package, DIST_PREINSTALL, false);
        cpmeta(package, DIST_POSTINSTALL, false);
        cpmeta(package, DIST_PREREMOVE, false);
        cpmeta(package, DIST_POSTREMOVE, false);
    }

    return 0;
}

/* MAIN */
int main(int argc, char **argv)
{
    int i = 0, long_idx = 0;
    int mode = S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH;
    struct dist *package = NULL;
    char *outdir = NULL;
    char *tmpdir = NULL;
    struct stat sb;

    setlinebuf(stdout);
    prog = basename(argv[0]);
    targetarch = getdefaultarch();

    /* create a new package structure */
    package = initpkg();
    assert(package != NULL);

    /* parse options */
    while ((i = getopt_long(argc, argv, "dt:m:p:v:r:l:R:a:D:M:L:E:o:z:s:cVh?", long_options, &long_idx)) != -1) {
        switch (i) {
            case 'd':
                package->debug = true;
                break;
            case 't':
                package->tree = strdup(optarg);
                break;
            case 'm':
                package->meta = strdup(optarg);
                break;
            case 'p':
                package->name = strdup(optarg);
                break;
            case 'v':
                package->ver = strdup(optarg);
                break;
            case 'r':
                package->rel = strtol(optarg, NULL, 10);

                if (errno == ERANGE) {
                    err(EXIT_FAILURE, "strtol");
                }

                break;
            case 'l':
                if (stat(optarg, &sb) == -1) {
                    err(EXIT_FAILURE, "stat");
                }

                if (!S_ISREG(sb.st_mode)) {
                    err(EXIT_FAILURE, "%s is not a regular file.", optarg);
                }

                package->plist = strdup(optarg);
                break;
            case 'R':
                package->repo = strdup(optarg);
                break;
            case 'a':
                /* check target arch and make sure it isn't bogus */
                if (check_targetarch(optarg)) {
                    err(EXIT_FAILURE, "Unsupported target architecture: %s", optarg);
                }

                free(targetarch);
                targetarch = strdup(optarg);
                break;
            case 'D':
                package->desc = strdup(optarg);
                break;
            case 'M':
                package->maintainer = strdup(optarg);
                break;
            case 'L':
                package->license = strdup(optarg);
                break;
            case 'E':
                package->depspath = strdup(optarg);
                break;
            case 'o':
                if (stat(optarg, &sb) == -1) {
                    if (errno == ENOENT) {
                        errx(EXIT_FAILURE, "%s does not exist.", optarg);
                    } else {
                        err(EXIT_FAILURE, "stat");
                    }
                }

                if (!S_ISDIR(sb.st_mode)) {
                    errx(EXIT_FAILURE, "%s is not a directory.", optarg);
                }

                outdir = strdup(optarg);
                break;
            case 'z':
                if (!strcasecmp(optarg, "bzip2") || !strcasecmp(optarg, "bz2")) {
                    package->compression = ARCHIVE_FILTER_BZIP2;
                } else if (!strcasecmp(optarg, "compress") || !strcasecmp(optarg, "Z")) {
                    package->compression = ARCHIVE_FILTER_COMPRESS;
                } else if (!strcasecmp(optarg, "gzip") || !strcasecmp(optarg, "gz")) {
                    package->compression = ARCHIVE_FILTER_GZIP;
                } else if (!strcasecmp(optarg, "zst") || !strcasecmp(optarg, "zstd")) {
                    package->compression = ARCHIVE_FILTER_ZSTD;
                } else if (!strcasecmp(optarg, "none")) {
                    package->compression = ARCHIVE_FILTER_NONE;
                } else {
                    errx(EINVAL, "unsupported compression type");
                }

                break;
            case 's':
                if (!strcasecmp(optarg, "md5")) {
                    package->sumtype = MD5SUM;
                } else if (!strcasecmp(optarg, "sha1")) {
                    package->sumtype = SHA1SUM;
                } else if (!strcasecmp(optarg, "sha256")) {
                    package->sumtype = SHA256SUM;
                } else if (!strcasecmp(optarg, "sha384")) {
                    package->sumtype = SHA384SUM;
                } else if (!strcasecmp(optarg, "sha512")) {
                    package->sumtype = SHA512SUM;
                } else {
                    errx(EXIT_FAILURE, "unknown option: %s", optarg);
                }

                break;
            case 'c':
                package->do_cleanup = true;
                break;
            case 'V':
                ver(prog);
                return EXIT_SUCCESS;
            case 'h':
            case '?':
                usage();
                return EXIT_SUCCESS;
            case 0:
                /* long options with no short equivalent */
                if (!strcmp(long_options[long_idx].name, "tmpdir")) {
                    /* if this directory already exists, do not remove it */
                    assert(optarg != NULL);

                    if (stat(optarg, &sb) == -1) {
                        if (errno == ENOENT) {
                            errx(EXIT_FAILURE, "%s does not exist.", optarg);
                        } else {
                            err(EXIT_FAILURE, "stat");
                        }
                    }

                    if (S_ISDIR(sb.st_mode)) {
                        opts |= OPTS_SET_TMPDIR;
                    }

                    tmpdir = optarg;
                }

                break;
            default:
                usage();
                return EXIT_FAILURE;
        }
    }

    /* verify we have required options */
    if (checknull(package->tree, REASON_TREE) || checknull(package->name, REASON_PKGNAME) || checknull(package->ver, REASON_PKGVERSION) ||
        checkneg(package->rel, REASON_PKGRELEASE) || checknull(package->plist, REASON_PLIST)) {
        freepkg(package);
        return EXIT_FAILURE;
    }

    /* the metadata directory must be different than the tmp directory */
    if (package->meta && !strcmp(_PATH_VARTMP, package->meta)) {
        errx(EXIT_FAILURE, "*** ERROR: meta data directory and tmp directory are the same");
    }

    /* find a suitable tmp dir if one was not specified */
    if (tmppath == NULL) {
        xasprintf(&tmppath, "%s/%s.%s-%s-%d.%s.XXXXXX", (tmpdir == NULL) ? _PATH_VARTMP : tmpdir, prog, package->name, package->ver, package->rel, targetarch);

        if ((tmppath = mkdtemp(tmppath)) == NULL) {
            err(EXIT_FAILURE, "Unable to find a suitable temporary directory, please use the --tmpdir option");
        }
    }

    /* make the tmp path if it doesn't exist */
    if (mkdirp(tmppath, mode)) {
        err(EXIT_FAILURE, "mkdirp");
    }

    /* staging directory */
    xasprintf(&staging, "%s/%s", tmppath, STAGING_SUBDIR);

    /* generate some filename strings for later */
    if (genstrings(package, targetarch, outdir)) {
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        err(EXIT_FAILURE, "genstrings");
    }

    /* remove package if it exists (just in case, our add_archive_member()
     * function doesn't handle replacement. */
    if (!stat(package->distname, &sb) && S_ISREG(sb.st_mode)) {
        if (remove(package->distname) == -1) {
            rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
            err(EXIT_FAILURE, "remove");
        }
    }

    /* create the package staging directory */
    if (mkstagedir(package)) {
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        err(EXIT_FAILURE, "mkstagedir");
    }

    printf("Creating payloads...");

    if (verbose) {
        printf("\n");
    }

    plist = read_plist(package->plist, package->tree);

    if (add_inodes(package->tree)) {
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        err(EXIT_FAILURE, "add_inodes");
    }

    /* pack up /etc, if it exists */
    if (create_etc(package)) {
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        err(EXIT_FAILURE, "create_etc");
    }

    /* copy & preserve -t to the staging reloc tree */
    if (create_reloc(package) == -1) {
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        err(EXIT_FAILURE, "create_reloc");
    }

    if (plist_checked != plist_length(plist)) {
        warnx("*** WARNING: plist and package root tree mismatch (%s-%s-%d)", package->name, package->ver, package->rel);
    }

    if (!verbose) {
        printf("DONE.\n");
    }

    /* add in package meta data */
    printf("Adding meta data...");

    if (dometa(package)) {
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        return EXIT_FAILURE;
    }

    printf("DONE.\n");

    /* pack everything up */
    printf("Writing %s...", package->distname);

    if (create_dist(package)) {
        rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
        err(EXIT_FAILURE, "create_dist");
    }

    printf("DONE.\n");

    /* cleanup */
    if (package->do_cleanup) {
        if (rmtree(staging, false, false)) {
            rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
            errx(EXIT_FAILURE, "rmtree");
        }
    }

    rmtree(tmppath, true, (opts & OPTS_SET_TMPDIR));
    free_inodes();
    freepkg(package);
    free(outdir);
    free(tmppath);
    free_plist(plist);

    return EXIT_SUCCESS;
}
