/* Remove a dist package */

/*
 * Copyright David Shea <david@reallylongword.org>
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright Chris Lumens <chris@bangmoney.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <err.h>
#include "dist-tools.h"
#include "db.h"
#include "opts.h"

extern char *optarg;
extern int optind, opterr, optopt;

/* Globals */
char *prepend = NULL;
uint64_t opts = 0;

/* Local prototypes */
static void usage(void);
static int remove_data(const char *, struct dist *);
static int remove_entry(const char *, const char *, struct dist *);

/* Local global variables */
static char *prog = NULL;

static struct option long_options[] = {
    { "force", no_argument, NULL, 'f' },
    { "force-script", no_argument, NULL, 'F' },
    { "quick", no_argument, NULL, 'q' },
    { "strict", no_argument, NULL, 's' },
    { "noscripts", no_argument, NULL, 'S' },
    { "root", required_argument, NULL, 'r' },
    { "verbose", no_argument, NULL, 'v' },
    { "version", no_argument, NULL, 'V' },
    { "help", no_argument, NULL, 'h' },
    { NULL, 0, NULL, 0 }
};

/* Usage screen displayed when the user passes --help as an argument. */
static void usage(void)
{
    printf("Usage: %s [OPTIONS] packagespec...\n", prog);
    puts("Remove a Peachtree Linux dist package\n" "\n"
         "  -f, --force                Force removal of all files\n"
         "  -F, --force-script         Continue removal if scripts fail\n"
         "  -q, --quick                Skip all file checks\n"
         "  -s, --strict               Perform attribute checks\n"
         "  -S, --noscripts            Do not run pre/post-remove scripts\n"
         "  -r PATH, --root=PATH       Use PATH for the root (/) directory\n"
         "  -v, --verbose              Show more detailed status messages\n"
         "  -V, --version              Display version information\n"
         "  -h, --help                 This help message\n");
    return;
}

/* Remove the on-disk contents of this package (files & directories). */
static int remove_data(const char *db_installed, struct dist *package)
{
    int errs = 0;
    char *query = NULL;
    finfolist_t *contents = NULL;
    strlist_t *paths = NULL;
    strlist_entry_t *pathentry = NULL;
    finfolist_t *dirs = NULL;
    finfolist_entry_t *fientry = NULL;
    sqlite3 *db = NULL;
    sqlite3_stmt *stmt = NULL;
    char *err = NULL;
    struct finfo *fi = NULL;

    assert(package != NULL);

    /* build a list of the package contents */
    contents = init_finfolist();

    /* initialize the paths list */
    paths = init_strlist();

    /* open a connection to the database */
    db = open_db(db_installed);

    if (db == NULL) {
        return -1;
    }

    if (sqlite3_exec(db, "PRAGMA synchronous = OFF", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        sqlite3_close(db);
        return -1;
    }

    if (sqlite3_exec(db, "PRAGMA journal_mode = MEMORY", NULL, NULL, &err) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", err);
        sqlite3_free(err);
        sqlite3_close(db);
        return -1;
    }

    /* collect contents */
    xasprintf(&query, PACKAGE_CONTENTS, package->id);

    if (sqlite3_prepare(db, query, -1, &stmt, 0) != SQLITE_OK) {
        warnx("sqlite3_prepare: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_close(db);
        return -2;
    }

    while (sqlite3_step(stmt) == SQLITE_ROW) {
        /* allocate a new struct finfo for each entry */
        fi = calloc(1, sizeof(*fi));
        assert(fi != NULL);

        /* build a struct finfo for each package entry */
        if (read_finfo(-1, package->sigver, stmt, fi, prepend)) {
            warnx("error reading finfo for some row");
            errs = -4;
        }

        pathentry = calloc(1, sizeof(*pathentry));
        assert(pathentry != NULL);
        pathentry->val = strdup(fi->filename);
        TAILQ_INSERT_TAIL(paths, pathentry, items);

        /* is this a directory?  if so, record a copy in the dirs list */
        if (fi->type == TYPE_DIR) {
            if (dirs == NULL) {
                dirs = init_finfolist();
            }

            dirs = sorted_finfolist_insert(dirs, fi);
        } else {
            contents = sorted_finfolist_insert(contents, fi);
        }
    }

    if (sqlite3_finalize(stmt) != SQLITE_OK) {
        warnx("sqlite3_finalize: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
        sqlite3_close(db);
        return -3;
    }

    sqlite3_close(db);

    free(query);
    query = NULL;

    /* remove everything except directories */
    if (contents != NULL) {
        TAILQ_FOREACH(fientry, contents, items) {
            errs += checkunlink(fientry->info, prepend, opts);
        }

        free_finfolist(contents);
    }

    /* remove dirs */
    if (dirs != NULL) {
        TAILQ_FOREACH_REVERSE(fientry, dirs, finfolist_s, items) {
            errs = checkunlink(fientry->info, prepend, opts);
        }

        free_finfolist(dirs);
    }

    free_strlist(paths);
    return errs;
}

/* Remove the database entry for this package and any on-disk metadata. */
static int remove_entry(const char *db_installed, const char *dbpath, struct dist *package)
{
    int err = 0;
    char *dberr = NULL;
    char *query = NULL;
    sqlite3 *db = NULL;

    assert(package != NULL);

    /* open a connection to the database */
    db = open_db(db_installed);

    if (db == NULL) {
        return -1;
    }

    if (sqlite3_exec(db, "PRAGMA synchronous = OFF", NULL, NULL, &dberr) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", dberr);
        sqlite3_free(dberr);
        sqlite3_close(db);
        return -1;
    }

    if (sqlite3_exec(db, "PRAGMA journal_mode = MEMORY", NULL, NULL, &dberr) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", dberr);
        sqlite3_free(dberr);
        sqlite3_close(db);
        return -1;
    }

    /* remove the package table entry */
    xasprintf(&query, "DELETE FROM packages WHERE id=%d", package->id);

    if (sqlite3_exec(db, query, NULL, 0, &dberr) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", dberr);
        sqlite3_free(dberr);
        sqlite3_close(db);
        return -2;
    }

    free(query);
    query = NULL;

    /* remove the contents table entries */
    xasprintf(&query, "DELETE FROM contents WHERE packageid=%d", package->id);

    if (sqlite3_exec(db, query, NULL, 0, &dberr) != SQLITE_OK) {
        warnx("sqlite3_exec: %s", dberr);
        sqlite3_free(dberr);
        sqlite3_close(db);
        return -3;
    }

    free(query);
    query = NULL;

    /* remove metadata */
    if (package->pkgdigest != NULL) {
        xasprintf(&query, "%s/%s", dbpath, package->pkgdigest);

        if (rmtree(query, true, false)) {
            warn("rmtree");
            return -4;
        }

        free(query);
    }

    sqlite3_close(db);
    return err;
}

/* MAIN */
int main(int argc, char **argv)
{
    int i = 0;
    int len = 0;
    int success = EXIT_SUCCESS;
    char *db_installed = NULL;
    char *dbpath = NULL;
    struct dist *package = NULL;
    struct stat sb;
    pkglist_t *pkglist = NULL;

    setlinebuf(stdout);
    prog = basename(argv[0]);

    while ((i = getopt_long(argc, argv, "fFqsSr:vVh?", long_options, NULL)) != EOF) {
        switch (i) {
            case 'f':
                opts |= OPTS_FORCE;
                break;
            case 'F':
                opts |= OPTS_FORCESCRIPT;
                break;
            case 'q':
                opts |= OPTS_QUICK;
                break;
            case 's':
                opts |= OPTS_STRICT;
                break;
            case 'S':
                opts |= OPTS_NOSCRIPTS;
                break;
            case 'r':
                if (*optarg != '/') {
                    printf("Root directory must be an absolute pathname\n");
                    return EXIT_FAILURE;
                }

                prepend = strdup(optarg);
                break;
            case 'v':
                opts |= OPTS_VERBOSE;
                break;
            case 'V':
                ver(prog);
                return EXIT_SUCCESS;
            case 'h':
            case '?':
                usage();
                return EXIT_SUCCESS;
            default:
                usage();
                return EXIT_FAILURE;
        }
    }

    /* use / if the user did not specify an alternate root */
    if (prepend == NULL) {
        prepend = strdup("/");
    }

    /* handle an alternative root path for installation */
    dbpath = get_path(prepend, DB_PATH);
    db_installed = get_path(prepend, DB_FILE);

    if (stat(db_installed, &sb) == -1 || !S_ISREG(sb.st_mode)) {
        errx(EXIT_FAILURE, "This system lacks a Peachtree Linux package database.");
    }

    /* did the user specify a package? */
    if (optind == argc) {
        errx(EXIT_FAILURE, "*** ERROR: You must specify a package to remove.");
    }

    /* iterate over the arguments specified */
    for (i = optind; i < argc; i++) {
        /* make a list of possible installed packages */
        pkglist = scanpkg(db_installed, argv[i]);

        if (pkglist == NULL) {
            warnx("No installed package matching '%s' found.", argv[i]);
            success = EXIT_FAILURE;
            continue;
        }

        /* require more specification */
        len = length_pkglist(pkglist);

        if (len > 1) {
            warnx("*** ERROR: More than one version of %s found, please specify one to remove.", argv[i]);
            free_pkglist(pkglist);
            success = EXIT_FAILURE;
            continue;
        } else if (len != 1) {
            warnx("*** ERROR: package %s not found", argv[i]);
            free_pkglist(pkglist);
            success = EXIT_FAILURE;
            continue;
        }

        package = TAILQ_FIRST(pkglist)->pkg;
        printf("Removing %s-%s-%d%s.%s\n", package->name, package->ver, package->rel, package->debug ? ".debug" : "", package->arch);

        /* run the preremove script */
        if (!(opts & OPTS_NOSCRIPTS) && !package->debug) {
            if (runscript(package, NULL, SCRIPT_PREREMOVE, false, prepend, NULL)) {
                printf("*** %s failed for %s, ", DIST_PREREMOVE, package->name);

                if (opts & OPTS_FORCESCRIPT) {
                    printf("continuing anyway");
                } else {
                    printf("skipping package");
                    free_pkglist(pkglist);
                    success = EXIT_FAILURE;
                    continue;
                }

                printf("\n");
            }
        }

        /* remove the package data */
        if (remove_data(db_installed, package)) {
            printf("Error removing %s-%s-%d%s.%s\n", package->name, package->ver, package->rel, package->debug ? ".debug" : "", package->arch);
            success = EXIT_FAILURE;

            if (!(opts & OPTS_FORCE)) {
                free_pkglist(pkglist);
                printf("Not removing package data\n");
                continue;
            }
        }

        /* run the postremove script */
        if (!(opts & OPTS_NOSCRIPTS) && !package->debug) {
            if (runscript(package, NULL, SCRIPT_POSTREMOVE, false, prepend, NULL)) {
                printf("*** %s failed for %s, ", DIST_POSTREMOVE, package->name);

                if (opts & OPTS_FORCESCRIPT) {
                    printf("continuing anyway");
                } else {
                    printf("skipping package");
                    free_pkglist(pkglist);
                    success = EXIT_FAILURE;
                    continue;
                }

                printf("\n");
            }
        }

        /* remove the package database entry */
        if (remove_entry(db_installed, dbpath, package)) {
            printf("Error removing %s-%s-%d%s.%s entry\n", package->name, package->ver, package->rel, package->debug ? ".debug" : "", package->arch);
            success = EXIT_FAILURE;
        }

        free_pkglist(pkglist);
    }

    free(dbpath);
    free(db_installed);
    return success;
}
