/* Merge in a package's /etc files to the main system. */

/*
 * Copyright David Cantrell <dcantrell@burdell.org>
 * Copyright David Shea <david@reallylongword.org>
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <ftw.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <paths.h>
#include <limits.h>
#include <err.h>
#include "dist-tools.h"
#include "opts.h"
#include "db.h"

extern char *optarg;
extern int optind, opterr, optopt;

/* Local prototypes */
static void usage(void);
static char _looping_prompt(const char *, const char *, char *);
static char _prompt_yn(const char *, ...);
static char _prompt_mis(const char *, ...);
static char _prompt_is(const char *, ...);
static char _prompt_ies(const char *, ...);
static int _merge_entry(const char *, const struct stat *, int, struct FTW *);
int load_list(char *, strlist_t *);

/* Prompts when etcmerge hits various conditions. */
#define YN_PROMPT "\n(Y/N) > "

#define MIS_PROMPT "\nSelect option:\n" \
                   "    m   Merge differences.\n" \
                   "    i   Install the new file.\n" \
                   "    s   Skip installing new file.\n" \
                   "> "

#define IS_PROMPT "\nSelect option:\n" \
                  "    i   Install the new file.\n" \
                  "    s   Skip installing new file.\n" \
                  "> "

#define IES_PROMPT "\nSelect option:\n" \
                   "    i   Install the new file.\n" \
                   "    e   Open the new file in an editor.\n" \
                   "    s   Skip installing new file.\n" \
                   "> "

/* Local global variables */
static char *prog = NULL;
static char *editor = NULL;
static char *db_path = NULL;
static char *db_installed = NULL;
static char *prepend = NULL;
static char *tmppath = NULL;
static char *destetc = NULL;
static strlist_t *install_list = NULL;
static strlist_t *skip_list = NULL;
static bool query_results = false;
static size_t pkgpathlen = 0;

/* Command line options */
/* Use this global opts variable to hold boolean options. */
static uint64_t opts = 0;

static struct option long_options[] = {
    { "install", no_argument, NULL, 'i' },
    { "merge", no_argument, NULL, 'm' },
    { "reloc", required_argument, NULL, 'r' },
    { "query", no_argument, NULL, 'q' },
    { "ignore-install", no_argument, NULL, 'I' },
    { "ignore-skip", no_argument, NULL, 'S' },
    { "tmpdir", required_argument, NULL, 0 },
    { "version", no_argument, NULL, 'V' },
    { "help", no_argument, NULL, 'h' },
    { NULL, 0, NULL, 0 }
};

/* Usage screen displayed when the user passes --help as an argument. */
static void usage(void)
{
    printf("Usage: %s [OPTION]... [packagespec ...]\n", prog);
    puts("Merge package configuration files with system configuration files\n"
         "\n"
         "  -i, --install              Install new configuration files\n"
         "  -m, --merge                Merge package configuration files with the\n"
         "                             configuration files on the main system\n"
         "  -q, --query                Exit indicating if a merge would occur\n"
         "  -I, --ignore-install       Ignore the "ETC_INSTALL" file\n"
         "  -S, --ignore-skip          Ignore the "ETC_SKIP" file\n"
         "  -r PATH, --root=PATH       Use PATH as the root directory\n"
         "  --tmpdir=PATH              Use PATH for temporary files");
    printf("                               (default: %s)\n", _PATH_VARTMP);
    puts("  -V, --version              Display version information\n"
         "  -h, --help                 This help message\n");
    return;
}

/* Display a Y/N prompt and loop until the user provides valid input. */
static char _prompt_yn(const char *format, ...)
{
    va_list args;
    char *prompt = NULL;
    char ret;

    va_start(args, format);
    xvasprintf(&prompt, format, args);
    va_end(args);

    ret = _looping_prompt(prompt, YN_PROMPT, "yn");
    free(prompt);
    return ret;
}

/* Display a merge prompt and loop until the user provides valid input. */
static char _prompt_mis(const char *format, ...)
{
    va_list args;
    char *prompt = NULL;
    char ret;

    va_start(args, format);
    xvasprintf(&prompt, format, args);
    va_end(args);

    ret = _looping_prompt(prompt, MIS_PROMPT, "mis");
    free(prompt);
    return ret;
}

/* Display a prompt for installing new symlinks. */
static char _prompt_is(const char *format, ...)
{
    va_list args;
    char *prompt = NULL;
    char ret;

    va_start(args, format);
    xvasprintf(&prompt, format, args);
    va_end(args);

    ret = _looping_prompt(prompt, IS_PROMPT, "is");
    free(prompt);
    return ret;
}

/* Display a prompt following an sdiff run. */
static char _prompt_ies(const char *format, ...) {
    va_list args;
    char *prompt = NULL;
    char ret;

    va_start(args, format);
    xvasprintf(&prompt, format, args);
    va_end(args);

    ret = _looping_prompt(prompt, IES_PROMPT, "ies");
    free(prompt);
    return ret;
}

/* Display a prompt and collect a valid answer. */
static char _looping_prompt(const char *info, const char *prompt, char *options)
{
    ssize_t i = 0;
    size_t j = 0;
    size_t len = 0;
    char *buf = NULL;
    char ret = '\0';
    bool found = false;

    assert(info != NULL);
    assert(prompt != NULL);
    assert(options != NULL);

    while (!found) {
        buf = NULL;
        printf(info);
        printf(prompt);

        /* fail if getline(3) didn't work */
        if ((i = getline(&buf, &len, stdin)) == -1) {
            warn("getline");
            return '\0';
        }

        /* require single character input followed by Return */
        if (i != 2) {
            warnx("*** ERROR: please enter a valid option.");
            free(buf);
            continue;
        }

        /* if the user entered a valid response and Return, we found it */
        if (buf[1] == '\n') {
            ret = tolower(buf[0]);
            len = strlen(options);

            for (j = 0; j < len; j++) {
                if (ret == options[j]) {
                    found = true;
                    break;
                }
            }
        }

        free(buf);
    }

    return ret;
}

/* Called by main(), handles merging individual files. */
static int _merge_entry(const char *src, const struct stat *sb, __attribute__((unused)) int tflag, __attribute__((unused)) struct FTW *ftwbuf)
{
    int havedest = 0, status = 0;
    int i;
    int mode = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
    char answer = 'n';
    bool edit = true;
    char *dest = NULL;
    char *sumsrc = NULL, *sumdest = NULL;
    char *diffoutput = NULL;
    char linksrc[PATH_MAX], linkdest[PATH_MAX];
    struct stat destsb;
    strlist_entry_t *entry = NULL;
    strlist_entry_t *install = NULL;
    strlist_entry_t *skip = NULL;
    char *sdiff_argv[] = {SDIFF, "-o", NULL, "-a", "-s", "-w", "80", NULL, NULL, NULL};
    char *editor_argv[] = {EDITOR, NULL, NULL};
    pid_t pid = 0;

    /* handle an alternate root install location */
    if (!strcmp(prepend, "/")) {
        dest = joinpath(src + pkgpathlen, NULL);
    } else {
        dest = joinpath(prepend, src + pkgpathlen, NULL);
    }

    havedest = stat(dest, &destsb);

    if (install_list != NULL) {
        TAILQ_FOREACH(entry, install_list, items) {
            if (!strcmp(src + pkgpathlen, entry->val)) {
                install = entry;
            }
        }
    }

    if (skip_list != NULL) {
        TAILQ_FOREACH(entry, skip_list, items) {
            if (!strcmp(src + pkgpathlen, entry->val)) {
                skip = entry;
            }
        }
    }

    if (S_ISDIR(sb->st_mode)) {
        /* directory */

        if (havedest == 0 && S_ISDIR(destsb.st_mode)) {
            free(dest);
            return 0;
        } else if (havedest == -1) {
            /* destination directory is missing */
            if (!(opts & OPTS_IGNORE_SKIP) && skip != NULL) {
                printf("Skipping %s\n", dest);
            } else if (((opts & OPTS_INSTALL) || (opts & OPTS_MERGE)) || ((!(opts & OPTS_IGNORE_INSTALL) && install != NULL))) {
                /* create directory */
                if (opts & OPTS_QUERY) {
                    query_results = true;
                } else {
                    if (mkdirp(dest, mode)) {
                        warn("mkdirp");
                        free(dest);
                        return -1;
                    }

                    if (chown(dest, sb->st_uid, sb->st_gid) == -1) {
                        warn("chown");
                        free(dest);
                        return -1;
                    }

                    if (chmod(dest, (sb->st_mode & S_INTER)) == -1) {
                        warn("chmod");
                        free(dest);
                        return -1;
                    }
                }
            }
        }
    } else if (S_ISREG(sb->st_mode)) {
        /* regular file */
        if (skip != NULL && !(opts & OPTS_IGNORE_SKIP)) {
            /* user wants to always skip merging this file */
            free(dest);
            return 0;
        } else if (install != NULL && !(opts & OPTS_IGNORE_INSTALL)) {
            /* user wants to always install this file */
            printf("Installing %s ...", dest);

            if (opts & OPTS_QUERY) {
                printf("simulated\n");
                query_results = true;
            } else {
                if (copyfile(src, dest, true, true)) {
                    printf("FAILED\n");
                    free(dest);
                    return -1;
                } else {
                    printf("done\n");
                }
            }
        }

        if (havedest == 0 && S_ISREG(destsb.st_mode)) {
            /* destination file exists */
            sumsrc = checksum(src, SHA1SUM);
            sumdest = checksum(dest, SHA1SUM);

            if (sumsrc == NULL || sumdest == NULL) {
                free(dest);
                return -1;
            }

            if (!strcmp(sumsrc, sumdest)) {
                /* src and dest are the same */
                free(sumsrc);
                free(sumdest);
                free(dest);
                return 0;
            }

            free(sumsrc);
            free(sumdest);

            /* are we merging these files? */
            if (opts & OPTS_QUERY) {
                /* in query mode, simulate installation only */
                answer = 'i';
            } else {
                if (opts & OPTS_MERGE) {
                    /* merge src and dest */
                    answer = 'm';
                } else {
                    answer = _prompt_mis("%s is different from the package-provided file.", dest);
                }
            }

            if (answer == 'm') {
                /* perform the merge */

                /* generate the sdiff output file */
                xasprintf(&diffoutput, "%s/sdiff.XXXXXX", tmppath);

                if ((status = mkstemp(diffoutput)) == -1) {
                    warn("mkstemp");
                    free(dest);
                    return -1;
                }

                if (close(status) == -1) {
                    warn("close");
                    free(dest);
                    return -1;
                }

                /* run sdiff */
                sdiff_argv[2] = diffoutput;
                sdiff_argv[7] = (char *) src;
                sdiff_argv[8] = dest;

                if ((pid = fork()) == 0) {
                    if (execv(SDIFF, sdiff_argv) == -1) {
                        err(EXIT_FAILURE, "execv");
                    }
                } else if (pid == -1) {
                    warn("fork");
                    free(dest);
                    return 0;
                } else {
                    free(dest);

                    if (waitpid(pid, &status, 0) == -1) {
                        warn("waitpid");
                        return -1;
                    }

                    if (!WIFEXITED(status) || WEXITSTATUS(status) != EXIT_SUCCESS) {
                        return 0;
                    } else {
                        return -1;
                    }
                }

                if (status == 1 && (opts & OPTS_QUERY)) {
                    query_results = true;
                    free(diffoutput);
                    free(dest);
                    return 0;
                }

                /* take appropriate action depending on sdiff exit value */
                while (edit) {
                    edit = false;

                    if (status == 0) {
                        printf("No changes detected in %s.\n", dest);
                    } else if (status == 1) {
                        answer = _prompt_ies("Changes to %s merged.");
                    } else {
                        warnx("sdiff: %d", status);
                        free(dest);
                        return -1;
                    }

                    if (answer == 'i') {
                        printf("Installing merged %s ...", dest);

                        if (!(opts & OPTS_QUERY)) {
                            if (copyfile(diffoutput, dest, true, true)) {
                                printf("FAILED\n");
                                free(dest);
                                return -1;
                            } else {
                                printf("done\n");
                            }
                        }
                    } else if (answer == 'v') {
                        if ((pid = fork()) == 0) {
                            editor_argv[1] = diffoutput;

                            if (execv(EDITOR, editor_argv) == -1) {
                                err(EXIT_FAILURE, "execv");
                            }
                        } else if (pid == -1) {
                            warn("fork");
                            free(dest);
                            return 0;
                        } else {
                            free(dest);

                            if (waitpid(pid, &status, 0) == -1) {
                                warn("waitpid");
                                return -1;
                            }

                            if (!WIFEXITED(status) || WEXITSTATUS(status) != EXIT_SUCCESS) {
                                return 0;
                            } else {
                                return -1;
                            }
                        }

                        edit = true;
                    } else if (answer == 's') {
                        if (unlink(diffoutput)) {
                            warn("unlink");
                            free(dest);
                            return -1;
                        }
                    }
                }

                free(diffoutput);
            } else if (answer == 'i') {
                /* install the new version of this file */
                printf("Installing new %s ...", dest);

                if (opts & OPTS_QUERY) {
                    printf("simulated\n");
                    query_results = true;
                } else {
                    if (copyfile(src, dest, true, true)) {
                        printf("FAILED\n");
                        free(dest);
                        return -1;
                    } else {
                        printf("done\n");
                    }
                }
            } else if (answer == 's') {
                /* skip this file */
                printf("Skipping new %s\n", dest);
            }
        } else if (havedest == -1) {
            if (opts & OPTS_INSTALL) {
                /* user wants to install new config files */
                printf("Installing new %s ...", dest);

                if (opts & OPTS_QUERY) {
                    printf("simulated\n");
                    query_results = true;
                } else {
                    if (copyfile(src, dest, true, true)) {
                        printf("FAILED\n");
                        free(dest);
                        return -1;
                    } else {
                        printf("done\n");
                    }
                }
            } else {
                /* prompt the user to install the new file */
                answer = _prompt_yn("%s is not present, install new one?", dest);

                if (answer == 'y') {
                    printf("Installing new %s ...", dest);

                    if (opts & OPTS_QUERY) {
                        printf("simulated\n");
                        query_results = true;
                    } else {
                        if (copyfile(src, dest, true, true)) {
                            printf("FAILED\n");
                            free(dest);
                            return -1;
                        } else {
                            printf("done\n");
                        }
                    }
                }
            }
        }
    } else if (S_ISLNK(sb->st_mode)) {
        /* symlink */
        if (skip != NULL && !(opts & OPTS_IGNORE_SKIP)) {
            /* user wants to always skip merging this symlink */
            free(dest);
            return 0;
        } else if (install != NULL && !(opts & OPTS_IGNORE_INSTALL)) {
            /* user wants to always install this symlink */
            printf("Installing new symlink %s ...", dest);

            if (opts & OPTS_QUERY) {
                printf("simulated\n");
                query_results = true;
            } else {
                if (copyfile(src, dest, true, true)) {
                    printf("FAILED\n");
                    free(dest);
                    return -1;
                } else {
                    printf("done\n");
                }
            }
        }

        if (havedest == 0 && S_ISLNK(destsb.st_mode)) {
            /* link exists */
            if ((i = readlink(src, linksrc, PATH_MAX)) == -1) {
                warn("readlink");
                free(dest);
                return -1;
            }

            /* make sure our readlink value is NULL terminated */
            linksrc[i >= PATH_MAX ? PATH_MAX - 1 : i] = '\0';

            if ((i = readlink(dest, linkdest, PATH_MAX)) == -1) {
                warn("readlink");
                free(dest);
                return -1;
            }

            /* make sure our readlink value is NULL terminated */
            linkdest[i >= PATH_MAX ? PATH_MAX - 1 : i] = '\0';

            if (!strcmp(linksrc, linkdest)) {
                /* symlinks are the same, do nothing */
                free(dest);
                return 0;
            }

            /* symlinks are different, so merge */
            if (opts & OPTS_MERGE) {
                /* merge src and dest */
                answer = 'm';
            } else {
                printf("Soft link difference:\n");
                printf("%s -> %s\n", src, linksrc);
                printf("New link points to %s\n", linkdest);
                answer = _prompt_is("%s link destination differs.", dest);
            }

            if (answer == 'i') {
                /* install the new version of this link */
                printf("Installing new symlink %s ...", dest);

                if (opts & OPTS_QUERY) {
                    printf("simulated\n");
                    query_results = true;
                } else {
                    if (copyfile(src, dest, true, true)) {
                        printf("FAILED\n");
                        free(dest);
                        return -1;
                    } else {
                        printf("done\n");
                    }
                }
            } else if (answer == 's') {
                /* skip this link */
                printf("Skipping new symlink %s\n", dest);
            }
        } else if (havedest == -1) {
            if (opts & OPTS_INSTALL) {
                /* user wants to install new symlinks */
                printf("Installing new symlink %s ...", dest);

                if (opts & OPTS_QUERY) {
                    printf("simulated\n");
                    query_results = true;
                } else {
                    if (copyfile(src, dest, true, true)) {
                        printf("FAILED\n");
                        free(dest);
                        return -1;
                    } else {
                        printf("done\n");
                    }
                }
            } else {
                /* prompt the user to install the new symlink */
                answer = _prompt_yn("%s is not present, install new one?", dest);

                if (answer == 'y') {
                    printf("Installing new symlink %s ...", dest);

                    if (opts & OPTS_QUERY) {
                        printf("simulated\n");
                        query_results = true;
                    } else {
                        if (copyfile(src, dest, true, true)) {
                            printf("FAILED\n");
                            free(dest);
                            return -1;
                        } else {
                            printf("done\n");
                        }
                    }
                }
            }
        }
    } else if (!S_ISCHR(sb->st_mode) && !S_ISBLK(sb->st_mode) && !S_ISFIFO(sb->st_mode) && !S_ISSOCK(sb->st_mode)) {
        warnx("*** ERROR: unsupported type: %s", src);
        return -1;
    }

    free(dest);
    return 0;
}

/* Loads up the always-install and always-skip lists. */
int load_list(char *listfile, strlist_t *list)
{
    char *buf = NULL;
    char *line = NULL;
    size_t len = 0;
    FILE *fp = NULL;
    strlist_entry_t *entry = NULL;

    assert(listfile != NULL);

    if ((fp = fopen(listfile, "r")) == NULL) {
        /* it's ok for the file to not exist */
        return 0;
    }

    while (getline(&buf, &len, fp) != -1) {
        /* skip blank lines and comments */
        if (*buf == '#' || *buf == '\n' || *buf == '\r') {
            free(buf);
            buf = NULL;
            continue;
        }

        /* strip trailing line ending characters */
        len = strcspn(buf, "\r\n");
        line = calloc(len + 1, sizeof(*line));
        assert(line != NULL);
        line = strncpy(line, buf, len);
        assert(line != NULL);

        if (!is_str_in_list(list, line)) {
            entry = calloc(1, sizeof(*entry));
            assert(entry != NULL);
            entry->val = strdup(line);
            TAILQ_INSERT_TAIL(list, entry, items);
        }

        free(line);
        free(buf);
        buf = NULL;
    }

    fclose(fp);
    return 0;
}

/* MAIN */
int main(int argc, char **argv)
{
    int i = 0, long_idx = 0;
    int mode = S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
    char *arch = NULL;
    char *listfile = NULL;
    char *pkgstr = NULL;
    char *srcetc = NULL;
    char *tmpdir = NULL;
    pkglist_t *pkglist = NULL;
    pkglist_entry_t *entry = NULL;
    sqlite3 *db = NULL;
    sqlite3_stmt *stmt = NULL;
    char *errmsg = NULL;
    struct stat sb;
    struct dist *pkg = NULL;

    setlinebuf(stdout);
    prog = basename(argv[0]);

    /* parse options */
    while ((i = getopt_long(argc, argv, "imr:qISTVh?", long_options, &long_idx)) != -1) {
        switch (i) {
            case 'i':
                opts |= OPTS_INSTALL;
                break;
            case 'm':
                opts |= OPTS_MERGE;
                break;
            case 'q':
                opts |= OPTS_QUERY;
                break;
            case 'I':
                opts |= OPTS_IGNORE_INSTALL;
                break;
            case 'S':
                opts |= OPTS_IGNORE_SKIP;
                break;
            case 'r':
                if (*optarg != '/') {
                    printf("Root directory must be an absolute pathname\n");
                    return EXIT_FAILURE;
                }

                prepend = strdup(optarg);
                break;
            case 'V':
                ver(prog);
                return EXIT_SUCCESS;
            case 'h':
            case '?':
                usage();
                return EXIT_SUCCESS;
            case 0:
                if (!strcmp(long_options[long_idx].name, "tmpdir")) {
                    /* if this directory already exists, do not remove it */
                    assert(optarg != NULL);

                    if (stat(optarg, &sb) == -1) {
                        if (errno == ENOENT) {
                            warnx("%s does not exist.", optarg);
                        } else {
                            warn("stat");
                        }

                        return EXIT_FAILURE;
                    }

                    if (S_ISDIR(sb.st_mode)) {
                        opts |= OPTS_SET_TMPDIR;
                    }

                    tmpdir = optarg;
                }

                break;
            default:
                usage();
                return EXIT_FAILURE;
        }
    }

    /* use / if the user did not specify a root path */
    if (prepend == NULL) {
        prepend = strdup("/");
    }

    /* make sure we have sdiff */
    if (access(SDIFF, F_OK | X_OK)) {
        warnx("Missing %s command, exiting.", SDIFF);
        return EXIT_FAILURE;
    }

    /* the user may have a different editor in their environment */
    if (getenv("EDITOR") == NULL) {
        editor = strdup(EDITOR);
    } else {
        editor = strdup(getenv("EDITOR"));
    }

    /* initialize the installed package database if we need to */
    if (initdb(prepend, DB_PATH, DB_FILE)) {
        err(EXIT_FAILURE, "initdb");
    }

    /* find a suitable tmp dir if one was not specified */
    if (tmppath == NULL) {
        arch = getdefaultarch();
        xasprintf(&tmppath, "%s/%s.%s.XXXXXX", (tmpdir == NULL) ? _PATH_VARTMP : tmpdir, prog, arch);
        free(arch);

        if ((tmppath = mkdtemp(tmppath)) == NULL) {
            err(EXIT_FAILURE, "Unable to find a suitable temporary directory, please use the --tmpdir option");
        }
    }

    /* make the tmp path if it doesn't exist */
    if (mkdirp(tmppath, mode)) {
        err(EXIT_FAILURE, "mkdirp");
    }

    /* handle an alternative root path for installation */
    db_path = get_path(prepend, DB_PATH);
    db_installed = get_path(prepend, DB_FILE);
    destetc = get_path(prepend, ETCDIR);

    /* populate the install and skip lists */
    if (!strcmp(prepend, "/")) {
        listfile = strdup(ETC_INSTALL);
    } else {
        listfile = joinpath(prepend, ETC_INSTALL, NULL);
    }

    install_list = init_strlist();

    if (load_list(listfile, install_list)) {
        opts |= OPTS_IGNORE_INSTALL;
    }

    free(listfile);

    if (!strcmp(prepend, "/")) {
        listfile = strdup(ETC_SKIP);
    } else {
        listfile = joinpath(prepend, ETC_SKIP, NULL);
    }

    skip_list = init_strlist();

    if (load_list(listfile, skip_list)) {
        opts |= OPTS_IGNORE_SKIP;
    }

    free(listfile);
    listfile = NULL;

    /* make sure the destination configuration file directory is there */
    if ((stat(destetc, &sb) == -1 || !S_ISDIR(sb.st_mode)) && !(opts & OPTS_QUERY)) {
        if (mkdirp(destetc, mode)) {
            warn("mkdirp");
            return -1;
        }
    }

    /* what packages are we merging? */
    if (optind == argc) {
        /* open the installed package database */
        db = open_db(db_installed);

        if (db == NULL) {
            errx(EXIT_FAILURE, "open_db");
        }

        if (sqlite3_exec(db, "PRAGMA synchronous = OFF", NULL, NULL, &errmsg) != SQLITE_OK) {
            warnx("sqlite3_exec: %s", errmsg);
            sqlite3_free(errmsg);
            sqlite3_close(db);
            return EXIT_FAILURE;
        }

        if (sqlite3_exec(db, "PRAGMA journal_mode = MEMORY", NULL, NULL, &errmsg) != SQLITE_OK) {
            warnx("sqlite3_exec: %s", errmsg);
            sqlite3_free(errmsg);
            sqlite3_close(db);
            return EXIT_FAILURE;
        }

        /* no packages specified, traverse all */
        if (sqlite3_prepare(db, ETCMERGE_ALLPKGS, -1, &stmt, 0) != SQLITE_OK) {
            warnx("sqlite3_prepare: %s (%d)", sqlite3_errmsg(db), sqlite3_errcode(db));
            sqlite3_close(db);
            return EXIT_FAILURE;
        }

        while (sqlite3_step(stmt) == SQLITE_ROW) {
            /* debug packages do not (or should not) have config files */
            if (sqlite3_column_int(stmt, 4)) {
                continue;
            }

            /* build the package db subdir */
            free(srcetc);
            srcetc = NULL;
            srcetc = joinpath(db_path, (char *) sqlite3_column_text(stmt, 5), ETCDIR, NULL);

            if ((stat(srcetc, &sb) == -1) || !S_ISDIR(sb.st_mode)) {
                continue;
            }

            /* package string */
            xasprintf(&pkgstr, "%s-%s-%d.%s",
                      (char *) sqlite3_column_text(stmt, 0),
                      (char *) sqlite3_column_text(stmt, 1),
                      sqlite3_column_int(stmt, 2),
                      (char *) sqlite3_column_text(stmt, 3));
            printf(">>> %s\n\n", pkgstr);
            free(pkgstr);

            /* merge all of the configuration files for this package */
            pkgpathlen = strlen(srcetc) - strlen(ETCDIR);

            if (nftw(srcetc, _merge_entry, 25, FTW_MOUNT | FTW_PHYS) == -1) {
                warn("nftw");
                continue;
            }
        }

        free(srcetc);
        sqlite3_finalize(stmt);
        sqlite3_free(errmsg);
        sqlite3_close(db);
    } else {
        /* specific packages requested by the user, iterate over those */
        for (i = optind; i < argc; i++) {
            /* we allow users to specify packages in multiple ways, and
             * that specification can yield multiple matches, so call scanpkg */
            pkglist = scanpkg(db_installed, argv[i]);

            TAILQ_FOREACH(entry, pkglist, items) {
                pkg = entry->pkg;

                /* debug packages do not (or should not) have config files */
                if (pkg->debug) {
                    continue;
                }

                /* build the package db subdir */
                free(srcetc);
                srcetc = NULL;
                srcetc = joinpath(db_path, pkg->pkgdigest, ETCDIR, NULL);

                if ((stat(srcetc, &sb) == -1) || !S_ISDIR(sb.st_mode)) {
                    continue;
                }

                /* package string */
                xasprintf(&pkgstr, "%s-%s-%d.%s", pkg->name, pkg->ver, pkg->rel, pkg->arch);
                printf(">>> %s\n\n", pkgstr);
                free(pkgstr);

                /* merge all of the configuration files for this package */
                pkgpathlen = strlen(srcetc) - strlen(ETCDIR);

                if (nftw(srcetc, _merge_entry, 25, FTW_MOUNT | FTW_PHYS) == -1) {
                    warn("nftw");
                    continue;
                }
            }

            free(srcetc);
            free_pkglist(pkglist);
        }
    }

    /* clean up */
    free_strlist(install_list);
    free_strlist(skip_list);
    free(editor);
    free(db_path);
    free(db_installed);
    free(prepend);
    free(destetc);

    if (opts & OPTS_QUERY) {
        if (query_results) {
            return EXIT_SUCCESS;
        } else {
            return EXIT_FAILURE;
        }
    } else {
        return EXIT_SUCCESS;
    }
}
