#!/bin/sh
# Verify we can install a package to a specific root tree

CWD="$(pwd)"

title=" >>> $(basename ${0}) <<< "
line="-$(echo ${title} | sed -e 's|.|-|g')-"
echo
echo "+${line}+"
echo "|${title}|"
echo "+${line}+"
echo

TREE="$(mktemp -d)"
META="$(mktemp -d)"
OUTPUT="$(mktemp -d)"
TMPDIST="$(mktemp -d)"
FAILS=0
EXPECTED_FAILS=0

# create a tree to package up
mkdir -p ${TREE}/etc
mkdir -p ${TREE}/usr/sbin
echo "System ready" > ${TREE}/etc/issue
cp -p ${DISTMK} ${TREE}/usr/sbin/distmk

# create plist
cat << EOF > ${META}/plist
etc root(0) root(0) 0755
etc/issue root(0) root(0) 0644
usr root(0) root(0) 0755
usr/sbin root(0) root(0) 0755
usr/sbin/distmk root(0) root(0) 0755
EOF

DESC="Test program"
MAINTAINER="J. Random Packager <someone@example.net>"
LICENSE="WTFPL"
PKG="test-1.0-1.$(uname -m).dist"

# create package helper
create_pkg() {
    if [ -z "${LEAKCHECK}" ]; then
        ${DISTMK} -t ${TREE} -p "test" -v "1.0" -r 1 -m ${META} -l ${META}/plist -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST} -D "${DESC}" -M "${MAINTAINER}" -L "${LICENSE}" >/dev/null 2>&1
    else
        ${LEAKCHECK} ${DISTMK} -t ${TREE} -p "test" -v "1.0" -r 1 -m ${META} -l ${META}/plist -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST} -D "${DESC}" -M "${MAINTAINER}" -L "${LICENSE}"
    fi
}

create_pkg

# install package helper
install_pkg() {
    [ -z "${LEAKCHECK}" ] && env ETCMERGE=${ETCMERGE} ${DISTADD} $* || env ETCMERGE=${ETCMERGE} ${LEAKCHECK} ${DISTADD} $*
}

# install to / should fail if we are not root
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r
echo -n "checking 'distadd PKG'..."

if [ $(id -u) -eq 0 ]; then
    echo "SKIPPING (cannot perform test as non-root)"
else
    EXPECTED_FAILS=$((${EXPECTED_FAILS} + 1))
    out="$(install_pkg -d ${OUTPUT}/${PKG} 2>&1)"
    r=$?
    echo "${out}" | grep -q "distadd: open: Permission denied" >/dev/null 2>&1
    s=$?
    if [ ${r} -eq 1 ] && [ ${s} -eq 0 ]; then
        FAILS=$((${FAILS} + 1))
        echo "passed."
    else
        echo "FAILED."
    fi
    [ -z "${LEAKCHECK}" ] || echo "${out}"
fi

# install to a staging root should work fine
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -r PATH PKG'..."
out="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} 2>&1)"
r=$?
if [ ${r} -eq 1 ]; then
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
else
    echo "passed."
fi
[ -z "${LEAKCHECK}" ] || echo "${out}"

# installing a package we already have should fail
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -r PATH PKG' on already installed PKG..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} 2>&1)"
out2="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} 2>&1)"
r=$?
out="$(echo "${out2}" | grep "test already installed")"
if [ ${r} -eq 1 ] && [ ! -z "${out}" ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

# force installing a package we already have should work
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -f -r PATH PKG' on already installed PKG..."
spec="$(basename ${PKG} .dist)"
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} 2>&1)"
out2="$(install_pkg -d -f -r ${OUTPUT}/r ${OUTPUT}/${PKG} 2>&1)"
r=$?
echo "${out2}" | grep -q "Force installed package ${spec} successfully" 2>/dev/null
a=$?
out="$(echo "${out2}" | tail -n 1)"
p1=$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM packages" | wc -l)
if [ ${r} -eq 0 ] && [ ${p1} -eq 2 ] && [ ${a} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

# installing a package and generating metadata
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -g' on new package..."
out="$(install_pkg -g -d -r ${OUTPUT}/r ${OUTPUT}/${PKG} 2>&1)"
r=$?
p1=$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM packages" 2>/dev/null | wc -l)
p2=$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM contents" 2>/dev/null | wc -l)
if [ ${r} -eq 0 ] && [ ${p1} -eq 1 ] && [ ${p2} -eq 3 ] && [ -x ${OUTPUT}/r/usr/sbin/distmk ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
[ -z "${LEAKCHECK}" ] || echo "${out}"

# invoking etcmerge to install initial configuration files
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -e' for initial config file install..."
out="$(install_pkg -d -e -r ${OUTPUT}/r ${OUTPUT}/${PKG} 2>&1)"
r=$?
out="$(echo "${out}" | grep "Installing new" | grep "/etc/issue")"
if [ ${r} -eq 0 ] && [ ! -z "${out}" ] && [ -r ${OUTPUT}/r/etc/issue ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
[ -z "${LEAKCHECK}" ] || echo "${out}"

rm -rf ${TREE}
rm -rf ${META}
rm -rf ${TMPDIST}
rm -rf ${OUTPUT}

# expected $EXPECTED_FAILS failures
[ ${FAILS} -eq ${EXPECTED_FAILS} ] && exit 0 || exit 1
