#!/bin/sh
# Verify distadd can install and upgrade multiversion packages.

title=" >>> $(basename ${0}) <<< "
line="-$(echo ${title} | sed -e 's|.|-|g')-"
echo
echo "+${line}+"
echo "|${title}|"
echo "+${line}+"
echo

TREE="$(mktemp -d)"
META="$(mktemp -d)"
OUTPUT="$(mktemp -d)"
TMPDIST="$(mktemp -d)"
FAILS=0
PKG1="test-1.0-1.$(uname -m).dist"
PKG2="test-2.0-1.$(uname -m).dist"
PKG3="test-2.0-2.$(uname -m).dist"

# create package helper
create_pkg() {
    if [ -z "${LEAKCHECK}" ]; then
        ${DISTMK} -t ${TREE} -p "test" -v "$1" -r $2 -m ${META} -l ${META}/plist -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST} -D "${DESC}" -M "${MAINTAINER}" -L "${LICENSE}" >/dev/null 2>&1
    else
        ${LEAKCHECK} ${DISTMK} -t ${TREE} -p "test" -v "$1" -r $2 -m ${META} -l ${META}/plist -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST} -D "${DESC}" -M "${MAINTAINER}" -L "${LICENSE}"
    fi
}

# install package helper
install_pkg() {
    [ -z "${LEAKCHECK}" ] && env ETCMERGE=${ETCMERGE} ${DISTADD} $* || env ETCMERGE=${ETCMERGE} ${LEAKCHECK} ${DISTADD} $*
}

# create a tree to package up
mkdir -p ${TREE}/usr/sbin
cp -p ${DISTMK} ${TREE}/usr/sbin/distmk

# create plist
cat << EOF > ${META}/plist
usr root(0) root(0) 0755
usr/sbin root(0) root(0) 0755
usr/sbin/distmk root(0) root(0) 0755
EOF

DESC="Test program"
MAINTAINER="J. Random Packager <someone@example.net>"
LICENSE="WTFPL"

create_pkg "1.0" "1"

# make a variant package to upgrade to
mv ${TREE}/usr/sbin/distmk ${TREE}/usr/sbin/distmk2
cat << EOF > ${META}/plist
usr root(0) root(0) 0755
usr/sbin root(0) root(0) 0755
usr/sbin/distmk2 root(0) root(0) 0755
EOF

create_pkg "2.0" "1"

# make another variant to upgrade to
cat << EOF > ${TREE}/usr/sbin/mkdist
#!/bin/sh
/usr/sbin/distmk2 $*
EOF
chmod 0755 ${TREE}/usr/sbin/mkdist

echo "usr/sbin/mkdist root(0) root(0) 0755" >> ${META}/plist

create_pkg "2.0" "2"

# multi version package install
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -m'..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
out2="$(install_pkg -d -m -r ${OUTPUT}/r ${OUTPUT}/${PKG2} 2>&1)"
r=$?
p1=$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM PACKAGES" | wc -l)
p2="$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM PACKAGES" | grep "1|test|1.0|1")"
p3="$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM PACKAGES" | grep "2|test|2.0|1")"
if [ ${r} -eq 0 ] && [ ${p1} -eq 2 ] && [ ! -z "${p2}" ] && [ ! -z "${p3}" ] && [ -x ${OUTPUT}/r/usr/sbin/distmk ] && [ -x ${OUTPUT}/r/usr/sbin/distmk2 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

# upgrade with -P, package already installed
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -u -P' reports package already installed..."
spec="$(basename ${PKG1} .dist)"
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
out2="$(install_pkg -d -u -P ${spec} -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
r=$?
echo "${out2}" | grep -q "Package test-1.0-1 already installed, skipping." >/dev/null 2>&1
a=$?
if [ ${r} -eq 0 ] && [ ${a} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

# upgrade with -P and force, package installs
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -u -f -P' installs the package..."
spec="$(basename ${PKG1} .dist)"
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
out2="$(${DISTADD} -d -u -f -P ${spec} -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
r=$?
echo "${out2}" | grep -q "Force upgraded package ${spec} to ${spec} successfully" >/dev/null 2>&1
a=$?
if [ ${r} -eq 0 ] && [ ${a} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

# upgrade with -P -f -m, package installs
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -u -f -m -P' installs the package..."
spec="$(basename ${PKG1} .dist)"
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
out2="$(${DISTADD} -d -u -f -m -P ${spec} -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
r=$?
echo "${out2}" | grep -q "Force upgraded package ${spec} to ${spec} successfully" >/dev/null 2>&1
if [ ${r} -eq 0 ] && [ ${a} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

# make sure the -P is required with -u -m
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -u -m' requires '-P'..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
out2="$(install_pkg -d -m -r ${OUTPUT}/r ${OUTPUT}/${PKG2} 2>&1)"
out3="$(install_pkg -d -u -m -r ${OUTPUT}/r ${OUTPUT}/${PKG3} 2>&1)"
r=$?
echo "${out3}" | grep -q "The -P or -v option is required when specifying -u and -m." >/dev/null 2>&1
a=$?
if [ ${a} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
    echo "${out3}"
fi

# make sure the -P and -v options cannot be used together
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -v' and 'distadd -P' cannot be used together..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
out2="$(install_pkg -d -m -r ${OUTPUT}/r ${OUTPUT}/${PKG2} 2>&1)"
out3="$(install_pkg -d -u -m -v 2.0 -P test-2.0 -r ${OUTPUT}/r ${OUTPUT}/${PKG3} 2>&1)"
r=$?
echo "${out3}" | grep -q "The -P and -v options cannot be used together." >/dev/null 2>&1
a=$?
if [ ${a} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
    echo "${out3}"
fi

# multi version package upgrade with package spec
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -u -m -P'..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
out2="$(install_pkg -d -m -r ${OUTPUT}/r ${OUTPUT}/${PKG2} 2>&1)"
out3="$(install_pkg -d -u -m -P test-2.0 -r ${OUTPUT}/r ${OUTPUT}/${PKG3} 2>&1)"
r=$?
echo "${out3}" | grep -q "Upgraded package test-2.0-1.$(uname -m) to test-2.0-2.$(uname -m) successfully" >/dev/null 2>&1
a=$?
if [ ${r} -eq 0 ] && [ ${a} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
    echo "${out3}"
fi

# multi version package upgrade with package version
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -u -m -v'..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
out2="$(install_pkg -d -m -r ${OUTPUT}/r ${OUTPUT}/${PKG2} 2>&1)"
out3="$(install_pkg -d -f -u -m -v "2.0" -r ${OUTPUT}/r ${OUTPUT}/${PKG3} 2>&1)"
r=$?
p1=$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM PACKAGES" | wc -l)
p2="$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM PACKAGES" | grep "1|test|1.0|1")"
p3="$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM PACKAGES" | grep "2|test|2.0|2")"
if [ ${r} -eq 0 ] &&
   [ -x ${OUTPUT}/r/usr/sbin/distmk ] &&
   [ -x ${OUTPUT}/r/usr/sbin/distmk2 ] &&
   [ -x ${OUTPUT}/r/usr/sbin/mkdist ] &&
   [ ${p1} -eq 2 ] &&
   [ ! -z "${p2}" ] &&
   [ ! -z "${p3}" ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
    echo "${out3}"
fi

# multi version package upgrade with package spec
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r

echo -n "checking 'distadd -u -m -P'..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
out2="$(install_pkg -d -m -r ${OUTPUT}/r ${OUTPUT}/${PKG2} 2>&1)"
out3="$(install_pkg -d -f -u -m -P "test-2.0" -r ${OUTPUT}/r ${OUTPUT}/${PKG3} 2>&1)"
r=$?
p1=$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM PACKAGES" | wc -l)
p2="$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM PACKAGES" | grep "1|test|1.0|1")"
p3="$(sqlite ${OUTPUT}/r/var/db/dist/installed.db "SELECT * FROM PACKAGES" | grep "2|test|2.0|2")"
if [ ${r} -eq 0 ] &&
   [ -x ${OUTPUT}/r/usr/sbin/distmk ] &&
   [ -x ${OUTPUT}/r/usr/sbin/distmk2 ] &&
   [ -x ${OUTPUT}/r/usr/sbin/mkdist ] &&
   [ ${p1} -eq 2 ] &&
   [ ! -z "${p2}" ] &&
   [ ! -z "${p3}" ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi

rm -rf ${TREE}
rm -rf ${META}
rm -rf ${TMPDIST}
rm -rf ${OUTPUT}

# expected 0 failures
[ ${FAILS} -eq 0 ] && exit 0 || exit 1
