#!/bin/sh
# Verify we can create a package with distmk

title=" >>> $(basename ${0}) <<< "
line="-$(echo ${title} | sed -e 's|.|-|g')-"
echo
echo "+${line}+"
echo "|${title}|"
echo "+${line}+"
echo

TREE="$(mktemp -d)"
META="$(mktemp -d)"
OUTPUT="$(mktemp -d)"
TMPDIST="$(mktemp -d)"
FAILS=0

# create a tree to package up
mkdir -p ${TREE}/etc
mkdir -p ${TREE}/usr/sbin
echo "System ready" > ${TREE}/etc/issue
cp -p ${DISTMK} ${TREE}/usr/sbin/distmk

# create plist
cat << EOF > ${META}/plist
etc root(0) root(0) 0755
etc/issue root(0) root(0) 0644
usr root(0) root(0) 0755
usr/sbin root(0) root(0) 0755
usr/sbin/distmk root(0) root(0) 0755
EOF

DESC="Test program"
MAINTAINER="J. Random Packager <someone@example.net>"
LICENSE="WTFPL"

create_pkg() {
    if [ -z "${LEAKCHECK}" ]; then
        ${DISTMK} -t ${TREE} -p "test" -v "1.0" -r 1 -m ${META} -l ${META}/plist -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST} -D "${DESC}" -M "${MAINTAINER}" -L "${LICENSE}" >/dev/null 2>&1
    else
        ${LEAKCHECK} ${DISTMK} -t ${TREE} -p "test" -v "1.0" -r 1 -m ${META} -l ${META}/plist -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST} -D "${DESC}" -M "${MAINTAINER}" -L "${LICENSE}" >/dev/null 2>&1
    fi
}

# create the package
create_pkg
r=$?
echo -n "checking distmk packaging ability..."
FAILS=$((${FAILS} + ${r}))
[ ${r} -ne 0 ] && echo "FAILED." || echo "passed."

# check the package contents
pkg="$(/bin/ls -1 ${OUTPUT}/*.dist 2>/dev/null)"
cd ${OUTPUT}
mkdir a
cd a
tar -xf ${pkg} >/dev/null 2>&1

echo -n "checking signature..."
if [ ! "$(cat signature 2>/dev/null)" = "Peachtree Linux PACKAGE v2.2" ]; then
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
else
    echo "passed."
fi

echo -n "checking bom.db contents table..."
sqlite bom.db "SELECT path FROM CONTENTS" | sort > contents
grep -v "^etc" ${META}/plist | cut -d ' ' -f 1 | sort > ${META}/plist.sorted
cmp -s ${META}/plist.sorted contents
r=$?
FAILS=$((${FAILS} + ${r}))
[ ${r} -eq 1 ] && echo "FAILED." || echo "passed."

echo -n "checking etc.payload..."
zstd -dc < etc.payload | tar -xf -
if [ ! "$(cat etc/issue 2>/dev/null)" = "System ready" ]; then
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
else
    echo "passed."
fi

echo -n "checking reloc.payload..."
zstd -dc < reloc.payload | tar -xf -
if [ -f usr/sbin/distmk ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi

# NOTE: the default digest type in distmk is SHA-256
echo "checking digests..."
sqlite bom.db "SELECT path,digest FROM contents WHERE digest NOT NULL" | \
while read line ; do
    path="$(echo "$line" | cut -d '|' -f 1)"
    digest="$(echo "$line" | cut -d '|' -f 2)"

    echo -n "    $path..."
    computed="$(sha512sum $path | cut -d ' ' -f 1)"

    if [ "$digest" = "$computed" ]; then
        echo "passed."
    else
        echo "FAILED."
        FAILS=$((${FAILS} + 1))
    fi
done

echo -n "checking pkginfo..."
info="$(sqlite bom.db "SELECT * FROM info")"
name="$(echo $info | cut -d '|' -f 1)"
version="$(echo $info | cut -d '|' -f 2)"
release="$(echo $info | cut -d '|' -f 3)"
desc="$(echo $info | cut -d '|' -f 8)"
maintainer="$(echo $info | cut -d '|' -f 9)"
license="$(echo $info | cut -d '|' -f 10)"
repo="$(echo $info | cut -d '|' -f 5)"
if [ ! "${name}" = "test" -o ! "${version}" = "1.0" -o ! "${release}" = "1" -o ! "${desc}" = "Test program" -o ! "${maintainer}" = "J. Random Packager <someone@example.net>" -o ! "${repo}" = "peachtree" -o ! "${license}" = "WTFPL" ]; then
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
else
    echo "passed."
fi

rm -rf ${TREE}
rm -rf ${META}
rm -rf ${TMPDIST}
rm -rf ${OUTPUT}

# expected zero failures
[ ${FAILS} -gt 0 ] && exit 1 || exit 0
