#!/bin/sh
# Verify distrm can run package removal scripts

title=" >>> $(basename ${0}) <<< "
line="-$(echo ${title} | sed -e 's|.|-|g')-"
echo
echo "+${line}+"
echo "|${title}|"
echo "+${line}+"
echo

TREE="$(mktemp -d)"
META="$(mktemp -d)"
OUTPUT="$(mktemp -d)"
TMPDIST="$(mktemp -d)"
FAILS=0
PKG1="test-1.0-1.$(uname -m).dist"
PKG2="test-2.0-1.$(uname -m).dist"
PKG3="test-3.0-1.$(uname -m).dist"
PKG4="test-4.0-1.$(uname -m).dist"

# create package helper
create_pkg() {
    if [ -z "${LEAKCHECK}" ]; then
        ${DISTMK} -t ${TREE} -p "test" -v "$1" -r 1 -m ${META} -l ${META}/plist -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST} -D "${DESC}" -M "${MAINTAINER}" -L "${LICENSE}" >/dev/null 2>&1
    else
        ${LEAKCHECK} ${DISTMK} -t ${TREE} -p "test" -v "$1" -r 1 -m ${META} -l ${META}/plist -R peachtree -o ${OUTPUT} --tmpdir=${TMPDIST} -D "${DESC}" -M "${MAINTAINER}" -L "${LICENSE}"
    fi
}

# install package helper
install_pkg() {
    [ -z "${LEAKCHECK}" ] && env ETCMERGE=${ETCMERGE} ${DISTADD} $* || env ETCMERGE=${ETCMERGE} ${LEAKCHECK} ${DISTADD} $*
}

# remove package helper
remove_pkg() {
    [ -z "${LEAKCHECK}" ] && ${DISTRM} $* || ${LEAKCHECK} ${DISTRM} $*
}

# create a tree to package up
mkdir -p ${TREE}/usr/sbin
cp -p ${DISTMK} ${TREE}/usr/sbin/distmk

# create plist
cat << EOF > ${META}/plist
usr root(0) root(0) 0755
usr/sbin root(0) root(0) 0755
usr/sbin/distmk root(0) root(0) 0755
EOF

DESC="Test program"
MAINTAINER="J. Random Packager <someone@example.net>"
LICENSE="WTFPL"

# PKG1 - has a functional preremove
cat << EOF > ${META}/preremove
#!/bin/sh
if [ "\${ACTION}" = "remove" ]; then
    echo "*** PREREMOVE RUNS ON REMOVE ***"
fi
EOF

chmod 0755 ${META}/preremove
create_pkg "1.0"
rm -f ${META}/preremove

# PKG2 - has a non-functional preremove
cat << EOF > ${META}/preremove
#!/bin/sh
exit 1
EOF

chmod 0755 ${META}/preremove
create_pkg "2.0"
rm -f ${META}/preremove

# PKG3 - has a functional postremove
cat << EOF > ${META}/postremove
#!/bin/sh
if [ "\${ACTION}" = "remove" ]; then
    echo "*** POSTREMOVE RUNS ON REMOVE ***"
fi
EOF

chmod 0755 ${META}/postremove
create_pkg "3.0"
rm -f ${META}/postremove

# PKG4 - has a non-functional postremove
cat << EOF > ${META}/postremove
#!/bin/sh
exit 1
EOF

chmod 0755 ${META}/postremove
create_pkg "4.0"
rm -f ${META}/postremove

#############
# preremove #
#############

# check that package runs preremove script on removal
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r
echo -n "checking distrm preremove script execution..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG1} 2>&1)"
out2="$(remove_pkg -r ${OUTPUT}/r test 2>&1)"
r=$?
echo "${out2}" | grep -q "*** PREREMOVE RUNS ON REMOVE ***" >/dev/null 2>&1
q=$?
if [ ${r} -eq 0 ] && [ ${q} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

# check that preremove failure halts distrm
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r
echo -n "checking distrm halts on preremove script failure..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG2} 2>&1)"
out2="$(remove_pkg -r ${OUTPUT}/r test 2>&1)"
r=$?
echo "${out2}" | grep -q "*** preremove failed for test, skipping package" >/dev/null 2>&1
q=$?
if [ ${r} -eq 1 ] && [ ${q} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

# check that preremove failure can be ignored (-F)
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r
echo -n "checking 'distrm -F' continues on preremove script failure..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG2} 2>&1)"
out2="$(remove_pkg -F -r ${OUTPUT}/r test 2>&1)"
r=$?
echo "${out2}" | grep -q "*** preremove failed for test, continuing anyway" >/dev/null 2>&1
q=$?
if [ ${r} -eq 0 ] && [ ${q} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

# check that preremove failure can be ignored by skipping scripts (-S)
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r
echo -n "checking 'distrm -S' continues on preremove script failure..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG2} 2>&1)"
out2="$(remove_pkg -S -r ${OUTPUT}/r test 2>&1)"
r=$?
echo "${out2}" | grep -q "Removing" >/dev/null 2>&1
q=$?
if [ ${r} -eq 0 ] && [ ${q} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

##############
# postremove #
##############

# check that package runs postremove script on removal
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r
echo -n "checking distrm postremove script execution..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG3} 2>&1)"
out2="$(remove_pkg -r ${OUTPUT}/r test 2>&1)"
r=$?
echo "${out2}" | grep -q "*** POSTREMOVE RUNS ON REMOVE ***" >/dev/null 2>&1
q=$?
if [ ${r} -eq 0 ] && [ ${q} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

# check that postremove failure halts distrm
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r
echo -n "checking distrm halts on postremove script failure..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG4} 2>&1)"
out2="$(remove_pkg -r ${OUTPUT}/r test 2>&1)"
r=$?
echo "${out2}" | grep -q "*** postremove failed for test, skipping package" >/dev/null 2>&1
q=$?
if [ ${r} -eq 1 ] && [ ${q} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

# check that postremove failure can be ignored (-F)
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r
echo -n "checking 'distrm -F' continues on postremove script failure..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG4} 2>&1)"
out2="$(remove_pkg -F -r ${OUTPUT}/r test 2>&1)"
r=$?
echo "${out2}" | grep -q "*** postremove failed for test, continuing anyway" >/dev/null 2>&1
q=$?
if [ ${r} -eq 0 ] && [ ${q} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

# check that postremove failure can be ignored by skipping scripts (-S)
rm -rf ${OUTPUT}/r
mkdir -p ${OUTPUT}/r
echo -n "checking 'distrm -S' continues on postremove script failure..."
out1="$(install_pkg -d -r ${OUTPUT}/r ${OUTPUT}/${PKG4} 2>&1)"
out2="$(remove_pkg -S -r ${OUTPUT}/r test 2>&1)"
r=$?
echo "${out2}" | grep -q "Removing" >/dev/null 2>&1
q=$?
if [ ${r} -eq 0 ] && [ ${q} -eq 0 ]; then
    echo "passed."
else
    FAILS=$((${FAILS} + 1))
    echo "FAILED."
fi
if [ ! -z "${LEAKCHECK}" ]; then
    echo "${out1}"
    echo "${out2}"
fi

rm -rf ${TREE}
rm -rf ${META}
rm -rf ${TMPDIST}
rm -rf ${OUTPUT}

# expected 0 failures
[ ${FAILS} -eq 0 ] && exit 0 || exit 1
