# functions.zsh
# Misc functions used by the installer and other config scripts in the
# distribution.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# Copyright Chris Lumens <chris@bangmoney.org>
# SPDX-License-Identifier: Apache-2.0

setopt sh_word_split

PATH=/usr/bin

# Remove possible '=' from long option argument values.  Needed to
# emulate the functionality of the GNU getopt(1) command which was
# used prior to zparseopts.
trim_equal_sign() {
    [ -z "$1" ] && return

    if [ "$(echo $1 | cut -c1)" = "=" ]; then
        echo $1 | cut -c1
    else
        echo $1
    fi
}

# The main thing called by _infobox and _msgbox.
# Arguments:  $1   The type of box.
#             $2   The message to display.
#             $3   The dialog box title.
_box() {
    local clarg=
    local height=
    local typ="$1"
    local title="$2"
    local msg="$3"

    if [ "$typ" = "msg" ]; then
        clarg="--clear"
        arg="--msgbox"
        height=7
    elif [ "$typ" = "info" ]; then
        arg="--infobox"
        height=5
    else
        exit 1
    fi

    dialog $clarg --backtitle "$BACKTITLE" \
        --title "$title" $arg "\n$msg" $height $((${#${msg}} + 4))
}

# Display a single line information message with no buttons.
# Arguments:  $1   The message to display.
#             $2   The dialog box title.
infobox() {
    local title="$1"
    local msg="$2"

    _box "info" "$title" "$msg"
}

# Display a single line information message and wait for input.
# Arguments:  $1   The message to display.
#             $2   The dialog box title.
msgbox() {
    local title="$1"
    local msg="$2"

    _box "msg" "$title" "$msg"
}

# Ask the user a yes/no question and collect the response
# Arguments:  $1   The question to ask.
#             $2   The default response ('y' or 'n').
#             $3   The dialog box title (optional).
#             $4   The dialog box height (optional, 7 by default).
#             $5   The dialog box width (optional, length of string+5 by default).
# Returns:    0 for Yes and 1 for No
questionbox() {
    local msg="$1"
    local default="$2"
    local title="$3"
    local height="$4"
    local width="$5"

    local defopt=
    local titleopt=
    local scr=
    local ret=
    local have=
    local ans=
    local newans=

    [ -z "$height" ] && height="7"
    [ -z "$width" ] && width=$((${#${msg}} + 5))

    [ "${(L)default}" = "n" ] && defopt="--defaultno"
    [ -z "$title" ] || titleopt="--title \"$title\""
    scr="dialog --clear --backtitle \"$BACKTITLE\" $titleopt $defopt --yesno \"\n${msg}?\" $height $width"
    eval $scr
    ret=$?
    if [ $ret -eq 255 ]; then
        [ "$default" = "y" ] && ret=0 || ret=1
    fi

    return $ret
}

# Prompts the user for a string of input.
# Arguments:  $1   The description string.
#             $2   Secure input (y/n, default 'n').
#             $3   File to write input to ($INPUTBOX).
#             $4   The dialog box title (optional).
#             $5   The default answer (optional).
# Returns:    0 on normal exit, 1 if user canceled the dialog box.  The input
#             box contents is written to the file pointed to by $INPUTBOX
# Important:  The caller has to extract the contents of $INPUTBOX and should
#             remove it.  This function will always remove it before asking
#             the user for a new value.
inputbox() {
    local msg="$1"
    local secure="$2"
    local inputbox="$3"
    local title="$4"
    local default="$5"

    local titleopt=
    local boxopt=
    local defopt=
    local scr=
    local ret=
    local response=

    [ -z "$title" ] || titleopt="--title \"$title\""

    if [ ! -z "$default" ] && [ "${(L)secure}" = "n" ]; then
        defopt="\"$default\""
    fi

    [ "${(L)secure}" = "y" ] && boxopt="--passwordbox" || boxopt="--inputbox"

    scr="dialog --clear --backtitle \"$BACKTITLE\" $titleopt $boxopt \"${msg}\" 8 $((${#${msg}} + 5)) $defopt 2>$inputbox"
    eval $scr
    ret=$?

    if [ $ret -eq 1 ] || [ $ret -eq 255 ]; then
        return 1
    fi

    return 0
}

# Validate IPv4 address
is_valid_ipv4_addr() {
    local result=$(echo $1 | awk -F . ' $0 ~ /^([0-9]{1,3}\.){3}[0-9]{1,3}$/ && $1 <=255 && $2 <= 255 && $3 <= 255 && $4 <= 255 ')
    [ -z "$result" ] && return 1 || return 0
}

# Validate CIDR prefix
is_valid_cidr_prefix() {
    local prefix="$1"
    [ -z "$prefix" ] && return 1
    echo "$prefix" | grep -q -E "^[0-9]+$" 2>/dev/null
    [ $? -eq 1 ] && return 1
    if [ $prefix -lt 1 ] || [ $prefix -gt 32 ]; then
        return 1
    fi
}

# Checks if the framebuffer device is a HiDPI display or not
is_hidpi() {
    if [ -r "/sys/class/graphics/fb0/virtual_size" ]; then
        vs="$(cat /sys/class/graphics/fb0/virtual_size 2>/dev/null)"

        # Add your favorite resolutions to this case block.  Here is where I get them:
        # https://en.wikipedia.org/wiki/Display_resolution
        #
        # 3840x2160 - UHD-1 (e.g., Dell 24" UHD displays)
        # 1920x1080 - HD 1080 (seen on the ThinkPad X1 Carbon)
        case "${vs}" in
            "3840,2160"|"1920,1080")
                return 0
                ;;
        esac
    fi

    return 1
}
