#!/bin/sh
#
# install-chroot.sh: Install Peachtree Linux to the $CHROOT path.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level distribution source directory.
#    $COLLECTIONS          The names of all the collections.
#    $USE_MINIROOT         (y/n) If 'y', include packages with MINIROOT_PKG=y
#    $RELEASE_DIST         The directory containing built packages.
#    $TARGET_MACH          The target architecture we are working with.
#    $CHROOT               The chroot path (cannot be '/').
# This script is called by the build system scripts and should not be called
# directly.

PATH=/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin:/usr/local/sbin

update_ld_so_cache() {
    if [ -x "${CHROOT}"/sbin/ldconfig ]; then
        chroot "${CHROOT}" /sbin/ldconfig 2>/dev/null
    elif [ -x "${CHROOT}"/usr/sbin/ldconfig ]; then
        chroot "${CHROOT}" /usr/sbin/ldconfig 2>/dev/null
    fi
}

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

if [ -z "${COLLECTIONS}" ] && [ -z "${USE_MINIROOT}" ]; then
    exit 2
fi

if [ -z "${RELEASE_DIST}" ] || [ ! -d "${RELEASE_DIST}" ]; then
    exit 3
fi

if [ -z "${CHROOT}" ]; then
    echo "*** You must specify the CHROOT= variable." >&2
    exit 4
fi

if [ -d "${CHROOT}" ] && [ "$(readlink -f "${CHROOT}")" = "/" ]; then
    exit 5
fi

if [ ! -d "${CHROOT}" ]; then
    mkdir -p "${CHROOT}"
    chmod 0755 "${CHROOT}"
fi

# Require all collections to be built or use the miniroot packages
if [ -z "${USE_MINIROOT}" ]; then
    for collection in ${COLLECTIONS} ; do
        if [ ! -d "${RELEASE_DIST}"/"${collection}" ]; then
            echo "*** The ${collection} collection has not been built." >&2
            exit 6
        fi
    done
fi

# For a chroot installation, the user must be root.
if [ "$(id -u)" -ne 0 ]; then
    echo "*** You must be root to perform a chroot installation." >&2
    exit 7
fi

# We need distadd.  Just assume the one we can exec is current enough.
if ! distadd --help >/dev/null 2>&1 ; then
    echo "*** distadd command not found" >&2
    exit 8
fi

# We need etcmerge.  Just assume the one we can exec is current enough.
if ! etcmerge --help >/dev/null 2>&1 ; then
    echo "*** etcmerge command not found" >&2
    exit 9
fi

TMPDIR="$(mktemp -d)"
PKGLIST="${TMPDIR}"/pkglist

# Create the chroot environment
if [ "${USE_MINIROOT}" = "y" ]; then
    # This is with USE_MINIROOT=y in the environment

    # collect all of the packages that are marked for inclusion in the miniroot
    find "${TOPSRC}"/base -type f -name "plist*" | grep -vE 'plist.*\.[cho5]$' | while read -r pkgdir ; do
        subdir="$(dirname "${pkgdir}")"
        pkgpath="$(echo "${subdir}" | sed -e "s|^${TOPSRC}/||g")"
        MINIROOT_PKG="$(make --no-print-directory -C "${subdir}" show var=MINIROOT_PKG)"

        # full path to built package
        if [ "${MINIROOT_PKG}" = "y" ]; then
            echo "${pkgpath}:$(make --no-print-directory -C "${subdir}" show var=FULLPKGPATH)" >> "${PKGLIST}"
        fi
    done

    # Install the packages in build dependency order
    if [ -s "${PKGLIST}" ]; then
        # eliminate duplicate package entries
        sort -u "${PKGLIST}" > "${PKGLIST}".n
        mv "${PKGLIST}".n "${PKGLIST}"

        # create a location to store ld.so.cache in the chroot
        mkdir -p "${CHROOT}"/etc
        chmod 0755 "${CHROOT}"/etc
        touch "${CHROOT}"/etc/ld.so.cache

        # install the miniroot packages
        env COLLECTIONS=base make --no-print-directory -C "${TOPSRC}" show order | while read -r srcpath ; do
            have="$(grep "^${srcpath}:" "${PKGLIST}" 2>/dev/null)"

            if [ -n "${have}" ]; then
                pkg="$(echo "${have}" | cut -d ':' -f 2)"

                # Do not run the kernel postinstall script (it makes an
                # initramfs, but we do not want that here)
                if [ "${srcpath}" = "base/kernel" ]; then
                    distadd -S -r "${CHROOT}" -e "${pkg}"
                else
                    distadd -r "${CHROOT}" -e "${pkg}"
                fi
            fi
        done
    fi

    # Update library cache after installing the base collection
    update_ld_so_cache
else
    # This is using the "${COLLECTIONS}" variable

    # Generate list of base packages to install
    rm -f "${PKGLIST}"
    for a in ${TARGET_MACH} noarch ; do
        if [ -f "${RELEASE_DIST}"/base/"${a}"/00-ORDER ]; then
            cat "${RELEASE_DIST}"/base/"${a}"/00-ORDER >> "${PKGLIST}"
        fi
    done

    # Install the base packages
    while read -r dist ; do
        pkg="${RELEASE_DIST}"/base/"${TARGET_MACH}"/"${dist}"
        [ -f "${pkg}" ] || pkg="${RELEASE_DIST}"/base/noarch/"${dist}"

        if [ ! -f "${pkg}" ]; then
            echo "*** Missing ${pkg}, try 'make world' again." >&2
            exit 10
        fi

        distadd -m -r "${CHROOT}" -d -e "${pkg}"
    done < "${PKGLIST}"

    # Update library cache after installing the base collection
    update_ld_so_cache

    # Install the remaining collections
    for collection in ${COLLECTIONS} ; do
        [ "${collection}" = "base" ] && continue

        for a in ${TARGET_MACH} noarch ; do
            while read -r pkg ; do
                pkg="${RELEASE_DIST}"/"${collection}"/"${a}"/"${dist}"

                if [ ! -f "${pkg}" ]; then
                    echo "*** Missing ${pkg}, try 'make world' again." >&2
                    exit 11
                fi

                distadd -m -r "${CHROOT}" -d -e "${pkg}"
            done < "${RELEASE_DIST}"/"${collection}"/"${a}"/00-ORDER
        done

        # Update the library cache after the collection is installed
        update_ld_so_cache
    done
fi

# Perform initial boot up things that /etc/rc normally would do.
# These are things that a freshly installed system would have happen at boot up.
# Initialize/update cache files, set up other control files, etc.

# Run without network in the miniroot
sed -i -e 's/^id:.*:initdefault:$/id:2:initdefault:/g' "${CHROOT}"/etc/inittab

# Blank out root's password
sed -i -e 's/^root:x:/root::/g' "${CHROOT}"/etc/passwd

# We need a unique UUID for dbus...probably?
if [ -x "${CHROOT}"/usr/bin/dbus-uuidgen ]; then
    chroot "${CHROOT}" /usr/bin/dbus-uuidgen --ensure
fi

# Despite no network, drop in a resolv.conf to appease glibc
if [ -f /etc/resolv.conf ]; then
    install -D -m 0644 /etc/resolv.conf "${CHROOT}"/etc/resolv.conf
fi

# These things are all irritating
if [ -x "${CHROOT}"/usr/bin/env ]; then
    if [ -x "${CHROOT}"/usr/bin/gdk-pixbuf-query-loaders ]; then
        chroot "${CHROOT}" /usr/bin/env -u LC_ALL -u LC_COLLATE -u LANG /usr/bin/gdk-pixbuf-query-loaders --update-cache
    fi

    if [ -x "${CHROOT}"/usr/bin/pango-query-modules ]; then
        chroot "${CHROOT}" /usr/bin/env -u LC_ALL -u LC_COLLATE -u LANG /usr/bin/pango-querymodules > "${TMPDIR}"/pango.modules
        install -D -m 0644 "${TMPDIR}"/pango.modules "${CHROOT}"/etc/pango/pango.modules
    fi

    if [ -x "${CHROOT}"/usr/bin/gio-querymodules ] && [ -d "${CHROOT}"/usr/lib/gio/modules ]; then
        chroot "${CHROOT}" /usr/bin/env -u LC_ALL -u LC_COLLATE -u LANG /usr/bin/gio-querymodules /usr/lib/gio/modules
    fi
fi

# Clean up
rm -rf "${TMPDIR}"
exit 0
