#!/bin/sh
#
# last-commit.sh: Looks at the most recent commit and if it contains any
#                 changes to packages, it builds them in the order they
#                 appear in the log output.  This target is called by the
#                 automated build system.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level distribution source directory.
#    $MAKE                 The make command.
# This script is called by the build system scripts and should not be called
# directly.
#
# These variables come from BuildBot:
#    $BUILD_MASTER
#    $BUILD_USER
#    $BRANCH
#    $INCOMING_DIR
#

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

[ -z "${MAKE}" ] && exit 2

# We mark our place in the repo so we know what we last built.  If it doesn't
# exist, we just go with the most recent commit.
BUILDBOT_TAG="buildbot"

TMPDIR="$(mktemp -d)"
makecmd="${MAKE} --no-print-directory"
trap 'rm -rf "${TMPDIR}"' EXIT

move_marker() {
    cd "${TOPSRC}" || exit 1
    git tag -f ${BUILDBOT_TAG}
}

cd "${TOPSRC}" || exit 1

# Gather all the package subdirectories since buildbot's last run
if git tag -l | grep -q ${BUILDBOT_TAG} >/dev/null 2>&1 ; then
    # We have a marker, gather commits since
    COMMITS="$(git rev-list --reverse ${BUILDBOT_TAG}..HEAD)"
else
    # No marker, go with most recent commit
    COMMITS="$(git rev-parse HEAD)"
fi

# No commits, exit now
[ -z "${COMMITS}" ] && exit 0

# Gather the packages to build
for commit in ${COMMITS} ; do
    git show --pretty="" --name-only "${commit}" | while read -r f ; do
        # Directory name of this file
        dirpath="$(dirname "${TOPSRC}"/"${f}")"
        [ -d "${dirpath}" ] || continue

        # Only pick it up if it's a package
        ispkg="$(${makecmd} -C "${dirpath}" ispkg 2>&1)"
        if [ "${ispkg}" = "yes" ]; then
            pkgpath="$(echo "${dirpath}" | sed -e "s|${TOPSRC}/||g")"
            echo "${pkgpath}" >> "${TMPDIR}"/pkgs
        fi
    done
done

# No packages changed
if [ ! -f "${TMPDIR}"/pkgs ]; then
    rm -rf "${TMPDIR}"
    move_marker
    exit 0
fi

# Unique the list and get it in build dependent order
BUILD_PKGS="^($(sort "${TMPDIR}"/pkgs | uniq | tr '\n' '|' | sed -e 's|\|$||g'))$"

# No packages found to build, exit now
[ -z "${BUILD_PKGS}" ] && exit 0

# If we have a host to upload to, do it
do_upload=0
if [ -n "${BUILD_MASTER}" ] && [ -n "${BUILD_USER}" ] && [ -n "${BRANCH}" ] && [ -n "${INCOMING_DIR}" ]; then
    do_upload=1
fi

# Build the packages
grep -h -E "${BUILD_PKGS}" "${TOPSRC}"/mk/order/*.list | while read -r p ; do
    # generate the new packages
    ${makecmd} -C "${p}" bundle || exit 3

    # it was successful so copy up the bundle
    if [ ${do_upload} -eq 1 ]; then
        wrkdir="$(${makecmd} -C "${p}" show var=WRKDIR)"
        [ -d "${wrkdir}" ] || exit 4

        # figure out if we need to accept the host key or not
        # accept it if we don't have a key for the build master
        SSHCMD="ssh"
        SCPCMD="scp"

        if [ ! -f "${HOME}"/.ssh/known_hosts ]; then
            SSHCMD="ssh -o StrictHostKeyChecking=no"
            SCPCMD="scp -o StrictHostKeyChecking=no"
        else
            BUILD_MASTER_IP="$(host "${BUILD_MASTER}" | rev | cut -d ' ' -f 1 | rev)"

            if [ -n "${BUILD_MASTER_IP}" ]; then
                grep "${BUILD_MASTER_IP}" "${HOME}"/.ssh/known_hosts
                HAVE_IP=$?
            fi

            grep "${BUILD_MASTER}" "${HOME}"/.ssh/known_hosts
            HAVE_HOST=$?

            if [ "${HAVE_IP}" -eq 1 ] && [ ${HAVE_HOST} -eq 1 ]; then
                SSHCMD="ssh -o StrictHostKeyChecking=no"
                SCPCMD="scp -o StrictHostKeyChecking=no"
            fi
        fi

        ${SSHCMD} "${BUILD_USER}"@"${BUILD_MASTER}" mkdir -p "${INCOMING_DIR}"/"${BRANCH}"
        ${SCPCMD} "${wrkdir}"/*.bundle "${BUILD_USER}"@"${BUILD_MASTER}":"${INCOMING_DIR}"/"${BRANCH}"
    fi

    # don't want the worker system to run out of disk space
    ${makecmd} -C "${p}" clean
done

# Builds have been uploaded to the master server, wipe things clean
${makecmd} -C "${TOPSRC}" realclean

# Clean up and exit
move_marker
exit 0
