#!/bin/sh
#
# make-release.sh: Create bootable release media from the staging tree.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level distribution source directory.
#    $DISTFILES            The location of upstream distfiles.
#    $OUTPUT               The directory to write the miniroot.img to.
#    $CHROOT               The chroot path (cannot be '/').
#    $TARGET_MACH          The target architecture.
#    $RELEASE              The built release subdirectory.
#    $RELEASE_DIST         The built release dist subdirectory.
#    $DISTRIBUTION_VERSION The version of Peachtree Linux.
#    $KERNEL_VERSION       The kernel version included in this tree.
#    $KERNEL_WRKSRC        The extracted kernel source directory.
# This script is called by the build system scripts and should not be called
# directly.

PATH=/usr/bin:/usr/sbin

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

if [ -z "${DISTFILES}" ] || [ ! -d "${DISTFILES}" ]; then
    exit 2
fi

if [ -z "${OUTPUT}" ] || [ ! -d "${OUTPUT}" ]; then
    exit 3
fi

[ -z "${CHROOT}" ] && exit 4

if [ ! -d "${CHROOT}" ] || [ "$(readlink -f "${CHROOT}")" = "/" ]; then
    exit 5
fi

[ -z "${TARGET_MACH}" ] && exit 6

if [ -z "${RELEASE}" ] || [ ! -d "${RELEASE}" ]; then
    exit 7
fi

if [ -z "${RELEASE_DIST}" ] || [ ! -d "${RELEASE_DIST}" ] || [ ! -d "${RELEASE_DIST}" ]; then
    exit 8
fi

[ -z "${DISTRIBUTION_VERSION}" ] && exit 9
[ -z "${KERNEL_VERSION}" ] && exit 10

# The name of the kernel image file depends on the target architecture
if [ "${TARGET_MACH}" = "x86_64" ]; then
    KERNEL="vmlinuz-${TARGET_MACH}"
elif [ "${TARGET_MACH}" = "ppc64" ]; then
    KERNEL="vmlinux-${TARGET_MACH}"
else
    KERNEL="vmlinux"
fi

MINIROOT="${OUTPUT}/miniroot-${TARGET_MACH}.img"
RELEASE_ISO="${RELEASE}/media/${TARGET_MACH}/release-${DISTRIBUTION_VERSION}-${TARGET_MACH}-$(date +%Y%m%d).iso"
RELEASE_LABEL=PLDIST
TMPDIR="$(mktemp -d -t make-release.XXX)"

# User must be root to create the miniroot
if [ "$(id -u)" -ne 0 ]; then
    echo "*** You must be root to perform a chroot installation." >&2
    exit 11
fi

# Clear out initramfs-root
rm -f "${CHROOT}/*"

# Expand grub module list with dependent modules
collect_grub_modules() {
    INPUT_LIST="$1"
    DEP_LIST="$2"
    OUTPUT_LIST="${TMPDIR}"/modules.list

    if [ ! -f "${INPUT_LIST}" ]; then
        echo "*** unable to read ${INPUT_LIST}" >&2
        exit 1
    fi

    if [ ! -f "${DEP_LIST}" ]; then
        echo "*** unable to read ${DEP_LIST}" >&2
        exit 1
    fi

    rm -f "${OUTPUT_LIST}"

    grep -v ^# "${INPUT_LIST}" | while read -r mod ; do
        # We want to use the for loop to read module names broken up
        # by $IFS.  We don't care about the whole line here.
        # shellcheck disable=SC2013
        for modfield in $(grep "^${mod}:" "${DEP_LIST}" 2>/dev/null) ; do
            echo "${modfield}" | sed -e 's/://g'
        done >> "${OUTPUT_LIST}"
    done

    sort "${OUTPUT_LIST}" | uniq > "${OUTPUT_LIST}".new
    mv "${OUTPUT_LIST}".new "${OUTPUT_LIST}"
    echo "${OUTPUT_LIST}"
}

# Create the checksum files.
gen_SUMS() {
    type=sha512
    output="$(echo ${type} | tr '[:lower:]' '[:upper:]')"

    find "${RELEASE}"/ -type d | while read -r dname ; do
        if [ "${dname}/" = "$(realpath "${RELEASE}" 2>/dev/null)/" ]; then
            openssl dgst -${type} -- * > "${TMPDIR}/${output}" 2>/dev/null
            mv "${TMPDIR}/${output}" "${RELEASE}"
            [ -s "${RELEASE}/${output}" ] || rm -f "${RELEASE}/${output}"
            continue
        fi

        ( cd "${dname}" || exit 1
          openssl dgst -${type} -- * > "${TMPDIR}/${output}" 2>/dev/null
          mv "${TMPDIR}/${output}" .
        )

        [ -s "${dname}/${output}" ] || rm -f "${dname}/${output}"
    done
}

# Write out the version information file for this release tree.
gen_VERSION() {
    # shellcheck disable=SC1091
    . "${CHROOT}"/etc/os-release
    echo "${NAME} version ${VER}" > "${RELEASE}"/VERSION
}

# Remove any empty directories that may exist in the dist/ tree
find "${RELEASE_DIST}"/ -depth -type d -print0 | xargs -0 rmdir -p --ignore-fail-on-non-empty 2>/dev/null

# Count packages we have against what's in the 00-ORDER files
# Look for those files in architecture specific subdirectories, which will
# make sure we don't error out if we're missing arch-specific collections
# on a different architecture.
cd "${RELEASE_DIST}" || exit 1
for c in $(find . -type f -name "00-ORDER" | grep "${TARGET_MACH}" | sort -u) ; do
    collection="$(basename "$(dirname "$(dirname "${c}")")")"
    cd "${RELEASE_DIST}"/"${collection}" || exit 1

    ( cd "${TARGET_MACH}" || exit 1 ; ls -1 -- *.dist > "${TMPDIR}"/ls.out )
    expected=$(wc -l "${TARGET_MACH}/00-ORDER" | cut -d ' ' -f 1)

    if [ -d noarch ]; then
        ( cd noarch || exit 1 ; ls -1 -- *.dist >> "${TMPDIR}"/ls.out )
        expected=$((expected + $(wc -l noarch/00-ORDER | cut -d ' ' -f 1)))
    fi

    found=$(wc -l "${TMPDIR}"/ls.out | cut -d ' ' -f 1)

    if [ "${expected}" -ne "${found}" ]; then
        echo "*** ERROR: package count mismatch in ${collection} collection" >&2
        echo "***            Expected: ${expected}" >&2
        echo "***            Found:    ${found}" >&2

        if [ -d noarch ]; then
            sort -u "${TARGET_MACH}"/00-ORDER noarch/00-ORDER > "${TMPDIR}"/00-ORDER.sorted
        else
            sort -u "${TARGET_MACH}"/00-ORDER > "${TMPDIR}"/00-ORDER.sorted
        fi

        sort -u "${TMPDIR}"/ls.out > "${TMPDIR}"/ls.out.sorted
        diff -u "${TMPDIR}"/00-ORDER.sorted "${TMPDIR}"/ls.out.sorted | while read -r line ; do
            echo "*** ${line}"
        done

        exit 12
    fi
done

# Update the repodata files
printf "%s" "Updating repodata files..."
makerepo "${RELEASE}"/dist/
printf "%s\n" "done."

# Remove any existing autogenerated release tree files
find "${RELEASE}"/ -type f -name SHA512 -print0 | xargs -0 rm 2>/dev/null
find "${RELEASE}"/ -type f -name PACKAGES -print0 | xargs -0 rm 2>/dev/null

# Generate release tree content
mkdir -p "${RELEASE}"/media/"${TARGET_MACH}"
mkdir -p "${RELEASE}"/source
cp -p "${TOPSRC}"/LICENSE "${RELEASE}"
cp -p "${TOPSRC}"/README "${RELEASE}"
[ -f "${TOPSRC}"/NEWS ] && cp -p "${TOPSRC}"/NEWS "${RELEASE}"
[ -f "${TOPSRC}"/UPDATING ] && cp -p "${TOPSRC}"/UPDATING "${RELEASE}"
install -m 0755 "${TOPSRC}"/doc/verify.sh "${RELEASE}"/verify.sh
sed -e "s|%VERSION%|${DISTRIBUTION_VERSION}|g" < "${TOPSRC}"/doc/INSTALL.in | fold -s -w 75 > "${RELEASE}"/INSTALL
gen_VERSION

# Collection lists for network install support
( cd "${RELEASE}"/dist/"${DISTRIBUTION_VERSION}" || exit 1 ; find . -maxdepth 1 -type f | cut -c3- | grep -v '^debug$' > "${TMPDIR}"/00-COLLECTIONS )
mv "${TMPDIR}"/00-COLLECTIONS "${RELEASE}"/dist/"${DISTRIBUTION_VERSION}"
( cd "${RELEASE}"/dist/"${DISTRIBUTION_VERSION}"/debug || exit 1 ; ls -1 > "${TMPDIR}"/00-COLLECTIONS )
mv "${TMPDIR}"/00-COLLECTIONS "${RELEASE}"/dist/"${DISTRIBUTION_VERSION}"/debug

# Generate the PACKAGES file(s)
find "${RELEASE}"/dist -type f -name "*.dist" -print0 | xargs -0 dirname | sort -u | while read -r c ; do
    { echo
      echo "Packages in $(echo "${c}" | sed -e "s|^$(realpath "${RELEASE}" 2>/dev/null)/||g")"
      date
      echo
    } > "${c}/header"

    find "${c}" -type f -name '*.dist' | while read -r p ; do
        echo
        echo "------------------------------------------------------------"
        echo
        distinfo -A "${p}" | sed -e "s|\ $(realpath "${RELEASE}" 2>/dev/null)/| |g"
    done >> "${c}/PACKAGES.n"

    echo "Total size of all packages:  $(grep "Package size:" "${c}/PACKAGES.n" | awk '{ total = total + $3 } END { print total }') bytes" >> "${c}/header"
    echo "Total installed size of all packages:  $(grep "Installed size:" "${c}/PACKAGES.n" | awk '{ total = total + $3 } END { print total }') bytes" >> "${c}/header"

    cat "${c}/header" "${c}/PACKAGES.n" > "${c}/PACKAGES"
    rm -f "${c}/header" "${c}/PACKAGES.n"
done

# Copy in distfiles (assumes prune-distfiles has been run).
rm -rf "${RELEASE}"/source/distfiles
mkdir -p "${RELEASE}"/source/distfiles
cp -pr "${DISTFILES}"/* "${RELEASE}"/source/distfiles

# If a tag for this release exists, use it to create the source archive,
# otherwise just create a source archive of the master branch.
cd "${TOPSRC}" || exit 1

TAG=
if ( cd "${TOPSRC}" ; git tag -l ) | grep -q -E "^v${DISTRIBUTION_VERSION}-${TARGET_MACH}$" >/dev/null 2>&1 ; then
    TAG="v${DISTRIBUTION_VERSION}-${TARGET_MACH}"
fi

if [ -z "${TAG}" ]; then
    if ( cd "${TOPSRC}" ; git tag -l ) | grep -q -E "^v${DISTRIBUTION_VERSION}$" >/dev/null 2>&1 ; then
        TAG="v${DISTRIBUTION_VERSION}"
    fi
fi

[ -z "${TAG}" ] && TAG="master"

printf "%s" "Creating src.tar.gz from ${TAG}..."
git archive --format=tar --prefix=src/ -o "${RELEASE}"/source/src.tar "${TAG}"

cd "${RELEASE}/source" || exit 1
gzip -9f src.tar
printf "%s\n" "DONE"

# We do not need the dist-tools database in the initramfs
rm -rf "${CHROOT}"/var/db/dist/*

# Check the initramfs to see if any required libraries are missing.
install -D -m 0700 "${TOPSRC}"/mk/share/check-ldd.sh "${CHROOT}"/check-ldd.sh
[ -d "${CHROOT}"/dev ] || mkdir -p "${CHROOT}"/dev
mount -o bind /dev "${CHROOT}"/dev
missing="$(chroot "${CHROOT}" /check-ldd.sh)"
umount "${CHROOT}"/dev

if [ -n "${missing}" ]; then
    echo "*** ERROR: missing required libraries in initramfs." >&2
    echo "${missing}" >&2
    exit 13
fi

rm -f "${CHROOT}"/check-ldd.sh

# Make sure all ELF objects in the miniroot are stripped
strip "${CHROOT}"/usr/bin/* 2>/dev/null
strip "${CHROOT}"/usr/sbin/* 2>/dev/null
strip "${CHROOT}"/usr/libexec/* 2>/dev/null
[ -d "${CHROOT}"/usr/libexec/getconf ] && strip "${CHROOT}"/usr/libexec/getconf/* 2>/dev/null
[ -d "${CHROOT}"/usr/libexec/grub ] && strip "${CHROOT}"/usr/libexec/grub/* 2>/dev/null
strip "${CHROOT}"/usr/lib/* 2>/dev/null
[ -d "${CHROOT}"/usr/lib/gawk ] && strip "${CHROOT}"/usr/lib/gawk/* 2>/dev/null
[ -d "${CHROOT}"/usr/lib/gconv ] && strip "${CHROOT}"/usr/lib/gconv/* 2>/dev/null
[ -d "${CHROOT}"/usr/lib/security ] && strip "${CHROOT}"/usr/lib/security/* 2>/dev/null
[ -d "${CHROOT}"/usr/lib/zsh ] && find "${CHROOT}"/usr/lib/zsh -type f -name "*.so" -print0 | xargs -0 strip

# Copy kernel to the /boot directory and remove it so it does not end
# up in the miniroot
mkdir -p "${RELEASE}"/boot
if [ "${TARGET_MACH}" = "x86_64" ]; then
    cp -p "${CHROOT}"/boot/vmlinuz-"${KERNEL_VERSION}" "${RELEASE}"/boot/"${KERNEL}"
elif [ "${TARGET_MACH}" = "ppc64" ]; then
    cp -p "${CHROOT}"/boot/vmlinux-"${KERNEL_VERSION}" "${RELEASE}"/boot/"${KERNEL}"
fi
cp "${CHROOT}"/boot/System.map-"${KERNEL_VERSION}" "${TMPDIR}"
rm -rf "${CHROOT:?}"/boot

# Create the boot images before we trim the miniroot (files used to do
# this may be trimmed before the miniroot is created)
if [ "${TARGET_MACH}" = "x86_64" ]; then
    # On x86_64, create hybrid boot ISO that uses grub for BIOS and
    # UEFI systems.  This gives us something that works as a CD, DVD,
    # or USB media.
    BOOT_LOADER_PLATFORM="x86_64-efi"

    # Create the /boot directory
    [ -d "${RELEASE}"/boot ] || mkdir -p "${RELEASE}"/boot

    # Generate an expanded module list for the BIOS grub image
    MOD_LIST="$(collect_grub_modules "${TOPSRC}"/inst/"${TARGET_MACH}"/bios/modules "${CHROOT}"/usr/lib/grub/i386-pc/moddep.lst)"

    # Generate the eltorito.img file for BIOS systems
    # We want the word splitting for the grep|awk command.
    # shellcheck disable=SC2046
    env LD_LIBRARY_PATH="${CHROOT}"/usr/lib \
    "${CHROOT}"/usr/bin/grub-mkimage -O i386-pc-eltorito \
                                     -d "${CHROOT}"/usr/lib/grub/i386-pc \
                                     -o "${RELEASE}"/boot/eltorito.img \
                                     -p /boot \
                                     --disable-shim-lock \
                                     -v \
                                     $(awk 'NF' ORS=' ' < "${MOD_LIST}")

    # Build the BIOS grub configuration file
    sed -e "s|%DISTRIBUTION_VERSION%|${DISTRIBUTION_VERSION}|g" \
        -e "s|%TARGET_MACH%|${TARGET_MACH}|g" \
        "${TOPSRC}"/inst/"${TARGET_MACH}"/bios/grub.cfg.in > "${RELEASE}"/boot/grub.cfg

    # Generate an expanded module list for the UEFI grub image
    MOD_LIST="$(collect_grub_modules "${TOPSRC}"/inst/"${TARGET_MACH}"/uefi/modules "${CHROOT}"/usr/lib/grub/"${BOOT_LOADER_PLATFORM}"/moddep.lst)"

    # Generate the efiboot.img filesystem tree for UEFI systems
    mkdir -p "${TMPDIR}"/efiboot/EFI/BOOT/fonts
    cp -p "${CHROOT}"/usr/share/grub/unicode.pf2 "${TMPDIR}"/efiboot/EFI/BOOT/fonts
    # We want the word splitting for the grep|awk command.
    # shellcheck disable=SC2046
    env LD_LIBRARY_PATH="${CHROOT}"/usr/lib \
    "${CHROOT}"/usr/bin/grub-mkimage -O ${BOOT_LOADER_PLATFORM} \
                                     -d "${CHROOT}"/usr/lib/grub/${BOOT_LOADER_PLATFORM} \
                                     -o "${TMPDIR}"/efiboot/EFI/BOOT/BOOTX64.EFI \
                                     -c "${TOPSRC}"/inst/"${TARGET_MACH}"/uefi/grub-efiboot.img.cfg \
                                     -C none \
                                     -p '/EFI/BOOT' \
                                     --disable-shim-lock \
                                     -v \
                                     $(awk 'NF' ORS=' ' < "${MOD_LIST}")
    sed -e "s|%DISTRIBUTION_VERSION%|${DISTRIBUTION_VERSION}|g" \
        -e "s|%TARGET_MACH%|${TARGET_MACH}|g" \
        "${TOPSRC}"/inst/"${TARGET_MACH}"/uefi/grub.cfg.in > "${TMPDIR}"/efiboot/EFI/BOOT/grub.cfg

    # We need an EFI boot tree on the actual iso filesystem because
    # the BOOTX64.EFI loader will read the embedded grub configuration
    # file which we have set to look for the PLDIST volume and use
    # that as its root, so /EFI/BOOT/grub.cfg on the iso filesystem
    # is the main one to use.
    cp -pr "${TMPDIR}"/efiboot/EFI "${RELEASE}"
    cp -p "${TOPSRC}"/inst/x86_64/uefi/grubenv "${RELEASE}"/EFI/BOOT

    # Compute the size of efiboot.img
    # 17408 is the size of the GPT tables that parted creates, but
    # there are 2 of them (so 34816, or 68 blocks). (Note from pjones
    # via the mkefiboot code in the lorax project.)  Take the disk
    # size of the staging directory for the image for use with dd(1).
    dd if=/dev/zero of="${RELEASE}"/boot/efiboot.img bs=1k count=2880
    mkfs.fat -F 12 -v "${RELEASE}"/boot/efiboot.img

    # load the loopback module if we need to
    if ! lsmod | grep -q loop >/dev/null 2>&1 ; then
        modprobe loop
    fi

    # mount the image and copy files
    tmpmount="$(mktemp -d -t efiboot.XXX)"
    mount -o loop "${RELEASE}"/boot/efiboot.img "${tmpmount}"
    df -h "${tmpmount}"
    mkdir -p "${tmpmount}"/EFI/BOOT
    cp -v "${TMPDIR}"/efiboot/EFI/BOOT/BOOTX64.EFI "${tmpmount}"/EFI/BOOT
    ( cd "${tmpmount}" || exit ; ls -lR )
    umount "${tmpmount}"
    rmdir "${tmpmount}"
elif [ "${TARGET_MACH}" = "ppc64" ]; then
    BOOT_LOADER_PLATFORM="powerpc-ieee1275"

    # Copy initial boot files in place
    mkdir -p "${RELEASE}"/boot/"${TARGET_MACH}"/${BOOT_LOADER_PLATFORM}
    cp -p "${TOPSRC}"/inst/"${TARGET_MACH}"/ofboot.b "${RELEASE}"/boot/"${TARGET_MACH}"/${BOOT_LOADER_PLATFORM}/ofboot.b

    # Create grub boot image and copy in grub modules
    env LD_LIBRARY_PATH="${CHROOT}"/usr/lib \
    "${CHROOT}"/usr/bin/grub-mkimage -O ${BOOT_LOADER_PLATFORM} \
                                 -d "${CHROOT}"/usr/lib/grub/${BOOT_LOADER_PLATFORM} \
                                 -p "()/boot/${TARGET_MACH}" \
                                 -o "${RELEASE}"/boot/"${TARGET_MACH}"/${BOOT_LOADER_PLATFORM}/core.elf \
                                 -v \
                                 "$(tr '\n' ' ' < "${TOPSRC}"/inst/"${TARGET_MACH}"/modules)"
    cp -p "${CHROOT}"/usr/lib/grub/${BOOT_LOADER_PLATFORM}/*.mod "${RELEASE}"/boot/"${TARGET_MACH}"/${BOOT_LOADER_PLATFORM}
    cp -p "${CHROOT}"/usr/lib/grub/${BOOT_LOADER_PLATFORM}/*.lst "${RELEASE}"/boot/"${TARGET_MACH}"/${BOOT_LOADER_PLATFORM}

    # Install our grub install media config file
    sed -e "s|%DISTRIBUTION_VERSION%|${DISTRIBUTION_VERSION}|g" \
        -e "s|%TARGET_MACH%|${TARGET_MACH}|g" \
        < "${TOPSRC}"/inst/"${TARGET_MACH}"/grub.cfg.in > "${RELEASE}"/boot/"${TARGET_MACH}"/grub.cfg
fi

# Retain boot loader files we need
if [ "${TARGET_MACH}" = "x86_64" ]; then
    cp -p "${CHROOT}"/usr/lib/grub/i386-pc/boot_hybrid.img "${TMPDIR}"
fi

# Trim unwanted files from the miniroot image
env TOPSRC="${TOPSRC}" USE_MINIROOT="y" CHROOT="${CHROOT}" make --no-print-directory -C "${TOPSRC}" trim-chroot

# Remove any broken symlinks
find -L "${CHROOT}"/ -type l -exec rm -f {} \;

# Copy in installer files and modify other default config files
ln -sf usr/sbin/init "${CHROOT}"/init
rm -f "${CHROOT}"/etc/issue
rm -f "${CHROOT}"/etc/issue.net
install -D -m 0644 "${RELEASE}"/INSTALL "${CHROOT}"/root/INSTALL
install -m 0755 "${TOPSRC}"/inst/inst "${CHROOT}"/usr/sbin/inst
install -D -m 0644 "${TOPSRC}"/inst/inst.8 "${CHROOT}"/usr/share/man/man8/inst.8
install -m 0644 "${TOPSRC}"/inst/issue "${CHROOT}"/etc/issue
ln -sf issue "${CHROOT}"/etc/issue.net
install -m 0644 "${TOPSRC}"/inst/motd "${CHROOT}"/etc/motd
install -m 0644 "${TOPSRC}"/inst/rc.conf "${CHROOT}"/etc/rc.conf
install -m 0644 "${TOPSRC}"/inst/inst.conf "${CHROOT}"/etc/inst.conf
install -m 0644 "${TOPSRC}"/inst/fstab "${CHROOT}"/etc/fstab

# Regenerate module dependencies in the miniroot
printf "%s" "Renegenerating kernel module dependencies for the miniroot..."
depmod -a -b "${CHROOT}" -F "${TMPDIR}"/System.map-"${KERNEL_VERSION}" "${KERNEL_VERSION}"
printf "%s\n" "done"

# Create miniroot file.
# Use the kernel scripts to generate an external compressed initramfs.
if [ -z "${KERNEL_WRKSRC}" ] || [ ! -d "${KERNEL_WRKSRC}" ]; then
    echo "*** Missing base/kernel \${WRKSRC} directory." >&2
    echo "*** Building the release requires the kernel initramfs tools." >&2
    echo "*** Do a 'make world' in base/kernel, then try the 'make release' again." >&2
    exit 14
fi

printf "%s" "Generating compressed initramfs image (this takes a long time)..."
cd "${KERNEL_WRKSRC}" || exit 1
sh ./usr/gen_initramfs.sh \
    -o "${OUTPUT}"/initramfs.cpio \
    -u squash -g squash \
    "${CHROOT}"
zstd -9 --long -c "${OUTPUT}"/initramfs.cpio > "${MINIROOT}"
cp "${MINIROOT}" "${RELEASE}"/boot
printf "%s\n" "done."

# Generate checksum files
gen_SUMS

# Create the boot ISO for each architecture
if [ "${TARGET_MACH}" = "x86_64" ]; then
    # Create the ISO image
    xorrisofs -o "${RELEASE_ISO}" -iso-level 3 -v -r -J \
              -V "${RELEASE_LABEL}" \
              -A "Peachtree Linux" \
              --grub2-mbr "${TMPDIR}"/boot_hybrid.img \
              -partition_offset 16 \
              -appended_part_as_gpt \
              -append_partition 2 C12A7328-F81F-11D2-BA4B-00A0C93EC93B "${RELEASE}"/boot/efiboot.img \
              -iso_mbr_part_type EBD0A0A2-B9E5-4433-87C0-68B6B72699C7 \
              -c boot.cat --boot-catalog-hide \
              -b boot/eltorito.img -no-emul-boot \
              -boot-load-size 4 -boot-info-table \
              -eltorito-alt-boot -eltorito-platform efi \
              -isohybrid-gpt-basdat \
              -e '--interval:appended_partition_2:all::' -no-emul-boot \
              -m "media" \
              -m "distfiles" \
              -m "source" \
              -m "debug" \
              -m "ppc64" \
              -graft-points \
              EFI="${RELEASE}"/EFI \
              boot="${RELEASE}"/boot \
              dist="${RELEASE}"/dist \
              INSTALL="${RELEASE}"/INSTALL \
              LICENSE="${RELEASE}"/LICENSE \
              NEWS="${RELEASE}"/NEWS \
              README="${RELEASE}"/README \
              SHA512="${RELEASE}"/SHA512 \
              UPDATING="${RELEASE}"/UPDATING \
              VERSION="${RELEASE}"/VERSION \
              verify.sh="${RELEASE}"/verify.sh
elif [ "${TARGET_MACH}" = "ppc64" ]; then
    # Create the ISO image
    xorrisofs -o "${RELEASE_ISO}" -iso-level 3 -r -J -v -N \
              -omit-period -full-iso9660-filenames \
              -V "${RELEASE_LABEL}" \
              -A "Peachtree Linux" \
              -chrp-boot-part \
              -untranslated-filenames \
              -hfsplus \
              -sysid PPC \
              -hfs-bless "${RELEASE}/boot/${TARGET_MACH}/${BOOT_LOADER_PLATFORM}" \
              -m "EFI" \
              -m "media" \
              -m "distfiles" \
              -m "source" \
              -m "debug" \
              -m "x86_64" \
              -graft-points \
              /boot="${RELEASE}"/boot \
              /dist="${RELEASE}"/dist \
              /INSTALL="${RELEASE}"/INSTALL \
              /LICENSE="${RELEASE}"/LICENSE \
              /NEWS="${RELEASE}"/NEWS \
              /README="${RELEASE}"/README \
              /SHA512="${RELEASE}"/SHA512 \
              /UPDATING="${RELEASE}"/UPDATING \
              /VERSION="${RELEASE}"/VERSION \
              /verify.sh="${RELEASE}"/verify.sh
else
    echo "*** Unsupported target architecture: ${TARGET_MACH}" >&2
    exit 15
fi

# Generate checksums for the media files
cd "${RELEASE}/media/${TARGET_MACH}" || exit 1
rm -f SHA512
openssl dgst -sha512 -- * > "${TMPDIR}/SHA512"
mv "${TMPDIR}/SHA512" "${RELEASE}/media/${TARGET_MACH}"

rm -rf "${TMPDIR}" || exit 1
exit 0
