#!/bin/sh
#
# merge-releases.sh: Combine multiple release trees in to $RELEASE
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $RELEASE              The destination /usr/release tree.
#    $OTHER_RELEASES       Comma-delimited list of other /usr/release trees.
# This script is called by the build system scripts and should not be called
# directly.

PATH=/usr/bin

if [ -z "${RELEASE}" ] || [ ! -d "${RELEASE}" ]; then
    exit 1
fi

if [ -z "${OTHER_RELEASES}" ]; then
    echo "*** Missing OTHER_RELEASES environment variable." >&2
    exit 2
fi

TMPDIR="$(mktemp -d)"

for other in $(echo "${OTHER_RELEASES}" | tr ',' '\n') ; do
    if [ ! -d "${other}" ]; then
        echo "*** ${other} is not a directory, skipping." >&2
        continue
    fi

    # Copy in all new and different files, backup existing ones to have the
    # same name but ending with ~
    rsync -abviuP "${other}"/ "${RELEASE}"/

    # Iterate over all files ending with ~ and eliminate dupes and combine
    # things we can combine (text files we know the format for).
    find "${RELEASE}"/ -type f -name "*~" | while read -r oldfile ; do
        newfile="$(echo "${oldfile}" | rev | cut -c2- | rev)"
        oldsha="$(openssl dgst -sha512 "${oldfile}" | rev | cut -d ' ' -f 1 | rev)"
        newsha="$(openssl dgst -sha512 "${newfile}" | rev | cut -d ' ' -f 1 | rev)"

        if [ "${oldsha}" = "${newsha}" ]; then
            rm "${oldfile}"
        else
            distinfo -Ss "${oldfile}" > "${TMPDIR}"/vars
            unset SIZE
            # shellcheck disable=SC1091
            . "${TMPDIR}"/vars
            oldsize="${SIZE}"

            distinfo -Ss "${newfile}" > "${TMPDIR}"/vars
            unset SIZE
            # shellcheck disable=SC1091
            . "${TMPDIR}"/vars
            newsize="${SIZE}"

            if [ "${oldsize}" = "${newsize}" ]; then
                rm "${oldfile}"
            else
                echo "${newfile} contains differences, unable to resolve"
            fi
        fi
    done
done

rm -rf "${TMPDIR}"
