#!/usr/bin/python3
#
# order.py: Sort all of the packages in the specified collection in
#           build dependency order and save that order to files.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level distribution source directory.
#    $COLLECTION           The name of the collection.
#    $MAKE                 The make command.
# This script is called by the build system scripts and should not be called
# directly.

import os
import sys
from graphlib import TopologicalSorter, CycleError


def get_all_pkgs(topsrc, collection):
    pkgs = set()

    for dname, subdirlist, flist in os.walk(os.path.join(topsrc, collection)):
        for fname in flist:
            if fname == "upstream":
                pkgs.add(dname)

    return list(pkgs)


if __name__ == "__main__":
    topsrc = os.getenv("TOPSRC")
    collection = os.getenv("COLLECTION")
    graph = {}

    if topsrc is None or not os.path.isdir(topsrc):
        sys.exit(1)

    if collection is None or not os.path.isdir(os.path.join(topsrc, collection)):  # noqa: E501
        sys.exit(2)

    # gather a list of all packages in the collection
    allpkgs = get_all_pkgs(topsrc, collection)

    # iterate over every package to collect build dependencies
    for pkg in allpkgs:
        buildreqs = os.path.join(pkg, "buildreqs")
        pname = pkg[len(topsrc) + 1:]

        if os.path.isfile(buildreqs):
            graph[pname] = set(open(buildreqs, "r").read().split())
        else:
            graph[pname] = set()

    # topological sort
    try:
        ts = TopologicalSorter(graph)
        order = tuple(ts.static_order())
    except CycleError:
        sys.stderr.write("\n*** build dependencies are circular.\n")
        sys.stderr.write("*** See 'make show-order COLLECTION=[name]' for more info.\n")  # noqa: E501
        sys.stderr.write(
            "*** Also, consider backing out recent buildreqs changes to see\n"
        )
        sys.stderr.write("*** where the loop was introduced.\n\n")
        sys.stderr.write(graph)
        sys.stderr.write("\n\n")
        sys.exit(3)

    # show the order
    for pkg in order:
        if pkg.startswith(collection + "/"):
            print("%s" % pkg)

    sys.exit(0)
