#!/bin/sh
#
# trim-chroot.sh: Trim a given $CHROOT path using trimlist files.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level distribution source directory.
#    $USE_MINIROOT         (y/n) If 'y', include packages with MINIROOT_PKG=y
#    $CHROOT               The chroot path (cannot be '/').
# This script is called by the build system scripts and should not be called
# directly.

PATH=/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin:/usr/local/sbin

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

if [ -z "${CHROOT}" ]; then
    echo "*** You must specify the CHROOT= variable." >&2
    exit 2
fi

if [ -d "${CHROOT}" ] && [ "$(readlink -f "${CHROOT}")" = "/" ]; then
    exit 3
fi

# For a chroot installation, the user must be root.
if [ "$(id -u)" -ne 0 ]; then
    echo "*** You must be root to perform a chroot installation." >&2
    exit 4
fi

TMPDIR="$(mktemp -d)"
TRIMLIST="${TMPDIR}"/trimlist
OBJ="$(env make --no-print-directory -C "${TOPSRC}" show var=OBJ)"

# Collect all trimlists for miniroot packages
find "${TOPSRC}"/base -type f -name "plist*" | grep -vE 'plist.*\.[cho5]$' | while read -r pkgdir ; do
    subdir="$(dirname "${pkgdir}")"
    MINIROOT_PKG="$(make --no-print-directory -C "${subdir}" show var=MINIROOT_PKG)"

    if [ "${USE_MINIROOT}" = "y" ] && [ ! "${MINIROOT_PKG}" = "y" ]; then
        continue
    fi

    # pick up any trim list entries that may exist
    if [ -f "${subdir}"/trimlist ]; then
        grep -vE '(^$|^#)' "${subdir}"/trimlist >> "${TRIMLIST}"
    fi
done

# Trim unwanted files
echo "Trimming unwanted miniroot files..."
sort -u "${TRIMLIST}" > "${TRIMLIST}".n
mv "${TRIMLIST}".n "${TRIMLIST}"

rm -f "${OBJ}"/trim-miniroot.log 2>/dev/null

while read -r trim ; do
    [ "${CHROOT}"/"${trim}" = "${CHROOT}"/ ] && continue
    # we want globbing here because we allow globs in trimlist lines
    # shellcheck disable=SC2086
    rm -rfv "${CHROOT:?}"/${trim} >> "${OBJ}"/trim-miniroot.log 2>&1
done < "${TRIMLIST}"

# Collect all used libraries
for subdir in /usr/lib /usr/libexec /usr/bin /usr/sbin ; do
    if [ -d "${CHROOT}"${subdir} ]; then
        env LD_LIBRARY_PATH="${CHROOT}"/usr/lib ldd "${CHROOT}"${subdir}/* 2>/dev/null | grep " => " | grep "${CHROOT}" | awk '{ print $1; }' >> "${TMPDIR}"/usedlibs
    fi
done

sort "${TMPDIR}"/usedlibs | uniq > "${TMPDIR}"/usedlibs.new
mv "${TMPDIR}"/usedlibs.new "${TMPDIR}"/usedlibs

# Remove development symlinks for shared libraries
find "${CHROOT}"/usr/lib/ -type l -name "*.so" -print0 | xargs -0 rm

# Keep the used libraries
mkdir -p "${CHROOT}"/usr/lib/keep

while read -r reqlib ; do
    cp "${CHROOT}"/usr/lib/"${reqlib}" "${CHROOT}"/usr/lib/keep
done < "${TMPDIR}"/usedlibs

# Remove unused libraries
rm -f "${CHROOT}"/usr/lib/lib*.so*

# Move used libraries back in place
mv "${CHROOT}"/usr/lib/keep/* "${CHROOT}"/usr/lib
rmdir "${CHROOT}"/usr/lib/keep

# Update library cache after trimming files
if [ -x "${CHROOT}/sbin/ldconfig" ]; then
    chroot "${CHROOT}" /sbin/ldconfig 2>/dev/null
elif [ -x "${CHROOT}/usr/sbin/ldconfig" ]; then
    chroot "${CHROOT}" /usr/sbin/ldconfig 2>/dev/null
fi

# Clean up
rm -rf "${TMPDIR}"
exit 0
