#!/bin/sh
#
# world-builder.sh: Collect BUILD_DEPENDS information for all packages in
#                   the specified collection, do dependency sorting, and
#                   build all of those packages in order.
#
# Copyright David Cantrell <dcantrell@burdell.org>
# Copyright Chris Lumens <chris@bangmoney.org>
# SPDX-License-Identifier: Apache-2.0

# This script returns zero on success, non-zero on error.
# The following environment variables need to be set for the script:
#    $TOPSRC               The top level distribution source directory.
#    $COLLECTION           The name of the collection.
#    $MAKE                 The make command.
#    $CACHE_SYSROOT        If set to 'y', preserve sysroot as builds run.
#    $RELEASE              The release tree directory of the project.
#    $RELEASE_DIST         The built package directory.
#    $FORCE                If set to 'y', do not skip built packages.
#    $TARGET_MACH          The target architecture (e.g., x86_64).
#    $DISTADD_TMPDIR       Temp directory for distadd(8) to use.
# This script is called by the build system scripts and should not be called
# directly.

if [ -z "${TOPSRC}" ] || [ ! -d "${TOPSRC}" ]; then
    exit 1
fi

if [ -z "${COLLECTION}" ] || [ ! -d "${TOPSRC}/${COLLECTION}" ]; then
    exit 2
fi

[ -z "${DISTADD_TMPDIR}" ] && exit 3
[ -z "${MAKE}" ] && exit 4
[ -z "${TARGET_MACH}" ] && exit 5

[ -z "${FORCE}" ] && FORCE="n"
[ -z "${CACHE_SYSROOT}" ] && CACHE_SYSROOT="n"

[ -d "${DISTADD_TMPDIR}" ] || mkdir -p "${DISTADD_TMPDIR}"

TMPDIR="$(mktemp -d)"
ALLPKGS="${TMPDIR}/allpkgs"
BUILT="${TMPDIR}/built"

makecmd="${MAKE} --no-print-directory"
distadd="/usr/sbin/distadd"

build_package() {
    pkgpath="${1}"

    border="+-$(echo "${pkgpath}" | sed -e 's|.|-|g')-+"
    echo "${border}"
    echo "| ${pkgpath} |"
    echo "${border}"

    if [ -f "${pkgpath}"/Makefile ]; then
        fullpkgpath="$(${makecmd} -C "${pkgpath}" show var=FULLPKGPATH)"
        wrksysroot="$(${makecmd} -C "${pkgpath}" show var=WRKSYSROOT)"
        skip="$(${makecmd} -C "${pkgpath}" show var=SKIP)"
        target_mach="$(${makecmd} -C "${pkgpath}" show var=TARGET_MACH)"

        if [ "$(echo "${skip}" | tr '[:upper:]' '[:lower:]')" = "y" ]; then
            echo "Skipping ${pkgpath}, set SKIP=n to include in world build."
            return
        fi

        ${makecmd} -C "${pkgpath}" clean || return 1

        if [ "${CACHE_SYSROOT}" = "n" ]; then
            ${makecmd} -C "${pkgpath}" usereqs || return 2
        fi

        if ! ${makecmd} -C "${pkgpath}" package ; then
            touch "${RELEASE}/.failed"
            fail_world_build "${pkgpath}"
        else
            ${makecmd} -C "${pkgpath}" show var=PACKAGE >> "${BUILT}"
            if [ "${CACHE_SYSROOT}" = "y" ]; then
                ${distadd} --tmpdir="${DISTADD_TMPDIR}" -S -d -u -f -r "${wrksysroot}" "${fullpkgpath}"
            fi
        fi
    else
        echo "*** WARNING: ${pkgpath} does not exist." >&2
    fi
}

fail_world_build() {
    pkgpath="${1}"
    retval="${2}"

    echo >&2
    echo "===> ${pkgpath} failed to build" >&2
    # shellcheck disable=SC2086
    exit ${retval}
}

have_package() {
    pkgpath="${1}"
    [ -z "${pkgpath}" ] && return 1
    output="$(${makecmd} -C "${pkgpath}" show var=PACKAGE,PKGCOLLECTION,TARGET_MACH)"
    package="$(echo "${output}" | cut -d ' ' -f 1)"
    collection="$(echo "${output}" | cut -d ' ' -f 2)"
    target_mach="$(echo "${output}" | cut -d ' ' -f 3)"
    [ -f "${RELEASE_DIST}/${collection}/${target_mach}/${package}" ] && return 0 || return 1
}

# Clear any previous failure indicator
rm -f "${RELEASE}"/.failed

# Generate a list of all packages in the collection.
rm -f "${BUILT}"
rm -f "${ALLPKGS}"
${makecmd} -C "${TOPSRC}" order COLLECTIONS="${COLLECTION}" > "${ALLPKGS}" 2>/dev/null

# Recreate these files for the collection.
find "${RELEASE_DIST}"/"${COLLECTION}" -type f -name 00-ORDER -exec rm {} \; 2>/dev/null

# Build all packages
build=1
while [ ${build} -eq 1 ]; do
    pkgpath="$(head -n 1 "${ALLPKGS}")"
    have_package "${pkgpath}"
    built=${?}

    # avoid SKIP=y packages
    skip="$(${makecmd} -C "${pkgpath}" show var=SKIP)"

    # update the 00-ORDER
    package="$(${makecmd} -C "${pkgpath}" show var=PACKAGE)"
    d="${RELEASE_DIST}/${COLLECTION}/${target_mach}"
    [ -d "${d}" ] || mkdir -p "${d}"
    [ "${skip}" = "y" ] || echo "${package}" >> "${d}/00-ORDER"

    if [ ${built} -eq 0 ]; then
        if [ "${FORCE}" = "y" ]; then
            # forcing a rebuild of an existing package
            build_package "${pkgpath}" || exit 1
        elif [ "${FORCE}" = "n" ]; then
            # skip packages we already have, speeds up world builds
            echo "Skipping ${pkgpath}, already built."
        fi
    elif [ ${built} -eq 1 ]; then
        # package doesn't exist
        build_package "${pkgpath}" || exit 1
    fi

    # the sysroot may be cached, so install the package
    if [ "${skip}" = "n" ]; then
        target_mach="$(${makecmd} -C "${pkgpath}" show var=TARGET_MACH)"
        if [ "${CACHE_SYSROOT}" = "y" ]; then
            if [ "${target_mach}" = "noarch" ] || [ "${TARGET_MACH}" = "${target_mach}" ]; then
                fullpkgpath="$(${makecmd} -C "${pkgpath}" show var=FULLPKGPATH)"
                wrksysroot="$(${makecmd} -C "${pkgpath}" show var=WRKSYSROOT)"

                ${distadd} --tmpdir="${DISTADD_TMPDIR}" -S -d -u -f -r "${wrksysroot}" "${fullpkgpath}"
            fi
        fi
    fi

    # pop the package we just built from the list and continue
    len=$(wc -l < "${ALLPKGS}")
    tail -n $((len - 1)) "${ALLPKGS}" > "${ALLPKGS}".new
    mv "${ALLPKGS}".new "${ALLPKGS}"

    len=$(wc -l < "${ALLPKGS}")
    [ "${len}" -eq 0 ] && build=0
done

if [ -s "${ALLPKGS}" ]; then
    echo
    echo "*** DEPENDENCY FAILURES"
    echo "*** THE FOLLOWING PACKAGES WERE NOT BUILT:"
    cat "${ALLPKGS}"
fi

rm -rf "${TMPDIR}"
echo

exit 0
