#!/bin/sh
#
# Verify the running kernel configuration contains minimum required
# settings.  Things that should be compiled in, set as modules, or
# disabled entirely in the kernel
#

PATH=/usr/bin:/usr/sbin
CWD="$(pwd)"
ARCH="$(uname -m)"
TEST="$0"
LOGNAME="$(basename $TEST .sh).log"
LOG="$CWD/$LOGNAME"
CONF="/proc/config.gz"
RET=0

_check_enabled() {
    val="$(gzip -dc $CONF | grep "${1}=")"
    setting="$(echo "$val" | rev | cut -c 1)"

    if [ ! "$setting" = "y" ]; then
        echo "Checking for ${1}..." >> $LOG
        echo "    Expected: ${1}=y" >> $LOG
        echo "       Found: $val" >> $LOG
        echo >> $LOG
        RET=1
    fi
}

_check_enabled_as_module() {
    val="$(gzip -dc $CONF | grep "${1}=")"
    setting="$(echo "$val" | rev | cut -c 1)"

    if [ ! "$setting" = "m" ]; then
        echo "Checking for ${1}..." >> $LOG
        echo "    Expected: ${1}=m" >> $LOG
        echo "       Found: $val" >> $LOG
        echo >> $LOG
        RET=1
    fi
}

_check_disabled() {
    val="$(gzip -dc $CONF | grep "${1}")"
    setting="$(echo "$val" | cut -c 1)"

    if [ ! "$setting" = "#" ]; then
        echo "Checking for ${1}..." >> $LOG
        echo "    Expected: # ${1} is not set" >> $LOG
        echo "       Found: $val" >> $LOG
        echo >> $LOG
        RET=1
    fi
}

rm -f $LOG

echo -n "checking for minimum required kernel settings..."

if [ ! -f "$CONF" ]; then
    echo "Missing $CONF, cannot continue." >> $LOG
    echo "FAIL"
    exit 1
fi

#####################
# ALL ARCHITECTURES #
#####################

# check for common options
for option in CONFIG_HID_GENERIC \
              CONFIG_KEYBOARD_ATKBD \
              CONFIG_USB_STORAGE \
              CONFIG_USB_UAS \
              CONFIG_INPUT_EVDEV \
              CONFIG_HID \
              CONFIG_USB_HID \
              CONFIG_USB_EHCI_HCD \
              CONFIG_USB_OHCI_HCD \
              CONFIG_BINFMT_MISC ; \
do
    _check_enabled_as_module $option
done

for option in CONFIG_BLK_DEV_INITRD \
              CONFIG_RD_GZIP \
              CONFIG_RD_BZIP2 \
              CONFIG_RD_ZSTD \
              CONFIG_RD_LZO \
              CONFIG_RD_LZ4 \
              CONFIG_BINFMT_ELF \
              CONFIG_COMPAT_BINFMT_ELF \
              CONFIG_BINFMT_SCRIPT ; \
do
    _check_enabled $option
done

##########
# x86_64 #
##########

if [ "$ARCH" = "x86_64" ]; then
    for option in CONFIG_SCSI \
                  CONFIG_EFI \
                  CONFIG_EFI_PARTITION \
                  CONFIG_FB_EFI \
                  CONFIG_CRYPTO_AES ; \
    do
        _check_enabled $option
    done

    for option in CONFIG_E1000 \
                  CONFIG_IGB \
                  CONFIG_IXGBE \
                  CONFIG_USB_XHCI_HCD \
                  CONFIG_USB_UHCI_HCD \
                  CONFIG_DRM_GMA500 \
                  CONFIG_CRYPTO_AES_TI \
                  CONFIG_CRYPTO_AES_NI_INTEL ; \
    do
        _check_enabled_as_module $option
    done
fi

#########
# ppc64 #
#########

if [ "$ARCH" = "ppc64" ]; then
    _check_enabled_as_module CONFIG_SUNGEM
    _check_enabled CONFIG_PPC_4K_PAGES
    _check_enabled_as_module CONFIG_DRM_RADEON
    _check_disabled CONFIG_PPC_64K_PAGES
fi

[ $RET -eq 0 ] && echo "pass" || echo "FAIL"
exit $RET
